/* ************************************************************************* */
/* addr_calc.c: Calculate the address of each instruction or state. This is  */
/*              not architecture dependent.                                  */
/*                                                                           */
/*    Copyright 2002-2006 Axis Communications AB.                            */
/*                                                                           */
/*                                                                           */
/*    This program is free software; you can redistribute it and/or modify   */
/*    it under the terms of the GNU General Public License as published by   */
/*    the Free Software Foundation; either version 2 of the License, or      */
/*    (at your option) any later version.                                    */
/*                                                                           */
/*    This program is distributed in the hope that it will be useful,        */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*    GNU General Public License for more details.                           */
/*                                                                           */
/*    You should have received a copy of the GNU General Public License      */
/*    along with this program; if not, write to the Free Software            */
/*    Foundation, Inc., 59 Temple Place - Suite 330,                         */
/*    Boston, MA 02111-1307, USA.                                            */
/*                                                                           */
/*                                                                           */
/*    Send bug reports and inquiries to Johan.Jorgensen@axis.com             */
/* ************************************************************************* */



/* ************************************************************************* */
/*                      INCLUDES USED BY THIS MODULE                         */
/* ************************************************************************* */
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "p_list.h"
#include "cg_fe.h"
#include "dyn_array.h"
#include "dyn_stack.h"
#include "labels.h"
#include "debug.h"
#include "iopasm_err.h"
#include "iopasm_global.h"
#include "stack_eval.h"



/* ************************************************************************* */
/*                     DEFINITIONS LOCAL TO THIS MODULE                      */
/* ************************************************************************* */



/* ************************************************************************* */
/*                       TYPES DEFINED IN THIS MODULE                        */
/* ************************************************************************* */



/* ************************************************************************* */
/*                              CONSTANT VARIABLES                           */
/* ************************************************************************* */



/* ************************************************************************* */
/*                    PROTOTYPES FOR STATIC FUNCTIONS                        */
/* ************************************************************************* */



/* ************************************************************************* */
/* PROTOTYPE   : void calculate_code_size(cg_context_t *ctxt,                */
/*                                        plist_t *a_list);                  */
/*                                                                           */
/* ARGUMENTS   : "ctxt" points to a context-descriptor. "a_list" points to a */
/*               parser list.                                                */
/*                                                                           */
/* FUNCTION    : Calculates the size of each instruction and sets the size-  */
/*               field for the current instruction                           */
/*                                                                           */
/* RETURN VALUE: Error handling within this module, as errors are dependant  */
/*               on the architecture specified by current context-descriptor.*/
/*               Thus the function does not return any value.                */
/* ************************************************************************* */
void calculate_code_size(cg_context_t *ctxt,plist_t *a_list)
{
  
  inst_t *first_inst;
  inst_t *last_inst;
  int i_size;
  int num_states;

      first_inst = get_first_inst(a_list);

      do {
	minimize_stacks(ctxt,first_inst,INTERMEDIATE_PASS);
	/* eval_stacks(ctxt,first_inst,INTERMEDIATE_PASS); */

	switch (get_inst_type(first_inst)) {

	case INST_ORIGIN:
	  set_inst_size(first_inst,0);
	  break;
	  

	case INST_ALIGN:
	  set_inst_size(first_inst,0);
	  break;


	case INST_SEQ: /* Sequential instruction */
	  i_size = cg_gen_seq_code(ctxt,first_inst,INTERMEDIATE_PASS,SCOPE_SEQ); 
	  set_inst_size(first_inst,i_size);
	  break;
	  

	case INST_STATE_BEGIN:
	  /* find last instruction in a state */
	  num_states++;
	  last_inst = get_next_inst(first_inst);
	  minimize_stacks(ctxt,last_inst,INTERMEDIATE_PASS);
	  first_inst = get_prev_inst(first_inst); /* Start at label */

	  
	  while((get_inst_type(last_inst) != INST_LABEL) && 
                (get_inst_type(last_inst) != INST_STATE_END)) {
	    last_inst = get_next_inst(last_inst);
	    minimize_stacks(ctxt,last_inst,INTERMEDIATE_PASS);
	  }
	  
	  last_inst = get_prev_inst(last_inst);
	  i_size = cg_gen_fsm_code(ctxt,first_inst,last_inst,INTERMEDIATE_PASS);
	  
	  
	  if (i_size == 0) { /* We've got an error */
	    /* Report error in state here */
	  }
	  
	  else {
	    set_inst_size(first_inst,i_size);
	  }
	  
	  first_inst = last_inst;
	  break;
	}

	
	first_inst = get_next_inst(first_inst);
      } while (first_inst != NULL);
      
      return ;
}



/* ************************************************************************* */
/* PROTOTYPE   : address_calculation cg_context_t *ctxt,p_list a_list);      */
/*                                                                           */
/* ARGUMENTS   : ctxt is the current context. a_list points to a list        */
/*               containing the required code                                */
/*                                                                           */
/* FUNCTION    : Calculates the absolute adresses of all labels              */
/*                                                                           */
/* RETURN VALUE: None. This function can only fail due to an error           */
/* ************************************************************************* */
void address_calculation(cg_context_t *ctxt,plist_t *a_list)
{
  unsigned int new_pc = 0;
  unsigned int pc = 0; /* Fake program counter */
  int row_len;
  int diff;
  inst_t *first;
  inst_t *next_inst = NULL;
  label_t *temp_label;
  label_t *real_label;
  arg_t *an_arg;
  char *arg_val;
  unsigned int value;
  char *temp_string;
  array_t *labels;
  int size,arg_type;
  int bit_dif,inc_val;
  assert((ctxt != NULL) && (a_list != NULL));
  
  labels = cg_get_label_array(ctxt);
  first = get_first_inst(a_list);
  row_len = cg_get_row_len(ctxt);
  
  while (first != NULL) {

    diff = 0;

    switch(get_inst_type(first)) {

      
    case INST_ORIGIN:
      an_arg =get_first_arg(first);
      arg_type = convert_stack_arg(ctxt,first,an_arg,INTERMEDIATE_PASS);
      if (arg_type != ARG_IMMEDIATE) { 
	iopasm_error(ctxt,first,ERR_NON_CONSTANT_ARG,
		     "(.org pseudo operator)");
      }
      arg_val = get_arg_val(an_arg);
      value = atol(arg_val);
      pc = value;
      new_pc = value;
      set_inst_addr(first,value);
      break;


    case INST_ALIGN:
      bit_dif = atoi(get_arg_val(get_first_arg(first)));
      inc_val = pc % (bit_dif / 8);
      set_inst_size(first,inc_val);
      pc += inc_val;
      set_inst_addr(first,pc);
      break;


    case INST_DWORD:
    case INST_BYTE:
    case INST_WORD:
    case INST_SEQ:
      size = get_inst_size(first);
      if (cg_check_memory_region(ctxt,pc,size)) { /* Is region free */
	/* No it was not */
	iopasm_error(ctxt,first,ERR_MEMORY_REGION_USED,
		     "New PC: %d, check size: %d",
		     pc,size);
      }
      
      cg_set_memory_region(ctxt,pc,size);
      set_inst_addr(first,pc);
      pc += size;
      break;
      

    case INST_LABEL:
      temp_string = get_inst(first);
      size = get_inst_size(first); /* We will need the size if we have FSM code */
      temp_label = create_label(temp_string,0,SCOPE_NO_SCOPE);

      if (temp_label == NULL) {
	iopasm_error(ctxt,first,ERR_REPORT_BUG,
		     "Reported from: %s:%d\n",__FILE__,__LINE__);
      }
      
      real_label = (label_t *) find_array_elem(labels,temp_label);

      if (real_label == NULL) {
	iopasm_error(ctxt,first,ERR_UNDEFINED_LABEL,"Label: %s",temp_string);
      }
      
      /* Change label value and set memory region
       * We do not change labels if it is a state 
       * State labels are set by the state-machine code generetion
       * As we do not know the size of these instructions prior to the 
       * generation of the state */

      delete_label(temp_label);

      if (get_inst_type(get_next_inst(first)) != INST_STATE_BEGIN) {
	set_inst_addr(first,pc);
	change_label_value(real_label,pc);
      }
      
      else { /* Current label is the beginning of a FSM state */
	diff = row_len - pc % row_len;
	diff = (diff < size) ? diff: 0;
	set_inst_addr(first,pc+diff+size);    
	change_label_value(real_label,pc+size+diff); 
	pc += diff;
	
	/* Find end of state */
	next_inst = get_next_inst(first);
	while ((get_inst_type(next_inst) != INST_STATE_END) &&
	       (get_inst_type(next_inst) != INST_LABEL)) {
	  next_inst = get_next_inst(next_inst);
	  if (next_inst == NULL) break;
	}

      } /* End of if (get_inst_type(get_next_inst(first)) ... */
      
      size = get_inst_size(first);

      if (cg_check_memory_region(ctxt,pc,size)) { /* Is region free */
	/* No it was not */
	iopasm_error(ctxt,first,ERR_MEMORY_REGION_USED,
		     "New PC: %d, check size: %d",pc,size);
      }
      
      cg_set_memory_region(ctxt,pc,size);
      /* Update PC */
      pc += size;

      /* Check if we processed a state before continuing */
      if (get_inst_type(get_next_inst(first)) == INST_STATE_BEGIN) {
	first = next_inst;
      }
      break;


    default:
      break;
    } /* End of switch */
    
    first = get_next_inst(first);
  }
}
