/* ************************************************************************* */
/* cg_fe.c: Code-generation frontend and architectural context functions     */
/*                                                                           */
/*    Copyright 2002-2006 Axis Communications AB.                            */
/*                                                                           */
/*                                                                           */
/*    This program is free software; you can redistribute it and/or modify   */
/*    it under the terms of the GNU General Public License as published by   */
/*    the Free Software Foundation; either version 2 of the License, or      */
/*    (at your option) any later version.                                    */
/*                                                                           */
/*    This program is distributed in the hope that it will be useful,        */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*    GNU General Public License for more details.                           */
/*                                                                           */
/*    You should have received a copy of the GNU General Public License      */
/*    along with this program; if not, write to the Free Software            */
/*    Foundation, Inc., 59 Temple Place - Suite 330,                         */
/*    Boston, MA 02111-1307, USA.                                            */
/*                                                                           */
/*                                                                           */
/*    Send bug reports and inquiries to Johan.Jorgensen@axis.com             */
/* ************************************************************************* */



/* ************************************************************************* */
/*                      INCLUDES USED BY THIS MODULE                         */
/* ************************************************************************* */
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "debug.h"
#include "dyn_array.h"
#include "iopasm_global.h"
#include "p_list.h"
#include "labels.h"
#include "etrax_fs_mpu_be.h"
#include "etrax_fs_spu_be.h"
#include "iopasm_err.h"
#include "iopasm_types.h"
#include "meta_object.h"



/* ************************************************************************* */
/*                     DEFINITIONS LOCAL TO THIS MODULE                      */
/* ************************************************************************* */
#define ROTATE_REGISTER1(POS,LEN,REGVAL) (REGVAL << (POS - LEN))
#define I_BUFLEN 4 /* Length of instruction buffer */
#define MAX_STATE_BITS 4
#define FSM_BUF_SIZE 256 /* Size of FSM code generation buffer */
#define MAX_REG_ENTRIES 8 /* Maximum number of register tables */
#define CG_CONTEXT_T_SIZE (sizeof(struct __cg_context_t))
#define GET_MEMORY_SIZE(N) (supported_arch[N].mem_size)
#define GET_BACKEND(N) (supported_arch[N].backend)
#define GET_ARCH_NAME(N) (supported_arch[N].arch_name)
#define GET_MEMORY_DESC(N) (supported_arch[N].mem_description)
#define GET_ROW_LEN(N) (supported_arch[N].row_len)
#define GET_SCOPE_NAME(N) (supported_arch[N].scope_name)



/* ************************************************************************* */
/*                       TYPES DEFINED IN THIS MODULE                        */
/* ************************************************************************* */
typedef struct __opt_t {
  char *opt_name;
  int opt_value;
} opt_t;



struct __cg_context_t {
  int arch;        /* architecture variant (SPU/MPU) */ 
  int seq_boundrary; /* Memory Boundrary for instructions */
  int fsm_boundrary;  /* Memeory boundrary for fsm-instructions */
  int fsm_size;       /* Optimal size for FSM-instructions (in bytes) */
  inst_table_t *inst_table;  /* Table associated with this architecture*/
  int mem_size;       /* Size of memory in bytes */
  int row_len;        /* Length of row in memory */
  int n_mem_banks;      /* Number of memory banks (if banked memory */
  reg_table_t *reg_table; /* Register table pointer */
  char **cpp_macros;      /* Predefined cpp macros for this architecture */
  int cpp_macro_count;    /* Number of cpp macros */
  array_t *labels;
  array_t *sources;       /* Index containing source files for this module */
  int source_index;       /* Next source index available */
  char *scope_name;       /* Scope name (used for c-macro generations) */
  char *memory_map;       /* memory_map - used to keep track of memory */
  unsigned char *code_buffer;      /* Code buffer, used to store generated code */
  char *namespace;        /* Label namespace (if defined) */
  opt_t *machine_options; /* Machine specific options (if any) */
  int n_machine_opts;     /* Number of machine specific options */
  cg_operator_t *be_operators;  /* Operator functions associated with arch */
  cg_backend_t *backend;     /* Backend (i.e. code generator */
};



struct __itable_collections {
  int arch_num;
  const char *arch_name;
  const char *scope_name;
  int mem_size;
  int row_len;
  const char *mem_description;
  const cg_backend_t *backend;
};



struct __cg_backend_t {
  int (*gen_seq_code)(cg_context_t *seq_context,inst_t *inst,int pass,int scope);
  int (*gen_fsm_code)(cg_context_t *context,inst_t *first, inst_t *last,int pass);
  int (*set_tables)(cg_context_t *ctxt);
  unsigned int (*pc_converter_seq)(unsigned int val);
  unsigned int (*pc_converter_fsm)(unsigned int val);
  void (*post_code_generation)(cg_context_t *ctxt,plist_t *a_list);
  plist_t *(*optimize_states)(cg_context_t *ctxt,plist_t *a_list);  
};
	      


/* ************************************************************************* */
/*                              CONSTANT VARIABLES                           */
/* ************************************************************************* */
const cg_backend_t guinness_spu_be = /* Defines the methods for architecure  */
  {                                  /* specific backend operations. This is */
                                     /* the SPU version                      */
    .gen_seq_code = etrax_fs_spu_gen_seq,
    .gen_fsm_code = etrax_fs_spu_gen_fsm,
    .set_tables =  etrax_fs_spu_set_tables,
    .pc_converter_seq = etrax_fs_spu_conv_pc_seq,
    .pc_converter_fsm = etrax_fs_spu_conv_pc_fsm,
    .post_code_generation = etrax_fs_spu_finish,
    .optimize_states = etrax_fs_spu_fsm_optimizer
  };



const cg_backend_t guinness_mpu_be = /* And this is the MPU version          */
  { 
    .gen_seq_code = etrax_fs_mpu_gen_seq,
    .gen_fsm_code = etrax_fs_mpu_gen_fsm,
    .set_tables = etrax_fs_mpu_set_tables,
    .pc_converter_seq = etrax_fs_mpu_conv_pc_seq,
    .pc_converter_fsm = etrax_fs_mpu_conv_pc_fsm,
    .post_code_generation = etrax_fs_mpu_finish,
    .optimize_states = etrax_fs_mpu_fsm_optimizer 
  };



const itable_table supported_arch[] = {            /* Bind backends to an   */
  { ARCH_SPU, "Guinness SPU v1", "SPU_", 4096, 32, /* Architecture-name and */
    "128 256-bit rows", &guinness_spu_be },         /* A memory layout.      */
  { ARCH_MPU, "Guinness MPU v1", "MPU_", 4096,  8, 
    "512 64-bit rows", &guinness_mpu_be },
  { 0, NULL, NULL, 0, 0, NULL, NULL }}; /* Last line _must_ look like this */



/* ************************************************************************* */
/*                    PROTOTYPES FOR STATIC FUNCTIONS                        */
/* ************************************************************************* */



/* ************************************************************************* */
/* PROTOTYPE   : cg_context_t *cg_context_create(int arch,array_t *labels);  */
/*                                                                           */
/* ARGUMENTS   : "arch" is a number specifying an architecture. "labels" is  */
/*               a pointer to an array containing labels for the current     */
/*               context.                                                    */
/*                                                                           */
/* FUNCTION    : Creates a code-generation context and binds architecture    */
/*               to label arrays, memory size and -layout etc.               */
/*                                                                           */
/* RETURN VALUE: A pointer to a valid context if it could be created, NULL   */
/*               if the context could not be created.                        */
/* ************************************************************************* */
cg_context_t *cg_context_create(int arch,array_t *labels)
{
  cg_context_t *temp_context;
  char *memory_map;
  int i;
  temp_context = (cg_context_t *) IOPASM_MALLOC(sizeof(cg_context_t));

  if (temp_context != NULL) {
    temp_context->arch = arch;
    temp_context->fsm_size = 0;
    temp_context->labels = labels;
    temp_context->source_index = 0;
    temp_context->namespace = NULL;
    temp_context->n_mem_banks = 0;
    i = 0;

    while ((supported_arch[i].arch_num != 0) &&
	   (arch != supported_arch[i].arch_num)) i++;
      
    if (supported_arch[i].arch_num == arch) {

      fprintf(stderr,"         Selected architecture: %s\n",GET_ARCH_NAME(i));
      fprintf(stderr,"         Available memory     : ");
      fprintf(stderr,"%d bytes organized as %s\n",
	      GET_MEMORY_SIZE(i),GET_MEMORY_DESC(i));
      
      temp_context->backend = (cg_backend_t *) GET_BACKEND(i);

      temp_context->sources = create_array(str_compare,str_destroy);
      assert(temp_context->sources != NULL);

      memory_map = (char *) IOPASM_MALLOC(sizeof(char) * GET_MEMORY_SIZE(i));
      assert(memory_map != NULL);
      temp_context->scope_name = (char *) GET_SCOPE_NAME(i);
      temp_context->cpp_macros = NULL;
      temp_context->be_operators = NULL;
      temp_context->cpp_macro_count = 0;
      temp_context->mem_size = GET_MEMORY_SIZE(i);
      temp_context->memory_map = memory_map;
      memset(memory_map,0,GET_MEMORY_SIZE(i));
      temp_context->code_buffer = (unsigned char *) IOPASM_MALLOC(sizeof(char) * \
								  GET_MEMORY_SIZE(i));
      temp_context->machine_options = NULL;
      temp_context->n_machine_opts = 0;
      temp_context->row_len = GET_ROW_LEN(i);
      
      if (temp_context->code_buffer == NULL) {
	iopasm_error(temp_context,NULL,ERR_REPORT_BUG,"%s:%d: Memory allocation failed\n",
		   __FILE__,__LINE__);
      }
      
      temp_context->backend->set_tables(temp_context);
      
    }
    
    else { /* Report error here */ 
      iopasm_error(temp_context,NULL,ERR_UNSUPPORTED_ARCH,NULL);
    }
  }
  
  return temp_context;

}



/* ************************************************************************* */
/* PROTOTYPE   : void cg_set_memory_bank_count(cg_context_t *ctxt,           */
/*                                             int n_banks);                 */
/*                                                                           */
/* ARGUMENTS   : "ctxt" points to a code generation context (cg-context.     */
/*               "n_banks" is the number of memory banks in selected arch.   */
/*                                                                           */
/* FUNCTION    : Sets the number of memory banks available                   */
/*                                                                           */
/* RETURN VALUE: None (void)                                                 */
/* ************************************************************************* */
void cg_set_memory_bank_count(cg_context_t *ctxt,int n_banks) 
{
  assert((ctxt != NULL) && (n_banks>0));
  ctxt->n_mem_banks = n_banks;
}



/* ************************************************************************* */
/* PROTOTYPE   : int cg_get_memory_bank_count(cg_context_t *ctxt);           */
/*                                                                           */
/* ARGUMENTS   : "ctxt" points to a code generation context (cg-context).    */
/*                                                                           */
/* FUNCTION    : Return the number of banks associated with the current arch.*/
/*                                                                           */
/* RETURN VALUE: Number of memory banks associated with arch                 */
/* ************************************************************************* */
int cg_get_memory_bank_count(cg_context_t *ctxt)
{
  assert(ctxt != NULL);
  return ctxt->n_mem_banks;
}



/* ************************************************************************* */
/* PROTOTYPE   : int cg_set_machine_option(cg_context_t ctxt,                */
/*                                         char machine_opt,int value);      */
/*                                                                           */
/* ARGUMENTS   : "ctxt" points to a code-generation context (cg-context).    */
/*               "machine_opt" is the textual representation of the machine  */
/*               specific option to set and "value" is an integer value.     */
/*                                                                           */
/* FUNCTION    : Sets the machine specific option pointed to by              */
/*               "machine_opt" to "value" in the cg-context pointed out by   */
/*               "ctxt".                                                     */
/*                                                                           */
/* RETURN VALUE: -1 if option cannot be found. Returns 0 if option was found */
/* ************************************************************************* */
int cg_set_machine_option(cg_context_t *ctxt,char *machine_opt,int value) 
{
  int i = 0;
  int ret_val = -1;

  assert ((ctxt != NULL) && (machine_opt != NULL));

  while(ctxt->machine_options[i].opt_name != NULL) {

    if (!strcmp(ctxt->machine_options[i].opt_name,machine_opt)) { 
      /* We found the option we want to change - change it */
      ctxt->machine_options[i].opt_value = value;
      ret_val = 0;
      i++;
    } /* End of if (!strcmp(ctxt->machine_options[i].opt_name,machine_opt)) */

    else { 
      i++;
    } /* End of !if (!strcmp(ctxt->machine_options[i].opt_name,machine_opt)) */
  } /* End of while(ctxt->machine_options[i].opt_name != NULL) */

  return ret_val;
}


/* ************************************************************************* */
/* PROTOTYPE   : int cg_check_machine_option(cg_context_t *ctxt,             */
/*                                           char *machine_opt,int wanted);  */
/*                                                                           */
/* ARGUMENTS   : "ctxt" is a valid pointer to a context. "machine_opt" is    */
/*               a pointer to the machine option which value should checked  */
/*               and "wanted" is the value the option is expected to have.   */
/*                                                                           */
/* FUNCTION    : Check whether or not the machine specific option            */
/*               "machine_opt" is set to the value "wanted" or not.          */
/*                                                                           */
/* RETURN VALUE: Returns 1 if the machine option "machine_opt" is set to     */
/*               the value indicated by "wanted". Otherwise it returns Zero. */
/* ************************************************************************* */
int cg_check_machine_option(cg_context_t *ctxt,char *machine_opt,
			    int wanted)
{
  
  int i=0;
  assert ((ctxt != NULL) && (machine_opt != NULL));
  
  while (ctxt->machine_options[i].opt_name != NULL) {
    if (!strcmp(ctxt->machine_options[i].opt_name,machine_opt)) { /* got it */
      return (wanted == ctxt->machine_options[i].opt_value);
    } /* end of if (!strcmp(.... */

    else {
      i++;
    } /* end of !if (!strcmp(.... */
  } /* End of while (ctxt->machine_options[i].opt_name != NULL) */
  
  return 0;
  
}




/* ************************************************************************* */
/* PROTOTYPE   : int cg_get_machine_option(cg_context_t *ctxt,               */
/*                                         char *machine_opt,int *err);      */
/*                                                                           */
/* ARGUMENTS   : "ctxt" is a valid pointer to a context. "machine_opt" is    */
/*               a pointer to the machine option which value should be       */
/*               and "err" is a pointer to an integer where an appropriately */
/*               encoded value used for error-indication is stored.          */
/*                                                                           */
/* FUNCTION    : Returns the value of the machine-option pointed to by       */
/*               "machine_opt".                                              */
/*                                                                           */
/* RETURN VALUE: Value associated with "machine_opt"                         */
/* ************************************************************************* */
int cg_get_machine_option(cg_context_t *ctxt,char *machine_opt,int *err)
{
  int i=0;
  
  assert ((ctxt != NULL) && (machine_opt != NULL) && (err != NULL));
  *err = -1;
  
  while (ctxt->machine_options[i].opt_name != NULL) {
    if (!strcmp(ctxt->machine_options[i].opt_name,machine_opt)) { /* got it */
      *err = 0;
      return ctxt->machine_options[i].opt_value;
    } /* end of if (!strcmp(.... */

    else {
      *err = -1;
      i++;
    } /* end of !if (!strcmp(.... */
  } /* End of while (ctxt->machine_options[i].opt_name != NULL) */

  return 0;
}



/* ************************************************************************* */
/* PROTOTYPE   : void cg_init_machine_options(cg_context_t *ctxt,            */
/*                                            char **options,int *opt_vals,  */
/* 			                      int n_opts);                   */
/*                                                                           */
/* ARGUMENTS   : "ctxt" is a valid context, "options" is an array of textual */
/*               machine-dependent options and "opt_vals" is the associated  */
/*               values. "n_opts"                                            */
/*                                                                           */
/* FUNCTION    : Initiate the machine dependent options for a certain        */
/*               architecture.                                               */
/*                                                                           */
/* RETURN VALUE: None                                                        */
/* ************************************************************************* */
void cg_init_machine_options(cg_context_t *ctxt,char **options,int *opt_vals,
			     int n_opts)
{
  opt_t *tmp_options;
  int i = 0;
  assert(ctxt !=NULL);
  
  if (n_opts > 0) {  /* We need to do some work to set up the options */
    tmp_options = (opt_t *) IOPASM_MALLOC(sizeof(opt_t) * n_opts);
    assert(tmp_options != NULL);

    for (i=0;i<n_opts;i++) {
      tmp_options[i].opt_name = NULL;
      tmp_options[i].opt_value = 0;
    }
    i = 0;
    while(options[i] != NULL) { /* Assign and set options */
      tmp_options[i].opt_name = strdup(options[i]);
      assert(tmp_options[i].opt_name != NULL);
      tmp_options[i].opt_value = opt_vals[i];
      i++;
    } /* End of while(i < n_opts) */
    
    ctxt->machine_options = tmp_options;
    ctxt->n_machine_opts = n_opts;
  } /* End of if (n_opts > 0) */


  else { /* Just set options to NULL and continue */
    ctxt->machine_options = NULL;
    ctxt->n_machine_opts = 0;
  } /* End of if (n_opts > 0) */
}



/* ************************************************************************* */
/* PROTOTYPE   : int cg_gen_fsm_code(cg_context_t *ctxt,inst_t *first,       */
/*                                inst_t *last)                              */
/*                                                                           */
/* ARGUMENTS   : "ctxt" points to a code-generation context (cg-context).    */
/*               "first" points to the definition of the state and "last"    */
/*               points to the end-of-state meta-instruction.                */
/*                                                                           */
/* FUNCTION    : Wrapper around call to backend in order to hide             */
/*               architecture specific functions within cg-context.          */
/*                                                                           */
/* RETURN VALUE: Size of code generated or 0 if code could not be generated  */
/* ************************************************************************* */
int cg_gen_fsm_code(cg_context_t *ctxt,inst_t *first,inst_t *last,int pass)
{
  assert (ctxt != NULL);
  return ctxt->backend->gen_fsm_code(ctxt,first,last,pass);
}



/* ************************************************************************* */
/* PROTOTYPE   : cg_get_row_len(cg_context_t *ctxt);                         */
/*                                                                           */
/* ARGUMENTS   : "ctxt" points to a code-generation context.                 */
/*                                                                           */
/* FUNCTION    : Returns the size of the current memory-row in bytes         */
/*                                                                           */
/* RETURN VALUE: Size of the current memory row in bytes                     */
/* ************************************************************************* */
int cg_get_row_len(cg_context_t *ctxt)
{
  assert(ctxt != NULL);
  return ctxt->row_len;
}



/* ************************************************************************* */
/* PROTOTYPE   : int cg_gen_seq_code(cg_context_t *ctxt,inst_t *inst);       */
/*                                                                           */
/* ARGUMENTS   : "ctxt" is a pointer to a cg-context. "inst" points a        */
/*               sequential instruction.                                     */
/*                                                                           */
/* FUNCTION    : Wrapper around call to backend in order to hide             */
/*               architecture  specific functions from architecture indepen- */
/*               dent parts of the assembler.                                */
/*                                                                           */
/* RETURN VALUE: Size of generated code in bytes 0 on error.                 */
/* ************************************************************************* */
int cg_gen_seq_code(cg_context_t *ctxt,inst_t *inst,int pass,int scope)
{
  assert (ctxt != NULL);
  return ctxt->backend->gen_seq_code(ctxt,inst,pass,scope);
}



/* ************************************************************************* */
/* PROTOTYPE   : void post_code_generation(cg_context_t *ctxt,               */
/*                                         plist_t *a_list);                 */
/*                                                                           */
/* ARGUMENTS   : "ctxt" is a pointer to a context. "a_list" is a pointer to  */
/*               a parser-list                                               */
/*                                                                           */
/* FUNCTION    : This function is called after all code has been generated.  */
/*               Allows the backend to perform transformations on            */
/*               the generated code. Currently used to (re)order the code    */
/*               prior to saving it.                                         */
/*                                                                           */
/* RETURN VALUE: None. The backend must take care of any error-handling.     */
/* ************************************************************************* */
void cg_post_code_generation(cg_context_t *ctxt,plist_t *a_list)
{
  assert((ctxt != NULL) && (a_list != NULL));
  ctxt->backend->post_code_generation(ctxt,a_list);
}



/* ************************************************************************* */
/* PROTOTYPE   : plist_t *cg_optimize_fsm_states(cg_context_t *ctxt,         */
/*                                               plist_t *a_list);           */
/*                                                                           */
/* ARGUMENTS   : "ctxt" points to a valid context. "a_list" points to a      */
/*               a valid parse list                                          */
/*                                                                           */
/* FUNCTION    : Optimizes the */
/*                                                                           */
/* RETURN VALUE: */
/* ************************************************************************* */
plist_t *cg_optimize_fsm_states(cg_context_t *ctxt,plist_t *a_list)
{
    assert((ctxt != NULL) && (a_list != NULL));
    return ctxt->backend->optimize_states(ctxt,a_list);
}



/* ************************************************************************* */
/* PROTOTYPE   : unsigned int cg_convert_seq_addr(cg_context_t *ctxt,        */
/*                                                unsigned int pc);          */
/*                                                                           */
/* ARGUMENTS   : "ctxt" points to a code generation context. "pc" is a value */
/*               to convert.                                                 */
/*                                                                           */
/* FUNCTION    : Internally the assembler stores all addr-references as byte */
/*               addresses. This function is a wrapper around the a backend  */
/*               function which converts a byte-address as required by the   */
/*               selected architecture. This is the version used for SEQ     */
/*               code.                                                       */
/*                                                                           */
/* RETURN VALUE: The converted value for PC                                  */
/* ************************************************************************* */
unsigned int cg_convert_seq_addr(cg_context_t *ctxt,unsigned int pc)
{
  assert (ctxt != NULL);
  return ctxt->backend->pc_converter_seq(pc);
}



/* ************************************************************************* */
/* PROTOTYPE   : unsigned int cg_convert_fsm_addr(cg_context_t *ctxt,        */
/*                                          unsigned int pc);                */
/*                                                                           */
/* ARGUMENTS   : "ctxt" points to a code generation context. "pc" is a value */
/*               to convert.                                                 */
/*                                                                           */
/* FUNCTION    : Internally the assembler stores all addr-references as byte */
/*               addresses. This function is a wrapper around the a backend  */
/*               function which converts a byte-address as required by the   */
/*               selected architecture. This is the version used for FSM     */
/*               code.                                                       */
/*                                                                           */
/* RETURN VALUE: The converted value for PC                                  */
/* ************************************************************************* */
unsigned int cg_convert_fsm_addr(cg_context_t *ctxt,unsigned int pc)
{
  assert (ctxt != NULL);
  return ctxt->backend->pc_converter_fsm(pc);
}



/* ************************************************************************* */
/* PROTOTYPE   : reg_table_t *cg_get_reg_table(cg_context_t *ctxt);          */
/*                                                                           */
/* ARGUMENTS   : "ctxt" is a pointer to a valid code-generation context.     */
/*                                                                           */
/* FUNCTION    : Returns the register-table for the selected architecture    */
/*                                                                           */
/* RETURN VALUE: The register table associated with the cg-context.          */
/* ************************************************************************* */
reg_table_t *cg_get_reg_table(cg_context_t *ctxt)
{
  assert(ctxt != NULL);
  return ctxt->reg_table;
}



/* ************************************************************************* */
/* PROTOTYPE   : void cg_set_inst_table(cg_context_t *ctxt,                  */
/*                                      const inst_table_t *inst_table)      */
/*                                                                           */
/* ARGUMENTS   : "ctxt" is a pointer to a context. "inst_table" is a pointer */
/*               to an instruction table.                                    */
/*                                                                           */
/* FUNCTION    : Set the instruction table in the current context. Used by   */
/*               backends to set instruction table.                          */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void cg_set_inst_table(cg_context_t *ctxt,const inst_table_t *inst_table)
{
  assert((ctxt != NULL) && (inst_table != NULL));
  ctxt->inst_table = (inst_table_t *) inst_table;
}



/* ************************************************************************* */
/* PROTOTYPE   : void cg_set_reg_table(cg_context_t *ctxt,                   */
/*                                     const reg_table_t *reg_table);        */
/*                                                                           */
/* ARGUMENTS   : "ctxt" is a pointer to a context. "reg_table" points to the */
/*               register table.                                             */
/*                                                                           */
/* FUNCTION    : Set the register table in the current context. Used by      */
/*               backends to set the register table.                         */
/*                                                                           */
/* RETURN VALUE: None                                                        */
/* ************************************************************************* */
void cg_set_reg_table(cg_context_t *ctxt,const reg_table_t *reg_table)
{
  assert ((ctxt != NULL) && (reg_table != NULL));
  ctxt->reg_table = (reg_table_t *)reg_table;
}



/* ************************************************************************* */
/* PROTOTYPE   : inst_table_t *cg_get_inst_table(cg_context_t *ctxt);        */
/*                                                                           */
/* ARGUMENTS   : "ctxt" points to a valid context.                           */
/*                                                                           */
/* FUNCTION    : returns a pointer to the instruction table associated with  */
/*               the current architecture.                                   */
/*                                                                           */
/* RETURN VALUE: A pointer to the current instruction table. NULL if un-     */
/*               defined.                                                    */
/* ************************************************************************* */
inst_table_t *cg_get_inst_table(cg_context_t *ctxt)
{
  assert(ctxt != NULL);
  return ctxt->inst_table;
}



/* ************************************************************************* */
/* PROTOTYPE   : array_t *cg_get_label_array(cg_context_t *ctxt);            */
/*                                                                           */
/* ARGUMENTS   : "ctxt" is pointer to a context.                             */
/*                                                                           */
/* FUNCTION    : Returns the label array associated with the current context */
/*                                                                           */
/* RETURN VALUE: The pointer to the label array. If not set NULL is returned */
/* ************************************************************************* */
array_t *cg_get_label_array(cg_context_t *ctxt)
{
  assert (ctxt != NULL);
  return ctxt->labels;
}



/* ************************************************************************* */
/* PROTOTYPE   : void cg_clear_memory_region(cg_context_t *ctxt,int pc,      */
/*                                           int size);                      */
/*                                                                           */
/* ARGUMENTS   : "ctxt" is a valid pointer to a context, "pc" is the start   */
/*               address and "size" is the number of bytes to clear          */
/*                                                                           */
/* FUNCTION    : Removes the marking of a memory region from pc and size     */
/*               bytes forward.                                              */
/*                                                                           */
/* RETURN VALUE: None                                                        */
/* ************************************************************************* */
void cg_clear_memory_region(cg_context_t *ctxt,int pc,int size)
{
  int i;
  assert(ctxt != NULL);
  
  for(i=pc;i<(pc + size);i++) {
    ctxt->memory_map[i] = 0;
  } /* End of for(i=pc;i<(pc + size);i++) */
}



/* ************************************************************************* */
/* PROTOTYPE   : int cg_check_memory_region(cg_context_t *ctxt,int pc,       */
/*                                       int size);                          */
/*                                                                           */
/* ARGUMENTS   : "ctxt" is a pointer to a context. "pc" is the value of the  */
/*               PC at which to begin the check. "size" is the size of the   */
/*               region to be checked measured in bytes.                     */
/*                                                                           */
/* FUNCTION    : Checks the memory region an reports the number of occupied  */
/*               bytes in the region.                                        */
/*                                                                           */
/* RETURN VALUE: The number of bytes occupied (i.e) the overlap measured in  */
/*               bytes.                                                      */
/* ************************************************************************* */
int cg_check_memory_region(cg_context_t *ctxt,int pc, int size)
{
  int tainted = 0;
  int i;
  
  assert(ctxt != NULL);

  if (pc > ctxt->mem_size) {
    fprintf(stderr,"Memory check failed!");
    exit(1);
  }

  for (i=pc;i<pc+size;i++) {

    if (ctxt->memory_map[i]) {
      tainted++;
    }

    else {
      ;
    }

  }
  
  return tainted;
}



/* ************************************************************************* */
/* PROTOTYPE   : int cg_set_memory_region(cg_context_t *ctxt,int pc,         */
/*                                        int size);                         */
/*                                                                           */
/* ARGUMENTS   : "ctxt" is a pointer to a context. "pc" is the start point   */
/*               in memory at which to start marking. "size" is the length   */
/*               (measured in bytes) to mark as occupied.                    */
/*                                                                           */
/* FUNCTION    : Marks a region of "size" bytes starting at "pc" as occupied */
/*                                                                           */
/* RETURN VALUE: Number of bytes marked as occupied.                         */
/* ************************************************************************* */
int cg_set_memory_region(cg_context_t *ctxt,int pc,int size)
{
  int i;

  assert(ctxt != NULL);

  for (i=pc;i<pc+size;i++) {
    ctxt->memory_map[i] = '1';
  }
  return size;
}



/* ************************************************************************* */
/* PROTOTYPE   : void cg_set_seq_boundrary(cg_context_t *ctxt,               */
/*                                         int boundrary);                   */
/*                                                                           */
/* ARGUMENTS   : "ctxt" is a context. "boundrary is the boundrary value in   */
/*               bytes.                                                      */
/*                                                                           */
/* FUNCTION    : Sets the boundrary of a sequential instruction to           */
/*               "boundrary" for the current architecture.                   */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void cg_set_seq_boundrary(cg_context_t *ctxt,int boundrary)
{
  assert(ctxt != NULL);
  ctxt->seq_boundrary = boundrary;
}



/* ************************************************************************* */
/* PROTOTYPE   : int cg_get_seq_boundrary(cg_context_t *ctxt);               */
/*                                                                           */
/* ARGUMENTS   : "ctxt" points to a code-generation context.                 */
/*                                                                           */
/* FUNCTION    : Returns the boundrary measured in bytes for the current     */
/*               architecture.                                               */
/*                                                                           */
/* RETURN VALUE: Boundrary of sequential instructions measured in bytes      */
/* ************************************************************************* */
int cg_get_seq_boundrary(cg_context_t *ctxt)
{
  assert(ctxt != NULL);
  return ctxt->seq_boundrary;
}



/* ************************************************************************* */
/* PROTOTYPE   : void cg_set_fsm_boundrary(cg_context_t *ctxt,               */
/*                                         int boundrary);                   */
/*                                                                           */
/* ARGUMENTS   : "ctxt" points to a valid context. "boundrary" is the new    */
/*               boundrary measured in bytes.                                */
/*                                                                           */
/* FUNCTION    : Set the boundrary of the FSM-instructions for the current   */
/*               architecture.                                               */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void cg_set_fsm_boundrary(cg_context_t *ctxt,int boundrary)
{
  assert(ctxt != NULL);
  ctxt->fsm_boundrary = boundrary;
}



/* ************************************************************************* */
/* PROTOTYPE   : int cg_get_fsm_boundrary(cg_context_t *ctxt);               */
/*                                                                           */
/* ARGUMENTS   : A valid context pointed to by "ctxt"                        */
/*                                                                           */
/* FUNCTION    : Returns the boundrary-address (byte address) for FSM        */
/*               instructions                                                */
/*                                                                           */
/* RETURN VALUE: FSM boundrary address as a byte address                     */
/* ************************************************************************* */
int cg_get_fsm_boundrary(cg_context_t *ctxt)
{
  assert(ctxt != NULL);
  return ctxt->fsm_boundrary;
}



/* ************************************************************************* */
/* PROTOTYPE   : void cg_set_fsm_size(cg_context_t *ctxt,int size);          */
/*                                                                           */
/* ARGUMENTS   : "ctxt" points to a valid context. size is the new maximum   */
/*                                                                           */
/* FUNCTION    : Sets the maximum size of an FSM instruction (in bytes) to   */
/*               "size".                                                     */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void cg_set_fsm_size(cg_context_t *ctxt,int size)
{
  assert(ctxt != NULL);
  ctxt->fsm_size = size;
}



/* ************************************************************************* */
/* PROTOTYPE   : int cg_get_fsm_size(cg_context_t *ctxt);                    */
/*                                                                           */
/* ARGUMENTS   : "ctxt" points to a valid context.                           */
/*                                                                           */
/* FUNCTION    : Returns the maximum size of an FSM instrctuion (in bytes)   */
/*                                                                           */
/* RETURN VALUE: Maximum size of an FSM instruction.                         */
/* ************************************************************************* */
int cg_get_fsm_size(cg_context_t *ctxt)
{
  assert(ctxt != NULL);
  return ctxt->fsm_size;
}



/* ************************************************************************* */
/* PROTOTYPE   : int cg_get_memory_size(cg_context_t *ctxt);                 */
/*                                                                           */
/* ARGUMENTS   : "ctxt" points to a valid context.                           */
/*                                                                           */
/* FUNCTION    : Returns the size (maximum byte-address) of the current      */
/*               context. (architecture dependent).                          */
/*                                                                           */
/* RETURN VALUE: The maximum byte address of the current arch memory         */
/* ************************************************************************* */
int cg_get_memory_size(cg_context_t *ctxt)
{
  assert(ctxt != NULL);
  return ctxt->mem_size;
}



/* ************************************************************************* */
/* PROTOTYPE   : void cg_set_code_buffer(cg_context_t *ctxt,                 */
/*                                    unsigned char *code_buffer);           */
/*                                                                           */
/* ARGUMENTS   : "ctxt" is a valid context. "code_buffer" points to the code */
/*               -buffer to be associated with the current context.          */
/*                                                                           */
/* FUNCTION    : Associates a code-buffer (storage area) with the current    */
/*               context.                                                    */
/*                                                                           */
/* RETURN VALUE: None                                                        */
/* ************************************************************************* */
void cg_set_code_buffer(cg_context_t *ctxt,unsigned char *code_buffer)
{
  assert(ctxt != NULL);
  ctxt->code_buffer = code_buffer;
}



/* ************************************************************************* */
/* PROTOTYPE   : unsigned char *cg_get_code_buffer(cg_context_t *ctxt);      */
/*                                                                           */
/* ARGUMENTS   : "ctxt" points to a valid context                            */
/*                                                                           */
/* FUNCTION    : Returns the code-buffer which is used to store the gene-    */
/*               rated code.                                                 */
/*                                                                           */
/* RETURN VALUE: A pointer to the code-buffer.                               */
/* ************************************************************************* */
unsigned char *cg_get_code_buffer(cg_context_t *ctxt)
{
  assert(ctxt != NULL);
  return ctxt->code_buffer;
}



/* ************************************************************************* */
/* PROTOTYPE   : void cg_delete(cg_context_t *ctxt);                         */
/*                                                                           */
/* ARGUMENTS   : "ctx" points to a valid context.                            */
/*                                                                           */
/* FUNCTION    : Delete the current context and free any memory implicitly   */
/*               allocated by context.                                       */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void cg_delete(cg_context_t *ctxt)
{
  assert(ctxt != NULL);

  delete_array(ctxt->sources);
  IOPASM_FREE(ctxt->memory_map);
  IOPASM_FREE(ctxt->code_buffer);
  IOPASM_FREE(ctxt);
  ctxt = NULL;
}



/* ************************************************************************* */
/* PROTOTYPE   : int cg_get_arch(cg_context_t *ctxt);                        */
/*                                                                           */
/* ARGUMENTS   : "ctxt" points to a valid context                            */
/*                                                                           */
/* FUNCTION    : return architecture number                                  */
/*                                                                           */
/* RETURN VALUE: Number associated with current architecture                 */
/* ************************************************************************* */
int cg_get_arch(cg_context_t *ctxt) 
{
  assert (ctxt != NULL);
  return ctxt->arch;
}
    


/* ************************************************************************* */
/* PROTOTYPE   : char *cg_get_scope_name(cg_context_t *ctxt);                */
/*                                                                           */
/* ARGUMENTS   : "ctxt" points to valid context                              */
/*                                                                           */
/* FUNCTION    : Returns the scope-name associated with the current archi-   */
/*               tecture.                                                    */
/*                                                                           */
/* RETURN VALUE: Pointer to a NULL-terminated string that matches the scope  */
/*               for this arch. (Used by in C macro generation)              */
/* ************************************************************************* */
char *cg_get_scope_name(cg_context_t *ctxt)
{
  assert(ctxt != NULL); 
  return ctxt->scope_name;
}


	
/* ************************************************************************* */
/* PROTOTYPE   :  char **cg_get_cpp_macros(cg_context_t *ctxt);              */
/*                                                                           */
/* ARGUMENTS   : "ctxt" points to a valid context.                           */
/*                                                                           */
/* FUNCTION    : Returns the pointer to a null terminated array containing   */
/*               architecture specific macros for the C pre-processor.       */
/*                                                                           */
/* RETURN VALUE: A null-terminated array of pointers.                        */
/* ************************************************************************* */
char **cg_get_cpp_macros(cg_context_t *ctxt) 
{
  assert(ctxt != NULL);
  return ctxt->cpp_macros;

}



/* ************************************************************************* */
/* PROTOTYPE   : void cg_set_cpp_macros(cg_context_t *ctxt,char **macros);   */
/*                                                                           */
/* ARGUMENTS   : "ctxt" points to a valid context. "macros" is a NULL-       */
/*               terminated array of pointers to strings defining the        */
/*               architecture specific macros.                               */
/*                                                                           */
/* FUNCTION    : Sets the architecture specific macros for the current       */
/*               architecture (used by backends).                            */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void cg_set_cpp_macros(cg_context_t *ctxt,char **macros)
{
  int i = 0;
  assert((ctxt != NULL) && (macros != NULL));

  while(macros[i] != NULL) { 
    i++; 
  };
  
  ctxt->cpp_macros = macros;
  ctxt->cpp_macro_count = i;
}



/* ************************************************************************* */
/* PROTOTYPE   : int cg_get_cpp_macro_count(cg_context_t *ctxt);             */
/*                                                                           */
/* ARGUMENTS   : "ctxt" is a valid context.                                  */
/*                                                                           */
/* FUNCTION    : Return the number of architecture-specific macros defined   */
/*               for this architecture                                       */
/* RETURN VALUE: Number of macros defined.                                   */
/* ************************************************************************* */
int cg_get_cpp_macro_count(cg_context_t *ctxt)
{
  assert (ctxt != NULL);
  return ctxt->cpp_macro_count;
}



/* ************************************************************************* */
/* PROTOTYPE   : int cg_get_file_index_count(cg_context_t *ctxt);            */
/*                                                                           */
/* ARGUMENTS   : ctxt is a valid context                                     */
/*                                                                           */
/* FUNCTION    : Return the number of source-files this context knows about. */
/*               (Multiple sourcefiles would just be found in objects)       */
/*                                                                           */
/* RETURN VALUE: number of sourcefiles registered with context.              */
/* ************************************************************************* */
int cg_get_file_index_count(cg_context_t *ctxt)
{
  assert(ctxt != NULL);
  return ctxt->source_index;
}



/* ************************************************************************* */
/* PROTOTYPE   : char *cg_get_file_by_index(cg_context_t *ctxt,int index);   */
/*                                                                           */
/* ARGUMENTS   : ctxt is a valid context and index is the index  to return   */
/*                                                                           */
/* FUNCTION    : Returns a pointer to the name of the file with index        */
/*                                                                           */
/* RETURN VALUE: String pointer or NULL if index does not exist              */
/* ************************************************************************* */
char *cg_get_file_by_index(cg_context_t *ctxt,int indx)
{
  assert(ctxt != NULL);
  if (ctxt->sources != NULL) {
    return (char *) get_array_elem(ctxt->sources,indx);
  }
  else {
    return NULL;
  }
}



/* ************************************************************************* */
/* PROTOTYPE   : char *cg_get_file_by_name(cg_context_t *ctxt,char *name);   */
/*                                                                           */
/* ARGUMENTS   : ctxt is a valid context and index is the index  to return   */
/*                                                                           */
/* FUNCTION    : Returns a pointer to the name of the file with index        */
/*                                                                           */
/* RETURN VALUE: String pointer or NULL if index does not exist              */
/* ************************************************************************* */
int cg_get_file_index_by_name(cg_context_t *ctxt,char *name)
{
  assert(ctxt != NULL);
  return find_array_elem_pos(ctxt->sources,name);
}



/* ************************************************************************* */
/* PROTOTYPE   : int cg_index_file(cg_context_t *ctxt,char *file_name);      */
/*                                                                           */
/* ARGUMENTS   : "ctxt" is a valid context and "file_name" is a valid        */
/*               filename                                                    */
/*                                                                           */
/* FUNCTION    : Register file with context.                                 */
/*                                                                           */
/* RETURN VALUE: Index number for the registered file.                       */
/* ************************************************************************* */
int cg_index_file(cg_context_t *ctxt,char *file_name)
{
  int i;
  void *elem;
  char *fn_copy;
  assert((ctxt != NULL) && (file_name != NULL));
  
  if (find_array_elem(ctxt->sources,(void *)file_name) != NULL) {
    i = 0;

    while((elem = get_array_elem(ctxt->sources,i)) != NULL) {
      if (!strcmp(elem,file_name)) {
	return i;
      }
      i++;
    } /* Matches while((elem = get_array_elem(ctxt->sources,i)) != NULL) */
  }
  
  else {
    fn_copy = (char *) strdup(file_name);
    assert(fn_copy != NULL);
    add_array_elem(ctxt->sources,(void *) fn_copy);
    return ctxt->source_index++;
  }
  return ctxt->source_index;
  
}



/* ************************************************************************* */
/* PROTOTYPE   : char *cg_get_namespace(cg_context_t *ctxt);                 */
/*                                                                           */
/* ARGUMENTS   : A pointer to a valid context.                               */
/*                                                                           */
/* FUNCTION    : Return the namespace associated with the current context    */
/*                                                                           */
/* RETURN VALUE: Pointer to the current namespace, NULL if no namespace has  */
/*               been set.                                                   */
/* ************************************************************************* */
char *cg_get_namespace(cg_context_t *ctxt)
{
  assert(ctxt != NULL);
  
  return ctxt->namespace;
}



/* ************************************************************************* */
/* PROTOTYPE   : void cg_set_namespace(cg_context_t *ctxt,char *namespace);  */
/*                                                                           */
/* ARGUMENTS   : 'ctxt' points to a valid context. "namespace" points to the */
/*               namespace wanted.                                           */
/*                                                                           */
/* FUNCTION    : Set the namespace for the current context.                  */
/*                                                                           */
/* RETURN VALUE: None                                                        */
/* ************************************************************************* */
void cg_set_namespace(cg_context_t *ctxt,char *namespace)
{
  
  assert((ctxt != NULL) && (namespace != NULL));
  
  ctxt->namespace = namespace;
}



/* ************************************************************************* */
/* PROTOTYPE   : void cg_set_be_operators(cg_context_t *context,             */
/*                                        const cg_operator_t *operators);   */
/*                                                                           */
/* ARGUMENTS   : "context" points to a valid context. "operators" points to  */
/*               the array of operators.                                     */
/*                                                                           */
/* FUNCTION    : */
/*                                                                           */
/* RETURN VALUE: */
/* ************************************************************************* */
void cg_set_be_operators(cg_context_t *context,const cg_operator_t *operators)
{
  assert(context != NULL);
  context->be_operators = (cg_operator_t *) operators;
}

  

/* ************************************************************************* */
/* PROTOTYPE   : cg_operator_t *cg_get_be_operators(cg_context_t *context);  */
/*                                                                           */
/* ARGUMENTS   : "context" points to a valid code generation context.        */
/*                                                                           */
/* FUNCTION    : Return the array of operator functions associated with a    */
/*               certain backend.                                            */
/*                                                                           */
/* RETURN VALUE: Pointer to the array of operators associated with a         */
/*               back-end. The function is deprecated/not used...            */
/* ************************************************************************* */
cg_operator_t *cg_get_be_operators(cg_context_t *context)
{
  assert(context != NULL);
  return context->be_operators;
}



/* ************************************************************************* */
/* PROTOTYPE   : int cg_eval_operator(cg_context_t *context,char *operator,  */
/*                                    int *meta_data,int meta_len,           */
/*                                    int *buffer,int buffer_len,            */
/*                                    inst_t *inst,int inum,arg_t *arg1,     */
/*                                    arg_t *arg2,int arg_no);               */
/*                                                                           */
/* ARGUMENTS   : "context" points to a valid context. "operator" is the      */
/*               operator to evaluate. "meta_data" is an integer used by the */
/*               operator evaluation function to store data in between calls.*/
/*               meta_len is the length of the meta_data. "buffer" and       */
/*               "buffer_len" are the buffer in which to store code and the  */
/*               length of that buffer respectively. "inst" is the           */
/*               instruction and "inum" is the number of the instrction in   */
/*               the backends instruction table. "arg1" and "arg2" are the   */
/*               operands associated with the operator and "arg_no" is the   */
/*               argument number as seen from the instruction.               */
/*                                                                           */
/* FUNCTION    : Evaluates an expression in the backend and generates code   */
/*               for that expression.                                        */
/*                                                                           */
/* RETURN VALUE: If the operator can be evaluated the corresponding value is */
/*               returned. If the operator is not defined. ILLEGAL_OPERATOR  */
/*               is returned.                                                */
/* ************************************************************************* */
int cg_eval_operator(cg_context_t *context,char *operator, int *meta_data,
		     int meta_len,int *buffer,int buffer_len,inst_t *inst,
		     int inum,arg_t *arg1,arg_t *arg2,int arg_no)
{
  int i = 0;
  cg_operator_t *op_table;
  assert ((context != NULL)  && (operator != NULL) && (meta_data != NULL) &
	  (buffer != NULL) && (arg1 != NULL));

  op_table = context->be_operators;
  while(op_table[i].operator != NULL) {

    if (strcmp(op_table[i].operator,operator) == 0) {

      if (op_table[i].operator_func == NULL) {
	return ILLEGAL_OPERATOR;
      } /* op_table[i].operator_func == NULL) { */
      
      else { 
	return op_table[i].operator_func(context,meta_data,meta_len,buffer,
					 buffer_len,inst,inum,arg1,arg2,
					 arg_no);
	  }
    } /* End of (if (strcmp(_op_table[i],operator) == 0) { */
    i++;

  } /* End of while(op_table[i].operator != NULL) */
  
    return ILLEGAL_OPERATOR;
}
