/* ************************************************************************* */
/* code_generator.c: Implementation of architecture independent code-        */
/*                   generation front end.                                   */
/*                                                                           */
/*    Copyright 2002-2006 Axis Communications AB.                            */
/*                                                                           */
/*                                                                           */
/*    This program is free software; you can redistribute it and/or modify   */
/*    it under the terms of the GNU General Public License as published by   */
/*    the Free Software Foundation; either version 2 of the License, or      */
/*    (at your option) any later version.                                    */
/*                                                                           */
/*    This program is distributed in the hope that it will be useful,        */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*    GNU General Public License for more details.                           */
/*                                                                           */
/*    You should have received a copy of the GNU General Public License      */
/*    along with this program; if not, write to the Free Software            */
/*    Foundation, Inc., 59 Temple Place - Suite 330,                         */
/*    Boston, MA 02111-1307, USA.                                            */
/*                                                                           */
/*                                                                           */
/*    Send bug reports and inquiries to Johan.Jorgensen@axis.com             */
/* ************************************************************************* */



/* ************************************************************************* */
/*                      INCLUDES USED BY THIS MODULE                         */
/* ************************************************************************* */
#include <assert.h>
#include <netinet/in.h>
#include <stdlib.h>
#include "cg_fe.h"
#include "dyn_array.h"
#include "labels.h"
#include "p_list.h"
#include "iopasm_err.h"
#include "iopasm_global.h"
#include "debug.h"
#include "stack_eval.h"



/* ************************************************************************* */
/*                     DEFINITIONS LOCAL TO THIS MODULE                      */
/* ************************************************************************* */



/* ************************************************************************* */
/*                       TYPES DEFINED IN THIS MODULE                        */
/* ************************************************************************* */



/* ************************************************************************* */
/*                              CONSTANT VARIABLES                           */
/* ************************************************************************* */



/* ************************************************************************* */
/*                    PROTOTYPES FOR STATIC FUNCTIONS                        */
/* ************************************************************************* */
static void generate_dword(cg_context_t *ctxt,inst_t *inst);



/* ************************************************************************* */
/* PROTOTYPE   : static void generate_dword(cg_context_t *ctxt,              */
/*                                          inst_t *inst);                   */
/*                                                                           */
/* ARGUMENTS   : "ctxt" points to a valid context and "inst" is the inst-    */
/*               ruction for which code is to be generated.                  */
/*                                                                           */
/* FUNCTION    : Generate code for a Dword instruction.                      */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
static void generate_dword(cg_context_t *ctxt,inst_t *inst)
{
  arg_t *arg;
  int arg_type;
  label_t *temp_label,*found_label;
  char *arg_val;
  array_t *label_array;
  unsigned int code_val;

  assert((ctxt != NULL) && (inst != NULL));
  label_array = cg_get_label_array(ctxt);
  
  arg = get_first_arg(inst);
      arg_type = convert_stack_arg(ctxt,inst,arg,FINAL_PASS);
      if ((arg_type != ARG_IMMEDIATE) && (arg_type != ARG_LABEL) && 
	  (arg_type != ARG_CURRENT_ADDR)) {
	iopasm_error(ctxt,inst,ERR_ARG_DWORD_EVAL,"(.dword statement)");
      }

      
      arg_val = get_arg_val(arg);
      if (arg_type == ARG_LABEL) {
       	temp_label = create_label(arg_val,0,SCOPE_NO_SCOPE);
	assert(temp_label != NULL);
	found_label = (label_t *) find_array_elem(label_array,

						  (label_t *) temp_label);
	if (found_label == NULL) {
	  iopasm_error(ctxt,inst,ERR_UNDEFINED_LABEL,"label is %s",arg_val);
	} /* End of if (found_label == NULL) */
	

	else {
	  code_val = get_label_value(ctxt,found_label);
	  IOPASM_FREE(temp_label);
	}
	  
      } /* End of if (arg_type == ARG_LABEL) */

	
      else {
	code_val = atol(arg_val);
      } /* End of !if (arg_type == ARG_LABEL) */

      code_val = htonl(code_val);
      set_inst_code(inst,(unsigned char*) &code_val,4);
}



/* ************************************************************************* */
/* PROTOTYPE   : unsigned char *generate_code(cg_context_t *ctxt,            */
/*                                            plist_t *a_list);              */
/*                                                                           */
/* ARGUMENTS   : "ctxt" is the current context-descriptor. "a_list" is the   */
/*               the pointer to the parser-list containing intermediate      */
/*               code.                                                       */
/*                                                                           */
/* FUNCTION    : Generates the code for the architecture specified by ctxt   */
/*                                                                           */
/* RETURN VALUE: Does not return a value, error handling must be implemented */
/*               in architecture-specific backend.                           */
/* ************************************************************************* */
unsigned char *generate_code(cg_context_t *ctxt,plist_t *a_list)
{
  inst_t *first;
  inst_t *last;
  inst_t *last_inst;
  unsigned char *code_buffer;
  int size = 0;
  int pc = 0;
  int new_pc = 0;
  int diff, row_len;

  assert((ctxt != NULL) && (a_list != NULL));

  code_buffer = cg_get_code_buffer(ctxt);

  if (code_buffer == NULL) {
    iopasm_error(ctxt,NULL,ERR_REPORT_BUG,"%s:%d : Memory allocation failed",__FILE__,
	       __LINE__);
  }
  
  /* Save code buffer in the context */
  cg_set_code_buffer(ctxt,code_buffer);

  row_len = cg_get_row_len(ctxt);

  first = get_first_inst(a_list);
 
  while (first != NULL) {

    diff = 0;
    minimize_stacks(ctxt,first,FINAL_PASS);
    switch (get_inst_type(first)) {


    case INST_ORIGIN:
      new_pc = atoi(get_arg_val(get_first_arg(first)));
      pc = new_pc;
      set_inst_addr(first,pc);
      break;
      

    case INST_ALIGN:
      pc = get_inst_addr(first);
      new_pc += get_inst_size(first);
      break;


    case INST_SEQ: /* Sequential instruction */
      pc = get_inst_addr(first);

      if (pc != new_pc) { /* Check that every instruction is aligned */
	;
      }

      if (pc % cg_get_seq_boundrary(ctxt)) {
        iopasm_error(ctxt,first,ERR_SEQUENTIAL_BOUNDRARY,"PC = %d",
		     pc);
      }

      new_pc += cg_gen_seq_code(ctxt,first,FINAL_PASS,SCOPE_SEQ);
      break;
    

    case INST_WORD:
      break;


    case INST_BYTE:
      break;


    case INST_DWORD:
      generate_dword(ctxt,first);
      pc = get_inst_addr(first);
      new_pc += get_inst_size(first);
      break;

  
    case INST_STATE_BEGIN:
      last = get_next_inst(first);
      first = get_prev_inst(first);
      pc = get_inst_addr(first);
      size = get_inst_size(first);
      
      diff = row_len - new_pc % row_len;
      diff = (diff < size) ? diff: 0;

      if (pc != (new_pc+diff+size)) { 
	iopasm_error(ctxt,first,ERR_REPORT_BUG,
		     "%s:%d : Code size discrepancy. This is an internal error and should be reported to Johan.Jorgensen@axis.com",
		   __FILE__,__LINE__);
      }
      
      if (pc % cg_get_fsm_boundrary(ctxt)) {
        iopasm_error(ctxt,first,ERR_FSM_BOUNDRARY,"PC = %d",pc);
      }

      while((last != NULL) && (get_inst_type(last) != INST_STATE_END)) {
	last = get_next_inst(last);
	  }
      last_inst = last;
      last = get_prev_inst(last);
      
      new_pc += diff + cg_gen_fsm_code(ctxt,first,last,FINAL_PASS);
      first = last_inst;
      break;
      
    }
    first = get_next_inst(first);
  }
  return NULL;
}
 
