/* ************************************************************************* */
/* dyn_array.c: Implementation of a dynamic array datatype.                  */
/*                                                                           */
/*    Copyright 2002-2006 Axis Communications AB.                            */
/*                                                                           */
/*                                                                           */
/*    This program is free software; you can redistribute it and/or modify   */
/*    it under the terms of the GNU General Public License as published by   */
/*    the Free Software Foundation; either version 2 of the License, or      */
/*    (at your option) any later version.                                    */
/*                                                                           */
/*    This program is distributed in the hope that it will be useful,        */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*    GNU General Public License for more details.                           */
/*                                                                           */
/*    You should have received a copy of the GNU General Public License      */
/*    along with this program; if not, write to the Free Software            */
/*    Foundation, Inc., 59 Temple Place - Suite 330,                         */
/*    Boston, MA 02111-1307, USA.                                            */
/*                                                                           */
/*                                                                           */
/*    Send bug reports and inquiries to Johan.Jorgensen@axis.com             */
/* ************************************************************************* */



/* ************************************************************************* */
/*                      INCLUDES USED BY THIS MODULE                         */
/* ************************************************************************* */
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include "dyn_array.h"
#include "debug.h"
#include "dyn_array.h"
#include "iopasm_global.h"



/* ************************************************************************* */
/*                       TYPES DEFINED IN THIS MODULE                        */
/* ************************************************************************* */
struct __array_t {
  void **array_data;                /* Array of pointers to elements         */
  int n_elem;                       /* Number of elements in array           */
  int free_elems;                   /* Number of free elements in array      */
  int (*cmp_func)(void *,void *);   /* Compare two elements -1 0 1           */
  void (*destructor)(void *);       /* Delete element from list              */
};



/* ************************************************************************* */
/*                     DEFINITIONS LOCAL TO THIS MODULE                      */
/* ************************************************************************* */
#define REALLOC_ELEM 64 /* Number of elements reserved, when realloc is done */



/* ************************************************************************* */
/*                              CONSTANT VARIABLES                           */
/* ************************************************************************* */



/* ************************************************************************* */
/*                    PROTOTYPES FOR STATIC FUNCTIONS                        */
/* ************************************************************************* */
static int realloc_elem(array_t *array);
static int find_elem(array_t *array,void *elem);
int find_elem_pos(array_t *array,void *elem);




/* ************************************************************************* */
/* PROTOTYPE   : static int array_realloc(array_t *array);                   */
/*                                                                           */
/* ARGUMENTS   : array points to an array                                    */
/*                                                                           */
/* FUNCTION    : Allocates another realloc elements to the current array     */
/*                                                                           */
/* RETURN VALUE: -1 on failure, positive value on success                    */
/* ************************************************************************* */
static int realloc_elem(array_t *array)
{
  void **tmp_data;

  if (array != NULL) {
    tmp_data = (void **) IOPASM_MALLOC((array->n_elem + REALLOC_ELEM)
			       *sizeof (void *));
    assert(tmp_data != NULL);
    memcpy(tmp_data,array->array_data,sizeof(void *) * array->n_elem);
    IOPASM_FREE(array->array_data);
    array->array_data = tmp_data;

    if (array->array_data == NULL) {
      DEBUG(1,"Could not allocate reallocate memory for array pool");
      IOPASM_FREE(array);
      return -1;
    } /* End of if (array->array_data == NULL) */

    else {
      DEBUG(3,"Reallocated %d bytes for array",REALLOC_ELEM*sizeof(void *));
      array->free_elems += REALLOC_ELEM;
      return 1;
    } /* End of !if (array->array_data == NULL) */
  } /* End of if (array != NULL) */
  
  return -1;

}



/* ************************************************************************* */
/* PROTOTYPE   : static int find_elem(array_t *array,void *elem);            */
/*                                                                           */
/* ARGUMENTS   : array is pointer to an array, elem is a pointer to an       */
/*               element                                                     */
/*                                                                           */
/* FUNCTION    : Search for an element in the array.                         */
/*                                                                           */
/*                                                                           */
/* RETURN VALUE: position of element in the array -1 if not found            */
/* ************************************************************************* */
static int find_elem(array_t *array,void *elem)
{
  int i = 0;
  
  assert(array != NULL);

  while(i < array->n_elem) {
    if ((array->cmp_func(array->array_data[i],elem)) == 0) {
      return i;
    } /* End of if ((array->cmp_func(array->array_data[i],elem)) == 0) */
    i++;
  } /* End of while(i < array->n_elem) */
  return -1;
}

/* ************************************************************************* */
/* PROTOTYPE   : static int find_array_elem_pos(array_t *array,void *elem);  */
/*                                                                           */
/* ARGUMENTS   : array is a pointer to an array and elem is a pointer to the */
/*               which we want to find in the array.                         */
/*                                                                           */
/* FUNCTION    : Returns the correct position of the element. Even if it is  */
/*               not found in the array the position where it should be will */
/*               be returned                                                 */
/*                                                                           */
/* RETURN VALUE: Correct position of the element pointed to by elem          */
/* ************************************************************************* */
int find_array_elem_pos(array_t *array,void *elem)
{
  int i;

  for (i=0;i<array->n_elem;i++) {
    if (!array->cmp_func(array->array_data[i],elem)) return i;
  } /* End of for (i=0;i<array->n_elem;i++) */
  return -1;
}

 

/* ************************************************************************* */
/* PROTOTYPE   : array_t *create_array((int cmp_func),int elem_size);        */
/*                                                                           */
/* ARGUMENTS   : cmp_func is a pointer to the comparison function.           */
/*               "elem_size" is the size of each element                     */
/*                                                                           */
/* FUNCTION    : Creates a new array and initializes it.                     */
/*                                                                           */
/* RETURN VALUE: pointer to the new array                                    */
/* ************************************************************************* */
array_t *create_array(int (*cmp_func)(void *,void *),
		      void (*destrudo)(void *))
{
  array_t *new_array = NULL;
  void *elem_ptr = NULL;
  
  new_array = (array_t *) IOPASM_MALLOC(sizeof(array_t));
  
  if (new_array == NULL) {
    DEBUG(1,"Cannot allocate memory for dynamic array");
  }
  
  else {
    elem_ptr = IOPASM_MALLOC ((size_t) (sizeof(void *) * REALLOC_ELEM));
  
    
    if (elem_ptr == NULL) {
      DEBUG(1,"Cannot allocate memory chunk for %d elements",REALLOC_ELEM);
    
      IOPASM_FREE(new_array);
      new_array = NULL;
    }
  
    else {
      new_array->array_data = elem_ptr;
      new_array->free_elems = REALLOC_ELEM;
      new_array->cmp_func = cmp_func;
      new_array->destructor = destrudo;
      new_array->n_elem = 0;
    }
  }
  return new_array;
}



/* ************************************************************************* */
/* PROTOTYPE   : int add_array_elem(array_t *array,void *elem);              */
/*                                                                           */
/* ARGUMENTS   : "array" is an array and "elem" is an element to be added    */
/*                                                                           */
/* FUNCTION    : Adds an element to an array at the correct position         */
/*                                                                           */
/* RETURN VALUE: Returns 1 if successful -1 otherwise                        */
/* ************************************************************************* */
int add_array_elem(array_t *array,void *elem)
{
  
  if (array == NULL) return -1; /* Array does not exist */
  
  if (array->free_elems==0) { 
    if (realloc_elem(array)<0) {
      DEBUG(1,"Cannot reallocate array");
      return -1;
    }
  }
  
  array->array_data[array->n_elem++] = elem; 
  array->free_elems--;

  return 1;
    
}



/* ************************************************************************* */
/* PROTOTYPE   : void *find_array_elem(array_t *array,void *elem);           */
/*                                                                           */
/* ARGUMENTS   : array is a pointer to an arrat elem is the [partial] elem   */
/*               that we want to find (only field required is the field used */
/*               by  cmp_func)                                               */
/*                                                                           */
/* FUNCTION    : Finds an element in the array and returns a pointer to it   */
/*                                                                           */
/* RETURN VALUE: Pointer to the element if found, null otherwise             */
/* ************************************************************************* */
void *find_array_elem(array_t *array,void *elem)
{
  int i;

  if (array==NULL) return NULL;
  
  i = find_elem(array,elem);

  if (i>=0) return (void *) array->array_data[i];
  else return NULL;
}



/* ************************************************************************* */
/* PROTOTYPE   : int delete_elem(array_t *array,void *elem)                  */
/*                                                                           */
/* ARGUMENTS   : "array" is the array. "elem" the element to delete          */
/*                                                                           */
/* FUNCTION    : Deletes an element "elem" from the array "array"            */
/*                                                                           */
/* RETURN VALUE: 1 if the element was deleted -1 otherwise                   */
/* ************************************************************************* */

int delete_elem(array_t *array,void *elem)
{
  int i=0;
  int j=0;

  if (array==NULL) return -1;

  i = find_elem(array,elem);
  
  array->destructor(array->array_data[i]);

  if (i<0) return -1;

  else { /* Compact array if necessary (probably)*/ 
    for (j=i;j<array->n_elem;j++) 
      array->array_data[j] = array->array_data[j+1];
    
    array->n_elem--;
    array->free_elems++;
  }
    
  return 1; 
}



/* ************************************************************************* */
/* PROTOTYPE   : int delete_array(array_t *array);                           */
/*                                                                           */
/* ARGUMENTS   : array is a pointer to an array                              */
/*                                                                           */
/* FUNCTION    : Deletes the entire array and frees up memory all memory.    */
/*               Elements stored int the array is also deleted.              */
/*                                                                           */
/* RETURN VALUE: 1 if deleted -1 if fail                                     */
/* ************************************************************************* */
int delete_array(array_t *array)
{
  int i;
  if (array==NULL) return -1;

  for(i=0;i<array->n_elem;i++) array->destructor(array->array_data[i]);

  IOPASM_FREE(array);
  array = NULL;
  
  return 1;

}



/* ************************************************************************* */
/* PROTOTYPE   : int delete_array(array_t *array);                           */
/*                                                                           */
/* ARGUMENTS   : array is a pointer to an array. "dst" is a pointer to the   */
/*               data-destination pointer used by the for_each function      */
/*                                                                           */
/* FUNCTION    : Deletes the entire array and frees up memory all memory.    */
/*               Elements stored int the array is also deleted.              */
/*                                                                           */
/* RETURN VALUE: 1 if deleted -1 if fail                                     */
/* ************************************************************************* */
void array_iterator(void *ctxt,array_t *array,void *dst,
		    int (*for_each)(void *a_context,void *dst,void *elem))
{
  int i;
  assert((array != NULL) && (for_each != NULL));

  for (i=0;i<array->n_elem;i++) {
    for_each(ctxt,dst,(array->array_data[i]));
  }

}



/* ************************************************************************* */
/* PROTOTYPE   : void *get_array_elem(array_t *array,int elem);              */
/*                                                                           */
/* ARGUMENTS   : "array" is a pointer to an array. "elem" is the number of   */
/*               the element we want to retrieve.                            */
/*                                                                           */
/* FUNCTION    : Returns a pointer to the element at position "elem" in      */
/*               array. if "elem" is out of bound NULL is returned           */
/*                                                                           */
/* RETURN VALUE: Pointer to element if "elem" is within bounds "NULL"        */
/*               otherwise.                                                  */
/* ************************************************************************* */
void *get_array_elem(array_t *array,int elem)
{
  assert ((array != NULL) && (elem >= 0));
  return (elem < array->n_elem) ? array->array_data[elem]: NULL;
}



/* ************************************************************************* */
/* PROTOTYPE   : unsigned int get_array_size(array_t *array);                */
/*                                                                           */
/* ARGUMENTS   : "array" is a pointer to an array                            */
/*                                                                           */
/* FUNCTION    : Returns the current number of elements in the array         */
/*                                                                           */
/* RETURN VALUE: The number of elements stored in the array                  */
/* ************************************************************************* */
unsigned int get_array_size(array_t *array)
{
  assert(array != NULL);
  return array->n_elem;
}



/* ************************************************************************* */
/* PROTOTYPE   : void set_array_elem(array_t *array,int pos,void *elem);     */
/*                                                                           */
/* ARGUMENTS   : "array" points to a valid array. "pos" is the position to   */
/*               change and "elem" is the element to add to the array at     */
/*               position "pos".                                             */
/*                                                                           */
/* FUNCTION    : Change an existing element in the array. if "elem" is NULL  */
/*               the element is deleted and the size of the array changes    */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void set_array_elem(array_t *array,int pos,void *elem) 
{
  int i = 0;
  
  assert(array != NULL);
  
  if (array->n_elem > pos) {

    if (elem == NULL) { /* Array should be compacted as this element disappears */
      for (i = pos;i<array->n_elem;i++) {
	array->array_data[i] = array->array_data[i+1];
      } /* for (i = pos;i<array->n_elems;i++) */
      array->n_elem--;
    } /* End of if (elem == NULL) */
    
    else {
      array->array_data[i] = elem;
    }
  } /* End of if (array->n_elems > pos) */
}
