/* ************************************************************************* */
/* dyn_stack.c: Implementation of a stack with dynamic depth                 */
/*                                                                           */
/*    Copyright 2002-2006 Axis Communications AB.                            */
/*                                                                           */
/*                                                                           */
/*    This program is free software; you can redistribute it and/or modify   */
/*    it under the terms of the GNU General Public License as published by   */
/*    the Free Software Foundation; either version 2 of the License, or      */
/*    (at your option) any later version.                                    */
/*                                                                           */
/*    This program is distributed in the hope that it will be useful,        */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*    GNU General Public License for more details.                           */
/*                                                                           */
/*    You should have received a copy of the GNU General Public License      */
/*    along with this program; if not, write to the Free Software            */
/*    Foundation, Inc., 59 Temple Place - Suite 330,                         */
/*    Boston, MA 02111-1307, USA.                                            */
/*                                                                           */
/*                                                                           */
/*    Send bug reports and inquiries to Johan.Jorgensen@axis.com             */
/* ************************************************************************* */




/* ************************************************************************* */
/*                      INCLUDES USED BY THIS MODULE                         */
/* ************************************************************************* */
#include <assert.h>
#include <netinet/in.h>
#include <stdint.h>
#include <stdlib.h>

#include "debug.h"
#include "dyn_stack.h"
#include "meta_object.h"
#include "iopasm_global.h"
/* ************************************************************************* */
/*                       TYPES DEFINED IN THIS MODULE                        */
/* ************************************************************************* */
struct __stack_t {
  void **stack_data;                /* Array of pointers to elements         */
  int n_elem;                       /* Number of elements in stack           */
  int free_elems;                   /* Number of free elements in stack      */
  int top;
  void (*elem_destroyer)(void *elem);
  int (*elem_saver)(FILE *out_file,void *object);
  void *(*elem_loader)(FILE *in_file);
};



/* ************************************************************************* */
/*                     DEFINITIONS LOCAL TO THIS MODULE                      */
/* ************************************************************************* */
#define REALLOC_ELEM 64 /* Number of elements reserved, when realloc is done */
#define STK_HOLE_SIZE 32 /* Rudimentary garbage collector */



/* ************************************************************************* */
/*                              CONSTANT VARIABLES                           */
/* ************************************************************************* */



/* ************************************************************************* */
/*                    PROTOTYPES FOR STATIC FUNCTIONS                        */
/* ************************************************************************* */
static dyn_stack_t *realloc_stack(dyn_stack_t *stack);



/* ************************************************************************* */
/* PROTOTYPE   : static dyn_stack_t *realloc_stack(dyn_stack_t *stack);      */
/*                                                                           */
/* ARGUMENTS   : "stack" points to a valid stack.                            */
/*                                                                           */
/* FUNCTION    : Reallocates stack pool data.                                */
/*                                                                           */
/* RETURN VALUE: Pointer to the new stack.                                   */
/* ************************************************************************* */
static dyn_stack_t *realloc_stack(dyn_stack_t *stack)
{ 

  assert(stack != NULL);
  
  stack->stack_data = realloc(*(stack->stack_data),(size_t)
			      (stack->n_elem+REALLOC_ELEM)*(sizeof(void *)));
  assert(stack->stack_data != NULL);

  stack->free_elems += REALLOC_ELEM;

  return stack;
}
  


/* ************************************************************************* */
/* PROTOTYPE   : dyn_stack_t  *create_stack(void (*destructor)(void *data)); */
/*                                                                           */
/* ARGUMENTS   : "destructor" is a valid pointer to the function used to     */
/*               delete stack elements. "loader" is a pointer to the loader  */
/*               function and "saver" is a pointer to a saver function.      */
/*                                                                           */
/* FUNCTION    : Creates a new dynamic stack and sets up the initial data    */
/*               structure.                                                  */
/*                                                                           */
/* RETURN VALUE: Pointer to the new stack                                    */
/* ************************************************************************* */
dyn_stack_t  *create_stack(void (*destructor)(void *),
			   void *(*loader)(FILE *),
			   int (*saver)(FILE *,void *))
{
  dyn_stack_t  *temp_stack;

  assert((destructor != NULL) && 
	 (loader != NULL) && 
	 (saver != NULL));

  temp_stack = IOPASM_MALLOC(sizeof(dyn_stack_t ));
  assert (temp_stack != NULL);
  
  temp_stack->stack_data = NULL;
  temp_stack->n_elem = 0;

  temp_stack->top = 0;
  temp_stack->elem_destroyer = destructor;
  temp_stack->elem_saver = saver;
  temp_stack->elem_loader = loader;
  temp_stack->stack_data = IOPASM_MALLOC((size_t) REALLOC_ELEM * sizeof(void *));
  assert(temp_stack->stack_data != NULL);
    temp_stack->free_elems += REALLOC_ELEM;
  return temp_stack;
}



/* ************************************************************************* */
/* PROTOTYPE   : void stack_push(dyn_stack_t *a_stack,void *an_elem);        */
/*                                                                           */
/* ARGUMENTS   : "a_stack" is a valid pointer to a stack and "an_elem" is a  */
/*               valid element.                                              */
/*                                                                           */
/* FUNCTION    : pushes the element "an_elem" onto the stack "a_stack".      */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void stack_push(dyn_stack_t *a_stack,void *an_elem)
{
  assert ((a_stack != NULL) && (an_elem != NULL));
  
  if (a_stack->free_elems == 0) {
    a_stack = realloc_stack(a_stack);
  }

  a_stack->stack_data[a_stack->top] = an_elem;
  a_stack->top++;
  a_stack->n_elem++;
  a_stack->free_elems--;
}



/* ************************************************************************* */
/* PROTOTYPE   : void *stack_pop(dyn_stack_t *a_stack);                      */
/*                                                                           */
/* ARGUMENTS   : "a_stack" points to a valid stack.                          */
/*                                                                           */
/* FUNCTION    : Returns the element most recently pushed onto the stack.    */
/*                                                                           */
/* RETURN VALUE: An element. NULL if the stack is empty.                     */
/* ************************************************************************* */
void *stack_pop(dyn_stack_t *a_stack) 
{
  void *ret_val=NULL;
  assert (a_stack != NULL);

  if (a_stack->n_elem > 0) {
    ret_val = a_stack->stack_data[a_stack->top - 1];
    a_stack->stack_data[a_stack->top - 1] = NULL;
    a_stack->top--;
    a_stack->n_elem--;
    a_stack->free_elems++;
  }

  return ret_val;
}



/* ************************************************************************* */
/* PROTOTYPE   : int get_stack_depth(dyn_stack_t *a_stack);                  */
/*                                                                           */
/* ARGUMENTS   : "a_stack" points to a valid stack.                          */
/*                                                                           */
/* FUNCTION    : Returns the number of elements currently in the stack (the  */
/*               stack-depth                                                 */
/*                                                                           */
/* RETURN VALUE: Number of elements in the stack.                            */
/* ************************************************************************* */
int get_stack_depth(dyn_stack_t *a_stack) 
{
  assert(a_stack != NULL);
  return a_stack->n_elem;
}



/* ************************************************************************* */
/* PROTOTYPE   : void *stack_peek(dyn_stack_t *a_stack);                     */
/*                                                                           */
/* ARGUMENTS   : "a_stack" points to a valid stack.                          */
/*                                                                           */
/* FUNCTION    : Returns the element on top of the stack, but does not       */
/*               remove it.                                                  */
/*                                                                           */
/* RETURN VALUE: Pointer to a stack element, NULL if the stack is empty.     */
/* ************************************************************************* */
void *stack_peek(dyn_stack_t *a_stack) 
{
  void *ret_val=NULL;
  assert (a_stack != NULL);

  if (a_stack->n_elem > 0) {
    ret_val = a_stack->stack_data[a_stack->top - 1];
  }

  return ret_val;
}



/* ************************************************************************* */
/* PROTOTYPE   : void delete_stack(dyn_stack_t *a_stack);                    */
/*                                                                           */
/* ARGUMENTS   : "a_stack" points to a valid stack.                          */
/*                                                                           */
/* FUNCTION    : Deletes a stack and all elements stored on it.              */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void delete_stack(dyn_stack_t *a_stack)
{
  int i;

  assert(a_stack != NULL);
 
  if ((a_stack->elem_destroyer != NULL) && (a_stack->n_elem >0)) {
    for (i=0;i<a_stack->n_elem;i++) {
      a_stack->elem_destroyer(a_stack->stack_data[i]);
    }
  }

  IOPASM_FREE(a_stack->stack_data);
  IOPASM_FREE(a_stack);
  a_stack = NULL;
}



/* ************************************************************************* */
/* PROTOTYPE   : void stack_iterator(dyn_stack_t *a_stack,void (*for_each)); */
/*                                                                           */
/* ARGUMENTS   : "a_stack" is a pointer to a stack. "for_each" points to a   */
/*               function which will be called once for each element stored  */
/*               on the stack.                                               */
/*                                                                           */
/* FUNCTION    : iterates through a stack and calls "for_each" for each and  */
/*               element in the stack.                                       */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void stack_iterator(dyn_stack_t *a_stack,
		    void (*for_each)(void *data,va_list *c_args),...)
{
  int i=0;
  va_list ap;

  assert((a_stack != NULL) && (for_each != NULL));
  
  va_start(ap,for_each);
  
  for(i=0;i<a_stack->n_elem;i++) {
    for_each(a_stack->stack_data[i],&ap);
  }

  va_end(ap);

}



/* ************************************************************************* */
/* PROTOTYPE   : void invert_stack(dyn_stack_t *a_stack);                    */
/*                                                                           */
/* ARGUMENTS   : "a_stack" points to a valid stack.                          */
/*                                                                           */
/* FUNCTION    : Inverts, i.e. turns the stack over so that the first        */
/*               element in the stack becomes the last and vice-versa.       */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void invert_stack(dyn_stack_t *a_stack)
{
  int i,j;
  void *tmp_data;
  assert(a_stack != NULL);
  j = a_stack->n_elem-1;
  
  for (i=0;i<(a_stack->n_elem/2);i++) {
    tmp_data = a_stack->stack_data[i];
    a_stack->stack_data[i] = a_stack->stack_data[j];
    a_stack->stack_data[j] = tmp_data;
    j--;
  }
}



/* ************************************************************************* */
/* PROTOTYPE   : int save_stack_object(FILE *out_file,dyn_stack_t *stack);   */
/*                                                                           */
/* ARGUMENTS   : "out_file" points to a valid file descriptor and "stack" is */
/*               a valid stack.                                              */
/*                                                                           */
/* FUNCTION    : Saves the stack "stack" as an object to the file pointed to */
/*               "out_file".                                                 */
/*                                                                           */
/* RETURN VALUE: Number of objects saved.                                    */
/* ************************************************************************* */
int save_stack_object(FILE *out_file,dyn_stack_t *stack)
{
  meta_object_t *temp_object;
  int i;
  uint32_t elems;
  
  assert((out_file !=NULL) && (stack != NULL));

  elems = htonl((uint32_t) stack->n_elem);

  temp_object = create_meta_object(ARG_STACK_ELEM,sizeof(uint32_t));
  assert(temp_object != NULL);

  save_meta_object(out_file,temp_object);
  delete_meta_object(temp_object);

  /* Now we can save the entire stack. This is simply done by iterating
   * over the array and calling "elem_saver" for each element. Starting from 0
   * means that the stack will be saved inverted, so that when we read it back 
   * we can simply use the supplied loader and push the elements onto the stack 
   * and we will end up with the same stack again */

  assert(fwrite(&elems,sizeof(uint32_t),1,out_file) == 1);

  i = 0;
  while(i < stack->n_elem) {
    stack->elem_saver(out_file,stack->stack_data[i]);
    i++;
  }

  return stack->n_elem; /* Corresponds to number of objects */
}



/* ************************************************************************* */
/* PROTOTYPE   : dyn_stack_t *load_stack_object(FILE *in_file,               */
/*			       void (*destructor)(void *data),               */
/*			       void *(*loader)(FILE *),                      */
/*			       void (*saver)(FILE *,void *));                */
/*                                                                           */
/* ARGUMENTS   : "in_file" points to a valid file descriptor. "destructor",  */
/*               "loader" and "saver" points to functions used to destroy,   */
/*               load and save elements stored on the stack.                 */
/*                                                                           */
/* FUNCTION    : Loads elements from "in_file" and stores them on a stack    */
/*                                                                           */
/* RETURN VALUE: Pointer to a stack populated with objects.                  */
/* ************************************************************************* */
dyn_stack_t *load_stack_object(FILE *in_file,
			       void (*destructor)(void *),
			       void *(*loader)(FILE *),
			       int (*saver)(FILE *,void *))
{
  void  *stack_elem;
  uint32_t elems;
  dyn_stack_t *the_stack;
  meta_object_t *stk_object;

  assert((in_file != NULL) && (destructor != NULL) && 
	 (loader != NULL) && (saver != NULL));

  the_stack = create_stack(destructor,
			   loader,
			   saver);
  assert(the_stack != NULL);
  /* We never really use the meta object for the stack - but it is still saved */
  /* Just throw it away */

  stk_object = load_meta_object(in_file);
  delete_meta_object(stk_object);

  assert(fread(&elems,sizeof(uint32_t),1,in_file) == 1);
  elems = ntohl(elems);
  
  while(elems > 0) {
    /* As all saver functions saves a meta_object regardless of how they are called
     * we expect to find a meta object in this part of the file as well. We just
     * read it and throw it away, as we do not need it for anything. We already 
     * now what to load right. */
    stk_object = load_meta_object(in_file);
    delete_meta_object(stk_object);
    stack_elem = loader(in_file);
    stack_push(the_stack,stack_elem);
    elems--;
  }
  
  return the_stack;
}



/* ************************************************************************* */
/* PROTOTYPE   : void dump_stack_object(FILE *in_file,void *object,          */
/*		                        void (*object_dump_func));           */
/*                                                                           */
/* ARGUMENTS   : "in_file" points to a valid file descriptor. "object" is    */
/*               the object to dump and "object_dump_func" is the function to*/
/*               use.                                                        */
/*                                                                           */
/* FUNCTION    : Dumps stack objects to stdout in human readable format.     */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void dump_stack_object(FILE *in_file,void *object,
		       void (*object_dump_func)(FILE *in_file,meta_object_t *object))
{

  int i;
  uint32_t elems;
  meta_object_t *an_object;
  
  assert((in_file != NULL) && (object != NULL) && (object_dump_func != NULL));

  assert(fread(&elems,sizeof(uint32_t),1,in_file) == 1);
  elems = ntohl(elems);
  fprintf(stdout,"Dumping stack object\n");
  fprintf(stdout,"     Object size      : %d\n",get_object_size((meta_object_t *)object));
  fprintf(stdout,"     Stack depth      : %d\n",elems);
  
  for (i=0;i<elems;i++) {
    an_object = load_meta_object(in_file);
    object_dump_func(in_file,an_object);
    delete_meta_object(an_object);
  }
  

  fprintf(stdout,"Stack ends here\n");
  fprintf(stdout,"--------------------------------------------------------\n");
}
