/* ************************************************************************* */
/* etrax_fs_mpu_be.c : implementation of ETRAX FS MPU specific code          */
/*                     generation backend.                                   */
/*                                                                           */
/*    Copyright 2002-2006 Axis Communications AB.                            */
/*                                                                           */
/*                                                                           */
/*    This program is free software; you can redistribute it and/or modify   */
/*    it under the terms of the GNU General Public License as published by   */
/*    the Free Software Foundation; either version 2 of the License, or      */
/*    (at your option) any later version.                                    */
/*                                                                           */
/*    This program is distributed in the hope that it will be useful,        */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*    GNU General Public License for more details.                           */
/*                                                                           */
/*    You should have received a copy of the GNU General Public License      */
/*    along with this program; if not, write to the Free Software            */
/*    Foundation, Inc., 59 Temple Place - Suite 330,                         */
/*    Boston, MA 02111-1307, USA.                                            */
/*                                                                           */
/*                                                                           */
/*    Send bug reports and inquiries to Johan.Jorgensen@axis.com             */
/* ************************************************************************* */



/* ************************************************************************* */
/*                      INCLUDES USED BY THIS MODULE                         */
/* ************************************************************************* */
#include <assert.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cg_fe.h"
#include "debug.h"
#include "dyn_array.h"
#include "global_rules.h"
#include "iopasm_err.h"
#include "iopasm_global.h"
#include "i_table.h"
#include "i_table_types.h"
#include "labels.h"
#include "p_list.h"
#include "stack_eval.h"



/* ************************************************************************* */
/*                        Definitions local to this module                   */
/* ************************************************************************* */
#define MPU_INST_LEN 8
#define IWIN_SIZE 4 /* Size of instruction window */
#define REG_SET_SPECIAL_BIT(OP,BIT) (OP = OP | (1 << BIT))



/* ************************************************************************* */
/*                              CONSTANT VARIABLES                           */
/* ************************************************************************* */



/* ************************************************************************* */
/*         Static prototypes defined and implemented in this module          */
/* ************************************************************************* */
static unsigned int convert_label_2addr(char *iname,unsigned int label);
static int pack_32bit_immediate(cg_context_t *ctxt,inst_t *inst,
				unsigned char *buf,int buf_size,arg_t *arg);



/* ************************************************************************* */
/* MPU Rule functions - These are the functions that rules are built from    */
/* ************************************************************************* */
static int validate_rr_inst(cg_context_t *ctxt,inst_t *inst,int i_num,
		    unsigned char *buf,int buf_size);

static int validate_rw_inst(cg_context_t *ctxt,inst_t *inst,int i_num,
		    unsigned char *buf,int buf_size);

static int validate_reti_inst(cg_context_t *ctxt,inst_t *inst,int i_num,
		    unsigned char *buf,int buf_size);


static int check_32bit_ds(cg_context_t *ctxt,inst_t *inst,int i_num,
		    unsigned char *buf,int buf_size);
static unsigned int eval_arg(cg_context_t *ctxt,inst_t *inst,arg_t *arg,
			     int i_num,int r_num,int arg_no,int pass,

			     unsigned int (*eval_reg_func)(cg_context_t *ctxt,
							   inst_t *,
							   int,int),

			     unsigned int (*eval_imm_func)(cg_context_t *ctxt,
							   inst_t *,
							   int,
							   char *));



/* ************************************************************************* */
/* MPU Rule sets - Define all rules that you want to use here                */
/* ************************************************************************* */
static const rule_set_t rrck = {
  "RR source register check and address correction",
  0, { /* validate_dest_register, */
       validate_rr_inst,
       validate_dest_register,
       NULL }
};

static const rule_set_t rwck = {
  "RW source register check and address correction", 
  0, { /* validate_dest_register, */
       validate_rw_inst,
       NULL }
};

static const rule_set_t dsok = { /* dsok =Delay Slot OK */
  "Validate delay slot in MPU", 
  0, { validate_delay_slot,
       NULL }
};

static const rule_set_t ds32 = {
  "",     
  0, { validate_delay_slot,
       check_32bit_ds,
       NULL } 
};

static const rule_set_t nofw = { /* no-forwarding only for cond. branch */
  "No forwarding for conditional branch",
  0, { validate_delay_slot,
       validate_dest_register,
       check_32bit_ds,
       NULL
  }
};

static const rule_set_t reti = { /* Check that no rw rwx or rwq after reti */
  "Warn for rw/rwq/rwx after reti",
  0, { validate_delay_slot,
       validate_reti_inst,
       check_32bit_ds,
       NULL 
  }
};

    

/* ************************************************************************* */
/*                      Operators implemented in backend                     */
/*                                                                           */
/* All operators in the table below are required. If this backend does not   */
/* a particular operator set the associated operator_func to NULL.           */ 
/* ************************************************************************* */
static const cg_operator_t etrax_fs_mpu_operators[] = { 
  { "init", NULL }, /* Initialization of operators - REQUIRED */
  { "+", NULL },    /* Addition - REQUIRED */
  { "-", NULL },    /* Subtraction - REQUIRED */
  { "*", NULL },    /* Multiplication - REQUIRED */
  { "/", NULL },    /* Division - REQUIRED */
  { "<<", NULL },   /* Shift left - REQUIRED */
  { ">>", NULL },   /* Shift Right - REQUIRED */
  { "^", NULL },    /* Xor - REQUIRED */
  { "|", NULL },    /* Or - REQUIRED */
  { "&", NULL },    /* And - REQUIRED */
  { "<", NULL },    /* Less than - REQUIRED */
  { ">", NULL },    /* Greater than - REQUIRED */
  { "end", NULL },  /* End  - REQUIRED */
  { NULL, NULL }    /* Table must end like this */
};



/* ************************************************************************* */
/*     MPU instruction table - describes how to generate MPU instructions    */
/* ************************************************************************* */
static const inst_table_t etrax_fs_mpu_seq[] = {
  { 
    .inst_name = "add",
    .num_args = 3,
    .arg_type = { R,R,R },
    .arg_field =  {20, 15, 25 },
    .field_len = { 5,  5,  5 },
    .inst_len = 4,
    .i_template = { 0x40,0x00,0x00,0x8C },
    .rule_set = &dsok

  },
  { "add",   3,R,R,S, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x04,0x8C, &dsok },
  { "add",   3,R,S,R, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x01,0x8C, &dsok },
  { "add",   3,R,S,S, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x05,0x8C, &dsok },
  { "add",   3,S,R,R, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x02,0x8C, &dsok },
  { "add",   3,S,R,S, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x06,0x8C, &dsok },
  { "add",   3,S,S,R, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x03,0x8C, &dsok },
  { "add",   3,S,S,S, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x07,0x8C, &dsok },

  { "addq",  3,R,I,R, 20, 15, 25,  5, 16,  5,  4, 0x10,0x00,0x00,0x00, &dsok },
  { "addq",  3,I,R,R, 15, 20, 25, 16,  5,  5,  4, 0x10,0x00,0x00,0x00, &dsok },

  { "addx",  3,I,R,R, 63, 20, 25, 32,  5,  5,  8, 0xC0,0x00,0x00,0x8C, &dsok },
  { "addx",  3,R,I,R, 20, 63, 25,  5, 32,  5,  8, 0xC0,0x00,0x00,0x8C, &dsok },
  { "addx",  3,I,S,R, 63, 20, 25, 32,  5,  5,  8, 0xC0,0x00,0x02,0x8C, &dsok },
  { "addx",  3,S,I,R, 20, 63, 25,  5, 32,  5,  8, 0xC0,0x00,0x02,0x8C, &dsok },

  { "addx",  3,I,R,S, 63, 20, 25, 32,  5,  5,  8, 0xC0,0x00,0x04,0x8C, &dsok },
  { "addx",  3,R,I,S, 20, 63, 25,  5, 32,  5,  8, 0xC0,0x00,0x04,0x8C, &dsok },
  { "addx",  3,I,S,S, 63, 20, 25, 32,  5,  5,  8, 0xC0,0x00,0x06,0x8C, &dsok },
  { "addx",  3,S,I,S, 20, 63, 25,  5, 32,  5,  8, 0xC0,0x00,0x06,0x8C, &dsok },

  { "and",   3,R,R,R, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x00,0x8A, &dsok },
  { "and",   3,R,R,S, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x04,0x8A, &dsok },
  { "and",   3,R,S,R, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x01,0x8A, &dsok },
  { "and",   3,R,S,S, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x05,0x8A, &dsok },
  { "and",   3,S,R,R, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x02,0x8A, &dsok },
  { "and",   3,S,R,S, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x06,0x8A, &dsok },
  { "and",   3,S,S,R, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x03,0x8A, &dsok },
  { "and",   3,S,S,S, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x07,0x8A, &dsok },
  
  { "andq",  3,R,I,R, 20, 15, 25,  5, 16,  5,  4, 0x08,0x00,0x00,0x00, &dsok },
  { "andq",  3,I,R,R, 15, 20, 25, 16,  5,  5,  4, 0x08,0x00,0x00,0x00, &dsok },

  { "andx",  3,R,I,R, 20, 63, 25,  5, 32,  5,  8, 0xC0,0x00,0x00,0x8A, &dsok },
  { "andx",  3,I,R,R, 63, 20, 25, 32,  5,  5,  8, 0xC0,0x00,0x00,0x8A, &dsok },
  { "andx",  3,I,S,R, 63, 20, 25, 32,  5,  5,  8, 0xC0,0x00,0x02,0x8A, &dsok },
  { "andx",  3,S,I,R, 20, 63, 25,  5, 32,  5,  8, 0xC0,0x00,0x02,0x8A, &dsok },
  { "andx",  3,I,R,S, 63, 20, 25, 32,  5,  5,  8, 0xC0,0x00,0x04,0x8A, &dsok },
  { "andx",  3,I,S,S, 63, 20, 25, 32,  5,  5,  8, 0xC0,0x00,0x06,0x8A, &dsok },
  { "andx",  3,R,I,S, 20, 63, 25,  5, 32,  5,  8, 0xC0,0x00,0x04,0x8A, &dsok },
  { "andx",  3,S,I,S, 20, 63, 25,  5, 32,  5,  8, 0xC0,0x00,0x06,0x8A, &dsok },

  { "ba",    1,I,0,0, 15,  0,  0, 16,  0,  0,  4, 0x60,0x00,0x00,0x00, &ds32 },
  { "bar",   1,R,0,0, 15,  0,  0,  5,  0,  0,  4, 0x62,0x00,0x00,0x00, &ds32 },
  { "bar",   1,S,0,0, 15,  0,  0,  5,  0,  0,  4, 0x63,0x00,0x00,0x00, &ds32 },
  
  { "bbc",   3,R,I,I, 20, 25, 15,  5,  5, 16,  4, 0x78,0x00,0x00,0x00, &nofw },
  { "bbs",   3,R,I,I, 20, 25, 15,  5,  5, 16,  4, 0x7C,0x00,0x00,0x00, &nofw },
  { "bnz",   2,R,I,0, 20, 15,  0,  5, 16,  0,  4, 0x74,0x40,0x00,0x00, &nofw },
  { "bmi",   2,R,I,0, 20, 15,  0,  5, 16,  0,  4, 0x7F,0xE0,0x00,0x00, &nofw },
  { "bpl",   2,R,I,0, 20, 15,  0,  5, 16,  0,  4, 0x7B,0xE0,0x00,0x00, &nofw },
  { "bz",    2,R,I,0, 20, 15,  0,  5, 16,  0,  4, 0x74,0x00,0x00,0x00, &nofw },

  { "di",    0,0,0,0,  0,  0,  0,  0,  0,  0,  4, 0x40,0x00,0x00,0x01, &dsok },
  { "ei",    0,0,0,0,  0,  0,  0,  0,  0,  0,  4, 0x40,0x00,0x00,0x03, &dsok },
  { "halt",  0,0,0,0,  0,  0,  0,  0,  0,  0,  4, 0x40,0x00,0x00,0x02, &dsok },

  { "jir",   1,I,0,0, 15,  0,  0, 16,  0,  0,  4, 0x60,0x20,0x00,0x00, &ds32 },
  { "jir",   1,R,0,0, 15,  0,  0,  5,  0,  0,  4, 0x62,0x20,0x00,0x00, &ds32 },
  { "jir",   1,S,0,0, 15,  0,  0,  5,  0,  0,  4, 0x63,0x20,0x00,0x00, &ds32 },

  { "jnt",   0,0,0,0,  0,  0,  0,  0,  0,  0,  4, 0x61,0x00,0x00,0x00, &ds32 },

  { "jsr",   1,I,0,0, 15,  0,  0, 16,  0,  0,  4, 0x60,0x40,0x00,0x00, &ds32 },
  { "jsr",   1,R,0,0, 15,  0,  0,  5,  0,  0,  4, 0x62,0x40,0x00,0x00, &ds32 },
  { "jsr",   1,S,0,0, 15,  0,  0,  5,  0,  0,  4, 0x63,0x40,0x00,0x00, &ds32 },

  { "lsl",   3,R,R,R, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x00,0x8E, &dsok },
  { "lsl",   3,R,R,S, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x04,0x8E, &dsok },
  { "lsl",   3,R,S,R, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x01,0x8E, &dsok },
  { "lsl",   3,R,S,S, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x05,0x8E, &dsok },
  { "lsl",   3,S,R,R, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x02,0x8E, &dsok },
  { "lsl",   3,S,R,S, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x06,0x8E, &dsok },
  { "lsl",   3,S,S,R, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x03,0x8E, &dsok },
  { "lsl",   3,S,S,S, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x07,0x8E, &dsok },

  { "lslq",  3,R,I,R, 20, 15, 25,  5, 16,  5,  4, 0x18,0x00,0x00,0x00, &dsok },

  { "lsr",   3,R,R,R, 20, 15 ,25,  5,  5,  5,  4, 0x40,0x00,0x00,0x8F, &dsok },
  { "lsr",   3,R,R,S, 20, 15 ,25,  5,  5,  5,  4, 0x40,0x00,0x04,0x8F, &dsok },
  { "lsr",   3,R,S,R, 20, 15 ,25,  5,  5,  5,  4, 0x40,0x00,0x01,0x8F, &dsok },
  { "lsr",   3,R,S,S, 20, 15 ,25,  5,  5,  5,  4, 0x40,0x00,0x05,0x8F, &dsok },
  { "lsr",   3,S,R,R, 20, 15 ,25,  5,  5,  5,  4, 0x40,0x00,0x02,0x8F, &dsok },
  { "lsr",   3,S,R,S, 20, 15 ,25,  5,  5,  5,  4, 0x40,0x00,0x06,0x8F, &dsok },
  { "lsr",   3,S,S,R, 20, 15 ,25,  5,  5,  5,  4, 0x40,0x00,0x03,0x8F, &dsok },
  { "lsr",   3,S,S,S, 20, 15 ,25,  5,  5,  5,  4, 0x40,0x00,0x07,0x8F, &dsok },

  { "lsrq",  3,R,I,R, 20, 15, 25,  5, 16,  5,  4, 0x1C,0x00,0x00,0x00, &dsok },

  { "lw",    2,I,R,0, 15, 20,  0, 16,  5,  0,  4, 0x64,0x40,0x00,0x00, &dsok },
  { "lw",    2,I,S,0, 15, 20,  0, 16,  5,  0,  4, 0x64,0x60,0x00,0x00, &dsok },
  { "lw",    2,R,R,0, 15, 20,  0,  5,  5,  0,  4, 0x66,0x40,0x00,0x00, &dsok },
  { "lw",    2,R,S,0, 15, 20,  0,  5,  5,  0,  4, 0x66,0x60,0x00,0x00, &dsok },
  { "lw",    2,S,R,0, 15, 20,  0,  5,  5,  0,  4, 0x67,0x40,0x00,0x00, &dsok },
  { "lw",    2,S,S,0, 15, 20,  0,  5,  5,  0,  4, 0x67,0x60,0x00,0x00, &dsok },
  { "lw",    3,R,I,R, 15,  7, 20,  5,  8,  5,  4, 0x66,0x40,0x00,0x00, &dsok },
  { "lw",    3,R,I,S, 15,  7, 20,  5,  8,  5,  4, 0x66,0x60,0x00,0x00, &dsok },
  { "lw",    3,S,I,R, 15,  7, 20,  5,  8,  5,  4, 0x67,0x40,0x00,0x00, &dsok },
  { "lw",    3,S,I,S, 15,  7, 20,  5,  8,  5,  4, 0x67,0x60,0x00,0x00, &dsok },

  { "move",  2,R,R,0, 15, 25,  0,  5,  5,  0,  4, 0x40,0x00,0x00,0x81, &dsok },
  { "move",  2,R,S,0, 15, 25,  0,  5,  5,  0,  4, 0x40,0x00,0x04,0x81, &dsok },
  { "move",  2,S,R,0, 15, 25,  0,  5,  5,  0,  4, 0x40,0x00,0x01,0x81, &dsok },
  { "move",  2,S,S,0, 15, 25,  0,  5,  5,  0,  4, 0x40,0x00,0x05,0x81, &dsok },
  
  { "moveq", 2,I,R,0, 15, 25,  0, 16,  5,  0,  4, 0x24,0x00,0x00,0x00, &dsok },
  { "moveq", 2,I,S,0, 15, 25,  0, 16,  5,  0,  4, 0x2C,0x00,0x00,0x00, &dsok },

  { "movex", 2,I,R,0, 63, 25,  0, 32,  5,  0,  8, 0xC0,0x00,0x00,0x81, &dsok },
  { "movex", 2,I,S,0, 63, 25,  0, 32,  5,  0,  8, 0xC0,0x00,0x04,0x81, &dsok },

  { "nop",   0,0,0,0,  0,  0,  0,  0,  0,  0,  4, 0x40,0x00,0x00,0x00, &dsok },

  { "not",  2,R,R,0, 15, 25,  0,  5,  5,  0,  4, 0x40,0x10,0x00,0x81, &dsok },
  { "not",  2,R,S,0, 15, 25,  0,  5,  5,  0,  4, 0x40,0x10,0x04,0x81, &dsok },
  { "not",  2,S,R,0, 15, 25,  0,  5,  5,  0,  4, 0x40,0x10,0x01,0x81, &dsok },
  { "not",  2,S,S,0, 15, 25,  0,  5,  5,  0,  4, 0x40,0x10,0x05,0x81, &dsok },

  { "or",    3,R,R,R, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x00,0x8B, &dsok },
  { "or",    3,R,R,S, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x04,0x8B, &dsok },
  { "or",    3,R,S,R, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x01,0x8B, &dsok },
  { "or",    3,R,S,S, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x05,0x8B, &dsok },
  { "or",    3,S,R,R, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x02,0x8B, &dsok },
  { "or",    3,S,R,S, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x06,0x8B, &dsok },
  { "or",    3,S,S,R, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x03,0x8B, &dsok },
  { "or",    3,S,S,S, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x07,0x8B, &dsok },

  { "orq",   3,R,I,R, 20, 15, 25,  5, 16,  5,  4, 0x0C,0x00,0x00,0x00, &dsok },
  { "orq",   3,I,R,R, 15, 20, 25, 16,  5,  5,  4, 0x0C,0x00,0x00,0x00, &dsok },

  { "orx",   3,R,I,R, 20, 63, 25,  5, 32,  5,  8, 0xC0,0x00,0x00,0x8B, &dsok },
  { "orx",   3,I,R,R, 63, 20, 25, 32,  5,  5,  8, 0xC0,0x00,0x00,0x8B, &dsok },
  { "orx",   3,S,I,R, 20, 63, 25,  5, 32,  5,  8, 0xC0,0x00,0x02,0x8B, &dsok },
  { "orx",   3,I,S,R, 63, 20, 25, 32,  5,  5,  8, 0xC0,0x00,0x02,0x8B, &dsok },
  
  { "orx",   3,R,I,S, 20, 63, 25,  5, 32,  5,  8, 0xC0,0x00,0x04,0x8B, &dsok },
  { "orx",   3,I,R,S, 63, 20, 25, 32,  5,  5,  8, 0xC0,0x00,0x04,0x8B, &dsok },
  { "orx",   3,S,I,S, 20, 63, 25,  5, 32,  5,  8, 0xC0,0x00,0x06,0x8B, &dsok },
  { "orx",   3,I,S,S, 63, 20, 25, 32,  5,  5,  8, 0xC0,0x00,0x06,0x8B, &dsok },

  { "ret",   0,0,0,0,  0,  0,  0,  0,  0,  0,  4, 0x63,0x00,0x30,0x00, &ds32 },

  { "reti",  0,0,0,0,  0,  0,  0,  0,  0,  0,  4, 0x63,0x60,0x28,0x00, &reti },

  { "rr",    2,I,R,0, 10, 25,  0, 11,  5,  0,  4, 0x50,0x00,0x00,0x00, &rrck },
  { "rr",    2,S,R,0, 20, 25,  0,  5,  5,  0,  4, 0x50,0x00,0x80,0x00, &rrck },
  
  { "rw",    2,R,I,0, 15, 10,  0,  5, 11,  0,  4, 0x56,0x00,0x00,0x00, &rwck },
  { "rw",    2,R,S,0, 15, 20,  0,  5,  5,  0,  4, 0x57,0x00,0x00,0x00, &rwck },

  { "rwq",   2,I,I,0, 26, 10,  0, 16, 11,  0,  4, 0x58,0x00,0x00,0x00, &rwck },
  { "rwq",   2,I,S,0, 15, 20,  0, 16,  5,  0,  4, 0x55,0x00,0x00,0x00, &dsok },

  { "rwx",   2,I,I,0, 63, 10,  0, 32, 11,  0,  8, 0xD4,0x00,0x00,0x00, &rwck },
  { "rwx",   2,I,S,0, 63, 20,  0, 32,  5,  0,  8, 0xD5,0x00,0x00,0x00, &rwck },

  { "sub",   3,R,R,R, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x00,0x8D, &dsok },
  { "sub",   3,R,R,S, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x04,0x8D, &dsok },
  { "sub",   3,R,S,R, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x01,0x8D, &dsok },
  { "sub",   3,R,S,S, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x05,0x8D, &dsok },
  { "sub",   3,S,R,R, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x02,0x8D, &dsok },
  { "sub",   3,S,R,S, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x06,0x8D, &dsok },
  { "sub",   3,S,S,R, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x03,0x8D, &dsok },
  { "sub",   3,S,S,S, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x07,0x8D, &dsok },

  { "subq",  3,R,I,R, 20, 15, 25,  5, 16,  5,  4, 0x14,0x00,0x00,0x00, &dsok },
  
  { "subx",  3,R,I,R, 20, 63, 25,  5, 32,  5,  8, 0xC0,0x00,0x00,0x8D, &dsok },
  { "subx",  3,S,I,R, 20, 63, 25,  5, 32,  5,  8, 0xC0,0x00,0x02,0x8D, &dsok },
  { "subx",  3,R,I,S, 20, 63, 25,  5, 32,  5,  8, 0xC0,0x00,0x04,0x8D, &dsok },
  { "subx",  3,S,I,S, 20, 63, 25,  5, 32,  5,  8, 0xC0,0x00,0x06,0x8D, &dsok },

  { "sw",    2,R,I,0, 20, 15,  0,  5, 16,  0,  4, 0x64,0x00,0x00,0x00, &dsok },
  { "sw",    2,S,I,0, 20, 15,  0,  5, 16,  0,  4, 0x64,0x20,0x00,0x00, &dsok },
  { "sw",    2,R,R,0, 20, 15,  0,  5,  5,  0,  4, 0x66,0x00,0x00,0x00, &dsok },
  { "sw",    2,S,R,0, 20, 15,  0,  5,  5,  0,  4, 0x66,0x20,0x00,0x00, &dsok },
  { "sw",    2,R,S,0, 20, 15,  0,  5,  5,  0,  4, 0x67,0x00,0x00,0x00, &dsok },
  { "sw",    2,S,S,0, 20, 15,  0,  5,  5,  0,  4, 0x67,0x20,0x00,0x00, &dsok },
  { "sw",    3,R,I,R, 20, 7, 15,  5,   8,  5,  4, 0x66,0x00,0x00,0x00, &dsok },
  { "sw",    3,S,I,R, 20, 7, 15,  5,   8,  5,  4, 0x66,0x20,0x00,0x00, &dsok },
  { "sw",    3,R,I,S, 20, 7, 15,  5,   8,  5,  4, 0x67,0x00,0x00,0x00, &dsok },
  { "sw",    3,S,I,S, 20, 7, 15,  5,   8,  5,  4, 0x67,0x20,0x00,0x00, &dsok },

  { "swx",   2,I,I,0, 63, 15,  0, 32, 16,  0,  8, 0xe4,0x00,0x00,0x00, &dsok },
  { "swx",   2,I,R,0, 63, 15,  0, 32,  5,  0,  8, 0xe6,0x00,0x00,0x00, &dsok },
  { "swx",   2,I,S,0, 63, 15,  0, 32,  5,  0,  8, 0xe7,0x00,0x00,0x00, &dsok },
  { "swx",   3,I,I,R, 63,  7, 15, 32,  8,  5,  8, 0xe6,0x00,0x00,0x00, &dsok },
  { "swx",   3,I,I,S, 63,  7, 15, 32,  8,  5,  8, 0xe7,0x00,0x00,0x00, &dsok },

  { "xor",   3,R,R,R, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x00,0x89, &dsok },
  { "xor",   3,R,R,S, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x04,0x89, &dsok },
  { "xor",   3,R,S,R, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x01,0x89, &dsok },
  { "xor",   3,R,S,S, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x05,0x89, &dsok },
  { "xor",   3,S,R,R, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x02,0x89, &dsok },
  { "xor",   3,S,R,S, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x06,0x89, &dsok },
  { "xor",   3,S,S,R, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x03,0x89, &dsok },
  { "xor",   3,S,S,S, 20, 15, 25,  5,  5,  5,  4, 0x40,0x00,0x07,0x89, &dsok },
  
  { "xor",  2,R,R,0, 15, 25,  0,  5,  5,  0,  4, 0x40,0x00,0x00,0x88, &dsok },
  { "xor",  2,R,S,0, 15, 25,  0,  5,  5,  0,  4, 0x40,0x00,0x04,0x88, &dsok },
  { "xor",  2,S,R,0, 15, 25,  0,  5,  5,  0,  4, 0x40,0x00,0x01,0x88, &dsok },
  { "xor",  2,S,S,0, 15, 25,  0,  5,  5,  0,  4, 0x40,0x00,0x05,0x88, &dsok },

  { "xorq",  3,R,I,R, 20, 15, 25,  5, 16,  5,  4, 0x04,0x00,0x00,0x00, &dsok },
  { "xorq",  3,I,R,R, 15, 20, 25, 16,  5,  5,  4, 0x04,0x00,0x00,0x00, &dsok },
  
  { "xorx",  3,R,I,R, 20, 63, 25,  5, 32,  5,  8, 0xC0,0x00,0x00,0x89, &dsok },
  { "xorx",  3,I,R,R, 63, 20, 25, 32,  5,  5,  8, 0xC0,0x00,0x00,0x89, &dsok },
  { "xorx",  3,S,I,R, 20, 63, 25,  5, 32,  5,  8, 0xC0,0x00,0x02,0x89, &dsok },
  { "xorx",  3,I,S,R, 63, 20, 25, 32,  5,  5,  8, 0xC0,0x00,0x02,0x89, &dsok },

  { "xorx",  3,R,I,S, 20, 63, 25,  5, 32,  5,  8, 0xC0,0x00,0x04,0x89, &dsok },
  { "xorx",  3,I,R,S, 63, 20, 25, 32,  5,  5,  8, 0xC0,0x00,0x04,0x89, &dsok },
  { "xorx",  3,S,I,S, 20, 63, 25,  5, 32,  5,  8, 0xC0,0x00,0x06,0x89, &dsok },
  { "xorx",  3,I,S,S, 63, 20, 25, 32,  5,  5,  8, 0xC0,0x00,0x06,0x89, &dsok },
  { NULL,    0,0,0,0,  0,  0,  0,  0,  0,  0,  0, 0x00,0x00,0x00,0x00, NULL }};
 


/* ************************************************************************* */
/*                            MPU Register values                            */
/* ************************************************************************* */
static const reg_table_t etrax_fs_mpu_regs[] = {
  {  "r0",  0, ARG_NORMAL_REG, ARG_NORMAL_REG }, 
  {  "r1",  1, ARG_NORMAL_REG, ARG_NORMAL_REG }, 
  {  "r2",  2, ARG_NORMAL_REG, ARG_NORMAL_REG }, 
  {  "r3",  3, ARG_NORMAL_REG, ARG_NORMAL_REG }, 
  {  "r4",  4, ARG_NORMAL_REG, ARG_NORMAL_REG }, 
  {  "r5",  5, ARG_NORMAL_REG, ARG_NORMAL_REG }, 
  {  "r6",  6, ARG_NORMAL_REG, ARG_NORMAL_REG }, 
  {  "r7",  7, ARG_NORMAL_REG, ARG_NORMAL_REG },
  {  "r8",  8, ARG_NORMAL_REG, ARG_NORMAL_REG }, 
  {  "r9",  9, ARG_NORMAL_REG, ARG_NORMAL_REG }, 
  { "r10", 10, ARG_NORMAL_REG, ARG_NORMAL_REG }, 
  { "r11", 11, ARG_NORMAL_REG, ARG_NORMAL_REG }, 
  { "r12", 12, ARG_NORMAL_REG, ARG_NORMAL_REG }, 
  { "r13", 13, ARG_NORMAL_REG, ARG_NORMAL_REG }, 
  { "r14", 14, ARG_NORMAL_REG, ARG_NORMAL_REG }, 
  { "r15", 15, ARG_NORMAL_REG, ARG_NORMAL_REG },
  { "pc",   2, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "wsts", 3, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "jaddr",4, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "irp",  5, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "srp",  6, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "t0",   8, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "t1",   9, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "t2",  10, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "t3",  11, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "i0",  16, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "i1",  17, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "i2",  18, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "i3",  19, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "i4",  20, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "i5",  21, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "i6",  22, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "i7",  23, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "i8",  24, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "i9",  25, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "i10", 26, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "i11", 27, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "i12", 28, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "i13", 29, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "i14", 30, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "i15", 31, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "p2",   2, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "p3",   3, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "p5",   5, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "p6",   6, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "p8",   8, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "p9",   9, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "p10", 10, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "p11", 11, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "p16", 16, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "p17", 18, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "p18", 18, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "p19", 19, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "p20", 20, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "p21", 21, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "p22", 22, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "p23", 23, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "p24", 24, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "p25", 25, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "p26", 26, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "p27", 27, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "p28", 28, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "p29", 29, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "p30", 30, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "p31", 31, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "p1",   1, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { "rega", 1, ARG_SPECIAL_REG, ARG_SPECIAL_REG }, 
  { NULL,   0, 0, 0 }};



/* ************************************************************************* */
/*                  MEMORY INFORMATION FOR ETRAX FS MPU                      */
/* ************************************************************************* */
static const mem_table_t etrax_fs_mpu_template = 
  { 4, 256, 64, 0, 0, 65535, 4095, 511, NULL, NULL, NULL, NULL };



/* ************************************************************************* */
/*                   MPU-SPECIFIC C MACROS & CONSTANTS                       */
/* ************************************************************************* */
static const char *etrax_fs_mpu_cpp_macros[] = {
  "-D__MPU__",
  "-D__ETRAX_FS__",
  "-DINV16(X)=(0x0000FFFF & (~X))",
  "-DINV32(X)=(0xFFFFFFFF & (~X))",
  NULL };



/* ************************************************************************* */
/* PROTOTYPE   : static unsigned int convert_label_2addr(char *iname,        */
/*                                                       unsigned int label);*/
/*                                                                           */
/* ARGUMENTS   : "iname" is the name of the instruction and "val" is the     */
/*               destination address.                                        */
/*                                                                           */
/* FUNCTION    : Converts a byte address to a dword address for selected     */
/*               instructions.                                               */
/*                                                                           */
/* RETURN VALUE: Correct address value.                                      */
/* ************************************************************************* */
static unsigned int convert_label_2addr(char *iname,unsigned int label)
{
  char *i_group[] = { "ba", "bbc", "bbs", "bnz", "bz", "jir",
		    "jsr","bmi","bpl", "lw", "sw", "swx", NULL };

  int i = 0;
  unsigned int val;

  assert (iname != NULL);
  val = label;

  while (i_group[i] != NULL) {
    if (!strcmp(iname,i_group[i])) {
      val = val / 4;
    }
    i++;
  }
  
  return val;
}



/* ************************************************************************* */
/* PROTOTYPE   : static int check_32bit_ds(cg_context_t *ctxt,inst_t *inst,  */
/*                                         int i_num,unsigned char *buf,     */
/*                                         int buf_size)                     */ 
/*                                                                           */
/* ARGUMENTS   : "ctxt" points to a valid context, "inst" is the current     */
/*               instruction. "i_num" is the instruction number according to */
/*               the instruction table. "buf" is the buffer where code is to */
/*               be stored and "buf_size" is the size of the aforementioned  */
/*               buffer.                                                     */
/*                                                                           */
/* FUNCTION    : Rule function used to check delay-slots for 32 bit          */
/*               instructions.                                               */
/*                                                                           */
/* RETURN VALUE: Returns "1" if the rule is obeyed otherwise it will report  */
/*               an error.                                                   */
/* ************************************************************************* */
static int check_32bit_ds(cg_context_t *ctxt,inst_t *inst,int i_num,
		    unsigned char *buf,int buf_size)
{

  inst_t *next_inst;
  int rel_addr,len;
  assert((ctxt != NULL) && (inst != NULL));

  next_inst = get_next_inst(inst);

  while (next_inst != NULL) { /* Need to check if it is not end of list */

    if (get_inst_type(next_inst) == INST_SEQ) { /* We have a sequential inst */
      rel_addr = get_inst_addr(next_inst) % 8; 
      len = get_inst_size(next_inst);
      
      if ((len == 8) && (rel_addr >0)) { /* We have an error */
	iopasm_error(ctxt,inst,ERR_32BIT_UNALIGNED_IN_DS,"");
      }
      return 0;
    } /* End of if (get_inst_type(next_inst) == INST_SEQ) */
	
    else {
      next_inst = get_next_inst(next_inst);
    } /* End of !if (get_inst_type(next_inst) == INST_SEQ) */

  } /* End of while (next_inst != NULL) */
  return 0;
}



/* ************************************************************************* */
/* PROTOTYPE   : static int validate_reti_inst(cg_context_t *ctxt,           */
/*                                             inst_t *inst,int i_num,       */
/*                    			       unsigned char *buf,           */
/*                                             int buf_size);                */
/*                                                                           */
/* ARGUMENTS   : See arguments for the check_32_bit_ds(...)                  */
/*                                                                           */
/* FUNCTION    : Rule function to check that no rw, rwx or rwq instruction   */
/*               is placed after a reti instruction.                         */
/*                                                                           */
/* RETURN VALUE: Returns 1 if the rule is obeyed. Otherwise it will report   */
/*               error or a warning and will probably exit afterwards.       */
/* ************************************************************************* */
static int validate_reti_inst(cg_context_t *ctxt,inst_t *inst,int i_num,
			      unsigned char *buf,int buf_size) 
{
  char *warn_for_these[] = { "rw", "rwx", "rwq", NULL };
  char *i_name;
  inst_t *next_inst;
  int i;
  next_inst = get_next_inst(inst);

  while((next_inst != NULL) && 
	(get_inst_type(next_inst) != INST_SEQ)) {
    next_inst = get_next_inst(next_inst);
  }  /* End of While */

  if (next_inst != NULL) { /* There is a  next instruction */
    i_name = get_inst(next_inst);
    i = 0 ;
    
    while(warn_for_these[i] != NULL) {
      if (!strcmp(warn_for_these[i],i_name)) {
	iopasm_error(ctxt,inst,WARN_RETI_VIOLATION,NULL);
      }
      i++;
    }
  } /* End of if (next_inst != NULL) */

    return 1; /* We did not find an offending instruction in the delay 
	       * slot of RETI */

}



/* ************************************************************************* */
/* PROTOTYPE   : static int validate_rr_inst(cg_context_t *ctxt,             */
/*                                           inst_t *inst,int i_num,         */
/*	                                     unsigned char *buf,             */
/*                                           int buf_size);                  */
/*                                                                           */
/* ARGUMENTS   : See arguments for the check_32_bit_ds(...)                  */
/*                                                                           */
/* FUNCTION    : Rule function to validate rr instruction                    */
/*                                                                           */
/* RETURN VALUE: Returns 1 if the rule is obeyed. Otherwise it will report   */
/*               error or a warning and will exit on error and               */
/*               return 1 if emitting a warning.                             */
/* ************************************************************************* */
static int validate_rr_inst(cg_context_t *ctxt,inst_t *inst,int i_num,
		    unsigned char *buf,int buf_size) 
{
  arg_t *first_arg;
  int len;
  unsigned int addr;
  char *arg_val;
  inst_t *prev_inst;
  inst_t *prev_inst2;
  char *prev_iname;
  char *prev_iname2;
  
  assert ((ctxt != NULL) && (inst != NULL) &&(buf != NULL) && (buf_size > 0));

  prev_inst = get_prev_inst(inst);

  while((prev_inst != NULL) && (get_inst_type(prev_inst) != INST_SEQ)) {
    prev_inst = get_prev_inst(prev_inst);
  }

  prev_inst2 = get_prev_inst(prev_inst);

  while((prev_inst2 != NULL) && (get_inst_type(prev_inst2) != INST_SEQ)) {
    prev_inst2 = get_prev_inst(prev_inst2);
  }

  if (prev_inst2 == NULL) {
    prev_inst2 = prev_inst;
  }

  if (prev_inst != NULL) { /* Check that it is not rw something */
    prev_iname = get_inst(prev_inst);
    prev_iname2 = get_inst(prev_inst2);
    if ((prev_iname[0] == 'r') && (prev_iname[1] == 'w')  && 
	(strcmp(prev_iname2,"ba")) && (strcmp(prev_iname2,"bar"))) {
      iopasm_error(ctxt,inst,ERR_ILLEGAL_INST_ORDER,
		 "\"%s\"  may not be preceeded by \"%s\"",
		 get_inst(inst),prev_iname);
    }
  }
  
  first_arg = get_first_arg(inst);
  
  /* Convert any stacks found in this instruction. They're dangerous! */
  if (get_arg_type(first_arg) == ARG_STACK) {
    convert_stack_arg(ctxt,inst,first_arg,FINAL_PASS);
  }
  
  arg_val = get_arg_val(first_arg);
  len = get_arg_len(first_arg);

  if (get_arg_type(first_arg) == ARG_IMMEDIATE) {
    addr = atol(arg_val);
    memset(arg_val,0,len);
    snprintf(arg_val,len+1,"%u",addr / 4);
    return 1;
  }

  else if (get_arg_type(first_arg) == ARG_SPECIAL_REG) {
    if (!strcmp("rega",arg_val)) {
      return 1;
    }
    else {
      iopasm_error(ctxt,inst,ERR_ILLEGAL_REGISTER_USE,
		   "The register %s may not be used here",arg_val);
    }
  }
  
  else {
    iopasm_error(ctxt,inst,ERR_ILLEGAL_REGISTER_USE,
		 "The register %s may not be used here",arg_val);

  }

  iopasm_error(NULL,NULL,ERR_REPORT_BUG,"Reported from %s:%d",__FILE__,__LINE__);
  return 0; /* To get rid of warning - will never be reached.
	     * This should be preceeded by an error causing a 
	     * bug-report because it should simply never happen.
	     * If not it is a bug in itself */
}



/* ************************************************************************* */
/* PROTOTYPE   : static int validate_rw_inst(cg_context_t *ctxt,             */
/*                                           inst_t *inst,int i_num,         */
/*	                                     unsigned char *buf,             */
/*                                           int buf_size);                  */
/*                                                                           */
/* ARGUMENTS   : See arguments for the check_32_bit_ds(...)                  */
/*                                                                           */
/* FUNCTION    : Rule function to validate rw instruction                    */
/*                                                                           */
/* RETURN VALUE: Returns 1 if the rule is obeyed. Otherwise it will report   */
/*               error or a warning and will exit on error and               */
/*               return 1 if emitting a warning.                             */
/* ************************************************************************* */
static int validate_rw_inst(cg_context_t *ctxt,inst_t *inst,int i_num,
		    unsigned char *buf,int buf_size)
{
  arg_t *first_arg;
  int len;
  unsigned int addr;
  char *arg_val;
  assert ((ctxt != NULL) && (inst != NULL) &&(buf != NULL) && (buf_size > 0));
  
  first_arg = get_first_arg(inst);
  first_arg = get_next_arg(first_arg); /* We can safely do this otherwise the
					* instruction would not have been 
					* found */

  /* Let us make sure that the argument is _not_ a stack. If it is */
  /* we will convert it to something different. */

  if (get_arg_type(first_arg) == ARG_STACK) {
    convert_stack_arg(ctxt,inst,first_arg,FINAL_PASS);
  }

  len = get_arg_len(first_arg);
  arg_val = get_arg_val(first_arg);

  if (get_arg_type(first_arg) == ARG_SPECIAL_REG) {
    if ((!strcmp("rega",arg_val))||(!strcmp("p1",arg_val))) {
      return 1;
    }
    
    else {
      iopasm_error(ctxt,inst,ERR_ILLEGAL_REGISTER_USE,
		   "The register %s may not be used here",arg_val);
    }
  }

  else if (get_arg_type(first_arg) == ARG_IMMEDIATE) {
    addr = atol(arg_val);
    memset(arg_val,0,len);
    snprintf(arg_val,len+1,"%u",addr / 4);
    return 1;
  }

  else {
    iopasm_error(ctxt,inst,ERR_ILLEGAL_REGISTER_USE,
		 "The register %s may not be used here",arg_val);
  }
  
  return 1;
}



/* ************************************************************************* */
/* PROTOTYPE   : static int pack_32bit_immediate(cg_context_t *ctxt,         */
/*                                               inst_t *inst,               */
/*      				         unsigned char *buf,         */
/*						 int buf_size,arg_t *arg);   */
/*                                                                           */
/* ARGUMENTS   : "ctxt" is a pointer to the current context and "inst" is    */
/*               the pointer to the current instruction. "buf" is the buffer */
/*               where the generated code is stored and "buf_size" is the    */
/*               size of "buf" measured in bytes. "arg" is the argument      */
/*               containing the 32-bit immediate.                            */
/*                                                                           */
/* FUNCTION    : Packs a 32 bit immediate at the right place in MPU memoruy. */
/*                                                                           */
/* RETURN VALUE: -1 on error. On success, the number of used bytes.          */
/* ************************************************************************* */
static int pack_32bit_immediate(cg_context_t *ctxt,inst_t *inst,
				unsigned char *buf,int buf_size,arg_t *arg)
{
  unsigned int imm_value;
  unsigned int conv_value;
  unsigned char *ptr;
  int arg_type;
  char *arg_val;
  char *imm_val;
  char txt_val[MAX_STRLEN];
  unsigned int label_val;
  label_t *temp_label;
  label_t *found_label;
  array_t *labels;
  char *iname;
  assert((buf != NULL) && (arg != NULL) && (ctxt != NULL) && (inst !=NULL));

  iname = get_inst(inst);
  labels = cg_get_label_array(ctxt);

  arg_type = get_arg_type(arg);
  arg_val = get_arg_val(arg);
  if (arg_type == ARG_LABEL) {
    temp_label = create_label(arg_val,0,SCOPE_NO_SCOPE);
    assert(temp_label != NULL);
    
    found_label = (label_t *) find_array_elem(labels,(void *) temp_label); 
    
    if ((found_label == NULL) && 
	(!check_option("allow-undefined-symbols",(int *)1,OPT_INTEGER)))
      { /* Cannot find label - return error and exit */
	iopasm_error(ctxt,inst,ERR_UNDEFINED_LABEL,"Label is %s\n",arg_val);
      }
    
    if (found_label == NULL) { /* Undefined symbols allowed and no label! */
      found_label = temp_label;
    }
    
    label_val = get_label_value(ctxt,found_label);
    label_val = convert_label_2addr(iname,label_val);
    snprintf(txt_val,MAX_STRLEN,"%d",label_val);
    imm_val = txt_val;

    IOPASM_FREE(temp_label);

  } /* End if arg_type == ARG_LABEL) */

  else {
    imm_val = get_arg_val(arg);
  }
 
  imm_value = (unsigned int) strtoul(imm_val,NULL,10);
  conv_value = htonl(imm_value);
  ptr = (unsigned char *) &conv_value;
  
  if (buf_size >= 4) {
    buf[0] = ptr[0];
    buf[1] = ptr[1];
    buf[2] = ptr[2];
    buf[3] = ptr[3];
    return 4;
  }

  else { return -1; }

}



/* ************************************************************************* */
/* PROTOTYPE   : static unsigned int eval_arg(cg_context_t *ctxt,            */
/*                                            inst_t *inst,arg_t *arg,       */
/* 			                      int i_num,int r_num,           */
/*                                            int arg_no,int pass,           */
/*                                                                           */
/* 			                      unsigned int (*eval_reg_func), */
/* 			                      unsigned int (*eval_imm_func));*/
/*                                                                           */
/* ARGUMENTS   : "ctxt" is the current code generation context. "inst" is a  */
/*               valid (current) instruction. "arg" is the argument to       */
/*               be evaluated, arg_no is the argument (operand) number, and  */
/*               "pass" is the current code-generation pass (intermediate    */
/*               or final), "eval_reg_func" and "eval_imm_func" are pointers */
/*               to functions used to evaluate register and immediate        */
/*               immediate operands respectively.                            */
/*                                                                           */
/* FUNCTION    : Generates the part of the op-code that is associated with   */
/*               operands. Called the same number of times as the number of  */
/*               instruction operands.                                       */
/*                                                                           */
/* RETURN VALUE: Returns the part of an op-code for the instruction "inst"   */
/*               that depends on "arg". The function emits hard errors if    */
/*               values etc. is out of range.                                */
/* ************************************************************************* */
static unsigned int eval_arg(cg_context_t *ctxt,inst_t *inst,arg_t *arg,
			     int i_num,int r_num,int arg_no,int pass,

			     unsigned int (*eval_reg_func)(cg_context_t *ctxt,
							   inst_t *,
							   int,int),

			     unsigned int (*eval_imm_func)(cg_context_t *ctxt,
							   inst_t *,
							   int,
							   char *))
{

  inst_table_t *i_table;
  reg_table_t *r_table;
  array_t *labels;
  label_t *temp_label;
  label_t *found_label;
  char *i_name;
  char *arg_val;
  unsigned int imm_val;
  int arg_type;
  int meta_data;

  unsigned char txt_buff[MPU_INST_LEN];

  assert((ctxt != NULL) && (inst != NULL) && (arg != NULL));
  arg_type = get_arg_type(arg);
  meta_data = 0;
  labels = cg_get_label_array(ctxt);
  i_table = cg_get_inst_table(ctxt);
  r_table = cg_get_reg_table(ctxt);
  i_name = get_inst(inst);

  if (arg_type == ARG_STACK) {
    imm_val = eval_arg_stack(ctxt,inst,i_num,arg,&meta_data,arg_no,pass);
    arg_type == get_arg_type(arg);
    arg_val = get_arg_val(arg);
  }

  arg_val = get_arg_val(arg);

  if ((arg_type == ARG_NORMAL_REG) || (arg_type == ARG_SPECIAL_REG)) {
    return eval_reg_func(ctxt,inst,i_num,r_num);
    } 

  else if (arg_type == ARG_IMMEDIATE) {
    return eval_imm_func(ctxt,inst,i_num,arg_val);
  }

  else if (arg_type == ARG_LABEL) {
    temp_label = create_label(arg_val,0,SCOPE_NO_SCOPE);
    assert(temp_label != NULL);

    found_label = (label_t *) find_array_elem(labels,(void *) temp_label); 

    if ((found_label == NULL) && 
	(!check_option("allow-undefined-symbols",(int *)1,OPT_INTEGER)))
      { /* Cannot find label - return error and exit */
	iopasm_error(ctxt,inst,ERR_UNDEFINED_LABEL,"Label is %s\n",arg_val);
      }
    
    if (found_label == NULL) { /* Undefined symbols allowed and no label! */
      found_label = temp_label;
    }
    
    imm_val = get_label_value(ctxt,found_label);
    IOPASM_FREE(temp_label);

    imm_val = convert_label_2addr(i_name,imm_val);
    snprintf(txt_buff,MPU_INST_LEN,"%d",imm_val);
    return eval_imm_func(ctxt,inst,i_num,txt_buff);
  } /* End of else if (arg_type == ARG_LABEL) */
  else { 
    iopasm_error(ctxt,inst,ERR_UNKNOWN_ARGUMENT,"Argument is: %s",arg_val);
  }

  iopasm_error(NULL,NULL,ERR_REPORT_BUG,"Reported from %s:%d",__FILE__,__LINE__);
  return 0; /* To get rid of warning - will never be reached.
	     * This should be preceeded by an error causing a 
	     * bug-report because it should simply never happen.
	     * If not it is a bug in itself */ 
}



/* ************************************************************************* */
/* PROTOTYPE   : unsigned int etrax_fs_mpu_conv_pc_fsm(unsigned int val);    */
/*                                                                           */
/* ARGUMENTS   : "val" is the PC which should be converted.                  */
/*                                                                           */
/* FUNCTION    : Converts a byte address to an FSM address.                  */
/*                                                                           */
/* RETURN VALUE: Always 0 as no FSM mode exists for the ETRAX FS             */
/* ************************************************************************* */
unsigned int etrax_fs_mpu_conv_pc_fsm(unsigned int val)
{
  return 0;
}
  


/* ************************************************************************* */
/* PROTOTYPE   : unsigned int etrax_fs_mpu_conv_pc_seq(unsigned int val);    */
/*                                                                           */
/* ARGUMENTS   : "val" is a value representing the sequential PC             */
/*                                                                           */
/* FUNCTION    : Converts a sequential PC byte address to the correct arch.  */
/*               representation.                                             */
/*                                                                           */
/* RETURN VALUE: */
/* ************************************************************************* */
unsigned int etrax_fs_mpu_conv_pc_seq(unsigned int val)
{
  return val;
}



/* ************************************************************************* */
/* PROTOTYPE   : etrax_fs_mpu_set_tables(cg_context_t *context)              */
/*                                                                           */
/* ARGUMENTS   : A pointer to a context in which tables are to be exported   */
/*                                                                           */
/* FUNCTION    : This function allowes for specific initialization of table  */
/*               constants prior to calling the cg_set_inst_table(...) and   */
/*               cg_set_reg_table(...) functions.                            */
/*                                                                           */
/* RETURN VALUE: returns a positive non-zero value on success and a negative */
/*               non-zero value on failure.                                  */
/* ************************************************************************* */
int etrax_fs_mpu_set_tables(cg_context_t *ctxt)
{
  char *options[] = { "auto-align-64bit-inst", NULL };
  int opt_val[] = { OPT_DEFAULT_OFF, OPTION_ENDS };
  /* This function allows for specific initialization of tables prior to 
   * assigning it to the context. In this case we are satisfied once the
   * instruction boundraries as well as the optimal size of the FSM
   * instruction has been set. We also set the machine-specific 
   * code-generation options, which can be supplied through the "-M" switch
   */

  cg_set_fsm_boundrary(ctxt,-1); /* Boundraries are _always_ measured in bytes */
  cg_set_seq_boundrary(ctxt,4); /* Boundraries are _always_ measured in bytes */

  cg_set_fsm_size(ctxt,32); /* Instruction-sizes are is _always_ measured 
			  * in bytes */

  cg_set_inst_table(ctxt,etrax_fs_mpu_seq);
  cg_set_reg_table(ctxt,etrax_fs_mpu_regs);
  cg_set_cpp_macros(ctxt,(char **)etrax_fs_mpu_cpp_macros);
  cg_set_be_operators(ctxt,etrax_fs_mpu_operators);
  cg_init_machine_options(ctxt,options,opt_val,sizeof(opt_val)/sizeof(int));
  cg_set_memory_bank_count(ctxt,1);
  return 1;
}  



/* ************************************************************************* */
/* PROTOTYPE   : int etrax_fs_mpu_gen_fsm(cg_context_t seq,char *buf,        */
/*                                        int buf_siz,inst_t *first,         */
/*                                        inst_t *last);                     */
/*                                                                           */
/*                                                                           */
/* ARGUMENTS   : "seq" is the current code-generation context. "first" is    */
/*               the startpoint of the current state and "last" is the       */
/*		 end-point.                                                  */
/*                                                                           */
/* FUNCTION    : Generates code for FSM state.                               */
/*                                                                           */
/* RETURN VALUE: Returns the size of the generated code. If buffer is to     */
/*               small the shortage of space is reported as a negative       */
/*               value. 0 is returned if code cannot be generated, i.e. some */
/*               non-syntactic error was discovered.                         */
/* ************************************************************************* */
int etrax_fs_mpu_gen_fsm(cg_context_t *seq,inst_t *first,inst_t *last,
			 int pass) 
{ 
  /* As the MPU does not allow FSM instructions we simply return 0 to indicate
   * that the instruction cannot be generated. 
   */
  return 0; 
}



/* ************************************************************************* */
/* PROTOTYPE   : int etrax_fs_mpu_gen_seq(cg_context_t *ctxt,inst_t *inst,   */
/*                                        int pass,int scope);               */
/*                                                                           */
/* ARGUMENTS   : "ctxt" is a valid context. "inst" is the instruction for    */
/*               which code is to be generated. "Pass" is the assembly pass  */
/*               and scope is the generation scope (FSM or SEQ).             */
/*                                                                           */
/* FUNCTION    : Generate the opcode for "inst"                              */
/*                                                                           */
/* RETURN VALUE: Size of instruction measured in bytes. 0 if instruction     */
/*               cannot be generated.                                        */
/* ************************************************************************* */
int etrax_fs_mpu_gen_seq(cg_context_t *ctxt,inst_t *inst,int pass,int scope)
{
  int i_number,r_number,sreg_number;
  arg_t *current_arg;
  int arg_type,i;
  inst_table_t *i_table;
  inst_table_t *current_row;
  reg_table_t *r_table;
  unsigned int op_code;
  char *text_inst;
  char *arg_text;
  int temp_val,i_size;
  int ret_val;
  int buffer_offset = 0;
  unsigned char *ptr;
  unsigned char buf[MPU_INST_LEN];
  int buf_size;
  int xarg_flag = 0;
  //  int error_code;
  unsigned int pc;
  int meta_data = 0;
  inst_t *tmp_inst;

  assert((ctxt != NULL) && (buf != NULL) && (inst != NULL));
  
  sreg_number = -1;
  r_number = -1;
  
  current_arg = get_first_arg(inst);
  r_table = cg_get_reg_table(ctxt);
  i_table = cg_get_inst_table(ctxt);
  current_row = create_itable_row(inst);

  buf_size = MPU_INST_LEN;

  i_number = get_inst_number(i_table,current_row,inst); 
  IOPASM_FREE(current_row);

  if (i_number < 0) {
    iopasm_error(ctxt,inst,ERR_UNKNOWN_INSTRUCTION,
		 "The instruction %s is unknown by the ETRAX fs MPU backend",
		 get_inst(inst));
  }

  pc = get_inst_addr(inst);

  
  /* Check if we have a 64-bit wide instruction and align it if auto-alignment
   * is enabled. Do this by inserting an ".align 64" instruction prior to
   * this instruction */
  i_size = get_inst_len(i_table,i_number);
  if ((i_size == 8) &&(cg_check_machine_option(ctxt,"auto-align-64bit-inst",
					       OPT_ON))) {
    iopasm_verbose_msg("\"%s\" in %s line %d has been auto-aligned to " 
		       "a 64-bit boundary",get_inst(inst),
		       cg_get_file_by_index(ctxt,get_inst_file_index(inst)),
		       get_inst_line(inst));
    tmp_inst = synth_align_inst(64); 
    insert_inst(inst,tmp_inst); 
  } /* End of if ((i_size ==8) &&(cg_get_machine_option.... */
  

  /* Next we check if this we need to continue generating code. 
   * If we are in the intermediate pass we do not have any addresses for labels
   * so we simply set the size for the instruction and we return the size 
   * of the instruction (in bytes) so that PC can be updated. */
  if ((pass == INTERMEDIATE_PASS) && (check_reloc_flag(inst,RF_RELOCATE))) {
    set_inst_size(inst,i_size);
    return (buf_size < i_size) ? buf_size - i_size: i_size;
  }

  
  if (i_number < 0) {
    text_inst = (char *) get_inst(inst);
    iopasm_error(ctxt,inst,ERR_UNKNOWN_INSTRUCTION,
	       "The instruction %s is unknown by the ETRAX fs MPU backend",
	       text_inst); /* Termination point */
  }

  
  /* At this point we know one thing: We can generate the entire instruction. 
   * The reason for this might be that either one (or both) of the following 
   * conditions have been fullfilled:
   *
   * 1. pass == FINAL_PASS e.g. values for labels have been calculated.
   * 2. Instruction is not relocatable, so we generate the op-code. */
 
  
  /* Get instruction size in bytes - this is important as we need to know how
   * to place any immediates that we find. */
  i_size = get_inst_len(i_table,i_number);


  /* Check if we have a 64-bit wide instruction and emit warning if 
   * unaligned - if enabled. */
  if ((pc % 8 != 0) && (i_size == 8)) {
    iopasm_error(ctxt,inst,WARN_UNALIGNED_64BIT_INST,NULL);
  }


  /* First initialize the opcode and set it to the template */
  op_code = generate_opcode(ctxt,i_table,inst,i_number,pass);

  i = 0;

  while((current_arg != NULL) && (i<3)) { 
    
    arg_type = get_arg_type(current_arg);

    if (arg_type == ARG_STACK) {
      if ((check_expression_type(get_arg_stack(current_arg)) == ARG_NORMAL_REG) ||
	  (check_expression_type(get_arg_stack(current_arg)) == ARG_SPECIAL_REG)) {
	text_inst = (char *) get_inst(inst);
	iopasm_error(ctxt,inst,ERR_UNKNOWN_INSTRUCTION,
		     "The instruction %s is unknown by the ETRAX FS MPU backend",
		   text_inst); /* Termination point */
      }
	else {
	  temp_val = eval_arg_stack(ctxt,inst,i_number,current_arg,&meta_data,i,pass);
	} /* End of !if ((check_expression_type(g... */
      } /* End of if (arg_type == ARG_STACK) { */
   
    arg_text = get_arg_val(current_arg);
    arg_type = get_arg_type(current_arg);

    switch(i) {
      

    case 0: /* This is source */
      if ((i_size == 8) && ((arg_type == ARG_IMMEDIATE)  || (arg_type == ARG_LABEL))
	&& (xarg_flag == 0)) { 
	/* Handle 32-bit immediate as bits 63-32 of instruction */
	pack_32bit_immediate(ctxt,inst,buf,buf_size,current_arg);
	xarg_flag = 1;
	buffer_offset += 4;
      }
    
    else {
      if ((arg_type == ARG_NORMAL_REG) ||(arg_type == ARG_SPECIAL_REG)) {
      r_number = get_reg_number((reg_table_t *)etrax_fs_mpu_regs,arg_text);
      }

      temp_val = eval_arg(ctxt,inst,current_arg,i_number,r_number,0,pass,
			  get_src_regval,eval_src_immval);
      op_code |= temp_val;
    }
    break;
      

    case 1: /* This is any */
    if ((i_size == 8) && ((arg_type == ARG_IMMEDIATE)  || (arg_type == ARG_LABEL))
	&& (xarg_flag == 0)) { 
      /* Handle 32-bit immediate as bits 63-32 of instruction */
      pack_32bit_immediate(ctxt,inst,buf,buf_size,current_arg);
      xarg_flag = 1;
      buffer_offset += 4;
      
    }
    else {
      if ((arg_type == ARG_NORMAL_REG) ||(arg_type == ARG_SPECIAL_REG)) {
	r_number = get_reg_number((reg_table_t *)etrax_fs_mpu_regs,arg_text);
      }

      temp_val = eval_arg(ctxt,inst,current_arg,i_number,r_number,1,pass,
			  get_any_regval,eval_any_immval);
      op_code |= temp_val;
    }
      break;
      

    case 2: /* This is dst */
      if ((arg_type == ARG_NORMAL_REG) ||(arg_type == ARG_SPECIAL_REG)) {
	r_number = get_reg_number((reg_table_t *)etrax_fs_mpu_regs,arg_text);
      }
      
      temp_val = eval_arg(ctxt,inst,current_arg,i_number,r_number,2,pass,
			  get_dst_regval,eval_dst_immval);
      op_code |= temp_val;
      break;
      
    }
    
    i++;
    current_arg = get_next_arg(current_arg);
  }

/* If we get this far we have generated an instruction, so we can return
 * the proper size -if there is enough space in the buffer */
  if (buf_size < i_size) ret_val = buf_size - i_size;

  else {
    ret_val = i_size;
    op_code = htonl(op_code);
    ptr = (unsigned char *) &op_code;
    buf[buffer_offset] = ptr[0];
    buf[buffer_offset+1] = ptr[1];
    buf[buffer_offset+2] = ptr[2];
    buf[buffer_offset+3] = ptr[3];
    buffer_offset += 4;
  }
      
      set_inst_code(inst,buf,ret_val);
      set_inst_size(inst,ret_val);
  return ret_val;
}



/* ************************************************************************* */
/* PROTOTYPE   : void etrax_fs_mpu_finish(cg_context_t *ctxt,                */
/*                                        plist_t *a_list);                  */
/*                                                                           */
/* ARGUMENTS   : "ctxt" is the pointer to the current context and "a_list"   */
/*               is the instruction (parse) list.                            */
/*                                                                           */
/* FUNCTION    : Performs post-code generation operations on the generated   */
/*               code. In this case fills memory with NOPs and rotate the    */
/*               instructions to the correct byte order.                     */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void etrax_fs_mpu_finish(cg_context_t *ctxt,plist_t *a_list)
{
  unsigned char *buffer;
  unsigned char *i_code;
  unsigned int pc;
  unsigned int hi_pc;
  int i_size;
  unsigned int offset,row,row_addr;
  inst_t *first;

  /* Because of a brain-dead programmer (Mr. Johan J) we will need to
   * repack all code here!  Fortunately for us this function was right
   * there to take care of it :-)
   */
  
  assert ((ctxt != NULL) && (a_list != NULL));
  
  first = get_first_inst(a_list);
  buffer = cg_get_code_buffer(ctxt);
  hi_pc = cg_get_memory_size(ctxt);

  /* Fill memory with NOPS */
  pc = 0;
  while (pc < 4096) {
    buffer[pc] = 0x40; /* NOP is 0x4000000 */
    pc += 4;
  }
    while (first != NULL) {

      if ((get_inst_type(first) == INST_SEQ) ||
	  (get_inst_type(first) == INST_DWORD) ||
	  (get_inst_type(first) == INST_WORD) ||
	  (get_inst_type(first) == INST_BYTE)) {

	pc = get_inst_addr(first);
	i_size = get_inst_size(first);
	i_code = get_inst_code(first);
	
	row = pc / 8; 
	row_addr = pc % 8;
	offset = row * 8 + (4 - row_addr);
	
	if (i_size == 4) { /* Simpe 32-bit wide instruction - just copy */
	  memcpy(&buffer[offset],i_code,i_size);
	}

	else if ((i_size == 8) && ((pc % 8) != 0)) { /* Unaligned 64-bit */
	  memcpy(&buffer[offset],&i_code[4],4);

	  pc += 4;
	  row = pc / 8; 
	  row_addr = pc % 8;
	  offset = row * 8 + (4 - row_addr);
	  memcpy (&buffer[offset],i_code,4);
	}

	else { /* 8-bit instruction aligned - just copy */
	  offset -= 4; /* We will use entire row! */
	  memcpy(&buffer[offset],i_code,i_size);
	}

      }

      first = get_next_inst(first);
    }

}



/* ************************************************************************* */
/* PROTOTYPE   : plist_t *etrax_fs_mpu_fsm_optimizer(cg_context_t *ctxt,     */
/*                                                   plist_t *a_list);       */
/*                                                                           */
/* ARGUMENTS   : "ctxt" is a pointer to a valid context and "a_list" is the  */
/*               instruction list to optimize.                               */
/*                                                                           */
/* FUNCTION    : Optimize an instruction list. (this does not happen in the  */
/*               MPU as FSM instructions are not supported by the MPU..      */
/*                                                                           */
/* RETURN VALUE: The newly optimized list (in this case the same as the      */
/*               inpt list "a_list"                                          */
/* ************************************************************************* */
plist_t *etrax_fs_mpu_fsm_optimizer(cg_context_t *ctxt,plist_t *a_list)
{

  assert((ctxt != NULL) && (a_list != NULL));
  IOPASM_INFO_MESSAGE("Cannot optimize MPU code -FSM not supported");
  return a_list;
}
