/* ************************************************************************* */
/* etrax_fs_spu_be.c : implementation of ETRAX FS SPU specific code          */
/*                     generation backend.                                   */
/*                                                                           */
/*    Copyright 2002-2006 Axis Communications AB.                            */
/*                                                                           */
/*                                                                           */
/*    This program is free software; you can redistribute it and/or modify   */
/*    it under the terms of the GNU General Public License as published by   */
/*    the Free Software Foundation; either version 2 of the License, or      */
/*    (at your option) any later version.                                    */
/*                                                                           */
/*    This program is distributed in the hope that it will be useful,        */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*    GNU General Public License for more details.                           */
/*                                                                           */
/*    You should have received a copy of the GNU General Public License      */
/*    along with this program; if not, write to the Free Software            */
/*    Foundation, Inc., 59 Temple Place - Suite 330,                         */
/*    Boston, MA 02111-1307, USA.                                            */
/*                                                                           */
/*                                                                           */
/*    Send bug reports and inquiries to Johan.Jorgensen@axis.com             */
/* ************************************************************************* */



/* ************************************************************************* */
/*                      INCLUDES USED BY THIS MODULE                         */
/* ************************************************************************* */
#include <assert.h>
#include <netinet/in.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "cg_fe.h"
#include "debug.h"
#include "dyn_array.h"
#include "global_rules.h"
#include "etrax_fs_spu_be.h"
#include "iopasm_err.h"
#include "i_table_types.h"
#include "i_table.h"
#include "labels.h"
#include "p_list.h"
#include "stack_eval.h"



/* ************************************************************************* */
/*                        Definitions local to this module                   */
/* ************************************************************************* */
#define SPU_SEQ_LEN 8  /* Maximum length of a sequential instruction (bytes) */
#define SPU_FSM_LEN 32 /* Maximum length of an FSM instruction */

#define SET_GO_SEQ(CODE) (CODE[0] = CODE[0] | 0x80)
#define IS_GO_SEQ_SET(CODE) (CODE[0] && 0x80)
#define SET_SEQ(CODE,VAL) (CODE[0] = (VAL) ? (CODE[0]|0x40): (CODE[0]&0xBF))
#define SET_DO_SEQ(CODE) (CODE[0] = CODE[0] | 0x20)
#define SET_TIMER(CODE,VAL) (CODE[0] = (VAL) ? (CODE[0]|0x10): (CODE[0]&0xEF))
#define SET_STATE_COUNT(CODE,TRANS) (CODE[0] = CODE[0] | (TRANS & 0x0F))
#define SET_INV_STROBE(CODE) (CODE[1] = CODE[1] | 0x40)
#define SET_STROBE(CODE,VAL) (CODE[1] = CODE[1] | ((VAL & 0x3) << 4))
#define SET_SEL_INPUTS(CODE,VAL) (CODE[2] = CODE[2] | (VAL & 0x0F))
#define SET_SEL_OUTPUTS(CODE,VAL) (CODE[3] = (VAL & 0xFF));
#define SET_ONLY(CODE) (CODE[1] = CODE[1] | 0x80)
#define SET_FSM_HALT(CODE) (CODE[1] = CODE[1] | 0x40)
#define SET_FSM_32BIT_SEQ(CODE) (CODE[0] = CODE[0] = CODE[0] | 0x58)
#define SET_REUPDATE(CODE) (CODE[1] = CODE[1] | 0x10)
#define SET_TIMER_KEEP(CODE) (CODE = CODE | 0x40)
#define SET_HIGH_ADDRESS_BIT(CODE,VAL) (CODE[1] = CODE[1] | ((VAL & 0x1) << 5))
#define SET_SRC_META_OP(CODE,OP) (CODE = CODE | (OP << 16))
#define SET_ANY_META_OP(CODE,OP) (CODE = CODE | (OP << 20))
#define SET_SRC_META_REG(CODE,REG) (CODE = CODE | (REG & 0xFF))
#define SET_ANY_META_REG(CODE,REG) (CODE = CODE | (((REG & 0xFF)<<8) & 0xFF00))
#define GET_ANY_META_REG(CODE) ((CODE & 0xFF00) >> 8)
#define GET_SRC_META_REG(CODE) ((CODE & 0xFF))
#define GET_ANY_META_OP(CODE) ((CODE & 0xF00000) >> 20)
#define GET_SRC_META_OP(CODE) ((CODE & 0xF0000) >> 16)
#define SET_OR_MASK(CODE) (CODE | 0x40)
#define SET_AND_MASK(CODE) (CODE)
#define SET_MASK_RS(CODE) (CODE | 0x20)
#define SET_MASK_RN(CODE) (CODE | 0x10)
#define SET_MASK_REGISTER(CODE,REG) (CODE | ((REG & 0xF) << 6))
#define META_OP_OR (0x1)
#define META_OP_AND (0x2)
#define META_OP_NOP (0x00)



/* ************************************************************************* */
/*                              CONSTANT VARIABLES                           */
/* ************************************************************************* */



/* ************************************************************************* */
/*         Static prototypes defined and implemented in this module          */
/* ************************************************************************* */
static int and_operator(cg_context_t *ctxt,void *meta_buffer,int mbuf_len,
			void *buffer,int buf_len,inst_t *inst,int inum,
			arg_t *arg1,arg_t *arg2,int arg_no);

static int or_operator(cg_context_t *ctxt,void *meta_buffer,int mbuf_len,
		       void *buffer,int buf_len,inst_t *inst,int inum,
		       arg_t *arg1,arg_t *arg2,int arg_no);

static int fsmq_rule_func(cg_context_t *ctxt,inst_t *inst,int i_num,
			  unsigned char *buf,int buf_size);

static int rwq_rule_func(cg_context_t *ctxt,inst_t *inst,int i_num,
			 unsigned char *buf,int buf_size);

static int rrq_rule_func(cg_context_t *ctxt,inst_t *inst,int i_num,
			 unsigned char *buf,int buf_size);

static unsigned int eval_arg(cg_context_t *ctxt,inst_t *inst,arg_t *arg,
			     int i_num,int *meta_data,int arg_no,int pass,
			     unsigned int (*eval_reg_func)(cg_context_t *,
							   inst_t *,
							   int,int),

			     unsigned int (*eval_imm_func)(cg_context_t *,
							   inst_t *,
							   int,
							   char *));



/* ************************************************************************* */
/* SPU Rule sets - Define all rules that you want to use here                */
/* ************************************************************************* */
static const rule_set_t rwqck = {
  "RWQ destination check and delay-slot rule",
  0, { rwq_rule_func,
       validate_delay_slot,
       NULL }
};

static const rule_set_t rrqck = {
  "RRQ destination check and delay-slot rule",
  0, { rrq_rule_func,
       validate_delay_slot,
       NULL }
};

static const rule_set_t rwck = {
  "RW destination check and delay-slot rule",
  0, { rwq_rule_func, 
       validate_delay_slot, 
       NULL }
};

static const rule_set_t rrck = {
  "RR(Q) source check and delay-slot rule",
  0, { rrq_rule_func, 
       validate_delay_slot, 
       validate_dest_register,
       NULL } 
};
  
static const rule_set_t dsok = {
  "Rule to check prescense of a delay slot",
  0, { validate_delay_slot,
       NULL }
};

static const rule_set_t nods = { 
  "Rule to check prescence of delay slot & register usage",
  0, { validate_delay_slot,
       validate_dest_register,
       NULL }
};

static const rule_set_t evnt = {
  "Rule to check forwarding for Move to/from event registers",
  0, 
  { validate_dest_register,
    NULL }
};
static const rule_set_t fsmq = {
  "Recalculate fsmq address",
  0, { validate_delay_slot,
       fsmq_rule_func,
       NULL }
};



/* ************************************************************************* */
/*                      Operators implemented in backend                     */
/*                                                                           */
/* All operators in the table below are required. If this backend does not   */
/* support a particular operator, set the associated operator_func to NULL.  */ 
/* ************************************************************************* */
static const cg_operator_t etrax_fs_spu_operators[] = { 
  { "init", NULL }, /* Initialization of operators - REQUIRED */
  { "+", NULL },    /* Addition - REQUIRED */
  { "-", NULL },    /* Subtraction - REQUIRED */
  { "*", NULL },    /* Multiplication - REQUIRED */
  { "/", NULL },    /* Division - REQUIRED */
  { "<<", NULL },   /* Shift left - REQUIRED */
  { ">>", NULL },   /* Shift Right - REQUIRED */
  { "^", NULL },    /* Xor - REQUIRED */
  { "|", or_operator },    /* Or - REQUIRED */
  { "&", and_operator },    /* And - REQUIRED */
  { "<", NULL },    /* Less than - REQUIRED */
  { ">", NULL },    /* Greater than - REQUIRED */
  { "end", NULL },  /* End  - REQUIRED */
  { NULL, NULL }    /* Table must end like this */
};



/* ************************************************************************* */
/*     SPU instruction table - describes how to generate SPU instructions    */
/* ************************************************************************* */
static const inst_table_t etrax_fs_spu_seq[] = {
  { "add",   3,R,R,R, 20, 15, 25,  5,  5,  5,  4, 0x00,0x00,0x00,0x01, &dsok },
  { "addq",  3,R,I,R, 20, 15, 25,  5, 16,  5,  4, 0x04,0x00,0x00,0x00, &dsok }, 
  { "addq",  3,I,R,R, 15, 20, 25, 16,  5,  5,  4, 0x04,0x00,0x00,0x00, &dsok }, 
  
  { "addx",  3,I,R,R, 31, 15, 25, 32,  5,  5,  8, 0x04,0x00,0x00,0x00, NULL },
  { "addx",  3,R,I,R, 15, 31, 25,  5, 32,  5,  8, 0x04,0x00,0x00,0x00, NULL },

  { "and",   3,R,R,R, 20, 15, 25,  5,  5,  5,  4, 0x00,0x00,0x00,0x02, &dsok },
  { "andq",  3,R,I,R, 20, 15, 25,  5, 16,  5,  4, 0x08,0x00,0x00,0x00, &dsok }, 
  { "andq",  3,I,R,R, 15, 20, 25, 16,  5,  5,  4, 0x08,0x00,0x00,0x00, &dsok },
  
  { "andx",  3,R,I,R, 20, 31, 25,  5, 32,  5,  8, 0x08,0x00,0x00,0x00, NULL }, 
  { "andx",  3,I,R,R, 31, 20, 25, 32,  5,  5,  8, 0x08,0x00,0x00,0x00, NULL },

  { "andqh", 3,R,I,R, 20, 15, 25,  5, 16,  5,  4, 0x14,0x00,0x00,0x00, &dsok }, 
  { "andqh", 3,I,R,R, 15, 20, 25, 16,  5,  5,  4, 0x14,0x00,0x00,0x00, &dsok }, 
  { "ba",    1,I,0,0, 15,  0,  0, 16,  0,  0,  4, 0xE0,0x00,0x00,0x00, &nods },
  { "bar",   1,R,0,0, 20,  0,  0,  5,  0,  0,  4, 0xF0,0x00,0x00,0x00, &nods }, 
  { "bbc",   3,R,I,I, 20, 25, 15,  5,  5, 16,  4, 0x80,0x00,0x00,0x00, &nods },
  { "bbs",   3,R,I,I, 20, 25, 15,  5,  5, 16,  4, 0x90,0x00,0x00,0x00, &nods },
  { "bnz",   2,R,I,0, 20, 15,  0,  5, 16,  0,  4, 0xC0,0x00,0x00,0x00, &nods },
  { "bmi",   2,R,I,0, 20, 15,  0,  5, 16,  0,  4, 0x93,0xE0,0x00,0x00, &nods },
  { "bpl",   2,R,I,0, 20, 15,  0,  5, 16,  0,  4, 0x83,0xE0,0x00,0x00, &nods },
  { "halt",  0,0,0,0,  0,  0,  0,  0,  0,  0,  4, 0xB0,0x00,0x00,0x00, &dsok },
  { "bz",    2,R,I,0, 20, 15,  0,  5, 16,  0,  4, 0xD0,0x00,0x00,0x00, &nods },
  { "fsm",   0,0,0,0,  0,  0,  0,  0,  0,  0,  4, 0xA0,0x11,0x00,0x00, &nods }, 
  { "fsmq",  1,I,0,0, 11,  0,  0, 12,  0,  0,  4, 0xA8,0x00,0x00,0x00, &nods },
  { "lsl",   3,R,R,R, 20, 15, 25,  5,  5,  5,  4, 0x00,0x00,0x00,0x03, &dsok },
  { "lslq",  3,R,I,R, 20, 13, 25,  5, 14,  5,  4, 0x0C,0x00,0x00,0x00, &dsok },
  { "lsr",   3,R,R,R, 20, 15, 25,  5,  5,  5,  4, 0x00,0x00,0x00,0x04, &dsok },
  { "lsrq",  3,R,I,R, 20, 13, 25,  5, 14,  5,  4, 0x10,0x00,0x00,0x00, &dsok },
  { "move",  2,R,R,0, 20, 25,  0,  5,  5,  0,  4, 0x00,0x00,0x00,0x05, &dsok },
  { "move",  2,S,R,0, 20, 25,  0,  5,  5,  0,  4, 0x00,0x00,0x00,0x08, &evnt },
  { "move",  2,R,S,0, 20, 25,  0,  5,  5,  0,  4, 0x00,0x00,0x00,0x07, &evnt },
  { "moveh", 3,R,I,R, 20, 15, 25,  5, 16,  5,  4, 0x20,0x00,0x00,0x00, &dsok },
  { "movel", 3,R,I,R, 20, 15, 25,  5, 16,  5,  4, 0x1C,0x00,0x00,0x00, &dsok },
  { "moveq", 2,I,R,0, 15, 25,  0, 16,  5,  0,  4, 0x18,0x00,0x00,0x00, &dsok },

  { "movex", 2,I,R,0, 31, 25,  0, 32,  5,  0,  8, 0x18,0x00,0x00,0x00, NULL },

  { "nop",   0,0,0,0,  0,  0,  0,  0,  0,  0,  4, 0x00,0x00,0x00,0x00, &dsok },
  { "not",   2,R,R,0, 20, 25,  0,  5,  5,  0,  4, 0x00,0x00,0x00,0x0F, &dsok },
  { "or",    3,R,R,R, 20, 15, 25,  5,  5,  5,  4, 0x00,0x00,0x00,0x0A, &dsok },
  { "orq",   3,R,I,R, 20, 15, 25,  5, 16,  5,  4, 0x28,0x00,0x00,0x00, &dsok },
  { "orq",   3,I,R,R, 15, 20, 25, 16,  5,  5,  4, 0x28,0x00,0x00,0x00, &dsok },

  { "orx",   3,R,I,R, 20, 31, 25,  5, 32,  5,  8, 0x28,0x00,0x00,0x00, NULL },
  { "orx",   3,I,R,R, 31, 20, 25, 32,  5,  5,  8, 0x28,0x00,0x00,0x00, NULL },

  { "rr",    2,R,R,0,  4, 25,  0,  0,  5,  0,  4, 0x48,0x00,0x00,0x00, &rrck },
  { "rr",    2,I,R,0, 10, 25,  0, 11,  5,  0,  4, 0x40,0x00,0x00,0x00, &rrck },
  { "rrm",   3,I,R,R, 10, 20, 25, 11,  5,  5,  4, 0x44,0x00,0x00,0x00, &rrck },
  { "rrm",   3,R,R,R,  0, 20, 25,  0,  5,  5,  4, 0x4C,0x00,0x00,0x00, &rrck },
  { "rrmh",  2,I,I,0, 10, 26,  0, 11, 16,  0,  4, 0x60,0x00,0x00,0x00, &rrqck},
  { "rrmh",  2,R,I,0,  0, 26,  0,  0, 16,  0,  4, 0x68,0x00,0x00,0x00, &rrqck},
  { "rrmq",  2,I,I,0, 10, 26,  0, 11, 16,  0,  4, 0x30,0x00,0x00,0x00, &rrqck},
  { "rrmq",  2,R,I,0,  0, 26,  0,  0, 16,  0,  4, 0x38,0x00,0x00,0x00, &rrqck},

  { "rrmx",  2,I,I,0, 10, 31,  0, 11, 32,  0,  8, 0x30,0x00,0x00,0x00, &rrqck},
  { "rrmx",  2,R,I,0,  0, 31,  0,  0, 32,  0,  8, 0x38,0x00,0x00,0x00, &rrqck},

  { "rw",    2,R,R,0, 20,  0,  0,  5,  0,  0,  4, 0x58,0x00,0x00,0x00, &rwck },
  { "rw",    2,R,I,0, 20, 10,  0,  5, 11,  0,  4, 0x50,0x00,0x00,0x00, &rwck },
  { "rwq",   2,I,R,0, 26, 15,  0, 16,  0,  0,  4, 0x78,0x00,0x00,0x00, &rwqck},
  { "rwq",   2,I,I,0, 26, 10,  0, 16, 11,  0,  4, 0x70,0x00,0x00,0x00, &rwqck},

  { "rwx",   2,I,I,0, 31, 10,  0, 32, 11,  0,  8, 0x70,0x00,0x00,0x00, &rwqck},
  { "rwx",   2,I,R,0, 31, 15,  0, 32,  5,  0,  8, 0x78,0x00,0x00,0x00, &rwqck},

  { "ssl",   3,R,R,R, 20, 15, 25,  5,  5,  5,  4, 0x00,0x00,0x00,0x0D, &dsok },
  { "sslq",  3,R,I,R, 20, 13, 25,  5, 14,  5,  4, 0x0C,0x00,0x80,0x00, &dsok },
  { "ssr",   3,R,R,R, 20, 15, 25,  5,  5,  5,  4, 0x00,0x00,0x00,0x0E, &dsok },
  { "ssrq",  3,R,I,R, 20, 13, 25,  5, 14,  5,  4, 0x10,0x00,0x80,0x00, &dsok },
  { "sub",   3,R,R,R, 20, 15, 25,  5,  5,  5,  4, 0x00,0x00,0x00,0x09, &dsok },
  { "subq",  3,R,I,R, 20, 15, 25,  5, 16,  5,  4, 0x24,0x00,0x00,0x00, &dsok },
  
  { "subx",  3,R,I,R, 20, 31, 25,  5, 32,  5,  8, 0x24,0x00,0x00,0x00, NULL },

  { "swap",  2,R,R,0, 20, 25,  0,  5,  5,  0,  4, 0x10,0x00,0x40,0x00, &dsok },
  { "swsrq", 3,R,I,R, 20, 15, 25,  5, 16,  5,  4, 0x10,0x00,0x40,0x00, &dsok },
  { "xor",   2,R,R,0, 20, 25,  0,  5,  5,  0,  4, 0x00,0x00,0x00,0x0C, &dsok },
  { "xor",   3,R,R,R, 20, 15, 25,  5,  5,  5,  4, 0x00,0x00,0x00,0x0B, &dsok },

  { "xorq",  3,R,I,R, 20, 15, 25,  5, 16,  5,  4, 0x2C,0x00,0x00,0x00, &dsok },
  { "xorq",  3,I,R,R, 15, 20, 25, 16,  5,  5,  4, 0x2C,0x00,0x00,0x00, &dsok },

  { "xorx",  3,R,I,R, 20, 31, 25,  5, 32,  5,  8, 0x2C,0x00,0x00,0x00, NULL },
  { "xorx",  3,I,R,R, 31, 20, 25, 32,  5,  5,  8, 0x2C,0x00,0x00,0x00, NULL },

  { NULL,    0,0,0,0,  0,  0,  0,  0,  0,  0,  0, 0x00,0x00,0x00,0x00, NULL }};
  


/* ************************************************************************* */
/*                            SPU Register values                            */
/* ************************************************************************* */
static const reg_table_t etrax_fs_spu_regs[] = {
  /* Register Base  Code Generation Architectural  */
  /* Name     Value interpretation  interpretation */
  {  "r0",      0, ARG_NORMAL_REG, ARG_NORMAL_REG }, 
  {  "r1",      1, ARG_NORMAL_REG, ARG_NORMAL_REG }, 
  {  "r2",      2, ARG_NORMAL_REG, ARG_NORMAL_REG }, 
  {  "r3",      3, ARG_NORMAL_REG, ARG_NORMAL_REG }, 
  {  "r4",      4, ARG_NORMAL_REG, ARG_NORMAL_REG }, 
  {  "r5",      5, ARG_NORMAL_REG, ARG_NORMAL_REG }, 
  {  "r6",      6, ARG_NORMAL_REG, ARG_NORMAL_REG }, 
  {  "r7",      7, ARG_NORMAL_REG, ARG_NORMAL_REG },
  {  "r8",      8, ARG_NORMAL_REG, ARG_NORMAL_REG }, 
  {  "r9",      9, ARG_NORMAL_REG, ARG_NORMAL_REG }, 
  { "r10",     10, ARG_NORMAL_REG, ARG_NORMAL_REG }, 
  { "r11",     11, ARG_NORMAL_REG, ARG_NORMAL_REG }, 
  { "r12",     12, ARG_NORMAL_REG, ARG_NORMAL_REG }, 
  { "r13",     13, ARG_NORMAL_REG, ARG_NORMAL_REG }, 
  { "r14",     14, ARG_NORMAL_REG, ARG_NORMAL_REG }, 
  { "r15",     15, ARG_NORMAL_REG, ARG_NORMAL_REG },
  { "p0",      16, ARG_NORMAL_REG, ARG_SPECIAL_REG }, 
  { "p1",      17, ARG_NORMAL_REG, ARG_SPECIAL_REG }, 
  { "p2",      18, ARG_NORMAL_REG, ARG_SPECIAL_REG }, 
  { "p3",      19, ARG_NORMAL_REG, ARG_SPECIAL_REG },
  { "p4",      20, ARG_NORMAL_REG, ARG_SPECIAL_REG }, 
  { "p5",      21, ARG_NORMAL_REG, ARG_SPECIAL_REG }, 
  { "p6",      22, ARG_NORMAL_REG, ARG_SPECIAL_REG }, 
  { "p7",      23, ARG_NORMAL_REG, ARG_SPECIAL_REG },
  { "p8",      24, ARG_NORMAL_REG, ARG_SPECIAL_REG }, 
  { "p9",      25, ARG_NORMAL_REG, ARG_SPECIAL_REG }, 
  { "p10",     26, ARG_NORMAL_REG, ARG_SPECIAL_REG }, 
  { "p11",     27, ARG_NORMAL_REG, ARG_SPECIAL_REG }, 
  { "p12",     28, ARG_NORMAL_REG, ARG_SPECIAL_REG }, 
  { "p13",     29, ARG_NORMAL_REG, ARG_SPECIAL_REG }, 
  { "p14",     30, ARG_NORMAL_REG, ARG_SPECIAL_REG }, 
  { "p15",     31, ARG_NORMAL_REG, ARG_SPECIAL_REG },
  { "seqpc",   16, ARG_NORMAL_REG, ARG_SPECIAL_REG }, 
  { "fsmpc",   17, ARG_NORMAL_REG, ARG_SPECIAL_REG }, 
  { "rega",    18, ARG_NORMAL_REG, ARG_SPECIAL_REG }, 
  { "fsm3_0",  19, ARG_NORMAL_REG, ARG_SPECIAL_REG },
  { "fsm7_4",  20, ARG_NORMAL_REG, ARG_SPECIAL_REG }, 
  { "gout",    21, ARG_NORMAL_REG, ARG_SPECIAL_REG }, 
  { "b0out",   22, ARG_NORMAL_REG, ARG_SPECIAL_REG }, 
  { "b1out",   23, ARG_NORMAL_REG, ARG_SPECIAL_REG },
  { "gin",     24, ARG_NORMAL_REG, ARG_SPECIAL_REG }, 
  { "b0in",    25, ARG_NORMAL_REG, ARG_SPECIAL_REG }, 
  { "b1in",    26, ARG_NORMAL_REG, ARG_SPECIAL_REG }, 
  { "statin",  27, ARG_NORMAL_REG, ARG_SPECIAL_REG },
  { "trigger", 28, ARG_NORMAL_REG, ARG_SPECIAL_REG }, 
  { "immhi",   28, ARG_NORMAL_REG, ARG_SPECIAL_REG }, 
  { "wsts",    29, ARG_NORMAL_REG, ARG_SPECIAL_REG }, 
  { "specs",    30, ARG_NORMAL_REG, ARG_SPECIAL_REG }, 
  { "statin",  27, ARG_NORMAL_REG, ARG_SPECIAL_REG },
  { "index",   31, ARG_NORMAL_REG, ARG_SPECIAL_REG },
  { "e0",       0, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "e1",       1, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "e2",       2, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "e3",       3, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "e4",       4, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "e5",       5, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "e6",       6, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "e7",       7, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "e8",       8, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "e9",       9, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "e10",      10, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "e11",      11, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "e12",      12, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "ecfg0",     0, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "ecfg1",     1, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "ecfg2",     2, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "ecfg3",     3, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "emask0",    4, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "emask1",    5, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "emask2",    6, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "emask3",    7, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "eval0",     8, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "eval1",     9, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "eval2",    10, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "eval3",    11, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { "eaddr",    12, ARG_SPECIAL_REG, ARG_SPECIAL_REG },
  { NULL,        0, 0, 0 }
}; /* Matches: static const reg_table_t etrax_fs_spu__regs[] */



/* ************************************************************************* */
/*                    MEMORY INFORMATION ETRAX FS SPU                        */
/* ************************************************************************* */
static const mem_table_t etrax_fs_spu_template = 
  { 4, 256, 64, 0, 0, 65535, 4095, 511, NULL, NULL, NULL, NULL };



/* ************************************************************************* */
/*                   SPU-SPECIFIC C MACROS & CONSTANTS                       */
/* ************************************************************************* */
static const char *etrax_fs_spu_cpp_macros[] = {
  "-D__SPU__",
  "-D__ETRAX_FS__",
  "-DINV16(X)=(0x0000FFFF & (~X))",
  "-DINV32(X)=(0xFFFFFFFF & (~X))",
  NULL };



/* ************************************************************************* */
/* PROTOTYPE   : static int fsmq_rule_func(cg_context_t *ctxt,inst_t *inst,  */
/*                                         int i_num, unsigned char *buf,    */
/*                                         int buf_size);                    */
/*                                                                           */
/* INPUT PARM's: "ctxt" is a valid context, "inst" is the instruction to be  */
/*               processed. "i_num" is the number of the instruction "inst", */
/*               "buf" is the buffer where code should be stored and         */
/*               "buf_size" is the size of "buf".                            */
/*                                                                           */
/* FUNCTION    : Rule function associated with the "fsmq" instruction. Used  */
/*               to recalculate the address value of an FSM-label..          */
/*                                                                           */
/* RETURN VALUE: returns 1 on success. If an error occurs it is reported by  */
/*               iopasm_error and the assembler exits.                       */
/* ************************************************************************* */
static int fsmq_rule_func(cg_context_t *ctxt,inst_t *inst,int i_num,
			  unsigned char *buf,int buf_size)
{
  array_t *labels;
  label_t *temp_label,*real_label;
  inst_table_t *inst_table;
  unsigned int temp_val;
  char *arg_val;
  char tmp_string[MAX_STRLEN];
  arg_t *arg;
  int i;
  assert ((ctxt != NULL) && (inst != NULL) && 
	  (buf != NULL) && (buf_size > 0));

  labels = cg_get_label_array(ctxt);
  inst_table = cg_get_inst_table(ctxt);

  arg = get_first_arg(inst);
  
  if (get_arg_type(arg) == ARG_LABEL) {
    arg_val = get_arg_val(arg);
    temp_label = create_label(arg_val,0,SCOPE_NO_SCOPE);
    real_label = (label_t *) find_array_elem(labels,(void *)temp_label);
    
    if ((real_label == NULL) && 
	(!check_option("allow-undefined-symbols",(int *)1,OPT_INTEGER)))
      { /* Cannot find label - return error and exit */
	iopasm_error(ctxt,inst,ERR_UNDEFINED_LABEL,"Label is %s\n",arg_val);
      }
    
    else if (real_label == NULL) { /* Undefined symbols allowed and no label! */
      temp_val = 0;
    }

    else { /* Symbol is defined */
      temp_val = get_label_value(ctxt,real_label);
    }
    /* addresses always stored in bytes */
    temp_val = (temp_val > 0) ? (temp_val / 8): 0;

    memset(tmp_string,0,MAX_STRLEN);
    arg_val = get_arg_val(arg);
    IOPASM_FREE(arg_val);
    i = snprintf(tmp_string,MAX_STRLEN,"%u",temp_val);

    arg_val = (char *) strdup(tmp_string);
    assert(arg_val != NULL);
    set_arg_val(arg,arg_val,strlen(tmp_string));
    set_arg_type(arg,ARG_IMMEDIATE);

    return 1;
  }

  else if (get_arg_type(arg) == ARG_IMMEDIATE) {
    arg_val = get_arg_val(arg);
    return 1;
  }
  
  else {
    iopasm_error(ctxt,inst,ERR_ILLEGAL_ARGUMENT,"Operand is: %s",arg_val);
  }
  return 1;
}



/* ************************************************************************* */
/* PROTOTYPE   : static int rwq_rule_func(cg_context_t *ctxt,inst_t *inst,   */
/*                                        int i_num,unsigned char *buf,      */
/*                                        int buf_size);                     */
/*                                                                           */
/* INPUT PARM's: "ctxt" is a valid context, "inst" is the instruction to be  */
/*               processed. "i_num" is the number of the instruction "inst", */
/*               "buf" is the buffer where code should be stored and         */
/*               "buf_size" is the size of "buf".                            */
/*                                                                           */
/* FUNCTION    : Check register usage and instruction sequencing for the     */
/*               "rwq" function.                                             */
/*                                                                           */
/* RETURN VALUE: Returns 1 on success. Errors are reported through           */
/*               calls to iopasm_error (which results in exit(...) being     */
/*               called).                                                    */
/* ************************************************************************* */
static int rwq_rule_func(cg_context_t *ctxt,inst_t *inst,int i_num,
			 unsigned char *buf,int buf_size)
{
  arg_t *first_arg;
  int len;
  unsigned int addr;
  char *arg_val;
  assert ((ctxt != NULL) && (inst != NULL) &&(buf != NULL) && (buf_size > 0));
  
  first_arg = get_first_arg(inst);
  first_arg = get_next_arg(first_arg); /* We can safely do this otherwise the
					* instruction would not have been 
					* found */
  
  len = get_arg_len(first_arg);
  arg_val = get_arg_val(first_arg);

  if (get_arg_type(first_arg) == ARG_STACK) {
      convert_stack_arg(ctxt,inst,first_arg,FINAL_PASS);
  }

  arg_val = get_arg_val(first_arg);

  if (get_arg_type(first_arg) == ARG_NORMAL_REG) {
    if ((!strcmp("rega",arg_val)) || (!strcmp("p2",arg_val))) {
      return 1;
    }
    
    else {
      iopasm_error(ctxt,inst,ERR_ILLEGAL_REGISTER_USE,
		   "The register %s may not be used here",arg_val);
    }
  }

  else if (get_arg_type(first_arg) == ARG_IMMEDIATE) {
    arg_val = get_arg_val(first_arg);
    len = get_arg_len(first_arg);
    addr = atol(arg_val);
    memset(arg_val,0,len);
    snprintf(arg_val,len+1,"%u",addr / 4);
    return 1;
  }

  else {
    iopasm_error(ctxt,inst,ERR_ILLEGAL_REGISTER_USE,
		 "The register %s may not be used here",arg_val);
  }
  
  return 1;

}



/* ************************************************************************* */
/* PROTOTYPE   : static int rrq_rule_func(cg_context_t *ctxt,inst_t *inst,   */
/*                                        int i_num,unsigned char *buf,      */
/*                                        int buf_size);                     */
/*                                                                           */
/* INPUT PARM's: "ctxt" is a valid context, "inst" is the instruction to be  */
/*               processed. "i_num" is the number of the instruction "inst", */
/*               "buf" is the buffer where code should be stored and         */
/*               "buf_size" is the size of "buf".                            */
/*                                                                           */
/* FUNCTION    : Check register usage and instruction sequencing for the     */
/*               "rrq" function.                                             */
/*                                                                           */
/* RETURN VALUE: Returns 1 on success. Errors are reported through           */
/*               calls to iopasm_error (which results in exit(...) being     */
/*               called).                                                    */
/* ************************************************************************* */
static int rrq_rule_func(cg_context_t *ctxt,inst_t *inst,int i_num,
			 unsigned char *buf,int buf_size)
{
  arg_t *first_arg;
  int len;
  unsigned int addr;
  char *arg_val;
  inst_t *prev_inst;
  inst_t *prev_inst2;
  char *prev_iname;
  char *prev_iname2;
  int state_found = 0;

  assert ((ctxt != NULL) && (inst != NULL) &&(buf != NULL) && (buf_size > 0));

  prev_inst = get_prev_inst(inst);

  while((prev_inst != NULL) &&
	(get_inst_type(prev_inst) != INST_SEQ) && 
	(state_found == 0)) {

    if (get_inst_type(prev_inst) == INST_STATE_END) {
      state_found = 1;
    }
    prev_inst = get_prev_inst(prev_inst);

  }
  prev_inst2 = get_prev_inst(prev_inst);

  if (state_found == 0) {
    prev_inst2 = get_prev_inst(prev_inst);
    
    while((prev_inst2 != NULL) &&
	  (get_inst_type(prev_inst2) != INST_SEQ) && 
	  (state_found == 0)) {
      if (get_inst_type(prev_inst) == INST_STATE_END) {
	prev_inst2 = prev_inst;
	break;
      }
      prev_inst2 = get_prev_inst(prev_inst2);
    }
    
  }
  if ((prev_inst != NULL) && (!state_found)) { /* Check that it is not rw something */
    prev_iname = get_inst(prev_inst);
    prev_iname2 = get_inst(prev_inst2);
    if ((prev_iname[0] == 'r') && (prev_iname[1] == 'w') && 
	(strcmp(prev_iname2,"ba")) && (strcmp(prev_iname2,"bar"))) {
      iopasm_error(ctxt,inst,ERR_ILLEGAL_INST_ORDER,
		 "The \"%s\" instruction may not be preceeded by \
the \"%s\" instruction", get_inst(inst),prev_iname);
    }
  }
  
  first_arg = get_first_arg(inst);
  arg_val = get_arg_val(first_arg);
  len = get_arg_len(first_arg);
  
  if (get_arg_type(first_arg) == ARG_STACK) {
    convert_stack_arg(ctxt,inst,first_arg,FINAL_PASS);
  }

  if (get_arg_type(first_arg) == ARG_IMMEDIATE) {
    arg_val = get_arg_val(first_arg);
    len = get_arg_len(first_arg);
    addr = atol(arg_val);
    memset(arg_val,0,len);
    snprintf(arg_val,len+1,"%u",addr / 4);
    return 1;
  }

  else if (get_arg_type(first_arg) == ARG_NORMAL_REG) {
    if ((!strcmp("rega",arg_val)) || (!strcmp("p2",arg_val))) {
      return 1;
    }
    else {
      iopasm_error(ctxt,inst,ERR_ILLEGAL_REGISTER_USE,
		   "The register %s may not be used here",arg_val);
    }
  }
  
  else {
    iopasm_error(ctxt,inst,ERR_ILLEGAL_REGISTER_USE,
		 "The register %s may not be used here",arg_val);
  }
  return 1;
}



/* ************************************************************************* */
/* PROTOTYPE   : static int check_masked_inst(inst_t *inst);                 */
/*                                                                           */
/* INPUT PARM's: "inst" points to a valid instruction                        */
/*                                                                           */
/* FUNCTION    : Check if the current instruction allows masked operands     */
/*                                                                           */
/* RETURN VALUE: 1 if masked operands are allowed -1 if not.                 */
/* ************************************************************************* */
static int check_masked_inst(inst_t *inst)
{
  char *valid_insts[] = { "add","and","lsl","lsr","move",
			  "or","sub", "xor","rxor","ssl", "ssr", NULL };
    int i = 0;
    int ret_val = -1;
    char *inst_string;
    
    assert(inst != NULL);
    
    inst_string = get_inst(inst);
    
    while(valid_insts[i] != NULL) {
      if (strcmp(inst_string,valid_insts[i]) == 0) {
	ret_val = 1;
      }

      i++;
    } /* End of while(valid_insts[i] != NULL) */
    
    return ret_val;
}


  
/* ************************************************************************* */
/* PROTOTYPE   : static int only_reg_check(char *val);                       */
/*                                                                           */
/* INPUT PARM's: "val" is a pointer to the name of a register.               */
/*                                                                           */
/* FUNCTION    : Check if a register name is allowed as operand to the only  */
/*               statement.                                                  */
/*                                                                           */
/* RETURN VALUE: 1 if the register is allowed -1 otherwise.                  */
/* ************************************************************************* */
static int only_reg_check(char *val) 
{
  int i=0;
  int hit = -1;
  char *allowed_regs[] = { "r0",   "r1",  "r2",  "r3", 
			   "r4",   "r5", " r6",  "r7",
			   "r8",   "r9", "r10", "r11", 
			   "r12", "r13", "r14", "r15",
			   "e12", "eaddr", NULL };

  while (allowed_regs[i] != NULL) {
    if (!strcmp(allowed_regs[i],val)) {
      hit = 1;
    }
    i++;
  }

  return hit;
}
  
 

/* ************************************************************************* */
/* PROTOTYPE   : static int or_operator(cg_context_t *ctxt,                  */
/*                                      void *meta_buffer,int mbuf_len,      */
/* 		                        void *buffer,int buf_len,            */
/*                                      inst_t *inst,int inum, arg_t *arg1,  */
/*                                      arg_t *arg2,int arg_no);             */
/*                                                                           */
/* INPUT PARM's: "ctxt" is a valid context. "meta_buffer" points to a buffer */
/*               with the length "mbuf_len" which is used to store info      */
/*               meta data such as mask register etc. "buffer" is the buffer */
/*               where the generated op-code is stored and "buf_len" is the  */
/*               length of that buffer. "inst" is the pointer to the current */
/*               instruction and "inum" is the number of that instruction.   */
/*               "arg1" points to the first operand for the "or" operation   */
/*               (performed by this function) and "arg2" is the second       */
/*               operand. "arg_no" is number of the argument associated with */
/*               the instruction.                                            */
/*                                                                           */
/* FUNCTION    : Evaluate and generate code for the "|" operand when         */
/*               registers are involved. SPU specific code.                  */
/*                                                                           */
/* RETURN VALUE: Returns 1 if successful. Errors are reported through        */
/*               iopasm_error(...)                                           */
/* ************************************************************************* */
static int or_operator(cg_context_t *ctxt,void *meta_buffer,int mbuf_len,
		       void *buffer,int buf_len,inst_t *inst,int inum,
		       arg_t *arg1,arg_t *arg2,int arg_no)
{
  
  int arg1_type,arg2_type;
  char *arg2_val,*arg1_val;
  int *op_code,rnum;
  int *meta_data;
  int reg_no;
  assert((buffer != NULL) && (buf_len > 0) && (inst != NULL) && 
	 (arg1 != NULL) && (arg2 != NULL));

  op_code = (int *) buffer;
  meta_data = (int *) meta_buffer;

  arg1_type = get_arg_type(arg1);
  arg2_type = get_arg_type(arg2);
  arg1_val = get_arg_val(arg1);
  arg2_val = get_arg_val(arg2);

  if (check_masked_inst(inst) == -1) {
    iopasm_error(ctxt,inst, ERR_UNKNOWN_INSTRUCTION,
		 "The %s instruction does not allow \
masked registers as  operands.",get_inst(inst));
  }

  arg1_type = get_arg_type(arg1);
  arg2_type = get_arg_type(arg2);
  arg1_val = get_arg_val(arg1);
  arg2_val = get_arg_val(arg2);

  if (arg2_type != ARG_NORMAL_REG) {
    iopasm_error(ctxt,inst,ERR_ILLEGAL_REGISTER_USE,
		 "The register %s may not be used here.\
 It is not a general register",arg2_val);
  }
      
  reg_no = get_reg_number((reg_table_t *)etrax_fs_spu_regs,arg2_val);
   
  if ((arg1_type != ARG_NORMAL_REG) &&
      (arg1_type != ARG_SPECIAL_REG)) {
    iopasm_error(ctxt,inst,ERR_ILLEGAL_REGISTER_USE,
		 "The register %s may not be used here.\
 It is neither a general nor a special register",arg1_val);
  }

  rnum = get_reg_number((reg_table_t *)etrax_fs_spu_regs,arg1_val);

  /* Let us check the op_code to see if what/if any mask instruction
   * has been specified 
   */
  switch(arg_no) {


  case 0:
    (*op_code) = (*op_code) | 0x20; /* Set mask on RS */
    (*op_code) = (*op_code) | ((reg_no & 0xF) << 7); /* Set mask register */
    (*op_code) = (*op_code) | 0x40; /* Select Or-operation */
    (*op_code) = (*op_code) | get_src_regval(ctxt,inst,
					     inum,rnum);
    SET_SRC_META_OP((*meta_data),META_OP_OR);
    SET_SRC_META_REG((*meta_data),reg_no);
    break;
    

  case 1:
    if (((GET_SRC_META_OP((*meta_data)) != META_OP_OR) ||
	 (GET_SRC_META_REG((*meta_data)) != reg_no)) && 
	(GET_SRC_META_OP(*meta_data) != META_OP_NOP)) {
      iopasm_error(ctxt,inst,ERR_ILLEGAL_MASK_REG_USE,
		   "The register %s may not be used as a mask register",
		   arg2_val);
    } /* End of if (GET_SRC_META_OP((* ... */
    (*op_code) = (*op_code) | 0x10; /* Set mask on Rn */
    (*op_code) = (*op_code) | ((reg_no & 0xF) << 7); /* Set mask register */
    (*op_code) = (*op_code) | 0x40; /* Select Or-operation */
    (*op_code) = (*op_code) | get_any_regval(ctxt,inst,
					     inum,rnum);
    SET_ANY_META_OP((*meta_data),META_OP_OR);
    SET_ANY_META_REG((*meta_data),reg_no);
/*     SET_MASK_RN((*op_code)); */
/*     SET_MASK_REGISTER((*op_code),reg_no); */
    break;


  case 2:
    iopasm_error(ctxt,inst,ERR_ILLEGAL_MASK_REG_USE,
		 "Masked register operands (%s | %s) are not allowed as \
3rd operand to the %s instruction",arg1_val,arg2_val,get_inst(inst)); 
    /* program exits here */
    break;
  } /* End of switch(arg_no) */
  return 1;
}



/* ************************************************************************* */
/* PROTOTYPE   : static int and_operator(cg_context_t *ctxt,                 */
/*                                       void *meta_buffer,int mbuf_len,     */
/* 			                 void *buffer,int buf_len,           */
/*                                       inst_t *inst,int inum, arg_t *arg1, */
/*                                       arg_t *arg2,int arg_no);            */
/*                                                                           */
/* INPUT PARM's: "ctxt" is a valid context. "meta_buffer" points to a buffer */
/*               with the length "mbuf_len" which is used to store info      */
/*               meta data such as mask register etc. "buffer" is the buffer */
/*               where the generated op-code is stored and "buf_len" is the  */
/*               length of that buffer. "inst" is the pointer to the current */
/*               instruction and "inum" is the number of that instruction.   */
/*               "arg1" points to the first operand for the "and" operation  */
/*               (performed by this function) and "arg2" is the second       */
/*               operand. "arg_no" is number of the argument associated with */
/*               the instruction.                                            */
/*                                                                           */
/* FUNCTION    : Evaluate and generate code for the "&" operand when         */
/*               registers are involved. SPU specific code.                  */
/*                                                                           */
/* RETURN VALUE: Returns 1 if successful. Errors are reported through        */
/*               iopasm_error(...)                                           */
/* ************************************************************************* */
static int and_operator(cg_context_t *ctxt,void *meta_buffer,int mbuf_len,
			void *buffer,int buf_len,inst_t *inst,int inum,
			arg_t *arg1,arg_t *arg2,int arg_no)
{
    
  int arg1_type,arg2_type;
  char *arg2_val,*arg1_val;
  int *op_code;
  int *meta_data,rnum;
  int reg_no;
  assert((buffer != NULL) && (buf_len > 0) && (inst != NULL) && 
	 (arg1 != NULL) && (arg2 != NULL));

  op_code = (int *) buffer;
  meta_data = (int *) meta_buffer;

  arg1_type = get_arg_type(arg1);
  arg2_type = get_arg_type(arg2);
  arg1_val = get_arg_val(arg1);
  arg2_val = get_arg_val(arg2);

  if (check_masked_inst(inst) == -1) {
    iopasm_error(ctxt,inst, ERR_UNKNOWN_INSTRUCTION,
		 "The %s instruction does not allow \
masked registers as  operands.",get_inst(inst));
  }

  
  if (arg2_type != ARG_NORMAL_REG) {
    iopasm_error(ctxt,inst,ERR_ILLEGAL_REGISTER_USE,
		 "The register %s may not be used here.\
 It is not a general register",arg2_val);    
  }

  reg_no = get_reg_number((reg_table_t *)etrax_fs_spu_regs,arg2_val);

  if ((arg1_type != ARG_NORMAL_REG) &&
      (arg1_type != ARG_SPECIAL_REG)) {
    iopasm_error(ctxt,inst,ERR_ILLEGAL_REGISTER_USE,
		 "The register %s may not be used here.\
 It is neither a general nor a special register",arg1_val);   
  }
  
  rnum = get_reg_number((reg_table_t *) etrax_fs_spu_regs,arg1_val);

  /* Let us check the op_code to see if what/if any mask instruction
   * has been specified 
   */
  switch(arg_no) {


  case 0:
    (*op_code) = (*op_code) | 0x20; /* Set mask on RS */
    (*op_code) = (*op_code) | ((reg_no & 0xF) << 7); /* Set mask register */
    (*op_code) = (*op_code) | get_src_regval(ctxt,inst,
					     inum,rnum);
    SET_SRC_META_OP((*meta_data),META_OP_AND);
    SET_SRC_META_REG((*meta_data),reg_no);
/*     SET_MASK_REGISTER(tmp_code,reg_no); */
    break;
    

  case 1:
    if (((GET_SRC_META_OP((*meta_data)) != META_OP_AND) ||
	 (GET_SRC_META_REG((*meta_data)) != reg_no)) && 
	(GET_SRC_META_OP(*meta_data) != META_OP_NOP)) {
      iopasm_error(ctxt,inst,ERR_ILLEGAL_MASK_REG_USE,
		   "The register %s may not be used as a mask register",
		   arg2_val);
    } /* End of check */
    
    (*op_code) = (*op_code) | 0x10; /* Set mask on RN */
    (*op_code) = (*op_code) | ((reg_no & 0xF) << 7); /* Set mask register */
    (*op_code) = (*op_code) | get_any_regval(ctxt,inst,
					     inum,rnum);
    SET_ANY_META_OP((*meta_data),META_OP_AND);
    SET_ANY_META_REG((*meta_data),reg_no);
    break;
    

  case 2:
    iopasm_error(ctxt,inst,ERR_ILLEGAL_MASK_REG_USE,
		 "Masked register operands (%s & %s) are not allowed as \
3rd operand to the %s instruction",arg1_val,arg2_val,get_inst(inst));
    /* program exits here */
    break;
  } /* End of switch(arg_no) */

  return 1;
}



/* ************************************************************************* */
/* PROTOTYPE   : static unsigned int eval_arg(cg_context_t *ctxt,            */
/*                                            inst_t *inst,arg_t *arg,       */
/* 			                      int i_num,int *meta_data,      */
/*                                            int arg_no,int pass,           */
/*                                            unsigned int (*eval_reg_func), */
/*			                      unsigned int (*eval_imm_func));*/
/*                                                                           */
/* INPUT PARM's: "ctxt" is a valid context, "inst" is the current instruction*/
/*               "arg" is the argument to evaluate and i_num is the number   */
/*               of the instruction. "meta_data" is used to store data       */
/*               between calls. "arg_no" is the number of the argument as    */
/*               seen from an instruction point of view. "pass" is the pass  */
/*               (intermediate or final. "eval_imm_func" and "eval_reg_finc" */
/*               are function pointers used to recalculate the value of a    */
/*               register or immedate value.                                 */
/*                                                                           */
/* FUNCTION    : Evaluate an argument passed to an instruction and generate  */
/*               the opcode accordingly.                                     */
/*                                                                           */
/* RETURN VALUE: The resulting changes to the opcode                         */
/* ************************************************************************* */
static unsigned int eval_arg(cg_context_t *ctxt,inst_t *inst,arg_t *arg,
			     int i_num,int *meta_data,int arg_no,int pass,
			     unsigned int (*eval_reg_func)(cg_context_t *,
							   inst_t *,
							   int,int),
			     unsigned int (*eval_imm_func)(cg_context_t *,
							   inst_t *,
							   int,
							   char *))
{

  inst_table_t *i_table;
  reg_table_t *r_table;
  array_t *labels;
  label_t *temp_label;
  label_t *found_label;
  unsigned int ret_val;
  unsigned int imm_val;
  int r_num;
  int arg_type;

  char *arg_val;
  dyn_stack_t *the_stack;

  char txt_buff[SPU_FSM_LEN];
  

  assert ((ctxt != NULL) && (inst != NULL));
  labels = cg_get_label_array(ctxt);
  i_table = cg_get_inst_table(ctxt);
  r_table = cg_get_reg_table(ctxt);
  
  arg_type = get_arg_type(arg);

  if ((arg_type == ARG_IMMEDIATE) && (get_inst_len(i_table,i_num) == 8) && 
      (get_field_size(i_table,i_num,arg_no) == 32)) {
    ret_val = 0;
    return ret_val;
  }
  arg_val = (arg_type != ARG_STACK) ? get_arg_val(arg): NULL;
  
  if (arg_type == ARG_STACK) {
    the_stack = get_arg_stack(arg);
    ret_val = eval_arg_stack(ctxt,inst,i_num,arg,meta_data,arg_no,pass);

    if (get_arg_type(arg) == ARG_STACK) { 
      return ret_val;
    }
    else {
      arg_type = get_arg_type(arg);
      arg_val = get_arg_val(arg);
    }
  } /* End of if (arg_type == ARG_STACK) { */
  
  
  if ((arg_type == ARG_NORMAL_REG)||(arg_type == ARG_SPECIAL_REG)) {
    r_num = get_reg_number(r_table,arg_val);
    return eval_reg_func(ctxt,inst,i_num,r_num);
    } 

  else if (arg_type == ARG_IMMEDIATE) {
    return eval_imm_func(ctxt,inst,i_num,arg_val);
  }

  else if (arg_type == ARG_LABEL) {
    temp_label = create_label(arg_val,0,SCOPE_NO_SCOPE);
    assert(temp_label != NULL);

    found_label = (label_t *) find_array_elem(labels,(void *) temp_label); 

    if ((found_label == NULL) && 
	(!check_option("allow-undefined-symbols",(int *)1,OPT_INTEGER))) 
      { /* Report error here */
	iopasm_error(ctxt,inst,ERR_UNDEFINED_LABEL,NULL,
		     "Label %s is undefined",arg_val);
      }

    if (found_label == NULL) {
      found_label = temp_label;
    }

    imm_val = get_label_value(ctxt,found_label);
    snprintf(txt_buff,SPU_FSM_LEN,"%d",imm_val);
    IOPASM_FREE(temp_label);
    return eval_imm_func(ctxt,inst,i_num,txt_buff);
    
  } /* Matches else if (arg_type == ARG_LABEL) */

  else { 
    iopasm_error(ctxt,inst,ERR_UNKNOWN_ARGUMENT,"The argument type of %s \
is not known by the assembler",arg_val);
  }

  iopasm_error(ctxt,inst,ERR_REPORT_BUG,
	       "Un-caught error in code generator (%s:%d)\n",
	       __FILE__,__LINE__);
  return 0;
}

  

/* ************************************************************************* */
/* PROTOTYPE   : static int set_input_muxval(cg_context_t *ctxt,             */
/*                                           inst_t *inst,char *fsm_code,    */
/*			                     char *val,int shift_val);       */
/*                                                                           */
/* INPUT PARM's: "ctxt" and "inst" are context and instruction pointers.     */
/*               "fsm_ocde" the buffer where the FSM code should be stored   */
/*               shift_val is the number of stages the value stored "val"    */
/*               should be shifted.                                          */
/*                                                                           */
/* FUNCTION    : Set the input mux part of an FSM instruction.               */
/*                                                                           */
/* RETURN VALUE: Always returns 1.                                           */
/* ************************************************************************* */
static int set_input_muxval(cg_context_t *ctxt,inst_t *inst,char *fsm_code,
			    char *val,int shift_val)
{
  int i_val;
  char *byte_ptr;
  
  i_val = atoi(val);
  if ((i_val < 0) || (i_val > 15)) { /* Report error here -out of range */
    iopasm_error(ctxt,inst,ERR_IMM_OUT_OF_RANGE,"Literal value %d is out of \
range. It must be within the range 0 - 15\n",i_val);
  }

  i_val = htonl(i_val); 
  byte_ptr = (char *) &i_val;

  fsm_code[2] = ((*(byte_ptr+3) << shift_val) | fsm_code[2]);
  return 1;
  
}



/* ************************************************************************* */
/* PROTOTYPE   : static int set_output_muxval(cg_context_t *ctxt,            */
/*                                            inst_t *inst,char *fsm_code,   */
/*                                            char *val);                    */
/*                                                                           */
/* INPUT PARM's: "ctxt" and "inst" are context and inst pointers. "fsm_code" */
/*               points to the buffer where the FSM op-code is stored. and   */
/*               "val" is the value to store in buffer.                      */
/*                                                                           */
/* FUNCTION    : Checks the range and stores (converts) the value for muxval */
/*               (see MDS) in fsm_code.                                      */
/*                                                                           */
/* RETURN VALUE: Always returns 1.                                           */
/* ************************************************************************* */
static int set_output_muxval(cg_context_t *ctxt,inst_t *inst,
			     char *fsm_code,char *val)
{
  int i_val;
  char *byte_ptr;
  i_val = atoi(val);
  if ((i_val < 0) || (i_val > 255)) { /* Report error here - out of range */
    iopasm_error(ctxt,inst,ERR_IMM_OUT_OF_RANGE,"Literal value %d is out of \
range. It must be within the range 0 - 255\n",i_val);
    
  }
  i_val = htonl(i_val); 
  byte_ptr = (char *) &i_val;
  fsm_code[3] = *(byte_ptr+3);
  return 1;
}



/* ************************************************************************* */
/* PROTOTYPE   : static int set_event_mask(char *fsm_code,char *val);        */
/*                                                                           */
/*                                                                           */
/* INPUT PARM's: "ctxt" and "inst" points to the context and instruction     */
/*               respectively. "fsm_code" points to the buffer where the FSM */
/*               op-code is stored and "val" is the value to encode.         */
/*                                                                           */
/* FUNCTION    : Checks the range for, and sets the event mask in a state    */
/*                                                                           */
/* RETURN VALUE: Always returns 1.                                           */
/* ************************************************************************* */
static int set_event_mask(cg_context_t *ctxt,inst_t *inst,char *fsm_code,
			  char *val)
{
  int i_val;
  
  i_val = atoi(val);
  if ((i_val < 0) || (i_val > 15)) {
    iopasm_error(ctxt,inst,ERR_IMM_OUT_OF_RANGE,"Literal value %d is out of \
range. It must be within the range 0 - 15\n",i_val);

 
  }
  fsm_code[2] = fsm_code[2] | ((i_val << 4) & 0xF0);
  return 1;
}



/* ************************************************************************* */
/* PROTOTYPE   : static int set_update_mask(cg_context_t *ctxt,inst_t *inst, */
/*                                          char *fsm_code,char *val);       */
/*                                                                           */
/*                                                                           */
/* INPUT PARM's: "ctxt" and "inst" points to a valid context and instruction */
/*               respectively. "fsm_code points to the buffer where the FSM  */
/*               op-code is stored and "val is the value to encode.          */
/*                                                                           */
/* FUNCTION    : Checks the range for, and sets the update mask in a state   */
/*                                                                           */
/* RETURN VALUE: Always returns 1.                                           */
/* ************************************************************************* */
static int set_update_mask(cg_context_t *ctxt,inst_t *inst,char *fsm_code,
			   char *val)
{
  int i_val;
  
  i_val = atoi(val);
  if ((i_val < 0) || (i_val > 15)) {
    iopasm_error(ctxt,inst,ERR_IMM_OUT_OF_RANGE,"Literal value %d is \
out of range. It must be within the range 0 - 15\n",i_val);
    
  }
  fsm_code[1] = fsm_code[1] | (i_val & 0x0F);
  return 1;
  
}



/* ************************************************************************* */
/* PROTOTYPE   : static unsigned char encode_in_symbol(char *symbols);       */
/*                                                                           */
/* INPUT PARM's: "symbols" points to a string containing a syntactically     */
/*               correct input string of an FSM statement.                   */
/*                                                                           */
/* FUNCTION    : Translate the symbol into correct op-code.                  */
/*                                                                           */
/* RETURN VALUE: Return the correct op-code for the FSM statement.           */
/* ************************************************************************* */
static unsigned char encode_in_symbol(char *symbols)
{
  int pos;
  int bit_no;
  unsigned char ret_val;

  bit_no = 3;
  ret_val = 0;

  while(bit_no != -1) {
    
    pos = 6 - (bit_no * 2);

    switch(symbols[pos]) {
      
    case '0':
      ret_val |= 1 << (4+bit_no); 
      ret_val |= 1 << (bit_no); 
      break;

    case '1':
      /* ret_val |= 0 << (4+bit_no); */
      ret_val |= 1 << (bit_no);
      break;

    case '?':
      /* This code is only only here for the sake of clarity */
      /* ret_val |= 0 << (4+bit_no); */
      /* ret_val |= 0 << (bit_no); */
      break;
    }
   
    bit_no--;
  }
  return ret_val;
}



/* ************************************************************************* */
/* PROTOTYPE   : static unsigned char encode_output_symbol(char *symbols);   */
/*                                                                           */
/* INPUT PARM's: "symbols" is a pointer to the output-statement of a state-  */
/*               transition or timer statement.                              */
/*                                                                           */
/* FUNCTION    : Generate the code for the output symbol in a FSM statement  */
/*                                                                           */
/* RETURN VALUE: The "output" part of the op-code associated FSM statement   */
/* ************************************************************************* */
static unsigned char encode_output_symbol(char *symbols)
{
  unsigned char tmp_encoded;
  int bit_no;
  int sym_no;
  
  bit_no = 3;
  sym_no = 0;
  tmp_encoded = 0;

  do {
    switch (symbols[sym_no]) {
      
    case '1':
      tmp_encoded |= (unsigned char )3 << (bit_no * 2);
      break;

    case '0':
      tmp_encoded |= (unsigned char) 2 << (bit_no * 2);
      break;

    case 'p':
      tmp_encoded |= (unsigned char) 1 << (bit_no * 2);
      break;

    case '?':
      tmp_encoded |= (unsigned char) 0 << (bit_no * 2);
      break;
    }

    sym_no += 2;
    bit_no--;

  } while (sym_no != 8);

  return tmp_encoded;

}



/* ************************************************************************* */
/* PROTOTYPE   : unsigned int etrax_fs_spu_conv_pc_fsm(unsigned int val);    */
/*                                                                           */
/* INPUT PARM's: "val" is a byte address.                                    */
/*                                                                           */
/* FUNCTION    : Converts a byte address to a valid FSM address              */
/*                                                                           */
/* RETURN VALUE: a valid FSM address.                                        */
/* ************************************************************************* */
unsigned int etrax_fs_spu_conv_pc_fsm(unsigned int val)
{
  return (val > 0) ? (val-1) & 0xFF8:0;
}



/* ************************************************************************* */
/* PROTOTYPE   : unsigned int etrax_fs_spu_conv_pc_seq(unsigned int val);    */
/*                                                                           */
/* INPUT PARM's: "val" is a byte address.                                    */
/*                                                                           */
/* FUNCTION    : Converts a byte address to a valid sequential address.      */
/*                                                                           */
/* RETURN VALUE:  A valid Sequential address.                                */
/* ************************************************************************* */
unsigned int etrax_fs_spu_conv_pc_seq(unsigned int val)
{
  return val;
}



/* ************************************************************************* */
/* PROTOTYPE   : etrax_fs_spu_set_tables(cg_context_t *context)              */
/*                                                                           */
/* INPUT PARM's: A pointer to a context in which tables are to be exported   */
/*                                                                           */
/* FUNCTION    : This function allowes for specific initialization of table  */
/*               constants prior to calling the cg_set_inst_table(...) and   */
/*               cg_set_reg_table(...) functions.                            */
/*                                                                           */
/* RETURN VALUE: returns a positive non-zero value on success and a negative */
/*               non-zero value on failure.                                  */
/* ************************************************************************* */
int etrax_fs_spu_set_tables(cg_context_t *ctxt)
{
  /* This function allows for specific initialization of tables prior to 
   * assigning it to the context. In this case we are satisfied once the
   * instruction boundraries as well as the optimal size of the FSM
   * instruction has been set 
   */

  cg_set_fsm_boundrary(ctxt,8); /* Boundraries are _always_ measured in bytes */
  cg_set_seq_boundrary(ctxt,4); /* Boundraries are _always_ measured in bytes */

  cg_set_fsm_size(ctxt,32); /* Instruction-sizes are is _always_ measured 
			     * in bytes */

  cg_set_inst_table(ctxt,etrax_fs_spu_seq);
  cg_set_reg_table(ctxt,etrax_fs_spu_regs);
  cg_set_cpp_macros(ctxt,(char **)etrax_fs_spu_cpp_macros);
  cg_set_be_operators(ctxt,etrax_fs_spu_operators);

  cg_set_memory_bank_count(ctxt,2); /* Set number of memory banks used by */
                                    /* This version of the SPU */
  return 1;
}  



/* ************************************************************************* */
/* PROTOTYPE   : int etrax_fs_spu_gen_fsm(cg_context_t *seq,inst_t *first,   */
/*                                        inst_t *last,int pass);            */
/*                                                                           */
/*                                                                           */
/* INPUT PARM's: "buf" is a buffer of "buf_siz" bytes in which the generated */
/*               code is to be stored. "first" is the first intruction       */
/*               describing the state and "last" is the last instruction     */
/*               describing the state. "labels" contains all the labels      */
/*               found during the parsing of the code and "fsm_symbols" is   */
/*               the array used to store all symbolic definitions for FSM    */
/*               signals.                                                    */
/*                                                                           */
/* FUNCTION    : Generates the initial portion of a FSM state. Manages such  */
/*               tasks as, sequential instructions, timers etc.              */
/*                                                                           */
/* RETURN VALUE: Returns the size of the generated code. If buffer is to     */
/*               small the shortage of space is reported as a negative       */
/*               value. 0 is returned if code cannot be generated, i.e. some */
/*               non-syntactic error was discovered.                         */
/* ************************************************************************* */
int etrax_fs_spu_gen_fsm(cg_context_t *ctxt,inst_t *first,
			inst_t *last,int pass)
{
  char *arg_val;
  int i;
  inst_t *this_inst;
  int seq_inst = 0;
  int go_seq_found = 0;
  int seq_size = 0;
  int timer_inst = 0;
  int max_xfers = 8;
  int fsm_index;
  int only_state = 0;
  int emask = 0;
  int umask = 0;
  int fsm_size = 4;
  int fsm_xfers = 0;
  int hi_bit_val = 0;
  int hi_bit_set = 0;
  array_t *labels;
  label_t *temp_label,*real_label;
  arg_t *option_arg;
  unsigned int temp_val = 0;
  unsigned short imm_val;
  unsigned char buf[SPU_FSM_LEN];
  int buf_size = SPU_FSM_LEN;


  assert(ctxt != NULL);
  assert(buf != NULL);

  memset(buf,0,SPU_FSM_LEN); /* Clear buffer, before filling it */
  
  buf_size = SPU_FSM_LEN;
  fsm_index = 4;

  labels = cg_get_label_array(ctxt);
  this_inst = first;
  
  
  while(get_inst_type(this_inst) != INST_STATE_END) {
    switch (get_inst_type(this_inst)) {


    case INST_SEQ:
      if (!seq_inst) {
	seq_inst++;
	fsm_size += 4;
      }
      else {
	seq_inst++;
	fsm_size = 0;
	iopasm_error(ctxt,this_inst,ERR_FSM_TOO_MANY_SEQS,
		   "Found %d sequential instructions, but only 1 is allowed.",
		     seq_inst);

      }
      
      DEBUG(3,"Found Sequential instruction, size is now: %d",fsm_size);
      DEBUG(3,"Number of sequential instructions: %d",seq_inst);
      break;

      
      case INST_TIMER:
      if (!timer_inst) {
        fsm_size += 4;
        timer_inst++;
      }

      else {
        fsm_size = 0;
        timer_inst++;
	iopasm_error(ctxt,this_inst,ERR_FSM_TOO_MANY_TIMERS,
		   "Found %d timer statements, but only 1 is allowed.",timer_inst);
      }

      DEBUG(3,"Found timer instruction, size is now: %d",fsm_size);
      DEBUG(3,"Number of timer instructions: %d",timer_inst);
      break;


    case INST_FSM_EVENT_MASK:
      emask = 1;
      break;


    case INST_FSM_UPDATE_MASK:
      umask = 1;
      break;


    case INST_ONLY_STATE:
      only_state++;
      if (fsm_xfers != 0) {
	iopasm_error(ctxt,this_inst,ERR_FSM_XFER_CONFLICT,
		   "Found \"only\" statement and also %d state transition(s). \
The \"only\" statement and state transitions cannot exist in the same state",
		     fsm_xfers);
      }
      if (timer_inst != 0) {
	iopasm_error(ctxt,this_inst,ERR_FSM_XFER_CONFLICT,
		   "Found \"only\" statement and also %d timer statements(s). \
The \"only\" statement and the timer statement cannot exist in the same state",
		     timer_inst);
      }
      break;

      
    case INST_STATE_XFER:
      max_xfers = ((timer_inst == 1) && (seq_inst == 1)) ? 6:8;
      if (fsm_xfers >= max_xfers) {
        fsm_size = 0;
        fsm_xfers++;
	iopasm_error(ctxt,this_inst,ERR_FSM_TOO_MANY_XFERS,
		   "Found %d state transitions. The maximum \
allowed number of state transitions is 8",fsm_xfers);
      } 

      else {
        fsm_size += 3;
        fsm_xfers++;
      }

      DEBUG(3,"Found state transition, size is now: %d",fsm_size);
      DEBUG(3,"Number of state transition: %d",fsm_xfers);
      break;
    }
    
    this_inst = get_next_inst(this_inst);
  }

  if (fsm_size == 0) { 
    iopasm_error(ctxt,this_inst,ERR_FSM_UNKNOWN_ERROR,
	       "Un-caught error in code generator (%s:%d)\n",
	       __FILE__,__LINE__);
  }

  /* Now we actually now the size of this state thus we can also calculate the start address 
   * which is entirely different from PC!  */
  
   
  
  /* Generate code for this state. What we do here is simply to generate the 
   * FSM instruction and set the appropriate flags. Once the stage is 
   * completed, the code will look like this:
   *
   * +----------+ 
   * | FSM INST |
   * +----------+
   *
   */ 

  this_inst = first; 
  SET_SEQ(buf,seq_inst);
  SET_TIMER(buf,timer_inst);
  SET_STATE_COUNT(buf,fsm_xfers);

  while (get_inst_type(this_inst) != INST_STATE_END) {
    
    switch(get_inst_type(this_inst)) {
      

    case INST_FSM_INP_FLAG:
      option_arg = get_first_arg(this_inst);
      convert_stack_arg(ctxt,this_inst,option_arg,pass);
      arg_val = get_arg_val(option_arg); 
      set_input_muxval(ctxt,this_inst,buf,arg_val,0);
      break;
	

    case INST_FSM_INV_STB:
      SET_INV_STROBE(buf); 
      break;
      

    case INST_FSM_STB:
      option_arg = get_first_arg(this_inst);
      convert_stack_arg(ctxt,this_inst,option_arg,pass);
      fprintf(stderr,
	      "Warning: STB statement is not supported by SPU(near line %d)\n",
	      get_inst_line(this_inst));
      arg_val = get_arg_val(option_arg);
      temp_val = atoi(arg_val);
      SET_STROBE(buf,temp_val);
      break;
      

    case INST_FSM_OUT_FLAG:
      option_arg = get_first_arg(this_inst);
      convert_stack_arg(ctxt,this_inst,option_arg,pass);
      arg_val = get_arg_val(option_arg);
      set_output_muxval(ctxt,this_inst,buf,arg_val);
      break;
      

    case INST_FSM_DO_SEQ:
      SET_DO_SEQ(buf);
      break;
      

    case INST_FSM_GO_SEQ:
      if (emask == 1) { /* IOPASM ERROR according to TR #23 */
	iopasm_error(ctxt,this_inst,ERR_FSM_EMASK_WITH_GO_SEQ,NULL);
      }
      SET_GO_SEQ(buf);
      go_seq_found = 1;
      break;
      

    case INST_FSM_EVENT_MASK:
      if (only_state) {
	iopasm_error(ctxt,this_inst,ERR_FSM_EMASK_IN_ONLY_XFER,NULL);
      }
      option_arg = get_first_arg(this_inst);
      convert_stack_arg(ctxt,this_inst,option_arg,pass);
      arg_val = get_arg_val(option_arg);
      set_event_mask(ctxt,this_inst,buf,arg_val);
      break;
      

    case INST_FSM_UPDATE_MASK:
      option_arg = get_first_arg(this_inst);
      convert_stack_arg(ctxt,this_inst,option_arg,pass);
      arg_val = get_arg_val(option_arg);
      set_update_mask(ctxt,this_inst,buf,arg_val);
      break;
      

    case INST_FSM_REUPDATE:
      SET_REUPDATE(buf);
      break;
	

    case INST_FSM_HALT:
      SET_FSM_HALT(buf);
      break;
    } /* End of switch(get_inst_type(this_inst)) */
    
    this_inst  = get_next_inst(this_inst);      
  } /* End of while (this_inst != last) */

    /* According to TR #29 we should check that we have some sort of state-transition */
    /* 2004-07-27 TR29 is incorrect from the very beginning */
/*   if ((timer_inst == 0) && (only_state == 0) &&  */
/*       (fsm_xfers == 0) && (seq_inst == 1)) { */
/*     IOPASM_ERR(ERR_FSM_NO_FSM_XFERS,first,""); */
/*   } */
  
  
  
  /* Generate code for the Sequential instruction (if required). 
   * As the sequential instruction must be placed before any other 
   * instructions (except the FSM instruction) this is the first 
   * that we try to generate. We must start from the beginning of the block 
   * After this stage is completed our generated code will look like this:
   *
   * +----------+----------+
   * | FSM INST | SEQ INST |
   * +----------+----------+
   */
 
  this_inst = first;
 
  if (seq_inst == 1) {
    while ((get_inst_type(this_inst) != INST_SEQ)) {
      this_inst = get_next_inst(this_inst);
    }
  
    seq_size = etrax_fs_spu_gen_seq(ctxt,this_inst,pass,SCOPE_FSM);
    
     fsm_size = (seq_size == 8) ? fsm_size+4: fsm_size;
    
    if ((seq_size == 8) && (timer_inst)) {

      iopasm_error(ctxt,this_inst,ERR_FSM_NO_TIMER_ALLOWED,NULL);
    }
     
    else if ((seq_size == 8) && (fsm_xfers > 6)) {
      iopasm_error(ctxt,this_inst,ERR_FSM_TOO_MANY_XFERS,
		 "Only 6 transfers allowed when a 32-bit immediate \
instruction is present in the state. This state contains %d",fsm_xfers);
    }

    
    else if (seq_size == 0) { 
      return 0;
    }
    
    else {
      if (seq_size == 8) {
	SET_FSM_32BIT_SEQ(buf);
      }
      memcpy(&buf[fsm_index],get_inst_code(this_inst),seq_size);
      fsm_index += seq_size;
    }
  }

  /* Next we generate the code for any timer instructions. Again we need to
   * place the timers before any state transitions so we do this now. Our 
   * "best case code" i.e. (fsm + seq + timer will look like this after this 
   * stage is completed):
   *
   * +----------+----------+------------+
   * | FSM INST | SEQ INST | TIMER INST |  
   * +----------+----------+------------+
   * 
   * We begin at the beginning of the block in order to find the timer 
   */


  this_inst = first;

  if (timer_inst == 1) {
    while ((get_inst_type(this_inst) != INST_TIMER)) {
      this_inst = get_next_inst(this_inst);
    } /* End while(...) */

    /* Minimize stacks in instruction */
    minimize_stacks(ctxt,this_inst,pass);

    /* this_inst now points to the timer instruction */
    option_arg = get_first_arg(this_inst);
    convert_stack_arg(ctxt,this_inst,option_arg,FINAL_PASS);
    arg_val = get_arg_val(option_arg);

    /* First we must check if argument is a register or not */
    if (get_arg_type(option_arg) == ARG_TIMER_KEEP) {
      SET_TIMER_KEEP(buf[fsm_index+2]);
    }

    else { 
      if (get_arg_type(option_arg) == ARG_NORMAL_REG) {
	
	i = get_reg_number(cg_get_reg_table(ctxt),arg_val);
	buf[fsm_index+2] = 0x10;
	buf[fsm_index+3] = (unsigned char ) i;
	
      } /* End of if (get_arg_type(option_arg) == ARG_TIMER_KEEP) { */
      
      else { 
	imm_val = (unsigned short) atoi(arg_val);
	
	if (imm_val > ((1 << 12) - 1)) { /* Report error here and exit */
	  iopasm_error(ctxt,this_inst,ERR_IMM_OUT_OF_RANGE,
		       "Literal value %lu is out of range. It must be within \
the range 0 - %lu",imm_val,((1 << 12) - 1));
	}
      
	imm_val &= 0xFFF;
	arg_val = (char *) &imm_val;
	buf[fsm_index+3] = arg_val[0];
	buf[fsm_index+2] = ((buf[fsm_index+2] & 0xF0) | (0x0F & arg_val[1]));
	
      } /* !if (get_arg_type(option_arg) == ARG_NORMAL_REG) */
    } /* End of !if (get_arg_type(option_arg) == ARG_TIMER_KEEP) { */
   
    /* 2nd: timer strobe */
    /* option_arg = get_next_arg(option_arg);
     * arg_val = get_arg_val(option_arg);
     *
     * if ((atoi(arg_val) > 3) || (atoi(arg_val) < 0)) { 
     * IOPASM_ERR(ctxt,ERR_IMM_OUT_OF_RANGE,this_inst,"Literal=%s (range 0-3)\n",
     * arg_val);
     * }
     * buf[fsm_index+2] |= (char) (((atoi(arg_val)) << 6) & 0xC0);
     */
    
    /* 3rd: Timer priority */
    option_arg = get_next_arg(option_arg);
    arg_val = get_arg_val(option_arg);
    
    if ((arg_val[0] != '1') && (arg_val[0] != '0')) { /* Report error here */
      iopasm_error(ctxt,this_inst,ERR_IMM_OUT_OF_RANGE,
		 "Literal value %s is out of range. It must be within the range \
0 or 1",arg_val);
    }

    if (arg_val[0] == '1') {
      buf[fsm_index+2] |= 0x20; 
    } 
    
    /* 4th: timer output */
    option_arg = get_next_arg(option_arg);
    arg_val = get_arg_val(option_arg);
    buf[fsm_index] = encode_output_symbol(arg_val);
    

    /* 5th: set next state */ 
    option_arg = get_next_arg(option_arg);
    arg_val = get_arg_val(option_arg); /* arg val now points to label */
    temp_label = create_label(arg_val,0,SCOPE_NO_SCOPE);
    real_label = (label_t *) find_array_elem(labels,(void *)temp_label);

    if ((real_label == NULL) && 
	(!check_option("allow-undefined-symbols",(int *)1,OPT_INTEGER)))
      { /* Cannot find label - return error and exit */
	iopasm_error(ctxt,this_inst,ERR_UNDEFINED_LABEL,"Label is %s\n",arg_val);
      }
    
    if (real_label == NULL) { /* Undefined symbols allowed and no label! */
      real_label = temp_label;
    }
    
    temp_val = get_label_value(ctxt,real_label);
    delete_label(temp_label); 
    
    /* addresses always stored in bytes */
    temp_val = (temp_val > 0) ? (temp_val / 8): 0;
    
    if (temp_val >= 0x100) {
      temp_val = (temp_val >= 0x100) ? temp_val - 256: temp_val;
      hi_bit_set = 1;
      hi_bit_val = 1;
      SET_HIGH_ADDRESS_BIT(buf,hi_bit_val);
    }
    
    else {
      hi_bit_set = 0;
      hi_bit_val = 0;
    }

    buf[fsm_index+1] = (char) (temp_val & 0xFF);
    fsm_index += 4; /* Size of timer instruction */
  } /* End if (timer_inst == 1) */


  /* Next we need to find and generate code for the state transitions
   * place the timers before any state transitions so we do this now. Our 
   * "best case code" i.e. (fsm + seq + timer will look like this after this 
   * stage is completed): 
   * 
   * +----------+----------+------------+-----------------+
   * | FSM INST | SEQ INST | TIMER INST | FSM TRANSITIONS |
   * +----------+----------+------------+-----------------+
   *  
   * We begin at the beginning of the block in order to find 
   * the first transition, but first we check for the only-state
   * operator.
   */

  this_inst = first; 
  
  if (only_state == 1) {
    while((get_inst_type(this_inst) != INST_ONLY_STATE) &&
	  (this_inst != last)) {
      this_inst = get_next_inst(this_inst);
    }
    
    /* Minimize stacks in only statement */
    minimize_stacks(ctxt,this_inst,pass);

    if (get_inst_type(this_inst) == INST_ONLY_STATE) {
      SET_ONLY(buf);
      option_arg = get_first_arg(this_inst);
      arg_val = get_arg_val(option_arg);

      if (get_arg_type(option_arg) == ARG_LABEL) {
	temp_label = create_label(arg_val,0,SCOPE_NO_SCOPE);
	real_label = (label_t *) find_array_elem(labels,(void *)temp_label);
	if ((real_label == NULL) && 
	    (!check_option("allow-undefined-symbols",(int *)1,OPT_INTEGER)))
	  { /* Cannot find label - return error and exit */
	    iopasm_error(ctxt,this_inst,ERR_UNDEFINED_LABEL,"Label is %s\n",
			 arg_val);
	  }
	
	else if (real_label == NULL) { /* Undefined symbols allowed and no label! */
	  temp_val = 0;
	}
	else { /* Symbol found */
	  temp_val = get_label_value(ctxt,real_label);
	}
	delete_label(temp_label); 
	
	/* addresses always stored in bytes */
	if (go_seq_found == 1) {
	  temp_val = (temp_val > 0) ? (temp_val / 4): 0;
	  buf[3] = (char) (temp_val & 0xff);
	  buf[2] = (buf[2] & 0xFC) | ((temp_val >> 8) & 0x3);
	}
	else {
	  temp_val = (temp_val > 0) ? (temp_val / 8): 0;
	  buf[3] = (char) temp_val & 0xFF;
	  buf[2] = (char) (buf[2] & 0xFC) | ((temp_val >> 8) & 0x3);
	}
      } /* end of if (get_arg_type(option_arg) == ARG_LABEL) { */
      
      else if ((get_arg_type(option_arg) == ARG_NORMAL_REG) ||
	       (get_arg_type(option_arg) == ARG_SPECIAL_REG)) {
	arg_val = get_arg_val(option_arg); 
	
	if (only_reg_check(arg_val) < 0) {
	  iopasm_error(ctxt,this_inst,ERR_ILLEGAL_REGISTER_USE,
		       "The register %s may not be used here",arg_val);
	}
	
	if ((strcmp(arg_val,"e12") == 0)||(strcmp(arg_val,"eaddr") == 0)) {
	  buf[3] = 16;
	}
	else {
	  buf[3] = get_reg_number((reg_table_t *)etrax_fs_spu_regs,arg_val);
	}
	
	buf[2] = buf[2] | 0x08;
	

      } /* Matches if ((get_arg_type(option_arg) == ARG_NORMAL_REG) || */

      else if (get_arg_type(option_arg) == ARG_IMMEDIATE) {
	arg_val = get_arg_val(option_arg);
	if (atol(arg_val) > 255) {
	  iopasm_error(ctxt,this_inst,ERR_IMM_OUT_OF_RANGE,
		       "Literal value %s is out of range. It must be within \
the range 0 - 255",arg_val);

	}
	buf[3] = atol(arg_val) & 0xFF;
      } /* Matches else if (get_arg_type(option_arg) == ARG_IMMEDIATE) { */

      else {
	iopasm_error(ctxt,this_inst,ERR_REPORT_BUG,"Called from %s:%d",
		     __FILE__,__LINE__); 
      }
    } /* Matches if (get_inst_type(this_inst) == INST_ONLY_STATE) */
  } /* Matches if (only_state == 1) */
    
  
  else {
    if (fsm_xfers == 0) { /* We do not have any state transitions */
      goto state_generation_ready;
    }

    while ((get_inst_type(this_inst) != INST_STATE_XFER) &&(this_inst != last)) {
      this_inst = get_next_inst(this_inst);
    } /* end while(get_inst_type(this_inst) != INST_STATE_XFER) */
    

    do {   
      /* Minimize stacks in state xter */
      minimize_stacks(ctxt,this_inst,pass);

      option_arg = get_first_arg(this_inst);
      arg_val = get_arg_val(option_arg);
      buf[fsm_index] = encode_in_symbol(arg_val);
      
      option_arg = get_next_arg(option_arg);
      arg_val = get_arg_val(option_arg);
      buf[fsm_index+1] = encode_output_symbol(arg_val);
      
      option_arg = get_next_arg(option_arg);
      arg_val = get_arg_val(option_arg);
      
      temp_label = create_label(arg_val,0,SCOPE_NO_SCOPE);
      real_label = (label_t *) find_array_elem(labels,(void *)temp_label);
      
      if ((real_label == NULL) && 
	(!check_option("allow-undefined-symbols",(int *)1,OPT_INTEGER)))
      { /* Cannot find label - return error and exit */
	iopasm_error(ctxt,this_inst,ERR_UNDEFINED_LABEL,"Label is %s\n",arg_val);
      }
    
      else if (real_label == NULL) { /* Undefined symbols allowed and no label! */
	temp_val = 0;
      }
      
      else {
	temp_val = get_label_value(ctxt,real_label);
      }
      
      /* Finally delete temporary label */
      delete_label(temp_label); 

      /* addresses always stored in bytes */

      temp_val = (temp_val > 0) ? (temp_val / 8): 0;

      if (((temp_val > 0xFF) && (hi_bit_set == 1) && (hi_bit_val == 0)) ||
	  ((temp_val <= 0xFF) && (hi_bit_set == 1) && (hi_bit_val == 1)))
	{ 
	  iopasm_error(ctxt,this_inst,ERR_FSM_ADDR_CONFLICT,
		       "State address is %d, but there are state transitions \
going to either the high or low bank as well",temp_val);
	}

      else {
	hi_bit_set = 1;
	hi_bit_val = (temp_val > 0xFF) ? 1: 0;
	SET_HIGH_ADDRESS_BIT(buf,hi_bit_val);
      }

      buf[fsm_index+2] = (char) (temp_val & 0xFF);
      fsm_index += 3;
      
      this_inst = get_next_inst(this_inst);
    
    }   while((get_inst_type(this_inst) == INST_STATE_XFER));
  } /* Matches  !if (only_state == 1) */
    
  state_generation_ready: 
  
    if (fsm_index == fsm_size) {
      /* now we should copy the code to the current instruction (label) */
      i = (fsm_size % 8) ? fsm_size + ( 8- (fsm_size % 8)): fsm_size;
      memset(&buf[fsm_size],0,i);
      set_inst_code(first,buf,i);

      return i;
      
    }
    
    else { 
      return 0; 
  }
}

  

/* ************************************************************************* */
/* PROTOTYPE   : int etrax_fs_spu_gen_seq(cg_context_t *ctxt,inst_t *inst,   */
/*                                        int pass,int scope);               */
/*                                                                           */
/* INPUT PARM's: "ctxt" and "inst" are poiners to a valid context and        */
/*               instruction respectively. "pass" indicates the pass         */
/*               (intermediate or final). "Scope" is the scope of the call   */
/*               (i.e. sequential of FSM).                                   */
/*                                                                           */
/* FUNCTION    : Generate the op-code for the instruction "inst"             */
/*                                                                           */
/* RETURN VALUE: Returns the size of the instruction in bytes. 0 on failure  */
/* ************************************************************************* */
int etrax_fs_spu_gen_seq(cg_context_t *ctxt,inst_t *inst,int pass,int scope)
{
  int i_number;
  inst_table_t *current_row;
  arg_t *current_arg;
  int arg_type,i;
  inst_table_t *i_table;
  reg_table_t *r_table;
  unsigned int op_code;
  char *text_arg = NULL;
  int temp_val;
  int ret_val;
  unsigned char buf[SPU_SEQ_LEN];
  unsigned int imm_32;

  int buf_size = SPU_SEQ_LEN;
  int meta_data;
  int i_size;
  assert((ctxt != NULL) && (buf != NULL) && (inst != NULL));
  
  current_arg = get_first_arg(inst);
  r_table = cg_get_reg_table(ctxt);
  i_table = cg_get_inst_table(ctxt);
  current_row = create_itable_row(inst);

  SET_SRC_META_OP(meta_data,META_OP_NOP);
  SET_ANY_META_OP(meta_data,META_OP_NOP);

  /* There's no difference between special registers and 
   * normal registers in the SPU so we patch any special 
   * registered found by the parser and say that they are
   * normal.  */
   
  /* i = sregs2regs(current_row); */
  i_number = get_inst_number(i_table,current_row,inst); 
  
  if (i_number < 0) {
    iopasm_error(ctxt,inst,ERR_UNKNOWN_INSTRUCTION,NULL);
  }
  
  /* First initialize the opcode and set it to the template */
  /* op_code = get_opcode(i_table,i_number); */
  
  i_size = get_inst_len(i_table,i_number);
  if ((i_size == 8) && (scope == SCOPE_SEQ)) {
    iopasm_error(ctxt,inst,ERR_32BIT_IMMEDIATE_IN_SEQ,NULL);
  }
  
  /* Check the relocation flags before proceeding any further.
   * There is no reason to generate an instruction for a relocatable instruction
   * in the intermediate pass as we will not know anything about labels etc */

  if ((get_reloc_flag(inst) == RF_RELOCATE) &&
      (pass == INTERMEDIATE_PASS)) {
    memset(buf,0x00,SPU_SEQ_LEN); /* Fill buffer with zero's - Generated NOP */
    set_inst_code(inst,buf,i_size); /* Set instruction to nop, as it cannot be generated */
    set_inst_size(inst,i_size);
    return  i_size;
   } /* End of if ((check_reloc_flag(inst,RF_RELOCATE)) && ... */

   
   minimize_stacks(ctxt,inst,pass);
   op_code = generate_opcode(ctxt,i_table,inst,i_number,pass);

   i =  0;
   meta_data = 0;

  while (current_arg) {

    arg_type = get_arg_type(current_arg);
    text_arg = get_arg_val(current_arg);
    
    /* There's no difference between special registers and 
     * normal registers in the SPU so we patch any special 
     * registered found by the parser and say that they are
     * normal */

    /* Initially the SPU did not have any special registers so we simply
     * converted them on the fly. Today the Parser only knows about registers
     * and uses special functions to map these registers to the current 
     * architecture, so we do not need to convert registers. They are already
     * OK. 
     *
     * if (arg_type == ARG_SPECIAL_REG) {
     *  set_arg_type(current_arg,ARG_NORMAL_REG);
     * arg_type = ARG_NORMAL_REG;
     * }
     */
    
    switch(i) {
      

    case 0: /* This is source */
	temp_val = eval_arg(ctxt,inst,current_arg,i_number,&meta_data,0,pass,
			    get_src_regval,eval_src_immval);

	arg_type = get_arg_type(current_arg);

	if ((i_size == 8) && (arg_type == ARG_IMMEDIATE) && 
	    (get_field_size(i_table,i_number,0) == 32)) { 
	  imm_32 = strtoul(get_arg_val(current_arg),(char **) NULL,10);
	  imm_32 = htonl(imm_32);
	  memcpy(buf+4,&imm_32,4);
	}
	else {
	  op_code |= temp_val;
	}
	break;
      
      
    case 1: /* This is any */
	temp_val = eval_arg(ctxt,inst,current_arg,i_number,&meta_data,1,pass,
			    get_any_regval,eval_any_immval);

	arg_type = get_arg_type(current_arg);

	if ((i_size == 8) && (arg_type == ARG_IMMEDIATE) && 
	    (get_field_size(i_table,i_number,1) == 32)) { 
	  imm_32 = strtoul(get_arg_val(current_arg),(char **) NULL,10);
	  imm_32 = htonl(imm_32);
	  memcpy(buf+4,&imm_32,4);
	}
	else {
	  op_code |= temp_val;
	}
	break;
    
    
    case 2: /* This is dst */
	temp_val = eval_arg(ctxt,inst,current_arg,i_number,&meta_data,2,pass,
			    get_dst_regval,eval_dst_immval);
	op_code |= temp_val;
    } /* Matches switch(i) */
    i++;
    current_arg = get_next_arg(current_arg);
  }

  /* Pack op_code in buffer */
  op_code = htonl(op_code);
  memcpy(buf,&op_code,4);
  
  /* If we get this far we have generated an instruction, so we can return
   * the proper size -if there is enough space in the buffer */

  if (buf_size < i_size) ret_val = buf_size - i_size;
  
  else {
    ret_val = i_size;
    set_inst_code(inst,(unsigned char *) buf,i_size);
  }

  IOPASM_FREE(current_row);
  return ret_val;
}



/* ************************************************************************* */
/* PROTOTYPE   : void etrax_fs_spu_finish(cg_context_t *ctxt,                */
/*                                        plist_t *a_list);                  */
/*                                                                           */
/* INPUT PARM's: "ctxt" is a pointer to a valid context. "a_list" is a valid */
/*               parse list containing code.                                 */
/*                                                                           */
/* FUNCTION    : Generic function called when _all_ code is generated. This  */
/*               function is currently used to re-pack code in memory.       */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void etrax_fs_spu_finish(cg_context_t *ctxt,plist_t *a_list)
{
  unsigned char *buffer;
  unsigned char *i_code;
  unsigned int pc;
  unsigned int hi_pc;
  int i_size,i;
  int a_val[4]; /* Address value for 64-bit alignment */
  int o_val[4]; /* Corresponding buffer index */
  unsigned int offset = 0;
  unsigned int row = 0;
  inst_t *first;
  int diff;

  /* Because of a brain-dead programmer (Mr. Johan J) we will need to
   * repack all code here!  Fortunately for us this function was right
   * there to take care of it :-)
   */
  
  assert ((ctxt != NULL) && (a_list != NULL));
  
  first = get_first_inst(a_list);
  buffer = cg_get_code_buffer(ctxt);
  hi_pc = cg_get_memory_size(ctxt);
  memset(buffer,0x00, hi_pc); /* This equals filling the memory with NOPs */
  
  while (first != NULL) {
    
    diff = 0;
    
    switch (get_inst_type(first)) {


    case INST_BYTE:
    case INST_WORD:
    case INST_DWORD:
    case INST_SEQ: /* We have a sequential instruction, let's relocate it! */
      pc = get_inst_addr(first);
      i_size = get_inst_size(first);
      i_code = get_inst_code(first);

      /* Now let us calculate the correct position for this code */
      row = (pc == 0) ? 0: pc / 32; 
      offset = (pc < 32) ? 28 - pc: 32 * row + (28 - pc % 32);

      memcpy(&buffer[offset],i_code,i_size);
      break;
     
      
    case INST_STATE_BEGIN: /* We found the beginning of a state */
      first = get_prev_inst(first); /* Start at label */

      pc = get_inst_addr(first);
      i_size = get_inst_size(first);
      i_code = get_inst_code(first);
      
      /* Calculate buffer indexes */
      row = (pc == 0) ? 0: (pc - 1) / 32;

	a_val[0] = (row + 1) * 32;
	a_val[1] = (row + 1) * 32 - 8;
	a_val[2] = (row + 1) * 32 - 16;
	a_val[3] = (row + 1) * 32 - 24;

	o_val[0] = (row + 1) * 32 - 32; 
	o_val[1] = (row + 1) * 32 - 24;
	o_val[2] = (row + 1) * 32 - 16;
	o_val[3] = (row + 1) * 32 - 8;

	for (i=0;i<4;i++) {
	  if (a_val[i] == pc) {
	    offset = o_val[i];
	    break;
	  }
	}
	      
      /* offset = (pc < 32) ? 32 - pc:32 * (row + 1) - (32 - pc % 32); */

      memcpy(&buffer[offset],i_code,i_size);

      first = get_next_inst(first);

      while ((get_inst_type(first) != INST_STATE_END) &&
	     (get_inst_type(first) != INST_LABEL)) {
	first = get_next_inst(first);
      }
      break;
    }

    first = get_next_inst(first);
  } /* end while(first != NULL) */
  
}
