/* ************************************************************************* */
/* etrax_fs_spu_fsm_optimizer.c: Implementation of the FSM memory optimizer  */
/*                               for the ETRAX FS SPU backend.               */
/*                                                                           */
/*    Copyright 2002-2006 Axis Communications AB.                            */
/*                                                                           */
/*                                                                           */
/*    This program is free software; you can redistribute it and/or modify   */
/*    it under the terms of the GNU General Public License as published by   */
/*    the Free Software Foundation; either version 2 of the License, or      */
/*    (at your option) any later version.                                    */
/*                                                                           */
/*    This program is distributed in the hope that it will be useful,        */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*    GNU General Public License for more details.                           */
/*                                                                           */
/*    You should have received a copy of the GNU General Public License      */
/*    along with this program; if not, write to the Free Software            */
/*    Foundation, Inc., 59 Temple Place - Suite 330,                         */
/*    Boston, MA 02111-1307, USA.                                            */
/*                                                                           */
/*                                                                           */
/*    Send bug reports and inquiries to Johan.Jorgensen@axis.com             */
/* ************************************************************************* */



/* ************************************************************************* */
/*                      INCLUDES USED BY THIS MODULE                         */
/* ************************************************************************* */
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cg_fe.h"
#include "debug.h"
#include "dyn_array.h"
#include "fsm_optimizer.h"
#include "iopasm_err.h" 
#include "iopasm_global.h"
#include "matrix.h"
#include "optimizer_sets.h"
#include "p_list.h"
#include "state_descriptor.h"



/* ************************************************************************* */
/*                        Definitions local to this module                   */
/* ************************************************************************* */
#define MAX_COLS (9)



/* ************************************************************************* */
/*         Static functions defined and implemented in this module           */
/* ************************************************************************* */
static int etrax_fs_spu_repack_states(cg_context_t *context,
				      plist_t *optimized_list,
				      optset_t *bank_set, 
				      optset_t *single_set,
				      int start_addr,int bank_size);



/* ************************************************************************* */
/* PROTOTYPE   : static int etrax_fs_spu_repack_states(cg_context_t *context,*/
/*                                                   plist_t *optimized_list,*/
/*				                     optset_t *bank_set,     */
/*				                     optset_t *single_set,   */
/*				                     int start_addr,         */
/*                                                   int bank_size);         */
/*                                                                           */
/* INPUT PARM's: "context" points to a valid context. "optimized_list points */
/*               to a valid list containing optimized code. "bank_set" is the*/
/*               set of states to be placed in the list. single_set is the   */
/*               set of states that can be placed anywhere. "start_addr" is  */
/*               start address of the bank, and "bank_size is the size of    */
/*               bank.                                                       */
/*                                                                           */
/* FUNCTION    : Repack the states and place them in the new plist, called   */
/*               optimized_list.                                             */
/*                                                                           */
/* RETURN VALUE: Returns the address where the last state was located.       */
/* ************************************************************************* */
static int etrax_fs_spu_repack_states(cg_context_t *context,
				      plist_t *optimized_list,
				      optset_t *bank_set, 
				      optset_t *single_set,
				      int start_addr,int bank_size)
{
  int opt_fsm_size;
  inst_t *state_start,*state_end;
  int fsm_addr;
  state_t *current_state;
  inst_t *new_inst;

  fsm_addr = start_addr;

  while( bank_size > 0) {
    
    opt_fsm_size = 32 - (fsm_addr - (32 * (fsm_addr / 32)));
    current_state = optset_find_sized_state(bank_set,opt_fsm_size,8);

    if (current_state == NULL) {
      current_state = optset_find_sized_state(single_set,opt_fsm_size,8);
    } /* End of if (current_state == NULL) */
    
    if (current_state == NULL) { /* All states are optimized */
      return fsm_addr;
    } /* End of if (current_state == NULL) */

    state_start = get_state_start(current_state);
    state_end = get_state_end(current_state);
    
    do {
      new_inst = copy_inst(state_start);
      add_inst(optimized_list,new_inst);
      state_start = get_next_inst(state_start);
    } while(state_start != get_next_inst(state_end));

    /* Delete the state once we are finished */
    delete_state(current_state);
    
    fsm_addr += get_state_size(current_state);
    bank_size -= get_state_size(current_state);
  } /* End of while (bank_size > 0) */

  return fsm_addr;
}



/* ************************************************************************* */
/* PROTOTYPE   : plist_t *etrax_fs_spu_fsm_optimizer(cg_context_t *ctxt,     */
/*                                                   plist_t *the_list);     */
/*                                                                           */
/* INPUT PARM's: "ctxt" is a pointer to the current context. "the_list"      */
/*               contains the list to optimize.                              */
/*                                                                           */
/* FUNCTION    : Minimize the memory footprint by reorganizing FSM states in */
/*               memory.                                                     */
/*                                                                           */
/* RETURN VALUE: a list containing the optimal state order.                  */
/* ************************************************************************* */
plist_t *etrax_fs_spu_fsm_optimizer(cg_context_t *ctxt,plist_t *the_list) 
{
  plist_t *optimized_list;
  inst_t *current_inst;
  inst_t *tmp_inst;
  int n_states;
  int i,j,k;
  matrix_t *state_matrix;
  optset_t *single_states;
  optset_t *current_set;
  optset_t *tmp_set;
  array_t *optimizer_sets;
  array_t *used_sets;
  int n_sets=0;
  int seq_size;
  int bank_remaining;
  int *set_location;

  int largest;
  int row_len = 0;
  int largest_index = 0;
  optset_t *bank1_set,*bank2_set;
  int fsm_addr;
  int bank_size;
  assert((ctxt !=NULL) && (the_list != NULL));

  optimized_list = create_plist();
  assert(optimized_list != NULL);

  used_sets = create_array(cmp_states,optset_delete);

  seq_size = 0; /* Size of sequential seqtion */
  current_inst = get_first_inst(the_list);

  /* Step 1: Pack all sequential blocks into one. We can assume the following:
  * 1. The optimizer will pack all sequential code and store it at the start of 
  * the SPU memory */

  seq_size = relocate_seq_code(ctxt,optimized_list,the_list);

  /* Next we will need to start building the state dependency matrix */
  n_states = get_inst_count(the_list,INST_STATE_BEGIN);
  iopasm_verbose_msg("         Optimizer found %d states\n",n_states);
  if (n_states == 0) { /* We do not have any states, so we will not optimize */
    
    return the_list; /* Return the same list as the old one */
  }
  
  state_matrix = build_dep_matrix(ctxt,the_list,n_states,MAX_COLS);

  /* If the assembler is running in verbose mode the state matrix should be 
   * dumped. So check the "verbose" parameter and call dump_state-matrix 
   * if necessary.  */

  if (check_option("verbose",(void *)1, OPT_INTEGER)) { 
    dump_state_matrix(state_matrix);
  }

  /* Start building optimizer sets here - That's an architecture independent 
   * function so we simply allocate single_states and optimizer_sets and
   * call the build function */
  single_states = optset_create();
  assert(single_states != NULL);

  optimizer_sets = create_array(cmp_states,optset_delete);
  assert(optimizer_sets != NULL); /* this will hold _all_ sets we find */
  n_sets = build_dep_sets(ctxt,state_matrix,optimizer_sets,single_states);
  iopasm_verbose_msg("\nState & state set statistics:");
  iopasm_verbose_msg("         Optimizer built %d dependency sets (and 1 independent set)",n_sets);
  /* All the code below is architecure dependent! */ 
  
  /* Now we need to add atleast one, perhaps two NOPS to our sequential code */
  i = seq_size % 8;

  i = (i == 0) ? 2: 1; /* If boundary is 64-bit, add two nops otherwise just one */
  seq_size += i * 4; /* Seq size is now adjusted to a 64 bit boundary */

  set_location = IOPASM_MALLOC(sizeof(int) *n_sets);
  assert(set_location != NULL);
  memset(set_location,0,sizeof(int)*n_sets);

  while(i > 0) {
    tmp_inst = create_inst("nop",INST_SEQ,0,RF_ABSOLUTE);
    assert(tmp_inst != NULL);
    add_inst(optimized_list,tmp_inst);
    i--;
  }
  /* End of optimizer sets here */
  /* Now we are ready to start the real optimization process */
  
  bank_remaining = cg_get_memory_size(ctxt) / cg_get_memory_bank_count(ctxt) - seq_size;
  bank_remaining = bank_remaining / cg_get_fsm_size(ctxt);

  i = 0;
  /* select sets to place in first memory bank */
  while ((bank_remaining >0) && (i < n_sets)) { 
    j = 0;
    largest = -1;
    while(j < n_sets) {
      current_set = get_array_elem(optimizer_sets,j);
      k = optset_get_size(current_set);
      if ((k > largest) && (set_location[j] == 0)) {
	largest = optset_get_size(current_set);
	row_len = optset_get_max_rows(current_set);
	largest_index = j;
	i++;
      }
      j++;
    } /* End of while (j<n_sets) */

    if (row_len < bank_remaining) {
      bank_remaining -= row_len;
      set_location[largest_index] = 1; /* Place set #largest_index in bank #1 */
    }
    else {
      bank_remaining = -1; /* Bank one is filled, the rest is 
			    * in bank #2 per default */
    }
  } /* End of while (bank_remaining != 0) */
  
  /* bank1_set and bank2_set contains the sets of states to be placed in bank #1
   * and two respectively. This is simply the union of the sets */

  bank1_set = optset_create();
  assert(bank1_set != NULL);

  bank2_set = optset_create();
  assert(bank2_set != NULL);

  /* place remaining sets in bank #2 - We have now partioned the sets */
  j = 0;
  while (j<n_sets) {
    if (set_location[j] == 0) {
      set_location[j] = 2; /* set is now to be placed in bank #2 */
    }

    if (set_location[j] == 1) {
      tmp_set = bank1_set;
      current_set = (optset_t *) get_array_elem(optimizer_sets,j);
      bank1_set = optset_union(bank1_set,current_set);
      optset_delete(tmp_set);
    }

    if (set_location[j] == 2) {
      tmp_set = bank2_set;
      current_set = (optset_t *) get_array_elem(optimizer_sets,j);
      bank2_set = optset_union(bank2_set,current_set);
      optset_delete(tmp_set);
    }
    j++;
    

  } /* End of while (j<n_sets) */
  

  /* Recalculate set sizes */
  calc_optset_size(ctxt,bank1_set);
  calc_optset_size(ctxt,bank2_set);

  i = optset_get_state_count(single_states);
  j = optset_get_state_count(bank1_set);
  k = optset_get_state_count(bank2_set);

  VERBOSE_MESSAGE(stderr,"         State distribution count: (indep   = %d)\n",i);
  VERBOSE_MESSAGE(stderr,"                                   (bank #1 = %d)\n",j);
  VERBOSE_MESSAGE(stderr,"                                   (bank #2 = %d)\n",k);
  VERBOSE_MESSAGE(stderr,"                                   (total   = %d) (%s)\n",
		  (i+j+k),((i+j+k) == n_states) ? "matches":"does not match");
  
  

  /* Now we can begin performing the actual state optimization */
  fsm_addr = seq_size;
  bank_size = cg_get_memory_size(ctxt)/cg_get_memory_bank_count(ctxt);
  bank_remaining = bank_size - seq_size;

  
  fsm_addr = etrax_fs_spu_repack_states(ctxt,optimized_list,
					bank1_set,single_states,
					fsm_addr,bank_remaining);

  current_inst = synth_org_inst(bank_size);
  add_origin(optimized_list,current_inst);

  fsm_addr = etrax_fs_spu_repack_states(ctxt,optimized_list,
					bank2_set,single_states,
					bank_size,bank_size);
  
  /* Make sure the last instruction points to NULL */
  current_inst = get_current_inst(optimized_list);
  set_next_inst(current_inst,NULL);
  return optimized_list;
 
} /* End of etrax_fs_spu_optimize_states(...) */
