/* ************************************************************************* */
/* fsm_optimizer.c: Front end for the FSM optimizer.                         */
/*                                                                           */
/*    Copyright 2002-2006 Axis Communications AB.                            */
/*                                                                           */
/*                                                                           */
/*    This program is free software; you can redistribute it and/or modify   */
/*    it under the terms of the GNU General Public License as published by   */
/*    the Free Software Foundation; either version 2 of the License, or      */
/*    (at your option) any later version.                                    */
/*                                                                           */
/*    This program is distributed in the hope that it will be useful,        */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*    GNU General Public License for more details.                           */
/*                                                                           */
/*    You should have received a copy of the GNU General Public License      */
/*    along with this program; if not, write to the Free Software            */
/*    Foundation, Inc., 59 Temple Place - Suite 330,                         */
/*    Boston, MA 02111-1307, USA.                                            */
/*                                                                           */
/*                                                                           */
/*    Send bug reports and inquiries to Johan.Jorgensen@axis.com             */
/* ************************************************************************* */



/* ************************************************************************* */
/*                      INCLUDES USED BY THIS MODULE                         */
/* ************************************************************************* */
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cg_fe.h"
#include "debug.h"
#include "dyn_array.h"
#include "iopasm_err.h" 
#include "iopasm_types.h"
#include "matrix.h"
#include "optimizer_sets.h"
#include "p_list.h"
#include "state_descriptor.h"



/* ************************************************************************* */
/* PROTOTYPE   : int match_rows_in_set(matrix_t *the_matrix,optset_t *a_set, */
/*                                     int start_row,int max_rows,           */
/*                                     int max_cols);                        */
/*                                                                           */
/* ARGUMENTS   : "the_matrix" point to a valid matrix. "a_set" is a valid    */
/*               pointer to an optimization set. "start_row" is the row in   */
/*               in the matrix where the matching will start. "max_rows" and */
/*               "max_cols" are the maximum number of rows and columns in    */
/*               the matrix respectively                                     */
/*                                                                           */
/* FUNCTION    : Matches a row in a set.                                     */
/*                                                                           */
/* RETURN VALUE:                                                             */
/* ************************************************************************* */
int match_rows_in_set(matrix_t *the_matrix,optset_t *a_set,int start_row,
		      int max_rows,int max_cols)
{
  int i = 0;
  int j;
  state_t *a_state;
  assert((the_matrix != NULL) && (a_set != NULL));
  
  for (i=start_row;i<max_rows;i++) {
    j = 1;
    while(get_matrix_elem(the_matrix,i,j) != NULL) {
      a_state = (state_t *) get_matrix_elem(the_matrix,i,j);
      if (optset_is_state_in_set(a_set,a_state)) {
	return i;
      }
      j++;
    } /* End of while(get_matrix_elem(the_matrix,i,j) != NULL) */
  } /* End of for (i=start_row;i<max_rows;i++) */
  return -1;
}



/* ************************************************************************* */
/* PROTOTYPE   : void calc_optset_size(cg_context_t *ctxt,optset_t *a_set);  */
/*                                                                           */
/* ARGUMENTS   : "ctxt" is the pointer to the current context, and "a_set"   */
/*               is an optimizer set which size we need to calculate.        */
/*                                                                           */
/* FUNCTION    : Calculate the size parameters for an optimizer set.         */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void calc_optset_size(cg_context_t *ctxt,optset_t *a_set)
{
  state_t *tmp_state;
  int n_states;
  int i;
  unsigned int tmp_size=0;
  int empty_max;  /* Size of empty space in bytes */
  int n_rows;  /* number of rows */
  int arch_row_len;
 
  assert((ctxt != NULL) && (a_set != NULL));
  n_states = optset_get_state_count(a_set);

  i = 0;
  while (i<n_states) {
    tmp_state = optset_get_state_by_index(a_set,i);
    tmp_size += get_state_size(tmp_state);
    i++;
  }
  
  optset_set_size(a_set,tmp_size);

  /* We also calculate some extra parameters */
  arch_row_len = cg_get_fsm_size(ctxt);
  empty_max = arch_row_len - (tmp_size %arch_row_len);
  empty_max = (empty_max == arch_row_len) ? 0: empty_max;
  n_rows = (tmp_size + empty_max) / arch_row_len;
  optset_set_max_rows(a_set,n_rows);
  optset_set_empty_max(a_set,empty_max);
}
  


/* ************************************************************************* */
/* PROTOTYPE   : void fill_opt_set(optset_t *the_set,matrix_t *a_matrix,     */
/*		                   int fill_row,int rows,int cols);          */
/*                                                                           */
/* ARGUMENTS   : "the_set" is the set to populate. "a_matrix" is the state-  */
/*               dependency matrix. "fill_row" is the starting row whereas   */
/*               "rows" and "cols" are the size of the matrix.               */
/*                                                                           */
/* FUNCTION    : Populate an optimizer set "the_set" with dependant states   */
/*               from the state-dependency matrix.                           */
/*                                                                           */
/* RETURN VALUE: None                                                        */
/* ************************************************************************* */
void fill_opt_set(optset_t *the_set,matrix_t *a_matrix,
		  int fill_row,int rows,int cols)
{
  int i,j;
  state_t *tmp_state;
  state_t *find_this;
  assert((the_set != NULL) && (a_matrix !=NULL) &&
	 (fill_row >= 0) && (rows >= 0) && (cols >= 0));
  
  j = 1;
  while (get_matrix_elem(a_matrix,fill_row,j)!= NULL) { /* Still need to find
							 * and add a state */
    find_this = (state_t *) get_matrix_elem(a_matrix,fill_row,j);
    for (i=0;i<rows;i++) {
      tmp_state = get_matrix_elem(a_matrix,i,0);
      if (!cmp_states(tmp_state,find_this)) { /* We found the state */
	optset_add_state(the_set,tmp_state);
	break;
      } /* End of if (!cmp_states(tmp_state,find_this)) */
    } /* End of: for (i=fill_row;i<n_states;i++) */

    j++;
  } /* End of while (get_matrix_elem(a_matrix,fill_row,j)!= NULL) */
}



/* ************************************************************************* */
/* PROTOTYPE   : void dump_state_matrix(matrix_t *a_matrix);                 */
/*                                                                           */
/* ARGUMENTS   : "a_matrix" points to a valid matrix                         */
/*                                                                           */
/* FUNCTION    : Dump contents of the state matrix to stdout. This is a      */
/*               debug function and is not used.                             */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void dump_state_matrix(matrix_t *a_matrix)
{
  int i=0; 
  int index;
  int k;
  char *state_name;
  char txt_buffer[512];
  state_t *tmp_state;
  int rows,cols;

  assert(a_matrix != NULL);

  iopasm_verbose_msg("Dumping State-dependency matrix:");
  
  rows = get_matrix_rows(a_matrix);
  cols = get_matrix_columns(a_matrix);
  iopasm_verbose_msg(" Row |  State                   | Depends on");
  iopasm_verbose_msg("-----+--------------------------+---------------------");

  while(i<rows) {
    index =0;
    memset(txt_buffer,0,512);
    tmp_state = (state_t *) get_matrix_elem(a_matrix,i,0);
    state_name = (char *) get_state_name(tmp_state);
    index = snprintf(txt_buffer,512," %3.3d | %24.24s | ",i,state_name);

    k = 1;
    tmp_state = (state_t *) get_matrix_elem(a_matrix,i,k);
    while ((tmp_state != NULL) && (k < cols)) {
      state_name = get_state_name(tmp_state);
      index += snprintf(txt_buffer+index,512-index,"%s",state_name);
      k++;
      tmp_state = (state_t *) get_matrix_elem(a_matrix,i,k);
      index += snprintf(txt_buffer+index,512-index,"%s", 
			(tmp_state != NULL) ? ", ": " ");
      iopasm_verbose_msg("%s",txt_buffer);
	
    } /* End of while ((tmp_state != NULL) && (k < cols)) */
    i++;
  } /* End of while(i<rows) */
}



/* ************************************************************************* */
/* PROTOTYPE   : void add_dependent_state(matrix_t *the_matrix,              */
/*                                        state_t *a_state,int row);         */
/*                                                                           */
/* ARGUMENTS   : "the_matrix" is a pointer to the state dependency matrix.   */
/*               "a_state" is the state to add and "row" is the row where it */
/*               [the state] is to be added.                                 */
/*                                                                           */
/* FUNCTION    : Adds a state to a certain row in the state dependency       */
/*               matrix.                                                     */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void add_dependent_state(matrix_t *the_matrix,state_t *a_state,int row)
{
  int n_rows,n_cols;
  int i;
  state_t *tmp_state;
  char *tmp_name;
  char *current_name;
  int found = 0;
  assert((the_matrix != NULL) && (a_state != NULL));
  
  n_rows = get_matrix_rows(the_matrix);
  n_cols = get_matrix_columns(the_matrix);

  current_name = get_state_name(a_state);

  i=1;
  tmp_state = (state_t *) get_matrix_elem(the_matrix,row,i);
  
  while ((tmp_state != NULL) && (i<n_cols)) {
    tmp_name = get_state_name(tmp_state);
    if (strcmp(tmp_name,current_name) == 0) {
      break;
      found = 1;
    } /* End of if (strcmp(tmp_name,current_name) == 0) */

    i++;
    tmp_state = (state_t *) get_matrix_elem(the_matrix,row,i);
  } /* End of while ((tmp_state != NULL) && (i<n_cols)) */
  
  if (found == 0) {
    add_matrix_elem(the_matrix,row,i,(void *) a_state);
  }
}



/* ************************************************************************* */
/* PROTOTYPE   : int get_row_count(matrix_t *a_matrix,int row,               */
/*                                 int start_col);                           */
/*                                                                           */
/* ARGUMENTS   : "a_matrix", points to a valid matrix. "row" is the row      */
/*               for which we count the number of elements. "start_col" is   */
/*               the column at which we start (i.e. we can skip elements     */
/*               prior to an arbitrary element).                             */
/*                                                                           */
/* FUNCTION    : Counts the number of elements in a given row, starting at an*/
/*               arbitrary column.                                           */
/*                                                                           */
/* RETURN VALUE: number of non-NULL elements starting from (row,col)         */
/* ************************************************************************* */
int get_row_count(matrix_t *a_matrix,int row,int start_col)
{
  int i=start_col;
  void *elem;
  int n_cols;

  assert(a_matrix != NULL);

  n_cols = get_matrix_columns(a_matrix);
  elem = get_matrix_elem(a_matrix,row,i);
 
 while ((elem != NULL) && (i < n_cols)) {
   i++;
   elem = get_matrix_elem(a_matrix,row,i);
  }

  return i-1;
}



/* ************************************************************************* */
/* PROTOTYPE   : int match_dependent_rows(matrix_t *the_matrix,              */
/*                                        int start_row, int n_rows,         */
/*                                        int n_cols);                       */
/*                                                                           */
/* ARGUMENTS   : "the_matrix" is a valid matrix. "start_row" is the row where*/
/*               the search for dependant states begin. "n_rows" and "n_cols"*/
/*               define the size of the matrix.                              */
/*                                                                           */
/* FUNCTION    : Finds the first row after "start_row" that contains 1 or    */
/*               more dependant states.                                      */
/*                                                                           */
/* RETURN VALUE: Matched row if found. -1 on failure.                        */
/* ************************************************************************* */
int match_dependent_rows(matrix_t *the_matrix,int start_row,
			 int n_rows,int n_cols)
{
  state_t *s1,*s2;
  char *state_name1,*state_name2;
  int i,j,k;
  int ndep_states; /* number of dependent states in row */
  int state_count=0;

  assert(the_matrix != NULL);
  
  state_count = get_row_count(the_matrix,start_row,1);

  if (state_count == 0) { /* No states in row, so d not do anything further */
    return -1;
  }
  
  /* Find first row with _any_ matching states */
  for (i=start_row+1;i<n_rows;i++) { 

    /* First count number of dependent states in search_row */
    ndep_states = get_row_count(the_matrix,i,1);
    
    /* Let's se if there are more than one unique state in _any_ of the two rows */
    if ((ndep_states > 1) || (state_count > 1)) {

      j = 1; /* Starting column in start_row */
      s1 = (state_t *) get_matrix_elem(the_matrix,start_row,j);

      while ((s1 != NULL) && (j<(n_cols - 1))) {

	k = 1; /* Starting column in row i */
	s2 = (state_t *) get_matrix_elem(the_matrix,i,k);
	state_name1 = get_state_name(s1);

	while((s2 != NULL) && (k<n_cols)) {
	  state_name2 = get_state_name(s2);
	  
	  if (strcmp(state_name1,state_name2) == 0) { /* We have a match */
	    return i; 
	  } /* End of if (strcmp(state_name1,state_name2) */
	  
	  k++;
	  s2 = (state_t *) get_matrix_elem(the_matrix,i,k);
	  
	} /* End of while((s2 != NULL) && (k<n_cols)) */
		
	j++;
	s1 = (state_t *) get_matrix_elem(the_matrix,start_row,j);
      } /* End of while ((s1 != NULL) && (j<(n_cols - 1))) */
      
    } /* End of if ((ndep_states > 1) || (state_count > 1)) */
  } /* End of for(k=start_row+1;k<n_rows;k++) */
  return -1;
}



/* ************************************************************************* */
/* PROTOTYPE   : int relocate_seq_code(cg_context_t *ctxt,plist_t *dest,     */
/*                                     plist_t *src);                        */
/*                                                                           */
/* ARGUMENTS   : "ctxt" is the current context. "dest" is the new list in    */
/*               which the optimized code is placed. "src" is the original   */
/*               instruction list.                                           */
/*                                                                           */
/* FUNCTION    : Moves all sequential code from "seq" to the beginning of    */
/*               "dest".                                                     */
/*                                                                           */
/* RETURN VALUE: Size of sequential code in bytes.                           */
/* ************************************************************************* */
int relocate_seq_code(cg_context_t *ctxt,plist_t *dest,plist_t *src)
{
  inst_t *tmp_inst;
  inst_t *current_inst;
  inst_t *next_inst;
  int seq_size;

  assert((ctxt != NULL) && (dest != NULL) && (src != NULL));

  seq_size = 0; /* Return the size of the sequential code in bytes */

  current_inst = get_first_inst(src);

  while (current_inst != NULL) {
    switch (get_inst_type(current_inst)) {
      
    case INST_ORIGIN: /* We should not copy alignment and origin instructions */
    case INST_ALIGN:
      break;

    case INST_LABEL:
      next_inst = get_next_inst(current_inst);
      if (get_inst_type(next_inst) == INST_STATE_BEGIN) {
	while(get_inst_type(next_inst) != INST_STATE_END) {
	  next_inst = get_next_inst(next_inst);
	} /* End of while(get_inst_type(next_inst) != INST_STATE_END) */
	current_inst = next_inst;
      } /* End of if (get_inst_type(next_inst) == INST_STATE_BEGIN) */
        
      else {
	tmp_inst = copy_inst(current_inst);
	assert(tmp_inst != NULL);
	add_inst(dest,tmp_inst);
	next_inst = current_inst;
      } /* End of !if (get_inst_type(next_inst) == INST_STATE_BEGIN) */
      break;

    case INST_SEQ:
    case INST_DWORD:
    case INST_BYTE:
    case INST_WORD:
      tmp_inst = copy_inst(current_inst);
      assert(tmp_inst != NULL);
      add_inst(dest,tmp_inst);
      seq_size += get_inst_size(tmp_inst);
      break;
    } /* End of switch(get_inst_type(current_inst)) */

    /* Get next instruction */
    current_inst = get_next_inst(current_inst);
    
  } /* End of while (current_inst != NULL) */
  
  return seq_size;
}



/* ************************************************************************* */
/* PROTOTYPE   : int is_entry_point(cg_context_t *ctxt,matrix_t *a_matrix,   */
/*                                  int this_row);                           */
/*                                                                           */
/* ARGUMENTS   : "ctxt" is the current code generation context. "a_matrix"   */
/*               is the state dependency matrix and "this_row" is the row to */
/*               check.                                                      */
/*                                                                           */
/* FUNCTION    : Check if a state (found in column 0 in the matrix) is an    */
/*               FSM entry point.                                            */
/*                                                                           */
/* RETURN VALUE: returns 1 if state is an entry point, 0 otherwise           */
/* ************************************************************************* */
int is_entry_point(cg_context_t *ctxt,matrix_t *a_matrix,int this_row)
{
  int rows,cols;
  int i,j;
  state_t *find_this;
  state_t *tmp_state;
  int entry_point; /* Return value 1= state is entry point */

  assert((ctxt != NULL) && (a_matrix != NULL));

  rows = get_matrix_rows(a_matrix);
  cols = get_matrix_columns(a_matrix);
  
  find_this = (state_t *) get_matrix_elem(a_matrix,this_row,0);

  i = 0;
  entry_point = 1; /* We will assume that it is an entry point
		    * per default */

  while (i < rows) {
    j = 1; 
    tmp_state = (state_t *) get_matrix_elem(a_matrix,i,j);
    while (tmp_state != NULL) {
      if (cmp_states(find_this,tmp_state) == 0) { /* Found it! */
	entry_point = 0; /* As it turned out it was not an entry point */
      } /* End of if (cmp_states(find_this,tmp_state) == 0) */
      j++;
      tmp_state = (state_t *) get_matrix_elem(a_matrix,i,j);
    } /* End of while (tmp_state != NULL) */

    i++;
  }  /* End of while (i < rows) */
  return entry_point;
} 



/* ************************************************************************* */
/* PROTOTYPE   : matrix_t *build_dep_matrix(cg_context_t *ctxt,              */
/*                                          plist_t *the_list,               */
/* 			                    int max_rows,int max_cols);      */
/*                                                                           */
/* ARGUMENTS   : "ctxt" is the current context, and "the_list is the un-     */
/*               optimized list. "max_rows" and "max_cols" is the size of    */
/*               of the matrix to build.                                     */
/*                                                                           */
/* FUNCTION    : Builds a state dependency matrix with size "max_rows" x     */
/*               "max_cols".                                                 */
/*                                                                           */
/* RETURN VALUE: Pointer to a state dependency matrix, (NULL on failure)     */
/* ************************************************************************* */
matrix_t *build_dep_matrix(cg_context_t *ctxt,plist_t *the_list,
			   int max_rows,int max_cols)
{
  inst_t *current_inst;
  matrix_t *tmp_dep_matrix;
  int row,col;
  inst_t *state_start,*state_end;
  int state_size;
  state_t *current_state,*dependent_state;
  arg_t *temp_arg;
  char *arg_val,*state_name;

  assert((ctxt != NULL) && (the_list != NULL));

    /* Now find all states and build a state transition matrix */
  tmp_dep_matrix = create_matrix(max_rows,max_cols,sizeof(state_t *), free);
  assert(tmp_dep_matrix != NULL);
  row = 0; col = 0;

  current_inst = get_first_inst(the_list);

  /* First and foremost we need to build a state-dependency matrix 
   * so that we know which states depend on each other */

  do {
    while(get_inst_type(current_inst) != INST_STATE_BEGIN && 
	  (current_inst!= NULL)) {
      
      current_inst = get_next_inst(current_inst);
    }
    
    /* If we have found a state let's add it to the matrix and analyze it's
     * dependencies */
    
    if (current_inst != NULL) { /* Find dependencies */
      state_start = get_prev_inst(current_inst);
      state_size = get_inst_size(state_start);
      
      state_end = get_next_inst(state_start);

      while(get_inst_type(state_end) != INST_STATE_END) {
	state_end = get_next_inst(state_end);
      }
            
      state_name = get_inst(state_start);

      current_state = create_state(state_name,state_start,
				   state_end,state_size);

      assert (current_state != NULL);
      add_matrix_elem(tmp_dep_matrix,row,col,current_state);
      col = 1;
      
      do {
	switch(get_inst_type(current_inst)) {


	case INST_STATE_XFER:
	  temp_arg = get_first_arg(current_inst);
	  temp_arg = get_next_arg(temp_arg);
	  temp_arg = get_next_arg(temp_arg);
	  arg_val = get_arg_val(temp_arg);
	  dependent_state = create_state(arg_val,NULL,NULL,0);
	  assert(dependent_state != NULL);
	  add_dependent_state(tmp_dep_matrix,dependent_state,row);
	  col++;
	  break;


	case INST_TIMER:
	  temp_arg = get_first_arg(current_inst);
	  while(get_next_arg(temp_arg) != NULL) {
	    temp_arg = get_next_arg(temp_arg);
	  }
	  arg_val = get_arg_val(temp_arg);
	  dependent_state = create_state(arg_val,NULL,NULL,0);
	  assert(dependent_state != NULL);
	  add_dependent_state(tmp_dep_matrix,dependent_state,row);
	  col++;
	  break;


	case INST_ONLY_STATE:
	  temp_arg = get_first_arg(current_inst);
	  arg_val = get_arg_val(temp_arg);
	  dependent_state = create_state(arg_val,NULL,NULL,0);
	  assert(dependent_state != NULL);
	  add_dependent_state(tmp_dep_matrix,dependent_state,row);
	  col++;
	  break;


	case INST_ORIGIN:
	  /* Remove .ORG statements from output code */
	  break;


	default:
	  break;
	} /* End of switch(get_inst_type(current_inst)) */
	current_inst = get_next_inst(current_inst);

      } while(get_inst_type(current_inst) != INST_STATE_END);
    } /* End of if (current_inst != NULL) */

    else { /* Matrix is ready. Do nothing */
      ;
    }
    col = 0; 
    row++;
  } while(current_inst != NULL);

  /* Now we have a complete state transition matrix - Return it to the 
   * caller */
  return tmp_dep_matrix;
}



/* ************************************************************************* */
/* PROTOTYPE   : int build_dep_sets(cg_context_t *ctxt,                      */
/*                                  matrix_t *state_matrix,                  */
/*				    array_t *found_sets,                     */
/*				    optset_t *indep_states);                 */
/*                                                                           */
/* ARGUMENTS   : "ctxt" is the current context and "state_matrix" is the     */
/*               state dependency matrix. "found sets" contains all sets     */
/*               of states that have been found. "indep_states" is the set   */
/*               of independent states.                                      */
/*                                                                           */
/* FUNCTION    : Identifies and creates the sets in the state-dependency     */
/*               matrix.                                                     */
/*                                                                           */
/* RETURN VALUE: The number of sets found (and stored in "found_sets").      */
/* ************************************************************************* */
int build_dep_sets(cg_context_t *ctxt,matrix_t *state_matrix,
		   array_t *found_sets,optset_t *indep_states)
{

  state_t *current_state;
  int row,n_states;
  int found_row;
  int max_cols;
  int i,j;
  optset_t *current_set;
  int n_sets;

  assert((ctxt != NULL) && (state_matrix != NULL) &&
	 (found_sets != NULL) && (indep_states != NULL));

  n_states = get_matrix_rows(state_matrix);
  max_cols = get_matrix_columns(state_matrix);
  row = 0;
  n_sets = 0;

  while (row < n_states) {

    /* The first we need to do is to check that the state is not an 
     * entry point. If it is, we need to add it to single states */
    if (is_entry_point(ctxt,state_matrix,row) == 1) { 
      current_state = (state_t *) get_matrix_elem(state_matrix,row,0);
      optset_add_state(indep_states,current_state);
    } /* End of if (is_entry_point(ctxt,state_matrix,row) == 0) */

    row++;

  } /* End of while (row < n_states) */

  row = 0;
  
  while(row < n_states) { /* walk through the matrix and find the sets 
			 * defined in the state_matrix */
    
    /* Now we start to search for any dependent states in the current row */
    found_row = match_dependent_rows(state_matrix,row,n_states,max_cols);
  
    /* Rows found_row and row is dependent if found_row is larger than zero.
     * if found_row is -1 and there is only one dependent state in "row", it 
     * should be added to indep_states */
    
    if (found_row == -1) { /* Check if there is only one dependent state and 
			    * added to single_list otherwise create a new set */
      i = 1;
      j = 0;
      current_set = NULL;
      while (get_matrix_elem(state_matrix,row,i) != NULL) {
	i++; j++;
      } /* End of while (get_matrix_elem(state_matrix,row,i) != NULL) */

      if (j==0) { /* No states to add just continue */
	;
      } /* End of !if (j==0) */
     
      else if (j>1) { /* We need to create and initialize a new optizer set and
		  * add this to our set of sets */
	current_set = optset_create();
	assert(current_set != NULL);
	fill_opt_set(current_set,state_matrix,row,n_states,max_cols);
      } /* End of if (j>1) */
      
      else { /* Add row to single states */

       	fill_opt_set(indep_states,state_matrix,row,n_states,max_cols);
      } /* End of !if (j>1) */
    } /* End of !if (found_row == -1) */

    
    else { /* We found a row - create a set containing "row" and "found_row"
	    * Continue to search until we  cant find anymore matching rows */
      current_set = optset_create();
      assert(current_set != NULL);
      fill_opt_set(current_set,state_matrix,row,n_states,max_cols);
      clear_matrix_row(state_matrix,row,max_cols);
      /* We need to look through the entire matrix in order to make sure
       * that we find all matching states */
      do {
      fill_opt_set(current_set,state_matrix,found_row,n_states,max_cols);
      
      /* Now we have added found row, as it is ahead of "row" we delete 
       * it and start searching from row again */
      clear_matrix_row(state_matrix,found_row,max_cols);
      found_row = match_rows_in_set(state_matrix,current_set,row,n_states,max_cols);
      } while(found_row != -1);
    } /* End of !if (found_row == -1) */

    if (current_set != NULL) { /* We found a set of states, so we 
				  added to our array of sets */
      add_array_elem(found_sets,current_set);
      n_sets++; 
      current_set = NULL;
    }

    /* We are now ready to process the next row in the matrix. 
     * So let's delete the the current row and repeat the process */
    clear_matrix_row(state_matrix,row,max_cols);
    row++;
  } /* End of while(row<n_states)  */

  /* Now all sets are stored in the dynamic array called found_sets.
   * "All" (not an easy feat) we need to reorder them in a more efficient 
   * way. */

  calc_optset_size(ctxt,indep_states);
  i = 0;
  while (get_array_elem(found_sets,i) != NULL) {
    current_set = (optset_t *) get_array_elem(found_sets,i);
    calc_optset_size(ctxt,current_set);
    i++;
  } /* end of while (get_array_elem(optimizer_sets,i) != NULL) */
  

  return n_sets;
}
