/* ************************************************************************* */
/* i_table.c: Various architecture independent functions used to manipulate  */
/*            instruction table entries.                                     */
/*                                                                           */
/*    Copyright 2002-2006 Axis Communications AB.                            */
/*                                                                           */
/*                                                                           */
/*    This program is free software; you can redistribute it and/or modify   */
/*    it under the terms of the GNU General Public License as published by   */
/*    the Free Software Foundation; either version 2 of the License, or      */
/*    (at your option) any later version.                                    */
/*                                                                           */
/*    This program is distributed in the hope that it will be useful,        */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*    GNU General Public License for more details.                           */
/*                                                                           */
/*    You should have received a copy of the GNU General Public License      */
/*    along with this program; if not, write to the Free Software            */
/*    Foundation, Inc., 59 Temple Place - Suite 330,                         */
/*    Boston, MA 02111-1307, USA.                                            */
/*                                                                           */
/*                                                                           */
/*    Send bug reports and inquiries to Johan.Jorgensen@axis.com             */
/* ************************************************************************* */



/* ************************************************************************* */
/*                      INCLUDES USED BY THIS MODULE                         */
/* ************************************************************************* */
#include <assert.h>
#include <ctype.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cg_fe.h"
#include "iopasm_err.h"
#include "iopasm_global.h"
#include "i_table.h"
#include "i_table_types.h"
#include "p_list.h"
#include "stack_eval.h"


/* ************************************************************************* */
/*                     DEFINITIONS LOCAL TO THIS MODULE                      */
/* ************************************************************************* */
#define SRC_FIELD 0
#define DST_FIELD 2
#define ANY_FIELD 1
#define MAX_INSTLEN 8

#define IS_SRC_REG_ALLOWED(TABLE,X) ((TABLE[X].arg_type[0] == R) || \
                                     (TABLE[X].arg_type[0] == S))

#define IS_DST_REG_ALLOWED(TABLE,X) ((TABLE[X].arg_type[2] == R) || \
                                     (TABLE[X].arg_type[2] == S))

#define IS_ANY_REG_ALLOWED(TABLE,X) ((TABLE[X].arg_type[1] == R) || \
                                     (TABLE[X].arg_type[1] == S))

#define IS_SRC_IMM_ALLOWED(TABLE,X) (TABLE[X].arg_type[0] == I)
#define IS_DST_IMM_ALLOWED(TABLE,X) (TABLE[X].arg_type[2] == I)
#define IS_ANY_IMM_ALLOWED(TABLE,X) (TABLE[X].arg_type[1] == I)

#define SRC_IMMEDIATE_MAX(ITABLE,INUM) ((unsigned int)((((unsigned long long) 1) << ITABLE[INUM].field_len[0]) - 1))
#define ANY_IMMEDIATE_MAX(ITABLE,INUM) ((unsigned int)((((unsigned long long) 1) << ITABLE[INUM].field_len[1]) - 1))
#define DST_IMMEDIATE_MAX(ITABLE,INUM) ((unsigned int)((((unsigned long long) 1) << ITABLE[INUM].field_len[2]) - 1))



/* ************************************************************************* */
/*                       TYPES DEFINED IN THIS MODULE                        */
/* ************************************************************************* */



/* ************************************************************************* */
/*                              CONSTANT VARIABLES                           */
/* ************************************************************************* */



/* ************************************************************************* */
/* All registers known to _any_ architecture should be defined in this table */
/* It is used to check that we really found a register and nothing else.     */
/* This should really not be necessary but as the parser only sees a         */
/* register and knows nothing about the selected architecture.               */
/* ************************************************************************* */
static const reg_table_t global_regs[] = {
  { "r0",   1 }, { "r1",   1 }, { "r2",   1 }, { "r3",   1 }, 
  { "r4",   1 }, { "r5",   1 }, { "r6",   1 }, { "r7",   1 }, 
  { "r8",   1 }, { "r9",   1 }, { "r10",  1 }, { "r11",  1 }, 
  { "r12",  1 }, { "r13",  1 }, { "r14",  1 }, { "r15",  1 }, 
  { "r16",  1 }, { "r17",  1 }, { "r18",  1 }, { "r19",  1 },
  { "r20",  1 }, { "r21",  1 }, { "r22",  1 }, { "r23",  1 },
  { "r24",  1 }, { "r25",  1 }, { "r26",  1 }, { "r27",  1 },
  { "r28",  1 }, { "r29",  1 }, { "r30",  1 }, { "r31",  1 }, 
  { "pc",   1 }, { "wsts", 1 }, { "irp",  1 }, { "srp",  1 },
  { "t0",   1 }, { "t1",   1 }, { "t2",   1 }, { "t3",   1 },
  { "i0",   1 }, { "i1",   1 }, { "i2",   1 }, { "i3",   1 },
  { "i4",   1 }, { "i5",   1 }, { "i6",   1 }, { "i7",   1 },
  { "i8",   1 }, { "i9",   1 }, { "i10",  1 }, { "i11",  1 },
  { "i12",  1 }, { "i13",  1 }, { "i14",  1 }, { "i15",  1 },
  { "p2",   1 }, { "p3",   1 }, { "p5",   1 }, { "p6",   1 },
  { "p8",   1 }, { "p9",   1 }, { "p10",  1 }, { "p11",  1 },
  { "p16",  1 }, { "p17",  1 }, { "p18",  1 }, { "p19",  1 },
  { "p20",  1 }, { "p21",  1 }, { "p22",  1 }, { "p23",  1 },
  { "p24",  1 }, { "p25",  1 }, { "p26",  1 }, { "p27",  1 },
  { "p28",  1 }, { "p29",  1 }, { "p30",  1 }, { "p31",  1 },
  { NULL,   0 }};



/* ************************************************************************* */
/*                    PROTOTYPES FOR STATIC FUNCTIONS                        */
/* ************************************************************************* */
static int is_register(char *reg_name);
static int to_upper(char *dst,char *src,int n);



/* ************************************************************************* */
/* PROTOTYPE   : static int to_upper(char *dst,char *src,int n);             */
/*                                                                           */
/* ARGUMENTS   : "dst" is the destination string, "src" is the source  and   */
/*               "n" is the maximal length.                                  */
/*                                                                           */
/* FUNCTION    : Converts a string from lower case to upper case.            */
/*                                                                           */
/* RETURN VALUE: Number of characters converted.                             */
/* ************************************************************************* */
static int to_upper(char *dst,char *src,int n)
{
  int i=0;
  assert ((dst != NULL) && (src != NULL) && (n > 0));

  while ((i<n) && (src[i] != 0)) {
    dst[i] = toupper(src[i]);
    i++;
  }
   return i;
}



/* ************************************************************************* */
/* PROTOTYPE   : static int is_register(char *reg_name);                     */
/*                                                                           */
/* ARGUMENTS   : "reg_name" points to a string containing a register name.   */
/*                                                                           */
/* FUNCTION    : Check if "reg_name" is a valid register name.               */
/*                                                                           */
/* RETURN VALUE: Returns >= 0 if "reg_name" is a valid register. -1 if not   */
/* ************************************************************************* */
static int is_register(char *reg_name)
{
  int i = 0;

  while (global_regs[i].reg_name != NULL) {
    if (!strcmp(global_regs[i].reg_name,reg_name)) { 
      return global_regs[i].reg_val;
    }

    else { 
      i++;
    }
  }
  
  return -1;
}



/* ************************************************************************* */
/* PROTOTYPE   : int reg_type_detect(reg_table_t *regs,                      */
/*                                reg_table_t *special_regs,char *reg_name); */
/*                                                                           */
/* ARGUMENTS   : "regs" is a valid register table containing the names of    */
/*               all normal registers. "special_regs" is the equivalent table*/
/*               containing all special registers, and "reg_name" is the     */
/*               register name which type is to be determined.               */
/*                                                                           */
/*                                                                           */
/* FUNCTION    : Determine if the register referenced by "reg_name" is a     */
/*               normal or special register.                                 */
/*                                                                           */
/* RETURN VALUE: Returns the type of register. -1 if the register is not     */
/*               known in the current architecture.                          */
/* ************************************************************************* */
int reg_type_detect(reg_table_t *regs,reg_table_t *special_regs,
		    char *reg_name)
{
  int i;
  
  i = 0;

  assert (regs != NULL);

  
  while(regs[i].reg_name != NULL) {

    if (!strcmp(reg_name,regs[i].reg_name)) {
      return REGULAR_REG; /* We found a normal register */
    }
    
    else { i++; }
  }

  i = 0;

  if (special_regs != NULL) { 
    while(special_regs[i].reg_name != NULL) {
      
      if (!strcmp(reg_name,special_regs[i].reg_name)) {
	return SPECIAL_REG; /* We found a normal register */
      }
      
      else { i++; }
    }
  }

  return -1; 
}



/* ************************************************************************* */
/* PROTOTYPE   : unsigned int get_src_regval(cg_context_t ctxt,inst_t *inst, */
/*                              	     int i_num,int reg_num);         */
/*                                                                           */
/* ARGUMENTS   : "inst" is the current instruction. "i_num" and "reg_num"    */
/*               are instruction number and register numbers respectively    */
/*               (table indexes).                                            */
/*                                                                           */
/* FUNCTION    : Calculate the part of the opcode for instruction i_num that */
/*               depends on reg_num as the source register.                  */
/*                                                                           */
/* RETURN VALUE: Partial op-code                                             */
/* ************************************************************************* */
unsigned int get_src_regval(cg_context_t * ctxt,inst_t *inst, 
			    int i_num,int reg_num)
{
  unsigned int r_val;
  unsigned int mask;
  inst_table_t *i_table;
  reg_table_t *reg_table;

  assert((ctxt != NULL) && (inst != NULL));
    
  /* Get the instruction and register tables from the context */
  i_table = cg_get_inst_table(ctxt);
  reg_table = cg_get_reg_table(ctxt);
  
  if (!IS_SRC_REG_ALLOWED(i_table,i_num)) { 
    iopasm_error(ctxt,inst,ERR_REGISTER_NOT_ALLOWED,
		 "Cannot use a register (%s) as source\n",
		 reg_table[reg_num].reg_name);
  }

  else if (i_table[i_num].field_len[SRC_FIELD] > 0) { 
    mask = (2 << (i_table[i_num].field_len[SRC_FIELD] - 1)) - 1;

    mask = mask << (i_table[i_num].arg_field[SRC_FIELD] - 
		    (i_table[i_num].field_len[SRC_FIELD] - 1));
    
    r_val = (reg_table[reg_num].reg_val << 
	     ((i_table[i_num].arg_field[SRC_FIELD]) - 
	     (i_table[i_num].field_len[SRC_FIELD] - 1)));

    return (r_val & mask);
  }

  else {
    return 0;
  }

  iopasm_error(NULL,NULL,ERR_REPORT_BUG,"Reported from %s:%d",__FILE__,__LINE__);
  return 0; /* To get rid of warning - will never be reached.
	     * This should be preceeded by an error causing a 
	     * bug-report because it should simply never happen.
	     * If not it is a bug in itself */
}



/* ************************************************************************* */
/* PROTOTYPE   : unsigned int get_dst_regval(cg_context_t ctxt,inst_t *inst, */
/*                              	     int i_num,int reg_num);         */
/*                                                                           */
/* ARGUMENTS   : "inst" is the current instruction. "i_num" and "reg_num"    */
/*               are instruction number and register numbers respectively    */
/*               (table indexes).                                            */
/*                                                                           */
/* FUNCTION    : Calculate the part of the opcode for instruction i_num that */
/*               depends on reg_num as the destination register.             */
/*                                                                           */
/* RETURN VALUE: Partial op-code                                             */
/* ************************************************************************* */
unsigned int get_dst_regval(cg_context_t *ctxt,inst_t *inst,
			    int i_num,int reg_num)
{
  unsigned int mask;
  unsigned int r_val;
  inst_table_t *i_table;
  reg_table_t *reg_table;

  assert((ctxt != NULL) && (inst != NULL));
    
  /* Get the instruction and register tables from the context */
  i_table = cg_get_inst_table(ctxt);
  reg_table = cg_get_reg_table(ctxt);

  if (!IS_DST_REG_ALLOWED(i_table,i_num)) { 
    iopasm_error(ctxt,inst,ERR_REGISTER_NOT_ALLOWED,
		 "Cannot use a register (%s) as source\n",
		 reg_table[reg_num].reg_name);
  }

  else if  (i_table[i_num].field_len[DST_FIELD] > 0) { 
    mask = (2 << (i_table[i_num].field_len[DST_FIELD] - 1)) - 1;

    mask = mask << (i_table[i_num].arg_field[DST_FIELD] - 
		    (i_table[i_num].field_len[DST_FIELD] - 1));

    r_val = (reg_table[reg_num].reg_val << 
	     ((i_table[i_num].arg_field[DST_FIELD]) -
	     (i_table[i_num].field_len[DST_FIELD] - 1)));
    return (r_val & mask);
  }

  else {
    return 0;
  }
  
  iopasm_error(NULL,NULL,ERR_REPORT_BUG,"Reported from %s:%d",__FILE__,__LINE__);
  return 0; /* To get rid of warning - will never be reached.
	     * This should be preceeded by an error causing a 
	     * bug-report because it should simply never happen.
	     * If not it is a bug in itself */
}



/* ************************************************************************* */
/* PROTOTYPE   : unsigned int get_any_regval(cg_context_t ctxt,inst_t *inst, */
/*                              	     int i_num,int reg_num);         */
/*                                                                           */
/* ARGUMENTS   : "inst" is the current instruction. "i_num" and "reg_num"    */
/*               are instruction number and register numbers respectively    */
/*               (table indexes).                                            */
/*                                                                           */
/* FUNCTION    : Calculate the part of the opcode for instruction i_num that */
/*               depends on reg_num as the "any" (middle-operand) register.  */
/*                                                                           */
/* RETURN VALUE: Partial op-code                                             */
/* ************************************************************************* */
unsigned int get_any_regval(cg_context_t *ctxt,inst_t *inst,
			    int i_num,int reg_num)
{
  unsigned int r_val;
  unsigned int mask;
  inst_table_t *i_table;
  reg_table_t *reg_table;

  assert((ctxt != NULL) && (inst != NULL));
    
  /* Get the instruction and register tables from the context */
  i_table = cg_get_inst_table(ctxt);
  reg_table = cg_get_reg_table(ctxt);

  if (!IS_ANY_REG_ALLOWED(i_table,i_num)) { 
        iopasm_error(NULL,inst,ERR_REGISTER_NOT_ALLOWED,
		     "Cannot use a register (%s) as the second operand\n",
		     reg_table[reg_num].reg_name);


  }

  else if (i_table[i_num].field_len[ANY_FIELD] > 0) { 
    mask = (2 << (i_table[i_num].field_len[ANY_FIELD] - 1)) - 1;

    mask = mask << (i_table[i_num].arg_field[ANY_FIELD] - 
		    (i_table[i_num].field_len[ANY_FIELD] - 1));

    r_val = (reg_table[reg_num].reg_val << 
	     ((i_table[i_num].arg_field[ANY_FIELD]) - 
	     (i_table[i_num].field_len[ANY_FIELD] - 1)));

    return (r_val & mask);
  }

  else {
    return 0;
  }


  iopasm_error(NULL,NULL,ERR_REPORT_BUG,"Reported from %s:%d",__FILE__,__LINE__);
  return 0; /* To get rid of warning - will never be reached.
	     * This should be preceeded by an error causing a 
	     * bug-report because it should simply never happen.
	     * If not it is a bug in itself */
}



/* ************************************************************************* */
/* PROTOTYPE   : unsigned int get_opcode(inst_table_t *i_table,int i_num);   */
/*                                                                           */
/* ARGUMENTS   : "i_table" is the instruction table associated with the cur- */
/*               rent architecture and "i_num" is the instruction number     */
/*               (table index).                                              */
/*                                                                           */
/* FUNCTION    : Returns the non-operand part of the op-code (refered to as  */
/*               the instruction template) for a specific instruction        */
/*                                                                           */
/* RETURN VALUE: Returns the instruction template.                           */
/* ************************************************************************* */
unsigned int get_opcode(inst_table_t *i_table,int i_num)
{
  unsigned int *ptr,template;
  
  assert(i_table != NULL);
  ptr = (unsigned int *) i_table[i_num].i_template;
  template = (unsigned int) ntohl(*ptr); /* For debugging purposes  */
  return template; 
}



/* ************************************************************************* */
/* PROTOTYPE   : unsigned int generate_opcode(void *ctxt,                    */
/*                                            inst_table_t *i_table,         */
/*                                            inst_t *inst,                  */
/*                     			      int i_num,int pass);           */
/*                                                                           */
/* ARGUMENTS   : "ctxt" is the current context" "i_table" is the instruction */
/*                table. "inst" points to the current instruction. "i_num" is*/
/*                the instruction number (table index) and "pass" is the     */
/*                assembler pass.                                            */
/*                                                                           */
/* FUNCTION    : Generates the op-code (instuction template) for an inst-    */
/*               ruction.                                                    */
/*                                                                           */
/* RETURN VALUE: Returns the instruction template.                           */
/* ************************************************************************* */
unsigned int generate_opcode(cg_context_t *ctxt,inst_table_t *i_table,
			     inst_t *inst,int i_num,int pass)
{
  int i;
  unsigned int *ptr,template;
  unsigned char buffer[MAX_INSTLEN];

  assert ((i_table != NULL) && 
	  (ctxt != NULL) && 
	  (inst != NULL));
  i = 0;
  ptr = (unsigned int *) i_table[i_num].i_template;
  template = (unsigned int) ntohl(*ptr);
  
  /* Should we check our rules, this is only done in the intermediate pass */
  if ((i_table[i_num].rule_set != NULL) && (pass == FINAL_PASS)) {  
    while (i_table[i_num].rule_set->rule_func[i] != NULL) {
      if ( (i+1) == i_table[i_num].rule_set->gen_func) {
	i_table[i_num].rule_set->rule_func[i](ctxt,inst,i_num,buffer,MAX_INSTLEN);
	ptr = (unsigned int *) buffer;
	template = *ptr;
      }
      else {
	i_table[i_num].rule_set->rule_func[i](ctxt,inst,i_num,buffer,MAX_INSTLEN);
      }
      
      i++;
    }

  } /* End of if ((i_table[i_num].rule_set != NULL) && ... */

  return template;
}



/* ************************************************************************* */
/* PROTOTYPE   : int get_inst_len(inst_table_t *i_table,int i_num);          */
/*                                                                           */
/* ARGUMENTS   : "i_table" is the instruction table and "i_num" is the table */
/*               index.                                                      */
/*                                                                           */
/* FUNCTION    : Finds the length of an instruction measured in bytes.       */
/*                                                                           */
/* RETURN VALUE: Returns the length of an instruction measured in bytes.     */
/* ************************************************************************* */
int get_inst_len(inst_table_t *i_table,int i_num)
{
  assert(i_table != NULL);
  return i_table[i_num].inst_len;
}

  


/* ************************************************************************* */
/* PROTOTYPE   : int get_reg_number(reg_table_t *reg_table,char *reg_name);  */
/*                                                                           */
/* ARGUMENTS   : "reg_table" points to a register table. "reg_name" points   */
/*               to a register name (string).                                */
/*                                                                           */
/* FUNCTION    : Returns the register number (index in table) where the      */
/*               register "reg_name" is found.                               */
/*                                                                           */
/* RETURN VALUE: Index in reg_table.                                         */
/* ************************************************************************* */
int get_reg_number(reg_table_t *reg_table,char *reg_name)
{
  int i=0;
  int known_reg;
  assert ((reg_table != NULL) && (reg_name != NULL));
  
  known_reg = is_register(reg_name);

  while(reg_table[i].reg_name != NULL) {

    if (!strcmp(reg_name,reg_table[i].reg_name)) {
      return i;
    }
    
    else { i++; }
    
  } /* end of while(...) */

  /* if we got here the register can exist but was not found 
   * We can report errors according to the following:
   * 
   * 1) Register is a global register but is not defined for this architecture.
   *    In this case i=-1 and known_reg is larger than 1
   *
   * 2) Register is not a global register nor a local register (defined for
   *    the current architecture. How come it was parsed as a register? Simple
   *    the parser screwed up and we must report a bug, this should naturally 
   *    never happen ;-) 
   */

  i = -1; 
  if ((known_reg > 0) && (i == -1 )) {
    iopasm_error(NULL,NULL,ERR_REGISTER_NOT_IN_ARCH,"Register is %s\n",reg_name);
  }

  else {
    iopasm_error(NULL,NULL,ERR_REPORT_BUG,"Reported from %s:%d",__FILE__,__LINE__);
  }


  iopasm_error(NULL,NULL,ERR_REPORT_BUG,"Reported from %s:%d",__FILE__,__LINE__);
  return 0; /* To get rid of warning - will never be reached.
	     * This should be preceeded by an error causing a 
	     * bug-report because it should simply never happen.
	     * If not it is a bug in itself */
}



/* ************************************************************************* */
/* PROTOTYPE   : int xlate_reg2asm(reg_table_t *table,char *reg_name);       */
/*                                                                           */
/* ARGUMENTS   : "table" points to a the register table. "reg_name" points   */
/*               to a register name.                                         */
/*                                                                           */
/* FUNCTION    : Returns the register type as seen from the assembler.       */
/*                                                                           */
/* RETURN VALUE: Type of register. -1 on failure.                            */
/* ************************************************************************* */
int xlate_reg2asm(reg_table_t *table,char *reg_name)
{
  int known_reg;
  int i = 0;
  assert (table != NULL);
  known_reg = is_register(reg_name);

  while(table[i].reg_name != NULL) {

    if (!strcmp(reg_name,table[i].reg_name)) {
      return table[i].cg_reg_type;
    }
    
    else { i++; }
  }

    return -1;
}



/* ************************************************************************* */
/* PROTOTYPE   : int xlate_reg2arch(reg_table_t *table,char *reg_name);      */
/*                                                                           */
/* ARGUMENTS   : "table" points to the architectural register table and      */
/*               "reg_name" points to the register name to be translated.    */
/*                                                                           */
/* FUNCTION    : Translates a register to the given architectural type       */
/*                                                                           */
/* RETURN VALUE: Register type. -1 on error.                                 */
/* ************************************************************************* */
int xlate_reg2arch(reg_table_t *table,char *reg_name)
{

  int known_reg;
  int i = 0;
  assert (table != NULL);
  known_reg = is_register(reg_name);

  while(table[i].reg_name != NULL) {

    if (!strcmp(reg_name,table[i].reg_name)) {
      return table[i].arch_reg_type;
    }
    
    else { i++; }
  }
  
  return -1;
}



/* ************************************************************************* */
/* PROTOTYPE   : int get_inst_number(inst_table_t *i_table,                  */
/*                                   inst_table_t *row,inst_t *inst);        */
/*                                                                           */
/* ARGUMENTS   : "i_table" points to a valid table, "row" points to the row  */
/*               representation of an instruction. "inst" is the instruction */
/*               represented as "row".                                       */
/*                                                                           */
/* FUNCTION    : Finds the index in the instruction table that matches "inst"*/
/*                                                                           */
/* RETURN VALUE:  Table index for "inst". -1 on error                         */
/* ************************************************************************* */
int get_inst_number(inst_table_t *i_table,inst_table_t *row,inst_t *inst) 
{
  int i=0;
  
  assert((i_table != NULL) && (row != NULL));

    while(i_table[i].inst_name != NULL) {

      if (!strcmp(i_table[i].inst_name,row->inst_name) && 
	  (i_table[i].num_args == row->num_args) &&
	  !memcmp(i_table[i].arg_type,row->arg_type,MAX_ARGS)) {
	return i;
      }
      
      else { 
	i++;
      }
    
  }
  return -1;


}



/* ************************************************************************* */
/* PROTOTYPE   : unsigned int eval_src_immval(cg_context_t *ctxt,            */
/*                                            inst_t *inst,int i_num,        */
/*                                            char *imm_val);                */
/*                                                                           */
/* ARGUMENTS   : "ctxt" points to a valid context, "inst" is the in-         */
/*               struction for which we wish to evaluate the immediate       */
/*               value "imm_val". "i_num" is the instruction number (table   */
/*               index).                                                     */
/*                                                                           */
/* FUNCTION    : Returns the part of an op-code for a given instruction      */
/*               based on the value imm_val as the first (source) operand.   */
/*                                                                           */
/* RETURN VALUE: Part of op-code. Never returns on error.                    */
/* ************************************************************************* */
unsigned int eval_src_immval(cg_context_t *ctxt,inst_t *inst,
			     int i_num,char *imm_val)
{
  unsigned int i_val;
  unsigned int r_val;
  unsigned int  mask;
  inst_table_t *i_table;
  
  assert((ctxt != NULL) && (inst != NULL) && (imm_val != NULL));
  i_table = cg_get_inst_table(ctxt);

  i_val = strtoul(imm_val,NULL,10); /* Decimal internal representation */
  r_val = SRC_IMMEDIATE_MAX(i_table,i_num);

  if (r_val < i_val) { /* ERROR Out of range */
    iopasm_error(ctxt,inst,ERR_IMM_OUT_OF_RANGE,
		 "%s\n","Immediate is out of range",imm_val);
  }
  
  else if (!IS_SRC_IMM_ALLOWED(i_table,i_num)) { 
    iopasm_error(ctxt,inst,ERR_IMM_NOT_ALLOWED,NULL);
  }
  
  else {
    mask = (2 << (i_table[i_num].field_len[SRC_FIELD] - 1)) - 1;
    
    mask = mask << (i_table[i_num].arg_field[SRC_FIELD] - 
		    (i_table[i_num].field_len[SRC_FIELD] - 1));
    
    r_val = (i_val << 
	     ((i_table[i_num].arg_field[SRC_FIELD]) - 
	     (i_table[i_num].field_len[SRC_FIELD] - 1)));
    
    return (r_val & mask);
  }

  iopasm_error(NULL,NULL,ERR_REPORT_BUG,"Reported from %s:%d",__FILE__,__LINE__);
  return 0; /* To get rid of warning - will never be reached.
	     * This should be preceeded by an error causing a 
	     * bug-report because it should simply never happen.
	     * If not it is a bug in itself */
}



/* ************************************************************************* */
/* PROTOTYPE   : unsigned int eval_any_immval(cg_context_t *ctxt,            */
/*                                            inst_t *inst,int i_num,        */
/*                                            char *imm_val);                */
/*                                                                           */
/* ARGUMENTS   : "ctxt" points to a valid context, "inst" is the in-         */
/*               struction for which we wish to evaluate the immediate       */
/*               value "imm_val". "i_num" is the instruction number (table   */
/*               index).                                                     */
/*                                                                           */
/* FUNCTION    : Returns the part of an op-code for a given instruction      */
/*               based on the value imm_val as the second (any) operand.     */
/*                                                                           */
/* RETURN VALUE: Part of op-code. Never returns on error.                    */
/* ************************************************************************* */
unsigned int eval_any_immval(cg_context_t *ctxt,inst_t *inst,
			     int i_num,char *imm_val)
{
  unsigned int i_val;
  unsigned int r_val;
  unsigned int mask;
  inst_table_t *i_table;
  
  assert ((ctxt != NULL) && (inst != NULL) && (imm_val != NULL));
  i_table = cg_get_inst_table(ctxt);

  i_val = strtoul(imm_val,NULL,10); /* Decimal internal representation */
  r_val = ANY_IMMEDIATE_MAX(i_table,i_num);

  if (ANY_IMMEDIATE_MAX(i_table,i_num) < i_val) { /* ERROR Out of range */
    iopasm_error(ctxt,inst,ERR_IMM_OUT_OF_RANGE,
		 "Immediate (%s) is out of range",
		 imm_val);
  }
  
  else if (!IS_ANY_IMM_ALLOWED(i_table,i_num)) { 
    iopasm_error(ctxt,inst,ERR_IMM_NOT_ALLOWED,
		 "Immediate (%s) is not allowed as secondary operand",
		 imm_val);
  }
  
  else {
    mask = (2 << (i_table[i_num].field_len[ANY_FIELD] - 1)) - 1;
    
    mask = mask << (i_table[i_num].arg_field[ANY_FIELD] - 
		    (i_table[i_num].field_len[ANY_FIELD] - 1));
    
    r_val = (i_val << 
	     ((i_table[i_num].arg_field[ANY_FIELD]) - 
	     (i_table[i_num].field_len[ANY_FIELD] - 1)));
    
    return (r_val & mask);
  }

  iopasm_error(NULL,NULL,ERR_REPORT_BUG,"Reported from %s:%d",__FILE__,__LINE__);
  return 0; /* To get rid of warning - will never be reached.
	     * This should be preceeded by an error causing a 
	     * bug-report because it should simply never happen.
	     * If not it is a bug in itself */
}



/* ************************************************************************* */
/* PROTOTYPE   :  unsigned int eval_dst_immval(cg_context_t *ctxt,           */
/*                                             inst_t *inst,int i_num,       */
/*                                             char *imm_val);               */
/*                                                                           */
/* ARGUMENTS   : "ctxt" points to a valid context, "inst" is the in-         */
/*               struction for which we wish to evaluate the immediate       */
/*               value "imm_val". "i_num" is the instruction number (table   */
/*               index).                                                     */
/*                                                                           */
/* FUNCTION    : Returns the part of an op-code for a given instruction      */
/*               based on the value imm_val as the third (destination)       */
/*               operand.                                                    */
/*                                                                           */
/* RETURN VALUE: Part of op-code. Never returns on error.                    */
/* ************************************************************************* */
 unsigned int eval_dst_immval(cg_context_t *ctxt,inst_t *inst,
			      int i_num,char *imm_val)
{
  unsigned int i_val;
  unsigned int mask;
  unsigned int r_val;
  inst_table_t *i_table;

  assert((ctxt != NULL) && (inst != NULL));
  i_table = cg_get_inst_table(ctxt);
  
  i_val = strtoul(imm_val,NULL,10); /* Decimal internal representation */
  r_val = DST_IMMEDIATE_MAX(i_table,i_num);
  if (DST_IMMEDIATE_MAX(i_table,i_num) < i_val) { /* ERROR Out of range */
    iopasm_error(ctxt,inst,ERR_IMM_OUT_OF_RANGE,
		 "Immediate (%s) is out of range\n",
		 imm_val);
  }
  
  else if (!IS_DST_IMM_ALLOWED(i_table,i_num)) { 
    iopasm_error(ctxt,inst,ERR_IMM_NOT_ALLOWED,
		 "Immediate (%s) is not allowed as destination operand",
		 imm_val);
  }
  
  else {
    mask = (2 << (i_table[i_num].field_len[DST_FIELD] - 1)) - 1;
    
    mask = mask << (i_table[i_num].arg_field[DST_FIELD] - 
		    (i_table[i_num].field_len[DST_FIELD] - 1));
    
    r_val = (i_val << 
	     ((i_table[i_num].arg_field[DST_FIELD]) - 
	     (i_table[i_num].field_len[DST_FIELD] - 1)));
    
    return (r_val & mask);
  }

  iopasm_error(NULL,NULL,ERR_REPORT_BUG,"Reported from %s:%d",__FILE__,__LINE__);

  return 0; /* To get rid of warning - will never be reached.
	     * This should be preceeded by an error causing a 
	     * bug-report because it should simply never happen.
	     * If not it is a bug in itself */
}



/* ************************************************************************* */
/* PROTOTYPE   : inst_table_t *create_itable_row(inst_t *inst) ;             */
/*                                                                           */
/* ARGUMENTS   : "inst" points to a valid instruction.                       */
/*                                                                           */
/* FUNCTION    : Creates an instruction table row, based on the instruction  */
/*               referenced by "inst" and it's operands.                     */
/*                                                                           */
/* RETURN VALUE: Pointer to a instruction table row. The row is generated    */
/*               using "inst" as a template.                                 */
/* ************************************************************************* */
inst_table_t *create_itable_row(inst_t *inst) 
{
  inst_table_t *temp_row;
  int i=0;
  arg_t *arg;
  char *arg_val;
  int arg_type;

  assert(inst != NULL);

  temp_row = (inst_table_t *) IOPASM_MALLOC(sizeof(inst_table_t));
  assert (temp_row != NULL);

  memset(temp_row,0,sizeof(inst_table_t));

  temp_row-> inst_name = (char *) strdup(get_inst(inst));
  arg = get_first_arg(inst);

  while((arg != NULL) && (i<3)) {
    arg_val = get_arg_val(arg);
    arg_type = get_arg_type(arg);

    if (arg_type == ARG_STACK) {
      arg_type = check_expression_type(get_arg_stack(arg));
    }
    switch(arg_type) {
      
    case ARG_LABEL:
    case ARG_IMMEDIATE:
      temp_row->arg_type[i] = I;
      i++;
      break;
      
    case ARG_NORMAL_REG:
      temp_row->arg_type[i] = R;
      i++;
      break;

    case ARG_SPECIAL_REG:
      temp_row->arg_type[i] = S;
      i++;
      break;
    }
    arg = get_next_arg(arg);
  }
  
  temp_row->num_args = i;
  return temp_row;
}



/* ************************************************************************* */
/* PROTOTYPE   : int sregs2regs(inst_table_t *row);                          */
/*                                                                           */
/* ARGUMENTS   : "row" points to a table row.                                */
/*                                                                           */
/* FUNCTION    : Converts special registers to registers in table row "row". */
/*                                                                           */
/* RETURN VALUE: Returns number of converted registers.                      */
/* ************************************************************************* */
int sregs2regs(inst_table_t *row)
{
  int i,converted;
  
  assert (row != NULL);
  i = 0;
  converted = 0;
  
  while(i < row->num_args) {

    if (row->arg_type[i] == S) {
      converted ++;
      row->arg_type[i] = R;
    }

    i++;

  }
  return converted;
}



/* ************************************************************************* */
/* PROTOTYPE   : void delete_itable_row(inst_table_t *row);                  */
/*                                                                           */
/* ARGUMENTS   : "row" points to a table row.                                */
/*                                                                           */
/* FUNCTION    : Deletes a table row and frees up any memory that was        */
/*               allocated.                                                  */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void delete_itable_row(inst_table_t *row)
{
  assert (row != NULL);
  IOPASM_FREE(row->inst_name);
  IOPASM_FREE(row);
}



/* ************************************************************************* */
/* PROTOTYPE   : int get_field_size(inst_table_t *i_table,int i_num,         */
/*                                  int field_no);                           */
/*                                                                           */
/* ARGUMENTS   : "i_table" points to a valid table "i_num" is the            */
/*               instruction number and "field_no" is the field which size   */
/*               we want to know.                                            */
/*                                                                           */
/* FUNCTION    : Return the field size for a given instruction.              */
/*                                                                           */
/* RETURN VALUE: Size of the field measured in bits. -1 on error             */
/* ************************************************************************* */
int get_field_size(inst_table_t *i_table,int i_num,int field_no) 
{
  assert((i_num > 0) && (field_no >= 0) && (field_no <= 2));
  if (i_table[i_num].num_args >= (field_no + 1)) {
    return i_table[i_num].field_len[field_no];
  }
  else { return -1; }

}



/* ************************************************************************* */
/* PROTOTYPE   : void save_register_macros(FILE *output_file,                */
/*                                         reg_table_t *registers,           */
/*                                         char *scope);                     */
/*                                                                           */
/* ARGUMENTS   : "output_file" points to a valid output file. "registers" is */
/*               the register table for which register macros are to be      */
/*               saved, and "scope" is the current scope (MPU or SPU).       */
/*                                                                           */
/* FUNCTION    : Saves the register table as C-macros.                       */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void save_register_macros(FILE *output_file,reg_table_t *registers,char *scope)
{
  int i =0;
  char work_string[MAX_STRLEN];
  assert ((output_file != NULL) && (registers != NULL));

  while(registers[i].reg_name != NULL) {
    
    memset(work_string,0,MAX_STRLEN);
    to_upper(work_string,registers[i].reg_name,MAX_STRLEN);
    assert (fprintf(output_file,"#define %s%s (0x%x)\n",scope,work_string,
		    registers[i].reg_val) > 0);
    i++;
  }
}



/* ************************************************************************* */
/* PROTOTYPE   : void save_inst_macros(FILE *output_file,                    */
/*                                     inst_table_t *i_table,char *scope);   */
/*                                                                           */
/* ARGUMENTS   : "output_file" points to a valid output file. "i_table" is   */
/*               the instruction table for which instruction macros are to   */
/*               be saved, and "scope" is the current scope (MPU or SPU).    */
/*                                                                           */
/* FUNCTION    : Saves the instruction table as C-macros.                    */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void save_inst_macros(FILE *output_file,inst_table_t *i_table,char *scope)
{
  
  int i =0;
  int j,k,l;
  int work_len;
  int work_len_imm;
  char work_string[MAX_STRLEN];
  char work_string_imm[MAX_STRLEN];
  char *macro_args3[] = { "S", "N", "D" };
  char *macro_args2[] = { "S", "D" };
  char *macro_args1[] = { "S" };
  char **current_args = NULL;
  assert ((output_file != NULL) && (i_table != NULL));
  
  while(i_table[i].inst_name != NULL) {
    
    memset(work_string,0,MAX_STRLEN);
    memset(work_string_imm,0,MAX_STRLEN);
   
    strcpy(work_string,scope);
    j = strlen(work_string);
    strcpy(work_string_imm,scope);
    
    /* Instruction name into upper case */
    for (l=0;l<strlen(i_table[i].inst_name);l++) { 
      work_string[j+l] = toupper(i_table[i].inst_name[l]);
      work_string_imm[j+l] = toupper(i_table[i].inst_name[l]);
    }

    j = strlen(work_string);
    
    if (i_table[i].num_args != 0) {
      work_string[j] = '_';
      work_string_imm[j++] = '_';
      k = 0;
      while( k <i_table[i].num_args) {
	work_string[j+k] = i_table[i].arg_type[k];
	work_string_imm[j+k] = i_table[i].arg_type[k];
	k++;
      }
    }
    
    j = strlen(work_string);
    if (i_table[i].inst_len == 8) {
      strcat(work_string_imm,"_IMM");
      strcat(work_string,"_INSTR");
    }
    switch (i_table[i].num_args) {
    case 3: /* Add 3 operands to macro */
      strcat(work_string,"(S,N,D) ");
      strcat(work_string_imm,"(S,N,D) ");
      break;

    case 2: 
      strcat(work_string,"(S,D) ");
      strcat(work_string_imm,"(S,D) ");
      break;
	
    case 1: /* Add 1 operand to macro */
      strcat(work_string,"(S) ");
      strcat(work_string_imm,"(S) ");
      break;
      
    case 0: /* We do not need to add any operands, change macro */
      strcat(work_string,"() ");
      strcat(work_string_imm,"() ");
      break;
    }

    work_len = strlen(work_string);
    work_len_imm = strlen(work_string_imm);

    work_len += snprintf(&work_string[work_len],MAX_STRLEN - work_len,
			 "(0x%2.2X%2.2X%2.2X%2.2X",
			 i_table[i].i_template[0],
			 i_table[i].i_template[1],
			 i_table[i].i_template[2],
			 i_table[i].i_template[3]);

    j = 0;
    
    while (j != i_table[i].num_args) {
      
      switch(i_table[i].num_args) {
	
      case 3:
	current_args = macro_args3;
	break;
	
      case 2:
	current_args = macro_args2;
	break;
	
      case 1:
	current_args = macro_args1;
	break;
	
      } /* Ends switch(i_table[i].num_args) */
      
      if ((i_table[i].num_args != 0) && (i_table[i].field_len[j]!=32)) {
	work_len += snprintf(&work_string[work_len],MAX_STRLEN - work_len,
			     " | ((%s & ((1 << %d) - 1)) << %d)",
			     current_args[j],
			     i_table[i].field_len[j],
			     ((i_table[i].arg_field[j] - \
			       i_table[i].field_len[j]) < 0) ? 0:
			     (i_table[i].arg_field[j] - \
			      i_table[i].field_len[j] + 1));

	if ((i_table[i].num_args-1) == j) {
	  strcat(work_string,")\n");
	}
	
	else {
	  strcat(work_string,"\\\n"); 
	}
      }
      
      if ((i_table[i].num_args != 0) && (i_table[i].field_len[j]==32)) {
	snprintf(&work_string_imm[work_len_imm],MAX_STRLEN - work_len_imm,
		 "(%s & 0xFFFFFFFF)\n",current_args[j]);
      }

      
      work_len = strlen(work_string);
      
      if ((strchr(work_string,'|') != NULL)&&(i_table[i].field_len[j]!=32)) {
	l = 0;
	while(work_string[l] != '|') { l++; }
	for (;l>0;l--) { work_string[work_len++] = ' '; };
      }

      work_len = strlen(work_string);
      j++; /* Process next argument */
    } /* Ends while (j != i_table[i].num_args) */

    if (i_table[i].num_args == 0) {
      strcat(work_string,")\n");
    }
    
    fprintf(output_file,"#define %s\n",work_string);
    if (i_table[i].inst_len == 8) {
      fprintf(output_file,"#define %s\n",work_string_imm);
    }

    i++; /* Process next instruction */
  }
}
