/* ************************************************************************* */
/* iopasm_err.c : Error Handling and reporting for I/O-processor assembler   */
/*                                                                           */
/*    Copyright 2002-2006 Axis Communications AB.                            */
/*                                                                           */
/*                                                                           */
/*    This program is free software; you can redistribute it and/or modify   */
/*    it under the terms of the GNU General Public License as published by   */
/*    the Free Software Foundation; either version 2 of the License, or      */
/*    (at your option) any later version.                                    */
/*                                                                           */
/*    This program is distributed in the hope that it will be useful,        */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*    GNU General Public License for more details.                           */
/*                                                                           */
/*    You should have received a copy of the GNU General Public License      */
/*    along with this program; if not, write to the Free Software            */
/*    Foundation, Inc., 59 Temple Place - Suite 330,                         */
/*    Boston, MA 02111-1307, USA.                                            */
/*                                                                           */
/*                                                                           */
/*    Send bug reports and inquiries to Johan.Jorgensen@axis.com             */
/* ************************************************************************* */



/* ************************************************************************* */
/*                      INCLUDES USED BY THIS MODULE                         */
/* ************************************************************************* */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <assert.h>
#include "cg_fe.h"
#include "p_list.h"
#include "iopasm_err.h"
#include "iopasm_global.h"
#include "debug.h"



/* ************************************************************************* */
/*                     DEFINITIONS LOCAL TO THIS MODULE                      */
/* ************************************************************************* */
#define ERR_MAX_STRLEN (1000)
#define ERR_LINE_BREAK (79)



/* ************************************************************************* */
/*                       TYPES DEFINED IN THIS MODULE                        */
/* ************************************************************************* */
struct __error_t {
  const int error_number; /* Error number as defined in iopasm_err.h         */
  int error_mode;         /* Error mode (On, Off or reported as warning      */
  const char *error_id;   /* String ID used to change mode                   */
  const char *error_text; /* Error message string                            */
  
};


 
typedef struct __error_t error_t;



/* ************************************************************************* */
/*                     DEFINITIONS LOCAL TO THIS MODULE                      */
/* ************************************************************************* */
static error_t iopasm_errors[] = {
  { .error_number = ERR_UNSUPPORTED_ARCH,
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Architecture not supported by iopasm" 
  },
  { .error_number = ERR_FORK_FAILED, 
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Could not create a new process" 
  },
  { .error_number = ERR_CPP_EXEC_FAILED, 
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Could not execute c-preprocessor" 
  },
  { .error_number = ERR_REPORT_BUG, 
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "There is a bug in iopasm, please report it to Johan.Jorgensen@axis.com" 
  },
  { .error_number = ERR_CPP_STAGE_FAILED,
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Preprocessing stage failed. You should check your source for errors" 
  },
  { .error_number = ERR_MEMORY_REGION_USED, 
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Memory region is already in use" 
  },
  { .error_number = ERR_CANNOT_WRITE_FILE, 
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Cannot open output-file for writing" 
  },
  { .error_number = ERR_SYNTAX_ERROR, 
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Syntax error" 
  },
  { .error_number = ERR_UNDEFINED_LABEL, 
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Label is not defined" 
  },
  { .error_number = ERR_UNDEFINED_FW_LABEL, 
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Cannot find forward reference to local label" 
  },
  { .error_number = ERR_UNDEFINED_BW_LABEL, 
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Cannot find backward reference to local label" 
  },
  { .error_number = ERR_ILLEGAL_LOCAL_ADDR_LABEL,
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Local addess label not allowed here" 
  },
  { 
    .error_number = ERR_DUPLICATE_LABEL,
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Label is already defined"
  },
  { .error_number = ERR_FILE_IS_EMPTY,
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Input file is empty" 
  },
  { .error_number = ERR_ILLEGAL_WARN_FLAG, 
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Warning/Error flag is not known by the assembler" 
  },
  { .error_number = ERR_ILLEGAL_WARN_MODE, 
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Cannot set fault-mode to requested value" 
  },
  { .error_number = ERR_PARSER_MODE, 
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Missing .fsm or .seq statement" 
  },
  { .error_number = ERR_PARSER_EXTRA_MODE_SWITCH, 
    .error_mode = ERROR_ON,
    .error_id = NULL, 
    .error_text = "Extraneous mode switch" 
  },
  { .error_number = ERR_INPUT_FILE_MISSING, 
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Input file is missing." 
  },
  { .error_number = ERR_ARG_EXPECTED, 
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Instruction operand expected" 
  },
  { .error_number = ERR_UNEXPECTED_ARG,
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Unexpected instruction operand" 
  },
  { .error_number = ERR_UNKNOWN_ARGUMENT, 
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Cannot determine operand type" 
  },
  { .error_number = ERR_ILLEGAL_ARGUMENT, 
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Operand is not allowed!" 
  },
  { .error_number = ERR_ARG_DWORD_EVAL, 
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Dword operand evaluates to a register in pseudo instruction" 
  },
  { .error_number = ERR_NON_CONSTANT_ARG, 
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Operand evaluates to a non-constant expression" 
  },
  { .error_number = ERR_IMM_OUT_OF_RANGE, 
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Literal value is out of range" 
  },
  { .error_number = ERR_IMM_NOT_ALLOWED, 
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Literal value is not expected here" 
  },
  { .error_number = ERR_FSM_TOO_MANY_SEQS, 
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text =  "Too many sequential instructions in FSM-statement" 
  },
  { .error_number = ERR_FSM_TOO_MANY_TIMERS,
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text =  "Too many timer statements found in FSM-statement" 
  },
  { .error_number = ERR_FSM_TOO_MANY_XFERS,
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Too many state-transitions found in FSM-statement" 
  },
  { .error_number = ERR_FSM_UNKNOWN_ERROR,
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Uncaught error in FSM code. Report this to Johan.Jorgensen@axis.com" 
  },
  { .error_number = ERR_FSM_XFER_CONFLICT,
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Ambigous state transition found in state" 
  },
  { .error_number = ERR_FSM_TIMER_CONFLICT,
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Timer statement is in conflict with \"only\"-statement" 
  },
  { .error_number = ERR_FSM_ADDR_CONFLICT,
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "State-address is invalid (Crosses bank-boundary" 
  },
  { .error_number = ERR_FSM_NO_TIMER_ALLOWED,
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Timer statement is not allowed together with 32-bit immediate sequential instruction" 
  },
  { .error_number = ERR_FSM_EMASK_IN_ONLY_XFER,
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Emask may not be changed in a state containing an 'only' transition" 
  },
  { .error_number = ERR_FSM_EMASK_WITH_GO_SEQ,
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Emask may not be changed in a state containing a 'go_seq' flag" 
  },
  { .error_number = ERR_FSM_NO_FSM_XFERS,
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "No state transitions nor any 'go_seq' flag found in state" 
  },
  { .error_number = ERR_UNKNOWN_REGISTER, 
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Unknown register name" 
  },
  { .error_number = ERR_REGISTER_NOT_ALLOWED, 
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Register not expected here" 
  },

  { .error_number = ERR_UNKNOWN_INSTRUCTION, 
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Unsupported instruction found" 
  },
  { .error_number = ERR_REGISTER_NOT_IN_ARCH, 
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Register not supported by architecture" 
  },
  { .error_number = ERR_ILLEGAL_REGISTER_USE, 
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Register not allowed as an operand to this instruction" 
  },
  { .error_number = ERR_ILLEGAL_MASK_REG_USE,
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Register is not allowed in masked instruction" 
  },
  { .error_number = ERR_ILLEGAL_INST_ORDER, 
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Illegal instruction sequence" 
  },
  { .error_number = ERR_SEQUENTIAL_BOUNDRARY,
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Sequential instruction is placed at illegal boundrary" 
  },
  { .error_number = ERR_32BIT_IMMEDIATE_IN_SEQ,
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Sequential instruction with 32-bit immediate outside of FSM statement" 
  },
  { .error_number = ERR_32BIT_UNALIGNED_IN_DS,
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Sequential 64-bit instruction (with 32bit immediate) is unaligned and resides in a delay slot" 
  },
  { .error_number = ERR_FSM_BOUNDRARY,
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "FSM-instruction is placed at illegal boundrary" 
  },
  { .error_number = ERR_UNKNOWN_MACHINE_OPTION,
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Machine option is not supported by architecture" 
  },
  { .error_number = ERR_UNKNOWN_SYMFILE_FORMAT,
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Symbol-filetype is not known by this verion of iopasm" 
  },
  { .error_number = ERR_ARCHITECTURE_MISMATCH,
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Processor Architecture specified in object file differs from previously loaded files" 
  },
  { .error_number = ERR_INCOMPATIBLE_OBJ_FILE,
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "This does not seem to be a valid object-file" 
  }, 
  { .error_number = ERR_NO_OBJECT_FILE_FOUND,
    .error_mode = ERROR_ON,
    .error_id = NULL,
    .error_text = "Cannot find object files" 
  },
  
  /* Configurable errors - simply set the ID to make it configurable. */
  { .error_number = WARN_DELAY_SLOT_USAGE,
    .error_mode = ERROR_OFF,    
    .error_id = "delay-slot-usage",
    .error_text = "Possible incorrect use of delay slot"
  },

  { .error_number = WARN_DEST_REG_USAGE,
    .error_mode = ERROR_OFF,
    .error_id = "dest-reg-usage",
    .error_text = "Possibly incorrect use of destination register" 
  },
  
  { .error_number = WARN_UNALIGNED_64BIT_INST,
    .error_mode = ERROR_OFF,
    .error_id = "unaligned-64bit-inst",
    .error_text = "Unaligned 64 bit instruction"
  },
  
  { .error_number = WARN_STATE_TERMINATED,
    .error_mode = ERROR_OFF,
    .error_id = "state-terminated",
    .error_text = "FSM state clause might have been prematurely terminated by \".fsm\" directive" 
  },

  { .error_number = WARN_RETI_VIOLATION,
    .error_mode = ERROR_ON,
    .error_id = "reti-rw-rr-sequence",
    .error_text = "RETI followed by rw/rwq/rwx instruction - potentially illegal instruction sequence."
  },
  /* Last entry must end like this */
  { .error_number = 0, 
    .error_mode = ERROR_OFF,
    .error_id = NULL,
    .error_text = NULL }
}; 



/* ************************************************************************* */
/*                    PROTOTYPES FOR STATIC FUNCTIONS                        */
/* ************************************************************************* */



/* ************************************************************************* */
/* PROTOTYPE   : static void format_output(char *a_string,int indent_len);   */
/*                                                                           */
/* ARGUMENTS   : "a_string" points to a string. "indent_len" is the amount   */
/*               of indentation to be used.                                  */
/*                                                                           */
/* FUNCTION    : Formats a string so that it can be printed on stdout/stderr */
/*               (pretty printer).                                           */   
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
static void format_output(char *a_string,int indent_len)
{  
  int i,j;
  int str_index;
  int remaining;
  int cpy_len;
  int tmp_len = 0 ;
  char output_buffer[ERR_MAX_STRLEN];

  i = 0;
  memset(output_buffer,0,ERR_MAX_STRLEN);
  memset(output_buffer,32,indent_len);
  str_index = indent_len;

  if (strlen(a_string) < (ERR_LINE_BREAK - indent_len)) {
    strncpy(output_buffer+str_index,a_string,strlen(a_string));
    strcat(output_buffer,"\n");
    remaining = 0;
  }
  
  else {
    remaining = strlen(a_string);
    tmp_len = remaining;
  }

  /* Set up start and maximum output length */
  i = 0;

  while (remaining > 0) {
    
    j = i + (ERR_LINE_BREAK - indent_len);
    j = (j > tmp_len) ? tmp_len: j;

    if (remaining > (ERR_LINE_BREAK - indent_len)) { 
      while(a_string[j] != ' ') {
	j--;
      } 
    } /* End of if ((strlen(a_string) - j)  ... */

    cpy_len = (i < j) ? j - i: i - j;

    strncat(output_buffer,&a_string[i],cpy_len);
    strncat(output_buffer,"\n",1);
    
    remaining = (remaining > (ERR_LINE_BREAK - indent_len)) ? 
      ((i > j) ? remaining - (i - j) : remaining - (j - i)): 0;

    i = j+1; /* Skip current character _and_ a space */
    j = (j >= tmp_len) ? 0: j;
    if (j == 0) {
      remaining = 0;
    }
    
    else {
      str_index = strlen(output_buffer);
      memset(output_buffer+str_index,32,indent_len);
    }
  } /* End of while (remaining > 0) */
  
  fprintf(stderr,"%s",output_buffer);

}



/* ************************************************************************* */
/* PROTOTYPE   : void iopasm_verbose_msg(char *fmt,...);                     */
/*                                                                           */
/* ARGUMENTS   : "fmt" points to the format string.                          */
/*                                                                           */
/* FUNCTION    : Displays a verbose message (if verbose is on).              */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void iopasm_verbose_msg(char *fmt,...) 
{
  va_list ap;
  char tmp_buffer[ERR_MAX_STRLEN];
  
  
  assert(fmt != NULL);
  
  va_start(ap,fmt);

  /* Clear the string buffer prior to use */
  memset(tmp_buffer,0,ERR_MAX_STRLEN);
  
  if (check_option("verbose",(void *) 1,OPT_INTEGER)) {
    snprintf(tmp_buffer,ERR_MAX_STRLEN,"iopasm INFO: ");
    vsnprintf(tmp_buffer + strlen(tmp_buffer),
	      ERR_MAX_STRLEN - strlen(tmp_buffer),
	      fmt,ap);
    format_output(tmp_buffer,0);
  }
  va_end(ap);
  
}



/* ************************************************************************* */
/* PROTOTYPE   : void iopasm_error(cg_context_t *ctxt,inst_t *inst,          */
/*                                 int err_no, char *extra_fmt,... );        */
/*                                                                           */
/* ARGUMENTS   : "ctxt" is the code-generation context. "inst" is the        */
/*               instruction where the error occured, "err_no" is the error  */
/*               number and "extra_fmt" is the extra (optional) format       */
/*               specifier.                                                  */
/*                                                                           */
/* FUNCTION    : Reports an error/warning or ignores it. Depends on the      */
/*               error mode.                                                 */
/*                                                                           */
/* RETURN VALUE: None -will exit on error and just return quietly on warning */
/* ************************************************************************* */
void iopasm_error(cg_context_t *ctxt,inst_t *inst,int err_no,
		  char *extra_fmt,... )
{

  va_list ap;
  char tmp_buffer[ERR_MAX_STRLEN];
  char *file_name = NULL;
  int line_no = 0;
  int file_index = 0;
  char *err_text = NULL;
  char *pre_fmt = NULL;
  int indent_len;
  int i=0;

  /* Take care of variable arguments */
  va_start(ap,extra_fmt);

 
  /* First we find the error we want to display */
  while (iopasm_errors[i].error_number != 0) {

    if (iopasm_errors[i].error_number == err_no) {
      err_text = (char *) iopasm_errors[i].error_text;
      break;
    } /* End of if (iopasm_errors[i].err_number == err_no) */
    i++;
  } /* End of while (iopasm_errors[i].err_number != 0) */

  /* Next we test if inst and context are NULL. If not we can 
   * Retrieve Line-number and filename information from the instruction */
  if (inst != NULL) {
    line_no = get_inst_line(inst);
    file_index = get_inst_file_index(inst);
    
  }
  
  if (ctxt != NULL) {
    file_name = cg_get_file_by_index(ctxt,file_index);
  }

  /* Clear the string buffer prior to use */
  memset(tmp_buffer,0,ERR_MAX_STRLEN);

  /* Now let us find out why iopasm_error was called: 
   * Reasons could be : Error, Warning or just ignore */
  
  switch (iopasm_errors[i].error_mode) {

  case ERROR_ON: /* We will report an error and exit */
    pre_fmt = ((ctxt != NULL) && (inst != NULL)) ?
      "IOPASM Error %d in %s, line %d: ":
      "IOPASM Error %d: ";
    break;


  case WARNING_ON:
    pre_fmt = ((ctxt != NULL) && (inst != NULL)) ?
      "IOPASM Warning %d in %s, line %d: ":
      "IOPASM Warning %d: ";
    break;

  case ERROR_OFF:
    goto end_of_error;
    break; /* Not reached */

  } /* End switch(iopasm_errors[i].error_mode) */

  if ((ctxt != NULL) && (inst != NULL)) {
    indent_len = snprintf(tmp_buffer,ERR_MAX_STRLEN,pre_fmt,err_no,
			  file_name,line_no);
  }  /* End of if ((ctxt != NULL) && (inst !=NULL)) */ 
  
  else {
    indent_len = snprintf(tmp_buffer,ERR_MAX_STRLEN,pre_fmt,err_no);
  }    
  
  /* Now we can display the error message */

  snprintf(tmp_buffer+indent_len,ERR_MAX_STRLEN,"%s",err_text);
  fprintf(stderr,"%s\n",tmp_buffer);
  memset(tmp_buffer,0,ERR_MAX_STRLEN);

  /* If we have any extra format we will begin to display this now,
   * but we will format it to fit in 80-columns and indent it until after
   * The colon in the standard error message. It looks better that way */
  
  if (extra_fmt != NULL) {
    vsnprintf(tmp_buffer,ERR_MAX_STRLEN,extra_fmt,ap);
    format_output(tmp_buffer,indent_len);
  }
  
  /* At last we find out if we should exit or just return */
  if (iopasm_errors[i].error_mode == ERROR_ON) {
    exit(EXIT_FAILURE);
  }

  /* Return as if nothing has happened */
 end_of_error: 
    va_end(ap);
    return ;
}



/* ************************************************************************* */
/* PROTOTYPE   : void set_error_mode(char *error_identifier,char *set_state);*/
/*                                                                           */
/* INPUT PARM's: "error_identifier" points to an error-identifier string     */
/*               (supplied from the cmd-line). "set_state" is the new state. */
/*                                                                           */
/* FUNCTION    : Changes the behaviour of the error associated with          */
/*               "error_identifier to "set_state".                           */
/*                                                                           */
/* RETURN VALUE: None. May exit if an illegal error_identifier is supplied   */
/* ************************************************************************* */
void set_error_mode(char *error_identifier,char *set_state)
{
  int i = 0;
  int j = 0;
  char *states[] = { "error", "off", "warning", NULL };

  /* First, find out if the configurable soft error (serror) is 
   * defined. */

  while(iopasm_errors[i].error_text != NULL) {

    if (iopasm_errors[i].error_id != NULL) {
      if (!strcmp(error_identifier,iopasm_errors[i].error_id)) {
	break;
      }
      i ++;
    }
    else {
      i ++;
    }

  } /* End of while */
  
  if (iopasm_errors[i].error_id == NULL) {
    iopasm_error(NULL,NULL,ERR_ILLEGAL_WARN_FLAG,
		 "Error identifier %s is not known",error_identifier);
  }

  /* Next find out if mode is set to something sane */
  while(states[j] != NULL) {
    if (!strcmp(states[j],set_state)) {
      break;
    }

    else {
      j++;
    }
  } /* End of second while */

  if (states[j] == NULL) { /* Report error */
        iopasm_error(NULL,NULL,ERR_ILLEGAL_WARN_MODE,
		 "Error state %s is not known",set_state);
  } /* Error caught, exit point */


  /* We are ready to change the mode */
  switch(j) {
  case 0: /* Should result in an error */
    iopasm_errors[i].error_mode = ERROR_ON;
      break;
    
  case 1: /* No error/warning should be reported */
    iopasm_errors[i].error_mode = ERROR_OFF;
    break;
    
    
  case 2: /* Warning should be reported */
    iopasm_errors[i].error_mode = WARNING_ON;
    break;
  } /* End of switch */
}
