/* ************************************************************************* */
/*    iopasm main function and associated command-line parameter handling    */
/*    functions.                                                             */
/*                                                                           */
/*    Copyright 2002-2006 Axis Communications AB.                            */
/*                                                                           */
/*                                                                           */
/*    This program is free software; you can redistribute it and/or modify   */
/*    it under the terms of the GNU General Public License as published by   */
/*    the Free Software Foundation; either version 2 of the License, or      */
/*    (at your option) any later version.                                    */
/*                                                                           */
/*    This program is distributed in the hope that it will be useful,        */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*    GNU General Public License for more details.                           */
/*                                                                           */
/*    You should have received a copy of the GNU General Public License      */
/*    along with this program; if not, write to the Free Software            */
/*    Foundation, Inc., 59 Temple Place - Suite 330,                         */
/*    Boston, MA 02111-1307, USA.                                            */
/*                                                                           */
/*                                                                           */
/*    Send bug reports and inquiries to Johan.Jorgensen@axis.com             */
/* ************************************************************************* */


/* ************************************************************************* */
/*                      INCLUDES USED BY THIS MODULE                         */
/* ************************************************************************* */
#include <assert.h>
#include <fnmatch.h>
#include <getopt.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include "addr_calc.h"
#include "cg_fe.h"
#include "code_generator.h"
#include "debug.h"
#include "dyn_array.h"
#include "iopasm_err.h"
#include "iopasm_global.h"
#include "labels.h"
#include "load_operations.h"
#include "parser.h"
#include "p_list.h"
#include "save_operations.h"
#include "stage_1.h"




/* ************************************************************************* */
/*                     DEFINITIONS LOCAL TO THIS MODULE                      */
/* ************************************************************************* */



/* ************************************************************************* */
/*                       TYPES DEFINED IN THIS MODULE                        */
/* ************************************************************************* */
struct __option_t {
  char *name;
  int flag;
  void *value;
};

typedef struct __option_t cmd_option_t;

/* ************************************************************************* */
/*                              CONSTANT VARIABLES                           */
/* ************************************************************************* */
DEBUG_INIT(); /* Initialize debugger macros                                  */

static struct option iopasm_opts[] = { { "verbose", 0, 0, 1 },
                                       { "arch", 1, 0, 0 },
				       { "output-file", 1, 0 ,0 },
				       { "output-format", 1, 0, 0 },
				       { "symbol-file", 1, 0, 0 },
				       { "include-path", 1, 0, 0 },
				       { "link-only", 0, 0, 1 },
				       { "list-file", 1, 0, 0 },
				       { "namespace", 1, 0, 0 },
                                       { "debug", 1, 0, 0 },
				       { "save-object",1, 0, 0},
				       { "symbol-file-format", 1, 0, 0 },
				       { "preload-object", 1, 0, 0 },
				       { "generate-c-macros", 1, 0, 0 },
				       { "preload-symbol-file", 1, 0, 0 },
				       { "assemble-only", 0, 0, 1 },
				       { "no-cpp", 0,0 ,1 },
				       { "cpp-only", 0, 0, 1 },
				       { "optimize-fsm", 0, 0, 1 },
				       { "with-builtin-includes", 0, 0, 1 },
				       { "allow-undefined-symbols", 0, 0, 1 },
                                       { NULL, 0, 0, 0 } };
  


static cmd_option_t cmdline_options[] = { 
  { "verbose", 0, NULL },
  { "arch", 0, NULL },
  { "cpp-only", 0, NULL },
  { "debug", 0, NULL },
  { "output-file",  0, NULL },
  { "output-format", 0, NULL },
  { "symbol-file", 0, NULL },
  { "include-path", 0, NULL },
  { "link-only", 0, NULL },
  { "list-file", 0, NULL },
  { "namespace", 0, NULL },
  { "save-object", 0,  NULL },
  { "symbol-file-format", 0, NULL },
  { "preload-object", 0, NULL },
  { "generate-c-macros", 0, NULL },
  { "preload-symbol-file", 0, NULL },
  { "assemble-only", 0, NULL },
  { "no-cpp", 0, NULL },
  { "optimize-fsm", 0, NULL },
  { "with-builtin-includes", 0, NULL },
  { "allow-undefined-symbols", 0, NULL },
  { NULL, 0, NULL } };


/* ************************************************************************* */
/*                    PROTOTYPES FOR STATIC FUNCTIONS                        */
/* ************************************************************************* */
static void iopasm_help();
static void preload_symbols(array_t *labels,array_t *files);
static void iopasm_greeting();
static void preload_objects(cg_context_t *ctxt,plist_t *a_list,array_t *files);


extern void parse_file(plist_t *a_list,char *f_name,cg_context_t *arch_context,
		       char *src_fn,int index);

/* ************************************************************************* */
/* PROTOTYPE   : static void iopasm_greeting();                              */
/*                                                                           */
/* ARGUMENTS   : none                                                        */
/*                                                                           */
/* FUNCTION    : Display a simple greeting on startup                        */
/*                                                                           */
/* RETURN VALUE: None                                                        */
/* ************************************************************************* */
static void iopasm_greeting()
{
  fprintf(stderr,"IOP assembler version %d.%d ",IOPASM_MAJOR_VERSION,
	  IOPASM_MINOR_VERSION);
  fprintf(stderr,"(C)2002 Axis Communications AB.\n");
  fprintf(stderr,"Built %s %s using gcc version %s\n\n", __DATE__, __TIME__,
	  __VERSION__);
  fprintf(stderr,"Send bug reports/inqueries to Johan.Jorgensen@axis.com\n\n");
}


void *iopasm_malloc(char *file,int line,size_t size)
{
  void *data;
  data = malloc(size);
  fprintf(stderr,"%s:%d  malloc(%d) -> %p\n",file,line,(int)size,data);
  return data;
}

void iopasm_free(char *file,int line,void *data)
{
  fprintf(stderr,"%s:%d free(%p)\n",file,line,data);
  free(data);
}



/* ************************************************************************* */
/* PROTOTYPE   : static void iopasm_help();                                  */
/*                                                                           */
/* ARGUMENTS   : None                                                        */
/*                                                                           */
/* FUNCTION    : Displays a brief help message listing the various command-  */
/*               line options.                                               */
/*                                                                           */
/* RETURN VALUE: None                                                        */
/* ************************************************************************* */
static void iopasm_help()
{

  fprintf(stderr,"Usage: iopasm [switches] -o output-file input-file.s\n");
  fprintf(stderr,"Switches: (short versions within parantheses)\n\n");
  fprintf(stderr,"          --allow-undefined-symbols      - Set undefined labels to 0\n");
  fprintf(stderr,"          --arch (-M) <arch>             - Select architecture [SPU|MPU]\n");
  fprintf(stderr,"          --assemble-only (-c)           - Only run through parsing stage\n");
  fprintf(stderr,"          --cpp-only                     - Only run through C-preprocessor\n");
  fprintf(stderr,"          -D <cpp-macro>                 - Pass -D to C pre-processor\n");
  fprintf(stderr,"          -e <condition>                 - Set <condition> to cause an error\n");
  fprintf(stderr,"                                           (e.g. -eunaligned-64bit-inst)\n");
  fprintf(stderr,"          --generate-c-macros <file>     - Generate instruction macros\n");
  fprintf(stderr,"          --include-path (-I) <path>     - Make cpp look in path also\n");
  fprintf(stderr,"          --list-file <file>             - Name of list-file\n");
  fprintf(stderr,"          --link-only, -l                - Run linker stage only\n");
  fprintf(stderr,"           -m <machine-option>           - Pass machine-specific option\n");
  fprintf(stderr,"                                           to code-generator\n");
  fprintf(stderr,"           -n <condition>                - Set <condition> to generate neither\n"); 
  fprintf(stderr,"                                           error nor warning (See -e and -w)\n");
  fprintf(stderr,"          --namespace <string>           - Set prefix of namepace in headers\n");
  fprintf(stderr,"                                           to <string>\n");
  fprintf(stderr,"          --preload-symbol-file <file>   - Load symbol table before parsing\n");
  fprintf(stderr,"          --no-cpp                       - Do not run C-preprocessor\n");
  fprintf(stderr,"          --optimize-fsm                 - Run FSM memory optimizer\n");
  fprintf(stderr,"                                           (off per default)\n");
  fprintf(stderr,"          --output-file, (-o) <file>     - Write output to file\n");
  fprintf(stderr,"          --output-format (-t)  <format> - Output file format [binary|vlog64|\n"); 
  fprintf(stderr,"                                           vlog256|object] (default: object)\n");
  fprintf(stderr,"          --preload-object <file>        - Load object file before parsing\n");
  fprintf(stderr,"          --symbol-file (-s) <file>      - Name of symbol file\n");
  fprintf(stderr,"          --symbol-file-format <format>  - Format of symbol file [header|plain]\n");
  fprintf(stderr,"                                           (default is plain)\n");
  fprintf(stderr,"          --verbose (-v)                 - Be verbose (show stages)\n");
  fprintf(stderr,"           -w <condition>                - Set <condition> to cause a warning\n");
  fprintf(stderr,"                                           e.g. (-wunaligned-64bit-inst)\n");
  fprintf(stderr,"          --with-builtin-includes        - Use built-in path to include files\n");
  return;
}


/* ************************************************************************* */
/* PROTOTYPE   : static cg_context_t *iopasm_create_context(                 */
/*                                         array_t *label_array);            */
/*                                                                           */
/* ARGUMENTS   : "label_array" is a dynamic array used to hold the labels    */
/*               found in the code.                                          */
/*                                                                           */
/* FUNCTION    : Creates a code-generation context                           */
/*                                                                           */
/* RETURN VALUE: Pointer to a valid code generation context                  */
/* ************************************************************************* */
static cg_context_t *iopasm_create_context(array_t *label_array)
{
  cg_context_t *seq_context;
  char *arch;

  arch = (char *) get_option("arch");
  
  if (check_option("arch","MPU",OPT_STRING)) {
    seq_context = cg_context_create(ARCH_MPU,label_array);
    assert(seq_context != NULL);
  }
  
  else if (check_option("arch","SPU",OPT_STRING)) {
    seq_context = cg_context_create(ARCH_SPU,label_array);
    assert(seq_context != NULL);
  }
  
  else {
    iopasm_error(NULL,NULL,ERR_UNSUPPORTED_ARCH,
		 "You tried to assemble a program for an architecture named\
 \"%s\". The assembler only supports \"SPU\" and \"MPU\" as architecture\
specifiers",arch);
    iopasm_help();
    exit(EXIT_FAILURE);
  }

  return seq_context;
}

/* ************************************************************************* */
/* PROTOTYPE   : static int str_compare(void *s1,void *s2);                  */
/*                                                                           */
/* ARGUMENTS   : "s1" is a pointer to a string and so is "s2".               */
/*                                                                           */
/* FUNCTION    : Compares the two strings. A simple wrapper around           */
/*               strcmp(...) used by the dynamic array ADT.                  */
/*                                                                           */
/* RETURN VALUE: 0 if the strings matches, <0 if s2>s1 and >0 if s1>s2       */
/* ************************************************************************* */
int str_compare(void *s1,void *s2) { return strcmp(s1,s2);  }



/* ************************************************************************* */
/* PROTOTYPE   : static void str_destroy(void *s1);                          */
/*                                                                           */
/* ARGUMENTS   : "s1" is a pointer to a string which will be destroyed       */
/*               (de-allocated)                                              */
/*                                                                           */
/* FUNCTION    : Destructor used by dyn_array ADT for destroying dynamically */
/*               allocated strings                                           */
/*                                                                           */
/* RETURN VALUE: None                                                        */
/* ************************************************************************* */
void str_destroy(void *s1) { free(s1); }


/* ************************************************************************* */
/* PROTOTYPE   : static void preload_symbols(array_t *labels,                */
/*                                           array_t *files);                */
/*                                                                           */
/* ARGUMENTS   : "labels" is a pointer to the array of labels defined for    */
/*               the current assembly-pass. "files" is an array containing   */
/*               filenames from which labels are to be loaded.               */
/*                                                                           */
/* FUNCTION    : Loads labels and assigns values to them prior to parsing    */
/*               the parsing stage.                                          */
/*                                                                           */
/* RETURN VALUE: None                                                        */
/* ************************************************************************* */
static void preload_symbols(array_t *labels,array_t *files)
{
  int i = 0;
  assert ((labels != NULL) && (files != NULL));

  while (get_array_elem(files,i) != NULL) {
    load_symbol_file(labels,(char *)get_array_elem(files,i));
    i++;
  }
}


/* ************************************************************************* */
/* PROTOTYPE   : static void preload_objects(cg_context_t *ctxt,             */
/*                                           plist_t *a_list,                */
/*                                           array_t *files);                */ 
/*                                                                           */
/* ARGUMENTS   : "ctxt" is a pointer to a context-descriptor. "a_list" is    */
/*               the pointer to the parse_list and "files" is an array       */
/*               containing the names of object files to be loaded.          */
/*                                                                           */
/* FUNCTION    : Loads an arbitrary number of object files prior to calling  */
/*               the parser stage                                            */
/*                                                                           */
/* RETURN VALUE: This function does not return any value. Errors are managed */
/*               internally.                                                 */
/* ************************************************************************* */
static void preload_objects(cg_context_t *ctxt,plist_t *a_list,array_t *files)
{
  int i = 0;

  assert ((ctxt != NULL) && (a_list != NULL) && (files != NULL));
  
  while (get_array_elem(files,i) != NULL) {
    load_object_file(ctxt,a_list,(char *)get_array_elem(files,i));
    i++;
  }
}




/* ************************************************************************* */
/* PROTOTYPE   : int check_option(char *option,void *option_val,             */
/*                                int interpretation)                        */
/*                                                                           */
/* ARGUMENTS   : "option" is the name of an option. "option_val" is a        */
/*               pointer pointing to a value. "interpretation" is the method */
/*               used for comparison (OPT_INTEGER|OPT_STRING)                */
/*                                                                           */
/* FUNCTION    : Checks if "option" is set to "option_val" using the         */
/*               comparison method specified by "interpretation"             */
/*                                                                           */
/* RETURN VALUE: Returns 1 if option is set as expected. Zero otherwise      */
/* ************************************************************************* */
int check_option(char *option,void *option_val,int interpretation)
{
  int i,len;

  i = 0; 
  len = 0;

  while(cmdline_options[i].name != NULL) {
    if (!strcmp(option,cmdline_options[i].name)) {
      
      if (cmdline_options[i].value == NULL) { 
	return 0; 
      }
      else if (interpretation == OPT_STRING) {
	len = strlen((char *) cmdline_options[i].value);
	return !strncasecmp((char *)cmdline_options[i].value,
			    (char *)option_val,len);
      }

      else if (interpretation == OPT_INTEGER) {
	if (((int )cmdline_options[i].flag) == ((int )option_val)) return 1;
      }
      
      else return 0;
    }

    i++;
  }
  return 0;
}


/* ************************************************************************* */
/* PROTOTYPE   : void *get_option(char * option);                            */
/*                                                                           */
/* ARGUMENTS   : "option" is a pointer to a string specifying the long-name  */
/*               of an option.                                               */
/*                                                                           */
/* FUNCTION    : Returns the current value of the option specified by        */
/*               "option".                                                   */
/*                                                                           */
/* RETURN VALUE: A pointer to the value of requested option. If option is    */
/*               undefined, NULL is returned.                                */
/* ************************************************************************* */
void *get_option(char * option)
{ 
  int i=0;
  
  while(cmdline_options[i].name != NULL) {

    if (!strcmp(option,cmdline_options[i].name)) {
      return cmdline_options[i].value; 
    }

    else { i++; }
  
  }

  return NULL;
}


/* ************************************************************************* */
/* PROTOTYPE   : static int set_option(char *opt_name,void *option_ptr);     */
/*                                                                           */
/* ARGUMENTS   : "opt_name" is a pointer to the name of an option.           */
/*               "option_ptr" points to a value which we want to associate   */
/*               with option                                                 */
/*                                                                           */
/* FUNCTION    : Associates the value pointed to by "option_ptr" to the      */
/*               option specified by "option"                                */
/*                                                                           */
/* RETURN VALUE: >0 if option was successfully set <0 if not                 */
/* ************************************************************************* */
static int set_option(char *opt_name,void *option_ptr)
{
  int i=0;
  
  while(cmdline_options[i].name != NULL) {
    if (!strcmp(opt_name,cmdline_options[i].name)) {
      cmdline_options[i].value = option_ptr;
      cmdline_options[i].flag = 1;
      return 1;
    }
    i++;
  }

  return -1;
}



/* ************************************************************************* */
/* PROTOTYPE   : int main(int argc,char **argv);                             */
/*                                                                           */
/* ARGUMENTS   : argc is the standard arg-count and argv is the arg-vector.  */
/*               Should not really nead to be explained                      */
/*                                                                           */
/* FUNCTION    : Main function as specified by ISO-C                         */
/*                                                                           */
/* RETURN VALUE: 0 on success. 1 if iopasm fails                             */
/* ************************************************************************* */
int main(int argc,char **argv)
{
  int c=0;
  int option_index;
  char *argument;
  const char *out_format_default = "binary";
  const char *out_format_object = "object";
  const char *default_symbol_format = "plain";
  char *out_format = NULL;
  char *src_fn;
  char *dst_fn;
  char *sym_file = NULL;
  char *list_file = NULL;
  char *output_file = NULL;
  char *macro_file = NULL;
  int opts_found=0;
  int i;
  int i_count = 0;
  plist_t *parsed_code=NULL;
  plist_t *optimized_code=NULL;
  plist_t *unoptimized_code=NULL;
  array_t *preloaded_objects;
  array_t *preloaded_symbols;
  array_t *include_paths;
  array_t *label_array;
  cg_context_t *seq_context = NULL;
  unsigned char *binary_buffer = NULL;
  char *namespace = NULL;
  char *new_arg;
  char *exec_name;
  int error_code;
  label_t *a_label;

  SET_DEBUG_LEVEL(4);
  setvbuf(stdout, (char *)NULL, _IONBF, 0);
  setvbuf(stderr, (char *)NULL, _IONBF, 0);
  
  iopasm_greeting();
  
  preloaded_objects = create_array(str_compare,str_destroy);
  preloaded_symbols = create_array(str_compare,str_destroy);
  include_paths = create_array(str_compare,str_destroy);

  assert ((preloaded_symbols != NULL) && 
	  (preloaded_objects != NULL) && 
	  (include_paths != NULL));

  /* Get executable name */
  exec_name = (char *) strdup(argv[0]);
  assert(exec_name != NULL);

  /* Set default options */
  set_option((char *) "symbol-file-format",(void *) default_symbol_format);
  set_option((char *) "output-format",(void *) out_format_default);
  set_option((char *) "verbose", (void *) 0);

  /* Create the label array so that it is ready to be used when the 
   * context is created */
  label_array = create_array(compare_labels,label_destructor);
  if (label_array == NULL) {
    DEBUG(1,"Cannot allocate memory for label array %p",label_array);
    perror("Could not allocate memory for label array");
    return 1;
  }


  /* Scan for -- and pass remaining options to the C preprocessor. 
   * This fixes TR #37 */
  c = 0;
  for (i=0;i<argc;i++) {
    if (!strncmp("--",argv[i],strlen(argv[i]))) {
      c = i;
    }
  }
  
  if (c != 0) {  /* We need to copy arguments */
    for(i=c+1;i<argc;i++) {
      new_arg = (char *)strdup(argv[i]);
      assert(new_arg != NULL);
      add_array_elem(include_paths,new_arg);
    }
  }
  /* End of fix for TR #37 */


  do {
    c = getopt_long (argc,argv,"vD:w:e:n:I:lcM:m:o:s:t:",iopasm_opts, &option_index);

    switch (c) {
      
    case 0:
    case 1:
      if (iopasm_opts[option_index].has_arg) {
	argument = (char *)strdup((char *)optarg);
	assert (argument != NULL);

	if (!strcmp(iopasm_opts[option_index].name,"preload-object")) {
	  add_array_elem(preloaded_objects,argument);
	}
	
	else if (!strcmp(iopasm_opts[option_index].name,"load-symbol-file")) {
	  add_array_elem(preloaded_symbols,argument);
	}

	else if (!strcmp(iopasm_opts[option_index].name,"include-path")) {
	  add_array_elem(include_paths,argument);
	}
	
	else if (!strcmp(iopasm_opts[option_index].name,"arch")) {
	  set_option((char *)iopasm_opts[option_index].name,(void *)argument);
	  seq_context = iopasm_create_context(label_array);
	}

	else {
	  set_option((char *)iopasm_opts[option_index].name,(void *)argument);
	}
      } /* !if (iopasm_opts[option_index].has_arg) ... */
      
      else {
	set_option((char *)iopasm_opts[option_index].name,(void *) 1);
      }
      
      opts_found++;
      break;


      
    case 'c': 
      set_option("assemble-only",(void *)1);
      set_option((char *) "output-format",(void *) out_format_object);
      opts_found++;
      break;

    
      /* Fix TR #40 by passing -D switch to CPP */
    case 'D':
      i = strlen(optarg) + 3;
      new_arg = (char *) malloc((sizeof(char) * i));
      assert(new_arg != NULL);
      memset(new_arg,0,i);
      strncpy(new_arg,"-D",2);
      strcat(new_arg,optarg);
      add_array_elem(include_paths,new_arg);
      break;

    
    case 'e': /* Set "serror" to report an error */
      set_error_mode(optarg,"error");
      break;


    case 'l':
      set_option("link-only",(void *)1);
      set_option("no-cpp",(void *)1);
      set_option("assemble-only",(void *)0);
      opts_found++;
      break;


    case 'm':
      seq_context = iopasm_create_context(label_array);
      if (seq_context == NULL) { /* No context has been created yet */
	iopasm_error(NULL,NULL,ERR_REPORT_BUG,NULL);
      }
      cg_get_machine_option(seq_context,optarg,&error_code);
      if (error_code < 0) {
	iopasm_error(seq_context,NULL,ERR_UNKNOWN_MACHINE_OPTION,
		     "You specified the machine specific option -m%s. The currently selected architecture does not know about that.",
		     optarg);
      }
      else {
	cg_set_machine_option(seq_context,optarg,1);
      }
      break;
      

    case 'M':
      argument = (char *)strdup((char *)optarg);
      set_option("arch",(void *) argument);
      seq_context = iopasm_create_context(label_array);
      opts_found++;
      break;


    case 'n':
      set_error_mode(optarg,"off");
      break;


    case 'o':
      output_file = (char *)strdup((char *)optarg);
      set_option("output-file",(void *) output_file);
      opts_found++;
      break;
      

    case 's':
      sym_file = (char *) strdup((char *)optarg);
      set_option("symbol-file",(void *) sym_file);
      opts_found++;
      break;


    case 'I':
      i = strlen(optarg) + 3; 
      new_arg = (char *) malloc((size_t) i);
      assert(new_arg != NULL);
      memset(new_arg,0,i);
      strncpy(new_arg,"-I",2);
      strcat(new_arg,optarg);
      add_array_elem(include_paths,new_arg);
      break;
    



    case 't':
      out_format = (char *) strdup((char *)optarg);
      set_option("output-format",(void *) out_format);
      opts_found++;
      break;
      

    case 'v':
      set_option("verbose",(void *) 0);
      opts_found++;
      break;
    

    case 'w':
      set_error_mode(optarg,"warning");
      break;
    }
    
    
  } while(c!=-1);

  
  if (opts_found == 0) {
    iopasm_help(); 
    return 0;
  }
  
  /* Find source file  */
  i = argc - 1;
     
  while(fnmatch("*.[sS]", argv[i],0) && i) { i--; }  src_fn = argv[i];
  dst_fn = create_tmpfile();
  
  DEBUG(3,"Source file is : %s",src_fn);
  DEBUG(3,"Destination file is : %s",dst_fn);
  
  iopasm_verbose_msg("Stage 1: Initializing assembler...");

  parsed_code = create_plist();
  if (parsed_code == NULL) {
    perror("Could not allocate memory for parse-list");
    return 1;
  }
  
  /* The assembler is now initialized and we are ready to proceed. First we must 
   * check that interdependencies between cmd-line arguments are ok. For instance:
   * --cpp-only imlies that no-cpp should be off or vice-versa. 
   * --cpp-only has precendence over --no-cpp as I assume people know what they're 
   * doing
   */
  
  if (check_option("cpp-only",(int *)1,OPT_INTEGER)) {
    set_option("no-cpp",(void *)0);
  }
  
  /* Now we will continue to process various "asynchronous" cmd-line options */

      
  
  /* Dump object file */
  if (get_option("debug") != NULL) {
    dump_object_file((char *)get_option("debug"),seq_context);
    return 0;
  }
  
  if (get_option("namespace") != NULL) {
    namespace = strdup((char *)get_option("namespace"));
    assert(namespace != NULL);
    cg_set_namespace(seq_context,namespace);
  }

  /* Below we should perform first check to see if we are generating C-macros.
   * If that is the case, we will simply just run the c-macro generation and then exit */
  macro_file = (char *) get_option("generate-c-macros");
  if (macro_file != NULL) { /* Yes - we should generate macros */
    iopasm_verbose_msg("Stage 2: Generating C-macros... (Only stage to run)");
    save_c_macros(macro_file,seq_context);
    return 0;
  }
    
  /* If we only need to link code, we simply just run through the preloading stage 
   * and then skip the compilation stage 
   */
  if (!check_option("link-only",(int *) 1,OPT_INTEGER)) {
    if (get_array_elem(preloaded_objects,0) != NULL) {
	iopasm_verbose_msg("Stage 2: Preloading objects...");
      preload_objects(seq_context,parsed_code,preloaded_objects); 
    }
    
    if (get_array_elem(preloaded_symbols,0) != NULL) {
      iopasm_verbose_msg("         Preloading symbols...");
      preload_symbols(label_array,preloaded_symbols);
    }

    /* Check if we allow undefined symbols */
    if (check_option("allow-undefined-symbols",(int *)1,OPT_INTEGER)) {
      new_arg = strdup("-D__ALLOW_UNDEFINED_SYMBOLS__");
      add_array_elem(include_paths,new_arg);
    }

    /* Now let us check if we want to run cpp or not - at this stage we 
     * need to check if we have a valid input file name */
    
    if (!strcmp(exec_name,src_fn)) { 
      iopasm_error(seq_context,NULL,ERR_INPUT_FILE_MISSING,NULL);
    }
    
    if (!check_option("no-cpp",(int *)1,OPT_INTEGER))
      {
	if (check_option("cpp-only",(int *)1,OPT_INTEGER)) {
	  dst_fn = (char *) get_option("output-file");
	  exec_cpp(seq_context,src_fn,dst_fn,include_paths);
	  goto clean_up;
	} /* End of if (check_option("cpp-only",(int *)1,OPT_INTEGER)) */
	
	else {
	  exec_cpp(seq_context,src_fn,dst_fn,include_paths);
	} /* End of !if (check_option("cpp-only",(int *)1,OPT_INTEGER)) */
      } /* End of if (!check_option("no-cpp",(int *)1,OPT_INTEGER)) */


    else {
    /* Make sure we get the right input */
      free(dst_fn);
      dst_fn = src_fn;
    }
    /* Initialize the parser list */
    if (seq_context == NULL) {
      iopasm_error(seq_context,NULL,ERR_UNSUPPORTED_ARCH,NULL);
    }
    iopasm_verbose_msg("Stage 2: Parsing source file...");
    i = cg_index_file(seq_context,src_fn);
    parse_file(parsed_code,dst_fn,seq_context,src_fn,i);

    /* Check that file is not empty (According to TR #15) */

    i_count = get_inst_count(parsed_code,INST_SEQ);
    i_count += get_inst_count(parsed_code,INST_TIMER);
    i_count += get_inst_count(parsed_code,INST_ONLY_STATE);
    i_count += get_inst_count(parsed_code,INST_STATE_XFER);
    i_count += get_inst_count(parsed_code,INST_DWORD);
    
    if (i_count == 0) {
      iopasm_error(seq_context,NULL,ERR_FILE_IS_EMPTY,
		   "(The source file is %s)",src_fn);
    }

  } /* end if (!check_option("link-only",(int *) 1,OPT_INTEGER); */
   
    
  else { /* We are linking so we just load our objects */
    iopasm_verbose_msg("Stage 2: Loading object files... (linking)");
    i = argc -1;
    do {
      if (!fnmatch("*.[oO]",argv[i],0)) {
	fprintf(stderr,"         Loading object: %s\n",argv[i]);
	load_object_file(seq_context,parsed_code,argv[i]);
      	/* dump_object_file(argv[i],seq_context); */
      } /* End of if (!fnmatch("*.[oO]",argv[i],FNM_FILE_NAME)) { */
      i--;

    } while(i>-1);
  } /* End of Linker stage */

  /* Check if parsing resulted in any code i.e. that we did not parse an empty file.
   * if that is the case, we simply emit an error and exit. 
   * 
   * Correction according to TR #15 
   */
  
 

   /* Check if we only need to compile code. If so we will skip code generation and
   * Just generate an object file with the name of the output-file.
   */
  if (!check_option("assemble-only",(int *)1,OPT_INTEGER)) { 

    /* Step 3: Calculate size of instructions */
    iopasm_verbose_msg("Stage 3: Generate intermediate code...");
    calculate_code_size(seq_context,parsed_code);
    
    /* Step 4: State optimization - not enabled per default.. */
    unoptimized_code = parsed_code;
    
    if (check_option("optimize-fsm",(int *) 1,OPT_INTEGER)) {
      iopasm_verbose_msg("Stage 4: Optimizing states..."); 
      optimized_code = cg_optimize_fsm_states(seq_context,parsed_code);
      
      /* Clear all labels - they do not have valid adresses any more */
      for (i=0;i<get_array_size(label_array);i++) {
	a_label = (label_t *) get_array_elem(label_array,i);
	change_label_value(a_label,0);
      } /* End of for (i=0;i<get_array_size(label_array);i++) */
      
      /* Delete all intermediate code and clear the allocated memory regions.
       * Everything has been invalidated */
      delete_code(optimized_code);
      cg_clear_memory_region(seq_context,0,cg_get_memory_size(seq_context));

      /* Switch the parsed code to use optimized code instead */
      parsed_code = optimized_code;
      /* Recalulate the code size, then we're ready with optimization */
      calculate_code_size(seq_context,parsed_code);
    }
    
    

    /* Step 5: Absolute address calculation */
    iopasm_verbose_msg("Stage 5: Calculating absolute addresses...");
    address_calculation(seq_context,parsed_code);
  
    /* Step 6: Code Generation */
    iopasm_verbose_msg("Stage 6: Generating code...");
    generate_code(seq_context,parsed_code);
    cg_post_code_generation(seq_context,parsed_code);
    
    binary_buffer = cg_get_code_buffer(seq_context);
  } /* if (!check_option("compile-only",...) */
  
  else { /* We are only compiling - let the user know it! */
    iopasm_verbose_msg("Stage 3: Generate intermediate code... (Skipped - only assembling)");
    iopasm_verbose_msg("Stage 4: Optimizing states... (Skipped - only assembling)");
    iopasm_verbose_msg("Stage 5: Calculating absolute addresses... (Skipped - only assembling)");
    iopasm_verbose_msg("Stage 6: Generating code... (Skipped - only assembling)");
    set_option("output_format",(void *) out_format_object);
  }


  /* Step 7: Now let us perform another check for output-filename options */
  if ((get_option("symbol-file") != NULL) && (sym_file == NULL)) {
    sym_file = (char *) get_option("symbol-file");
  }
  
  if ((get_option("output-file") != NULL) && (output_file == NULL)) {
    output_file = (char *) get_option("output-file");
  }
  
  if ((get_option("list-file") != NULL) && (list_file == NULL)) {
    list_file = (char *) get_option("list-file");
  }
  
  /* Step 7A: Find out what output-format to use. And save the output
   * of the assembler using that format
   */
  iopasm_verbose_msg("Stage 7: Saving files: ");
  if ((get_option("output-format") != NULL)) {
    out_format = (char *) get_option("output-format");
  }
  
  else if (check_option("assemble-only",(int *)1,OPT_INTEGER)) {
    out_format = (char *) out_format_object;
    set_option("output-format",(void *) out_format);
  }
  
  else {
    out_format = (char *) out_format_default;
    set_option("output-format",(void *) out_format);
  }
  


  if (check_option("output-format","vlog64",OPT_STRING)) {
    iopasm_verbose_msg("         Output file: %s (Verilog 64-bit format)\n",output_file);
    save_vlog64_file(output_file,seq_context,parsed_code,binary_buffer);
  }
  
  else if (check_option("output-format","vlog256",OPT_STRING)) {
    iopasm_verbose_msg("         Output file: %s (Verilog 256-bit format)\n",output_file);
    save_vlog256_file(output_file,seq_context,parsed_code,binary_buffer);
  }
  
  else if (check_option("output-format","object",OPT_STRING)) {
    iopasm_verbose_msg("         Output file: %s (Object)\n",output_file);
    save_object_file(output_file,seq_context,parsed_code);
  }
  
  else if (check_option("output-format", "binary",OPT_STRING)) {
    iopasm_verbose_msg("         Output file: %s (Binary)\n",output_file);
    save_binary_file(output_file,seq_context,parsed_code,binary_buffer);
  }
  
  else {
        iopasm_verbose_msg("         Not saving output file\n");
  }
  
  /* Step 8: Write symbol and list files -if desired */
  if (sym_file != NULL) {
    if (check_option("symbol-file-format","header",OPT_STRING)) {
      iopasm_verbose_msg("         Symbol file: %s (C-Header)",sym_file);
      save_symbol_file(sym_file,SYM_FILE_FMT_HEADER,seq_context,parsed_code);
    }
    
    else if (check_option("symbol-file-format","plain",OPT_STRING)) {
      iopasm_verbose_msg("         Symbol file: %s (Plain)",sym_file);
      save_symbol_file(sym_file,SYM_FILE_FMT_SYMBOLS,seq_context,parsed_code);
    }
    else {
      iopasm_verbose_msg("Warning: Not saving symbol file - unknown format");
    }
  }

  if (list_file != NULL) {
    iopasm_verbose_msg("         List file    : %s\n",list_file);
    save_list_file(list_file,seq_context,unoptimized_code,optimized_code);
  }
  
  
  
  /* Step 9: Clean up! */
  clean_up:
  delete_array(label_array);
/*   delete_plist(parsed_code); */
  cg_delete(seq_context);
  if ((!check_option("no-cpp",(int *)1,OPT_INTEGER)) && 
      (!check_option("cpp-only",(int *)1,OPT_INTEGER)))
    {
      unlink(dst_fn);
    }
  iopasm_verbose_msg("Stage 9: IOPASM completed successfully...");
  return 0;
}

