/* ************************************************************************* */
/* labels.c : Handles symbolic labels.                                       */
/*                                                                           */
/*    Copyright 2002-2006 Axis Communications AB.                            */
/*                                                                           */
/*                                                                           */
/*    This program is free software; you can redistribute it and/or modify   */
/*    it under the terms of the GNU General Public License as published by   */
/*    the Free Software Foundation; either version 2 of the License, or      */
/*    (at your option) any later version.                                    */
/*                                                                           */
/*    This program is distributed in the hope that it will be useful,        */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*    GNU General Public License for more details.                           */
/*                                                                           */
/*    You should have received a copy of the GNU General Public License      */
/*    along with this program; if not, write to the Free Software            */
/*    Foundation, Inc., 59 Temple Place - Suite 330,                         */
/*    Boston, MA 02111-1307, USA.                                            */
/*                                                                           */
/*                                                                           */
/*    Send bug reports and inquiries to Johan.Jorgensen@axis.com             */
/* ************************************************************************* */



/* ************************************************************************* */
/*                      INCLUDES USED BY THIS MODULE                         */
/* ************************************************************************* */
#include <assert.h>
#include <ctype.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cg_fe.h"
#include "debug.h"
#include "iopasm_err.h"
#include "iopasm_global.h"
#include "labels.h"
#include "meta_object.h"

/* ************************************************************************* */
/*                     DEFINITIONS LOCAL TO THIS MODULE                      */
/* ************************************************************************* */
#define MAX_LABEL_SIZE 32 /* Labels cannot be more than 32-characters long   */
#define LABEL_T_SIZE (sizeof(struct __label_t))



/* ************************************************************************* */
/*                       TYPES DEFINED IN THIS MODULE                        */
/* ************************************************************************* */
struct __label_t {
  char name[MAX_LABEL_SIZE];
  uint32_t scope;
  uint32_t value;
};



/* ************************************************************************* */
/*                    PROTOTYPES FOR STATIC FUNCTIONS                        */
/* ************************************************************************* */



/* ************************************************************************* */
/* PROTOTYPE   : label_t *create_label(char *name,unsigned_int value);       */
/*                                                                           */
/* ARGUMENTS   : name is the name of the label, value is the initial value   */
/*                                                                           */
/* FUNCTION    : Allocates memory for and returns a pointer to a new label   */
/*                                                                           */
/* RETURN VALUE: Pointer to a new label, NULL on failure                     */
/* ************************************************************************* */
label_t *create_label(char *label_name,unsigned int label_value,
		      int label_scope)
{
  label_t *new_label;

  new_label = (label_t *) IOPASM_MALLOC(sizeof(label_t));
  assert(new_label != NULL);
  
  memset(new_label->name,0,MAX_LABEL_SIZE); 
  strncpy(new_label->name,label_name,MAX_LABEL_SIZE);
  new_label->value = (uint32_t) label_value;
  new_label->scope = (uint32_t) label_scope;


  return new_label;
  
}



/* ************************************************************************* */
/* PROTOTYPE   : int change_label_value(label_t *label,unsigned int val);    */
/*                                                                           */
/* ARGUMENTS   : "label" is a pointer to a label. "val" is the new value     */
/*                                                                           */
/* FUNCTION    : Changes the value of a label                                */
/*                                                                           */
/* RETURN VALUE: 1 if value was changed, -1 on failure                       */
/* ************************************************************************* */
int change_label_value(label_t *label,unsigned int val)
{
  if (label==NULL) return -1;
  else {
    label->value = (uint32_t) val;
    return 1;
  }
}



/* ************************************************************************* */
/* PROTOTYPE   : int delete_label(label_t *label);                           */
/*                                                                           */
/* ARGUMENTS   : label is a pointer to a label                               */
/*                                                                           */
/* FUNCTION    : Frees up memory                                             */
/*                                                                           */
/* RETURN VALUE: 1 if label deleted -1 on failure                            */
/* ************************************************************************* */
int delete_label(label_t *label)
{
  if (label == NULL) return -1;
  IOPASM_FREE(label);
  return 1;
}



/* ************************************************************************* */
/* PROTOTYPE   : int compare_labels(void *label1,void *label2)               */
/*                                                                           */
/* ARGUMENTS   : "label1" & "label2" are pointers to labels to be compared   */
/*                                                                           */
/* FUNCTION    : Compares two labels.                                        */
/*                                                                           */
/* RETURN VALUE: Returns -ve if label1<label2, 0 if label1 = label2 and +ive */
/*               if label1<label2. Useful for callbacks                      */
/* ************************************************************************* */
int compare_labels(void *label1,void *label2) 
{
  label_t *l1,*l2;

  assert((label1 != NULL) && (label2 != NULL));

  l1 = (label_t *) label1;
  l2 = (label_t *) label2;

  return strcmp(l1->name,l2->name);
}



/* ************************************************************************* */
/* PROTOTYPE   : label_destructor(void *label)                               */
/*                                                                           */
/* ARGUMENTS   : "label" is a pointer to a label. "val" is the new value     */
/*                                                                           */
/* FUNCTION    : Changes the value of a label                                */
/*                                                                           */
/* RETURN VALUE: 1 if value was changed, -1 on failure                       */
/* ************************************************************************* */
void label_destructor(void *data)
{
  label_t *label;
  assert (data != NULL);
  
  label = (label_t *) data; 
  delete_label(label);
}



/* ************************************************************************* */
/* PROTOTYPE   : unsigned int get_label_value(char *label_name,int *error)   */
/*                                                                           */
/* ARGUMENTS   : "label_name" is the name of the label whose value is to be  */
/*               returned. "error" is a pointer to an integer which contains */
/*               the error status of the operation                           */
/*                                                                           */
/* FUNCTION    : Searches for, and returns the value of the label "label"    */
/*                                                                           */
/* RETURN VALUE: "error" is set to a negative value if the label could was   */
/*               not found. Otherwise "error" will be positive               */
/* ************************************************************************* */
unsigned int get_label_value(cg_context_t *context,label_t *label)
{
  unsigned int ret_val = 0;

  if ((check_option("allow-undefined-symbols",(int *)1,OPT_INTEGER)) && 
      (label == NULL)) { /* Undefined labels are allowed return 0 */
    return 0;
  }
  else if ((!check_option("allow-undefined-symbols",(int *)1,OPT_INTEGER)) &&
	   (label == NULL)) {
    iopasm_error(context,NULL,ERR_UNDEFINED_LABEL,NULL);
  }

    
  else if (label->scope == SCOPE_SEQ_ADDR) {
    ret_val = cg_convert_seq_addr(context,label->value);
  }
  
  else if (label->scope == SCOPE_FSM_ADDR) {
    ret_val = cg_convert_fsm_addr(context,label->value);
  }
  
  else {
    ret_val = 0;
  }
  return ret_val;
}



/* ************************************************************************* */
/* PROTOTYPE   : char *get_label_name(label_t *label);                       */
/*                                                                           */
/* ARGUMENTS   : "label" is a pointer to a label.                            */
/*                                                                           */
/* FUNCTION    : Returns the name of a label.                                */
/*                                                                           */
/* RETURN VALUE: Returns a pointer to the label. If label is invalid, NULL   */
/*               is returned instead.                                        */
/* ************************************************************************* */
char *get_label_name(label_t *label)
{
  
  return (label != NULL) ? label->name:NULL;
}



/* ************************************************************************* */
/* PROTOTYPE   : int write_label_as_symbols(void *dst,void *elem);           */
/*                                                                           */
/* ARGUMENTS   : dst is a pointer to the destination. This is the output of  */
/*               the function. elem is a pointer to an element of data.      */
/*                                                                           */
/* FUNCTION    : This function will save a label to a file specified by      */
/*               "dst". The label is stored in "elem". The function is in-   */
/*               tended for use with the array_iterator function.            */
/*                                                                           */
/* RETURN VALUE: Returns the number of bytes written. -1 if the operation    */
/*               failed                                                      */
/* ************************************************************************* */
int write_label_as_symbols(void *context,void *dst,void *elem)
{
  label_t *label;
  FILE *out_file;

  assert ((dst != NULL) && (elem != NULL));

  out_file = (FILE *) dst;
  label = (label_t *) elem;
  return fprintf(out_file,"%s 0x%08x\n",label->name, label->value);

}



/* ************************************************************************* */
/* PROTOTYPE   : int write_label_as_header(void *dst,void *elem);            */
/*                                                                           */
/* ARGUMENTS   : dst is a pointer to the destination. This is the output of  */
/*               the function. elem is a pointer to an element of data.      */
/*                                                                           */
/* FUNCTION    : This function will save a label to a file specified by      */
/*               "dst". The label is stored in "elem". The function is in-   */
/*               tended for use with the array_iterator function.            */
/*                                                                           */
/* RETURN VALUE: Returns the number of bytes written. -1 if the operation    */
/*               failed                                                      */
/* ************************************************************************* */
int write_label_as_header(void *context,void *dst,void *elem)
{
  label_t *label;
  FILE *out_file;
  char *temp_string;
  int i;
  cg_context_t *ctxt;

  assert ((dst != NULL) && (elem != NULL) &&  (context != NULL));
  
  ctxt = (cg_context_t *) context;

  out_file = (FILE *) dst;
  label = (label_t *) elem;

  temp_string = strdup(label->name);
  assert (temp_string != NULL);
  
  for(i=0;i<strlen(temp_string);i++) {
    temp_string[i] = toupper(temp_string[i]);
  }
  if (label->scope == SCOPE_FSM_ADDR) {
    i = fprintf(out_file,"#define %s (0x%x) /* FSM Label */\n",temp_string, 
		get_label_value(ctxt,label));
  }

  else if (label->scope == SCOPE_SEQ_ADDR) {
    i = fprintf(out_file,"#define %s (0x%x) /* SEQ Label */\n",temp_string, 
		get_label_value(ctxt,label));
  }

  else {
    i = fprintf(out_file,"#define %s (0x%x) /* No Scope set */\n",temp_string, 
		get_label_value(ctxt,label));
  }

  IOPASM_FREE(temp_string);
  return i;

}



/* ************************************************************************* */
/* PROTOTYPE   : int write_label_as_object(void *dst,void *elem);            */
/*                                                                           */
/* ARGUMENTS   : dst is a pointer to the destination. This is the output of  */
/*               the function. elem is a pointer to an element of data.      */
/*                                                                           */
/* FUNCTION    : This function will save a label to a file specified by      */
/*               "dst". The label is stored in "elem". The function is in-   */
/*               tended for use with the array_iterator function.            */
/*                                                                           */
/* RETURN VALUE: Returns the number of bytes written. -1 if the operation    */
/*               failed                                                      */
/* ************************************************************************* */
int write_label_as_object(void *context,void *dst,void *elem) 
{
  meta_object_t *my_object;
  label_t *label;
  label_t *saved_label;
  FILE *out_file;

  assert((dst != NULL) && (elem != NULL));
  out_file = (FILE *) dst;
  label = (label_t *) elem;

  saved_label = create_label(label->name,
			     htonl((uint32_t)label->value),
			     htonl((uint32_t)label->scope));

  assert(saved_label != NULL);

  my_object = create_meta_object(LABEL_ELEM,sizeof(label_t));
  save_meta_object(out_file,my_object);
  delete_meta_object (my_object);

  return fwrite(saved_label,sizeof(label_t),1,out_file);
}



/* ************************************************************************* */
/* PROTOTYPE   : int print_label_stdout(cg_context_t *ctxt,                  */
/*                                      void *dst,void *elem);               */
/*                                                                           */
/* ARGUMENTS   : "ctxt" points to a valid code generation context. "dst" is  */
/*               a possible destination pointer and elem is the element to   */
/*               print to stdout.                                            */
/*                                                                           */
/* FUNCTION    : This function prints labels to stdout. It is only supposed  */
/*               to be used by one of the iterator functions.                */
/*                                                                           */
/* RETURN VALUE: See printf()                                                */
/* ************************************************************************* */
int print_label_stdout(cg_context_t *ctxt,void *dst,void *elem) 
{
  label_t *a_label;
  assert((dst != NULL) && (elem != NULL));
  a_label = (label_t *) elem;

  return printf("Label Name: %s, Value: %d\n",a_label->name,
		 a_label->value);
} 



/* ************************************************************************* */
/* PROTOTYPE   : int print_label(cg_context_t *ctxt,void *dst,void *elem);   */
/*                                                                           */
/* ARGUMENTS   : "ctxt" points to a valid code generation context. "dst" is  */
/*               a possible destination pointer and elem is the element to   */
/*               print to stdout.                                            */
/*                                                                           */
/* FUNCTION    : This function prints labels to stdout. It is only supposed  */
/*               to be used by one of the iterator functions.                */
/*                                                                           */
/* RETURN VALUE: see fprintf()                                               */
/* ************************************************************************* */
int print_label(cg_context_t *ctxt,void *dst,void *elem)
{
  label_t *a_label;
  assert((dst != NULL) && (elem != NULL));
  a_label = (label_t *) elem;

  return fprintf((FILE *) dst,"Label Name: %s, Value: %d\n",a_label->name,
		 a_label->value);
}



/* ************************************************************************* */
/* PROTOTYPE   : label_t *load_label_object(FILE *input_file)                */
/*                                                                           */
/* ARGUMENTS   : "input file is a pointer to an open FILE                    */
/*                                                                           */
/* FUNCTION    : Loads a label object from an object file and creates a      */
/*               label.                                                      */
/*                                                                           */
/* RETURN VALUE: Pointer to the newly loaded label                           */
/* ************************************************************************* */
label_t *load_label_object(FILE *input_file)
{
  label_t *a_label;

  assert (input_file != NULL);
  
  a_label = (label_t *) IOPASM_MALLOC(LABEL_T_SIZE);
  assert (a_label != NULL);
  assert (fread(a_label,LABEL_T_SIZE,1,input_file) == 1);
  
  a_label->value = ntohl(a_label->value);
  a_label->scope = ntohl(a_label->scope);
  
  return a_label;
}



/* ************************************************************************* */
/* PROTOTYPE   : void dump_label_object(FILE *input_file);                   */
/*                                                                           */
/* ARGUMENTS   : "input_file" points to the an open file                     */
/*                                                                           */
/* FUNCTION    : dumps the content of a label object to stdout. Only used    */
/*               for debugging purposes.                                     */
/*                                                                           */
/* RETURN VALUE: none.                                                       */
/* ************************************************************************* */
void dump_label_object(FILE *input_file)
{
  label_t a_label;

  assert (input_file != NULL);
  assert (fread(&a_label,LABEL_T_SIZE,1,input_file) == 1);

  a_label.value = ntohl(a_label.value);
  a_label.value = ntohl(a_label.scope);

  fprintf(stdout,"     Label name  : %s\n",a_label.name);
  fprintf(stdout,"     Label value : %d\n",a_label.value);
  fprintf(stdout,"     Label scope : %d\n",a_label.scope);
  fprintf(stdout,"--------------------------------------------------------\n");
}

  
  
