/* ************************************************************************* */
/* matrix.c : Implementation of dynamic matrix                               */
/*                                                                           */
/*    Copyright 2002-2006 Axis Communications AB.                            */
/*                                                                           */
/*                                                                           */
/*    This program is free software; you can redistribute it and/or modify   */
/*    it under the terms of the GNU General Public License as published by   */
/*    the Free Software Foundation; either version 2 of the License, or      */
/*    (at your option) any later version.                                    */
/*                                                                           */
/*    This program is distributed in the hope that it will be useful,        */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*    GNU General Public License for more details.                           */
/*                                                                           */
/*    You should have received a copy of the GNU General Public License      */
/*    along with this program; if not, write to the Free Software            */
/*    Foundation, Inc., 59 Temple Place - Suite 330,                         */
/*    Boston, MA 02111-1307, USA.                                            */
/*                                                                           */
/*                                                                           */
/*    Send bug reports and inquiries to Johan.Jorgensen@axis.com             */
/* ************************************************************************* */



/* ************************************************************************* */
/*                      INCLUDES USED BY THIS MODULE                         */
/* ************************************************************************* */
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "iopasm_global.h"


/* ************************************************************************* */
/*                       TYPES DEFINED IN THIS MODULE                        */
/* ************************************************************************* */
struct __matrix_t {
  int n_rows;
  int n_columns;
  int data_size;
  void (*destructor_func)(void *);
  void **matrix_data;

};

typedef struct __matrix_t matrix_t;



/* ************************************************************************* */
/* PROTOTYPE   : matrix_t *create_matrix(int rows,int cols,int data_size,    */
/*			                 void (*destructor)(void *));        */
/*                                                                           */
/* ARGUMENTS   : "rows" and "cols" specifies the number of rows and columns  */
/*               in the matrix. "data_size" is the size of the data element  */
/*               stored in the matrix. Finally "destructor" is an iterator   */
/*               function called to destroy matrix elements when a matrix    */
/*               is deleted.                                                 */
/*                                                                           */
/* FUNCTION    : Creates a matrix with "rows" rows and "cols" columns        */
/*                                                                           */
/* RETURN VALUE: Pointer to the matrix. NULL on failure                      */
/* ************************************************************************* */
matrix_t *create_matrix(int rows,int cols,int data_size,
			void (*destructor)(void *))
{
  matrix_t *temp_matrix;
  int i;

  temp_matrix = (matrix_t *) IOPASM_MALLOC(sizeof(matrix_t));
  assert(temp_matrix != NULL);
  
  for (i=0;i<rows;i++) {
    temp_matrix->matrix_data = (void *) IOPASM_MALLOC(sizeof(void *) * rows * cols);
    assert(temp_matrix->matrix_data != NULL);
  }

  temp_matrix->n_rows = rows;
  temp_matrix->n_columns = cols;
  temp_matrix->data_size = data_size;
  temp_matrix->destructor_func = destructor;

  return temp_matrix;
}

/* ************************************************************************* */
/* PROTOTYPE   : int get_matrix_rows(matrix_t *a_matrix);                    */
/*                                                                           */
/* ARGUMENTS   : "a_matrix" points to a valid matrix                         */
/*                                                                           */
/* FUNCTION    : returns the number of rows in the matrix pointed to by      */
/*               "a_matrix"                                                  */
/*                                                                           */
/* RETURN VALUE: Number of rows                                              */
/* ************************************************************************* */
int get_matrix_rows(matrix_t *a_matrix) 
{
  assert(a_matrix != NULL);
  return a_matrix->n_rows;
}
/* ************************************************************************* */
/* PROTOTYPE   : int get_matrix_columns(matrix_t *a_matrix);                 */
/*                                                                           */
/* ARGUMENTS: "a_matrix" points to a valid matrix                            */
/*                                                                           */
/* FUNCTION    : Returns the number of columns in the matrix pointed to by   */
/*               "a_matrix"                                                  */
/*                                                                           */
/* RETURN VALUE: Number of columns                                           */
/* ************************************************************************* */
int get_matrix_columns(matrix_t *a_matrix) 
{
  assert(a_matrix != NULL);
  return a_matrix->n_columns;
}

/* ************************************************************************* */
/* PROTOTYPE   : void add_matrix_elem(matrix_t *a_matrix,int row,int col,    */
/*                                    void *elem);                           */
/*                                                                           */
/* ARGUMENTS   : "a_matrix" points to the active matrix. "row" specifies a   */
/*               row in the matrix and "col" a column. "elem" is the element */
/*               to be added                                                 */
/*                                                                           */
/* FUNCTION    : Inserts the element "elem" at "row"x"col"                   */
/*                                                                           */
/* RETURN VALUE: None                                                        */
/* ************************************************************************* */
void add_matrix_elem(matrix_t *a_matrix,int row,int col,void *elem) 
{
  
  int pos;
  assert(a_matrix != NULL);

 
  if ((a_matrix->n_rows < row) || (a_matrix->n_columns<col)) {
    ; /* We will not do anything here */
  }

  else {
    pos = (row-1) * a_matrix->n_columns + (col-1); 
    a_matrix->matrix_data[pos] = elem;
  }
}

/* ************************************************************************* */
/* PROTOTYPE   :  void *get_matrix_elem(matrix_t *a_matrix,int row,int col); */
/*                                                                           */
/* ARGUMENTS   : "a_matrix" points to the active matrix. "row" specifies a   */
/*               row in the matrix and "col" a column.                       */
/*                                                                           */
/* FUNCTION    : Returns the element at "row"x"col" found in the matrix      */ 
/*               "a_matrix"                                                  */
/*                                                                           */
/* RETURN VALUE: Pointer to the element. Null if empty.                      */
/* ************************************************************************* */
 void *get_matrix_elem(matrix_t *a_matrix,int row,int col)
{
  int pos;
  void *temp_elem;
  assert(a_matrix != NULL);
  
  if ((a_matrix->n_rows < row) || (a_matrix->n_columns<col)) {
    temp_elem = NULL;
  }

  else {
    pos = (row - 1)  * a_matrix->n_columns + (col-1);
    temp_elem = (void *) a_matrix->matrix_data[pos];
  }
  return temp_elem;
}


/* ************************************************************************* */
/* PROTOTYPE   : void clear_matrix_row(matrix_t *the_matrix,                 */
/*                                     int clear_this_row,int cols);         */
/*                                                                           */
/* ARGUMENTS   : "the_matrix" is a valid matrix. "clear_this_row" specifies  */
/*               the row to clear and "cols" are the number of columns       */
/*                                                                           */
/* FUNCTION    : Removes all the elements up to and including "cols" in the  */
/*               row specified by "clear_this_row"                           */
/*                                                                           */
/* RETURN VALUE: None                                                        */
/* ************************************************************************* */
void clear_matrix_row(matrix_t *the_matrix,int clear_this_row,int cols)
{
  int i=1;
  void *elem;
  assert(the_matrix != NULL);
  
  while((i < cols) && (get_matrix_elem(the_matrix,clear_this_row,i) != NULL)) {
    elem = get_matrix_elem(the_matrix,clear_this_row,i);

    /* delete the matrix element */
    the_matrix->destructor_func(elem);
    /* Set element to NULL to mark it empty */
    add_matrix_elem(the_matrix,clear_this_row,i,NULL);
    i++;
  } /* while(get_matrix_elem(the_matrix,clear_this_row,i) != NULL) */
}


/* ************************************************************************* */
/* PROTOTYPE   : void delete_matrix(matrix_t *the_matrix);                   */
/*                                                                           */
/* ARGUMENTS   : "a_matrix" points to a valid matrix                         */
/*                                                                           */
/* FUNCTION    : Deletes the matrix "the_matrix" and all elements stored     */
/*               in it.                                                      */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void delete_matrix(matrix_t *the_matrix)
{
  int row,col,pos;
  int i = 0 ;
  void *elem;
  assert(the_matrix != NULL);
  
  for (row=0;row<the_matrix->n_rows;row++) { 

    for (col=0;col<the_matrix->n_columns;col++) {
      pos = i * the_matrix->n_columns + (col-1); 
      elem = the_matrix->matrix_data[pos];
      if (elem != NULL) { 
	the_matrix->destructor_func(elem);
      } /* End of if (elem != NULL) */
    } /* End of for (col=0;col<the_matrix->n_columns;col++) */
  } /* End of for(row=0;row<the_matrix->n_rows;row++) */

  IOPASM_FREE(the_matrix);

} /* End of void delete_matrix(matrix_t *the_matrix) */
  
