/* ************************************************************************* */
/* meta_object.c : Meta object type used to tag objects before saving them   */
/*                                                                           */
/*    Copyright 2002-2006 Axis Communications AB.                            */
/*                                                                           */
/*                                                                           */
/*    This program is free software; you can redistribute it and/or modify   */
/*    it under the terms of the GNU General Public License as published by   */
/*    the Free Software Foundation; either version 2 of the License, or      */
/*    (at your option) any later version.                                    */
/*                                                                           */
/*    This program is distributed in the hope that it will be useful,        */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*    GNU General Public License for more details.                           */
/*                                                                           */
/*    You should have received a copy of the GNU General Public License      */
/*    along with this program; if not, write to the Free Software            */
/*    Foundation, Inc., 59 Temple Place - Suite 330,                         */
/*    Boston, MA 02111-1307, USA.                                            */
/*                                                                           */
/*    Send bug reports and inquiries to Johan.Jorgensen@axis.com             */
/*                                                                           */
/*                                                                           */
/*   Date      Description                                          Sign.    */
/*  =======================================================================  */
/*  2003-04-03 Module implemented for debug purposes                JohanJ   */
/*  2004-06-19 Now uses IOPASM_MALLOC and IOPASM_FREE macros         JohanJ  */
/* ************************************************************************* */
#include <assert.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdint.h>
#include "cg_fe.h"
#include "iopasm_err.h"
#include "iopasm_global.h"
#include "iopasm_types.h"
#include "meta_object.h"



/* ************************************************************************* */
/*                       TYPES DEFINED IN THIS MODULE                        */
/* ************************************************************************* */
struct __meta_object_t {
  uint32_t obj_type;
  uint32_t obj_size;
};


struct __object_header_t {
  uint32_t magic;
  uint32_t major_version;
  uint32_t minor_version;
  uint32_t arch;
  uint32_t n_objects;
};


/* ************************************************************************* */
/*                     DEFINITIONS LOCAL TO THIS MODULE                      */
/* ************************************************************************* */
#define MAGIC (0x224488AA)
#define OBJECT_HEADER_SIZE (sizeof(struct __object_header_t))


/* ************************************************************************* */
/*                              CONSTANT VARIABLES                           */
/* ************************************************************************* */



/* ************************************************************************* */
/*                    PROTOTYPES FOR STATIC FUNCTIONS                        */
/* ************************************************************************* */



/* ************************************************************************* */
/* PROTOTYPE   : meta_object_t *create_meta_object(int type,int size);       */
/*                                                                           */
/* ARGUMENTS   : "type" specifies the type of object to create, and "size"   */
/*               is the size of the object measured in bytes.                */
/*                                                                           */
/* FUNCTION    : Creates a meta object with specified type and size          */
/*                                                                           */
/* RETURN VALUE: Pointer to the newly created meta object. NULL on failure   */
/* ************************************************************************* */
meta_object_t *create_meta_object(int type,int size)
{
  meta_object_t *temp_object;
  
  temp_object = (meta_object_t *) IOPASM_MALLOC(sizeof(meta_object_t));
  assert(temp_object != NULL);

  temp_object->obj_type = htonl((uint32_t) type);
  temp_object->obj_size = htonl((uint32_t) size);
  return temp_object;
}


/* ************************************************************************* */
/* PROTOTYPE   : int get_object_size(meta_object_t *object);                 */
/*                                                                           */
/* ARGUMENTS   : "object" points to a valid meta object                      */
/*                                                                           */
/* FUNCTION    : returns the size of the descibed object (measured in bytes) */
/*                                                                           */
/* RETURN VALUE: Size of object in bytes.                                    */
/* ************************************************************************* */
int get_object_size(meta_object_t *object)
{
  assert (object != NULL);
  return (int) ntohl(object->obj_size);
}


/* ************************************************************************* */
/* PROTOTYPE   : int get_object_type(meta_object_t *object);                 */
/*                                                                           */
/* ARGUMENTS   : "object" points to a valid meta object                      */
/*                                                                           */
/* FUNCTION    : Returns the type of object represented by the "object"      */
/*                                                                           */
/* RETURN VALUE: Type of object.                                             */
/* ************************************************************************* */
int get_object_type(meta_object_t *object)
{
  assert (object != NULL);
  return (int) ntohl(object->obj_type);
}


/* ************************************************************************* */
/* PROTOTYPE   : void delete_meta_object(meta_object_t *object);             */
/*                                                                           */
/* ARGUMENTS   : "object" points to a valid object                           */
/*                                                                           */
/* FUNCTION    : Deletes a meta object and frees the allocated memory.       */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void delete_meta_object(meta_object_t *object)
{
  assert(object != NULL);
  IOPASM_FREE(object);
}


/* ************************************************************************* */
/* PROTOTYPE   : void save_meta_object(FILE *file,meta_object_t *object);    */
/*                                                                           */
/* ARGUMENTS   : */
/*                                                                           */
/* FUNCTION    : */
/*                                                                           */
/* RETURN VALUE: */
/* ************************************************************************* */
void save_meta_object(FILE *file,meta_object_t *object) 
{
  assert ((file != NULL) && (object != NULL));
  assert (fwrite(object,META_OBJECT_SIZE,1,file) == 1);
}

	  
  
/* ************************************************************************* */
/* PROTOTYPE   : */
/*                                                                           */
/* ARGUMENTS   : */
/*                                                                           */
/* FUNCTION    : */
/*                                                                           */
/* RETURN VALUE: */
/* ************************************************************************* */
meta_object_t *load_meta_object(FILE *file)
{
  meta_object_t *temp_object;
  
  assert (file != NULL);

  temp_object = (meta_object_t *) IOPASM_MALLOC(META_OBJECT_SIZE);
  assert(temp_object !=  NULL);

  assert (fread(temp_object,META_OBJECT_SIZE,1,file) == 1);
  return temp_object;
}


/* ************************************************************************* */
/* PROTOTYPE   : void save_object_header(FILE *out_file,cg_context_t *ctxt,  */
/*                                       unsigned int n_objects);            */
/*                                                                           */
/* ARGUMENTS   : "out_file" is a valid output file. "ctxt" is the current    */
/*               code-generation context and "n_objects" is the number of    */
/*               objects stored in the object file                           */
/*                                                                           */
/* FUNCTION    : Saves the object header to "out_file"                       */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void save_object_header(FILE *out_file,cg_context_t *ctxt,
			unsigned int n_objects)
{
  object_header_t the_header;
  


  assert ((out_file != NULL) && (ctxt != NULL));
  rewind(out_file);

  the_header.magic = htonl((uint32_t)MAGIC);
  the_header.major_version = htonl((uint32_t)IOPASM_MAJOR_VERSION);
  the_header.minor_version = htonl((uint32_t)IOPASM_MINOR_VERSION);
  the_header.arch = htonl((uint32_t)cg_get_arch(ctxt));
  the_header.n_objects = htonl((uint32_t)n_objects);

  assert(fwrite(&the_header,OBJECT_HEADER_SIZE,1,out_file) == 1);
  fseek(out_file,0L,SEEK_END);
}
  

/* ************************************************************************* */
/* PROTOTYPE   : unsigned int chk_object_header(char *file_name,             */
/*                                              FILE *in_file,               */
/*                                              cg_context_t *ctxt);         */
/*                                                                           */
/* ARGUMENTS   : "file_name" is a pointer to a valid filename and "in_file"  */
/*               points to the name of the input file. "ctxt" is the current */
/*               code generation context.                                    */
/*                                                                           */
/* FUNCTION    : Verify that the object file header is ok                    */
/*                                                                           */
/* RETURN VALUE: The number of object found in a valid object file.          */
/* ************************************************************************* */
unsigned int chk_object_header(char *file_name,FILE *in_file, 
			       cg_context_t *ctxt)
{
  object_header_t the_header;

  assert ((in_file != NULL) && (ctxt != NULL));
  assert (fread(&the_header,OBJECT_HEADER_SIZE,1,in_file) == 1);
  
  if ((ntohl(the_header.magic) != MAGIC) ||
      (ntohl(the_header.major_version) != IOPASM_MAJOR_VERSION) ||
      (ntohl(the_header.minor_version) != IOPASM_MINOR_VERSION)) {
    iopasm_error(ctxt,NULL,ERR_INCOMPATIBLE_OBJ_FILE,"Filename: %s",file_name);
  }

  if (ntohl(the_header.arch) != cg_get_arch(ctxt)) {
    iopasm_error(ctxt,NULL,ERR_ARCHITECTURE_MISMATCH,NULL);
  }

  return (unsigned int) ntohl(the_header.n_objects);
}


/* ************************************************************************* */
/* PROTOTYPE   : unsigned int dump_object_header(FILE *in_file,              */
/*                                               cg_context_t *ctxt);        */
/*                                                                           */
/* ARGUMENTS   : "in_file" is a valid input file. "ctxt" is the current code */
/*               generation context.                                         */
/*                                                                           */
/* FUNCTION    : Dumps the object header to stderr.                          */
/*                                                                           */
/* RETURN VALUE: Number of objects in the object file.                       */
/* ************************************************************************* */
unsigned int dump_object_header(FILE *in_file,cg_context_t *ctxt)
{
  object_header_t the_header;

  assert ((in_file != NULL) && (ctxt != NULL));
  assert (fread(&the_header,OBJECT_HEADER_SIZE,1,in_file) == 1);

  fprintf(stdout,"Object File Header :\n");

  fprintf(stdout,"     Magic Number   : %x (%s)\n",ntohl(the_header.magic),
	  (ntohl(the_header.magic) == MAGIC) ? "correct":"not correct");

  fprintf(stdout,"     IOPASM version : %d.%d (this version %d.%d)\n",
	  ntohl(the_header.major_version),ntohl(the_header.minor_version),
	  IOPASM_MAJOR_VERSION,IOPASM_MINOR_VERSION);
  fprintf(stdout,"     Objects in file: %d\n",
	  ntohl(the_header.n_objects));
  fprintf(stdout,"     Architecture   : %d (%s)\n",
	  ntohl(the_header.arch), 
	  (ntohl(the_header.arch) == cg_get_arch(ctxt)) ? "correct":"incorrect");

    if ((ntohl(the_header.magic) == MAGIC) && 
      (ntohl(the_header.major_version) == IOPASM_MAJOR_VERSION) &&
	(ntohl(the_header.minor_version) == IOPASM_MINOR_VERSION)) {
      fprintf(stdout,"     Version check is OK\n");
    }
    
    else {
      fprintf(stdout,"     Version check FAILED\n");
    }
   
  
    if (ntohl(the_header.arch) != cg_get_arch(ctxt)) {
      fprintf(stdout,"     Architecture check is OK\n");
    }

    else {
            fprintf(stdout,"     Architecture check FAILED\n");
    }
    fprintf(stdout,"--------------------------------------------------------\n");
  return the_header.n_objects;
}


/* ************************************************************************* */
/* PROTOTYPE   : char *load_obj_string(FILE *file,meta_object_t *object);    */
/*                                                                           */
/* ARGUMENTS   : "file" points to an open fiel. "object" points to a meta-   */
/*               object describing the file.                                 */
/*                                                                           */
/* FUNCTION    : Load a text/binary string with a length specified by the    */
/*               meta object.                                                */
/*                                                                           */
/* RETURN VALUE: Pointer to newly textual/binary string                      */
/* ************************************************************************* */
char *load_obj_string(FILE *file,meta_object_t *object)
{
  char *string;
  assert((file != NULL) && (object != NULL));
  assert((string = IOPASM_MALLOC(ntohl(object->obj_size) * sizeof(char))) != NULL);
  assert(fread(string,ntohl(object->obj_size),1,file) == 1);
  return string;
}


/* ************************************************************************* */
/* PROTOTYPE   : void dump_meta_object(meta_object_t *object);               */
/*                                                                           */
/* ARGUMENTS   : "object" points to a valid object                           */
/*                                                                           */
/* FUNCTION    : Dumps a meta object to stderr.                              */
/*                                                                           */
/* RETURN VALUE: None                                                        */
/* ************************************************************************* */
void dump_meta_object(meta_object_t *object) 
{
  assert(object != NULL);
  fprintf(stderr,"Dumping meta object: \n");
  fprintf(stderr,"     Object type   : %d\n",ntohl(object->obj_type));
  fprintf(stderr,"     Object size   : %d\n",ntohl(object->obj_size));
}

 
