/* ************************************************************************* */
/* optimizer_sets.c : Implementation of sets used by FSM optimizers          */
/*                                                                           */
/*    Copyright 2002-2006 Axis Communications AB.                            */
/*                                                                           */
/*                                                                           */
/*    This program is free software; you can redistribute it and/or modify   */
/*    it under the terms of the GNU General Public License as published by   */
/*    the Free Software Foundation; either version 2 of the License, or      */
/*    (at your option) any later version.                                    */
/*                                                                           */
/*    This program is distributed in the hope that it will be useful,        */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*    GNU General Public License for more details.                           */
/*                                                                           */
/*    You should have received a copy of the GNU General Public License      */
/*    along with this program; if not, write to the Free Software            */
/*    Foundation, Inc., 59 Temple Place - Suite 330,                         */
/*    Boston, MA 02111-1307, USA.                                            */
/*                                                                           */
/*                                                                           */
/*    Send bug reports and inquiries to Johan.Jorgensen@axis.com             */
/*                                                                           */
/*   Date      Description                                          Sign.    */
/*  =======================================================================  */
/*  2004-10-22 Module implementation began                          JohanJ   */
/*  2004-12-08 Module cleanup                                       JohanJ   */
/* ************************************************************************* */



/* ************************************************************************* */
/*                      INCLUDES USED BY THIS MODULE                         */
/* ************************************************************************* */
#include <assert.h>
#include <stdlib.h>
#include "cg_fe.h"
#include "dyn_array.h"
#include "p_list.h"
#include "iopasm_err.h"
#include "iopasm_global.h"
#include "iopasm_types.h"
#include "state_descriptor.h"
#include "optimizer_sets.h"

/* ************************************************************************* */
/*                       TYPES DEFINED IN THIS MODULE                        */
/* ************************************************************************* */
struct __optset_t__ {
  array_t *state_array;
  int code_size;
  int n_states;        /* number of states in set */
  int fragmentation;   /* Number of rows fragmented (fragmentation level) */
  int empty_max;       /* Max empty space in bytes */
  int n_rows;          /* Number of rows used in arch */
};


/* ************************************************************************* */
/*                     DEFINITIONS LOCAL TO THIS MODULE                      */
/* ************************************************************************* */



/* ************************************************************************* */
/*                              CONSTANT VARIABLES                           */
/* ************************************************************************* */



/* ************************************************************************* */
/*                    PROTOTYPES FOR STATIC FUNCTIONS                        */
/* ************************************************************************* */



/* ************************************************************************* */
/* PROTOTYPE   : optset_t *optset_create();                                  */
/*                                                                           */
/* ARGUMENTS   : None                                                        */
/*                                                                           */
/* FUNCTION    : Create an empty optimizer set.                              */
/*                                                                           */
/* RETURN VALUE: Pointer to a new set.                                       */
/* ************************************************************************* */
optset_t *optset_create()
{
  optset_t *temp_set;
  
  temp_set = (optset_t *) IOPASM_MALLOC(sizeof(optset_t));
  assert(temp_set != NULL);

  temp_set->state_array = create_array(cmp_states,delete_state);
  assert(temp_set->state_array != NULL);
  temp_set->code_size = 0;
  temp_set->n_states = 0;
  temp_set->fragmentation = 0;
  temp_set->empty_max = 0;
  temp_set->n_rows = 0;
  return temp_set;
}



/* ************************************************************************* */
/* PROTOTYPE   : void optset_delete(void *a_set);                            */
/*                                                                           */
/* ARGUMENTS   : "a_set" points to a valid set.                              */
/*                                                                           */
/* FUNCTION    : Delete a set and free allocated memory used by the and the  */
/*               the elements stored therein.                                */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void optset_delete(void *a_set)
{
  
  optset_t *tmp_set;

  assert(a_set != NULL);
  tmp_set = (optset_t *) a_set;
  delete_array(tmp_set->state_array);
  IOPASM_FREE(tmp_set);
}



/* ************************************************************************* */
/* PROTOTYPE   : int optset_add_state(optset_t *a_set,state_t *a_state);     */
/*                                                                           */
/* ARGUMENTS   : "a_set" points to a set and "a_state" points to a state     */
/*               which is to be added to the current set.                    */
/*                                                                           */
/* FUNCTION    : Add a state as an element to the set.                       */
/*                                                                           */
/* RETURN VALUE: number of elements in the set after addition.               */
/* ************************************************************************* */
int optset_add_state(optset_t *a_set,state_t *a_state)
{
  int i=0;
  int ret_val = -1;
  state_t *tmp_state;
  assert((a_set != NULL)&&(a_state != NULL));

  /* You can never add the same state twice to a set */
  while (i<a_set->n_states) {
    tmp_state = (state_t *) get_array_elem(a_set->state_array,i);
    if (!cmp_states(tmp_state,a_state)) {
      return 0;
    }
    i++;
  } /* End of while (i<a_set->n_states) */

  ret_val = add_array_elem(a_set->state_array,(void *) a_state);
  assert(ret_val > 0);
  a_set->n_states++;
  return a_set->n_states;

}


  
/* ************************************************************************* */
/* PROTOTYPE   : void optset_dump(optset_t *a_set);                          */
/*                                                                           */
/* ARGUMENTS   : "a_set" points to a valid set.                              */
/*                                                                           */
/* FUNCTION    : Dumps a set to stdout                                       */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void optset_dump(optset_t *a_set)
{
  int i=0;
  state_t *tmp_state;
  char *state_name;
  printf("Beginning of set dump ...\n ");
  printf("       Number of states in set: %d\n",a_set->n_states);
  printf("       Set size               : %d\n",a_set->code_size);
  printf("       Number of rows required: %d\n",a_set->n_rows);
  printf("       Longest empty row      : %d\n",a_set->empty_max);
  printf("-------------------------------------------\n");
  while(get_array_elem(a_set->state_array,i) != NULL) {
    tmp_state = (state_t *) get_array_elem(a_set->state_array,i);
    state_name = get_state_name(tmp_state);
    printf("     %s\n",state_name); 
    i++;
  }
  printf("End of set dump ...\n");
}



/* ************************************************************************* */
/* PROTOTYPE   : int optset_is_state_in_set(optset_t *a_set,                 */
/*                                          state_t *a_state);               */
/*                                                                           */
/* ARGUMENTS   : "a_set" is a valid set and "a_state is a valid state        */
/*                                                                           */
/* FUNCTION    : Check if "a_state" is already in the set "a_set"            */
/*                                                                           */
/* RETURN VALUE: 1 if the set is in the state 0 otherwise.                   */
/* ************************************************************************* */
int optset_is_state_in_set(optset_t *a_set,state_t *a_state)
{
  int i=0;
  state_t *tmp_state;
  assert((a_set != NULL) && (a_state != NULL));
  while(i<a_set->n_states) {
    tmp_state = (state_t *)get_array_elem(a_set->state_array,i);
    if (cmp_states(tmp_state,a_state)==0) {
      return 1;
    }
    i++;
  }
  return 0;
}



/* ************************************************************************* */
/* PROTOTYPE   : unsigned int optset_get_size(optset_t *a_set);              */
/*                                                                           */
/* ARGUMENTS   : "a_set" is a valid set.                                     */
/*                                                                           */
/* FUNCTION    : Calculate the code size of the set "a_set"                  */
/*                                                                           */
/* RETURN VALUE: The size of the code represented by the set a_set           */
/* ************************************************************************* */
unsigned int optset_get_size(optset_t *a_set)
{
  assert(a_set != NULL);
  return a_set->code_size;
}



/* ************************************************************************* */
/* PROTOTYPE   : int optset_get_state_count(optset_t *a_set);                */
/*                                                                           */
/* ARGUMENTS   : "a_set" points to a valid set.                              */ 
/*                                                                           */
/* FUNCTION    : Returns the number of states in the set                     */
/*                                                                           */
/* RETURN VALUE: Number of states in the set.                                */
/* ************************************************************************* */
int optset_get_state_count(optset_t *a_set)
{
  
  assert(a_set != NULL);
  return a_set->n_states;
}


/* ************************************************************************* */
/* PROTOTYPE   : state_t *optset_get_state_by_index(optset_t *a_set,         */
/*                                                  int which);              */
/*                                                                           */
/* ARGUMENTS   : "a_set" points to a set. "index" is the index of the        */
/*               element to return.                                          */
/*                                                                           */
/* FUNCTION    : Returns the state indexed by "index" i "a_set"              */
/*                                                                           */
/* RETURN VALUE: Pointer to a state.                                         */
/* ************************************************************************* */
state_t *optset_get_state_by_index(optset_t *a_set,int which)
{
  assert(a_set != NULL);
  return (state_t *) get_array_elem(a_set->state_array,which);
}



/* ************************************************************************* */
/* PROTOTYPE   : void optset_set_size(optset_t *a_set,unsigned int size);    */
/*                                                                           */
/* ARGUMENTS   : "a_set" is a pointer to a valid set and "size" is the       */
/*               calculated size represented in bytes.                       */
/*                                                                           */
/* FUNCTION    : Sets the code size of the set "a_set" to "size"             */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void optset_set_size(optset_t *a_set,unsigned int size)
{

  assert((a_set != NULL) && (size >= 0));
  a_set->code_size = size;
}



/* ************************************************************************* */
/* PROTOTYPE   : void opset_set_fragmentation(optset_t *a_set,               */
/*                                            int fragmentation);            */
/*                                                                           */
/* ARGUMENTS   : "a_set" points to a valid set and "fragmentation" is the    */
/*               fragmentation level of the current set.                     */
/*                                                                           */
/* FUNCTION    : Sets the fragmentation level associated with a given set    */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void opset_set_fragmentation(optset_t *a_set,int fragmentation)
{
  assert(a_set != NULL);
  a_set->fragmentation = fragmentation;
}


/* ************************************************************************* */
/* PROTOTYPE   : void optset_set_max_rows(optset_t *a_set,int rows);         */
/*                                                                           */
/* ARGUMENTS   : "a_set" points to a valid set and "rows" is the maximum     */
/*               number of rows in the matrix.                               */
/*                                                                           */
/* FUNCTION    : Set the number of maximum rows in an optimizer set to "rows"*/
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void optset_set_max_rows(optset_t *a_set,int rows)
{
  assert(a_set != NULL);
  a_set->n_rows = rows;
}


/* ************************************************************************* */
/* PROTOTYPE   : void optset_set_empty_max(optset_t *a_set,int empty);       */
/*                                                                           */
/* ARGUMENTS   : "a_set" points to a valid set. "empty" is the maximum       */
/*               number of free bytes.                                       */
/*                                                                           */
/* FUNCTION    : Sets the maximum size of free memory in the current set.    */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void optset_set_empty_max(optset_t *a_set,int empty)
{
  assert(a_set != NULL);
  a_set->empty_max = empty;
}



/* ************************************************************************* */
/* PROTOTYPE   : int optset_get_empty_max(optset_t *a_set);                  */
/*                                                                           */
/* ARGUMENTS   : "a_set" points to a valid set.                              */
/*                                                                           */
/* FUNCTION    : Returns the maximum empty space available in the set        */
/*                                                                           */
/* RETURN VALUE: # bytes in the maximum empty space.                         */
/* ************************************************************************* */
int optset_get_empty_max(optset_t *a_set)
{
  assert(a_set != NULL);
  return a_set->empty_max;
}


/* ************************************************************************* */
/* PROTOTYPE   : int optset_get_max_rows(optset_t *a_set);                   */
/*                                                                           */
/* ARGUMENTS   : "a_set" points to a valid set.                              */
/*                                                                           */
/* FUNCTION    : Returns the maximum number of rows required by the set.     */
/*                                                                           */
/* RETURN VALUE: Maximum number of rows required without optimization        */
/* ************************************************************************* */
int optset_get_max_rows(optset_t *a_set)
{
  assert(a_set != NULL);
  return a_set->n_rows;
}



/* ************************************************************************* */
/* PROTOTYPE   : int opset_get_fragmentation(optset_t *a_set);               */
/*                                                                           */
/* ARGUMENTS   : "a_set" points to a valid set.                              */
/*                                                                           */
/* FUNCTION    : returns the number of fragmented rows.                      */
/*                                                                           */
/* RETURN VALUE: # of fragmented rows in the set.                            */
/* ************************************************************************* */
int opset_get_fragmentation(optset_t *a_set)
{
  assert(a_set != NULL);
  return a_set->fragmentation;
}


/* ************************************************************************* */
/* PROTOTYPE   : optset_t *optset_union(optset_t *set_1,optset_t *set_2);    */
/*                                                                           */
/* ARGUMENTS   : "set_1" and "set_2" are pointers to valid sets.             */
/*                                                                           */
/* FUNCTION    : calculates the union of the two sets (S1 U S2)              */
/*                                                                           */
/* RETURN VALUE: returns a new set containing the union of set_1 and set_2   */
/*               (S1 U S2)                                                   */
/* ************************************************************************* */
optset_t *optset_union(optset_t *set_1,optset_t *set_2)
{
  optset_t *new_set;
  state_t *tmp_state;
  state_t *cpy_state;
  int i;

  new_set = optset_create();
  assert(new_set != NULL);

  i = 0;
  while (i<set_1->n_states) {
    tmp_state = optset_get_state_by_index(set_1,i);
    cpy_state = copy_state(tmp_state);
    optset_add_state(new_set,cpy_state);
    i++;
  }
  
  i = 0;
  while (i<set_2->n_states) {
    tmp_state = optset_get_state_by_index(set_2,i);
    cpy_state = copy_state(tmp_state);
    optset_add_state(new_set,cpy_state);
    i++;
  }

  new_set->code_size = 0;
  new_set->n_states = set_1->n_states + set_2->n_states;

  return new_set;
}


/* ************************************************************************* */
/* PROTOTYPE   : optset_t *optset_intersection(optset_t *set1,               */
/*                                             optset_t *set2);              */
/*                                                                           */
/* ARGUMENTS   : "set_1" and "set_2" are pointers to valid sets.             */
/*                                                                           */
/* FUNCTION    : calculates the intersection of "set_1" and "set_2"          */
/*                                                                           */
/* RETURN VALUE: a new set containing the intersection of "set_1" and "set_2"*/
/* ************************************************************************* */
optset_t *optset_intersection(optset_t *set1,optset_t *set2)
{
  return NULL;

}


/* ************************************************************************* */
/* PROTOTYPE   : state_t *optset_find_sized_state(optset_t *set1,            */
/*                                                int mx_size,               */
/*                                                int decrement);            */
/*                                                                           */
/* ARGUMENTS   : "set1" is a valid set. "mx_size" is the maximum size of the */
/*               wanted state measured in bytes. and "decrement" is the      */
/*               factor by which "mx_size" is decremented if a state of the  */
/*               initially specified size ("mx_size") is not found.          */
/*                                                                           */
/* FUNCTION    : Finds a state with size lessthan or equal to "mx_size"      */
/*                                                                           */
/* RETURN VALUE: pointer to a state with size <= "mx_size"                   */
/* ************************************************************************* */
state_t *optset_find_sized_state(optset_t *set1,int mx_size,int decrement) 
{
  int i;
  int this_size;
  state_t *a_state;
  assert(set1 != NULL);

  this_size = mx_size;
  while (this_size > 0) {

    for(i=0;i<set1->n_states;i++) {
      a_state = get_array_elem(set1->state_array,i);
     
      if (get_state_size(a_state) == this_size) {
	set_array_elem(set1->state_array,i,NULL);
	set1->n_states--;
	return a_state;
      } /* End of if (get_state_size(a_state) == mx_size) */
    } /* End of for(i=0;i<set1->n_states;i++) */
    this_size -= decrement;
  } /* End of while (this_size > 0) */
  return NULL;
}
