/* ************************************************************************* */
/* p_list.c: Stores the parsed representation of the program. This is the    */
/*           last step before code-generation.                               */
/*                                                                           */
/*    Copyright 2002-2006 Axis Communications AB.                            */
/*                                                                           */
/*                                                                           */
/*    This program is free software; you can redistribute it and/or modify   */
/*    it under the terms of the GNU General Public License as published by   */
/*    the Free Software Foundation; either version 2 of the License, or      */
/*    (at your option) any later version.                                    */
/*                                                                           */
/*    This program is distributed in the hope that it will be useful,        */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*    GNU General Public License for more details.                           */
/*                                                                           */
/*    You should have received a copy of the GNU General Public License      */
/*    along with this program; if not, write to the Free Software            */
/*    Foundation, Inc., 59 Temple Place - Suite 330,                         */
/*    Boston, MA 02111-1307, USA.                                            */
/*                                                                           */
/*                                                                           */
/*    Send bug reports and inquiries to Johan.Jorgensen@axis.com             */
/*                                                                           */
/* Short revision history:                                                   */
/*                                                                           */
/*   Date      Description                                           Sign.   */
/*  =======================================================================  */
/*  2002-10-04 Module implementation began                           JohanJ  */
/*  2002-11-11 Added get_arg_count() & changed get_first_arg()       JohanJ  */
/*  2003-01-15 Added add_inst_arg() function                         JohanJ  */
/*  2003-01-18 Added get_inst_line() and set_inst_line() functions   JohanJ  */
/*  2003-10-15 Added copy_arg() function                             JohanJ  */
/*  2004-06-11 Added copy_inst() function                            JohanJ  */
/*  2004-06-19 Now uses IOPASM_MALLOC and IOPASM_FREE macros         JohanJ  */
/*  2004-07-20 added get_list_count(...) function                    JohanJ  */
/* ************************************************************************* */



/* ************************************************************************* */
/*                      INCLUDES USED BY THIS MODULE                         */
/* ************************************************************************* */
#include <assert.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "debug.h"
#include "cg_fe.h"
#include "dyn_stack.h"
#include "p_list.h"
#include "meta_object.h"
#include "iopasm_global.h"



/* ************************************************************************* */
/*                       TYPES DEFINED IN THIS MODULE                        */
/* ************************************************************************* */
struct __arg_t {
  int arg_type; /* Type of argument, Immediate, Register etc */
  int arg_len;  /* Length of argument 0-255 */
  int raw_len;  /* Length of raw argument string */
  char *raw_arg; /* raw argument */
  union {
    char *value; /* Pointer to the textual argument */
    dyn_stack_t *arg_stack;
  } arg_val;
  arg_t *next; /* Next argument for this instruction */
};



struct __inst_t {
  char *inst;           /* Instruction text */
  int i_type;           /* Instruction type */
  unsigned int mem_loc; /* Memory Location */
  unsigned int line_no; /* Line number in source-file */
  int file_index;       /* Index of source file */
  int reloc_flags;      /* Right now only RELOCATABLE or ABSOLUTE */
  int mem_size;         /* Instruction size in bytes once built */
  arg_t *argument;      /* Pointer to first argument */
  arg_t *current_arg;   /* Current argument, Next one returned */
  unsigned char *code;
  inst_t *prev;
  inst_t *next;
};



struct __plist_t {
  unsigned int n_origins;
  unsigned int n_instructions;
  inst_t *org_list; /* List with pointers to special "origin" list" */ 
  inst_t *first_inst; /* Pointer to first element in instruction list */
  inst_t *current_inst; /* Current instruction */
};



struct __descriptor_t { /* This is only used for debugging purposes */
  int type_val;
  char *name;
};



typedef struct __descriptor_t descriptor_t;
    


/* ************************************************************************* */
/*                     DEFINITIONS LOCAL TO THIS MODULE                      */
/* ************************************************************************* */
#define INST_T_SIZE (sizeof (struct __inst_t))
#define ARG_T_SIZE (sizeof (struct __arg_t))



/* ************************************************************************* */
/*                              CONSTANT VARIABLES                           */
/* ************************************************************************* */
static const  descriptor_t inst_types[] = {
  { INST_UNKNOWN, "Unknown instruction type" },
  { INST_SEQ, "Sequential instruction" },
  { INST_ORIGIN, "Origin operator" },
  { INST_STATE_BEGIN, "Beginning of state (FSM)" },
  { INST_STATE_END, "End of state (FSM)" },
  { INST_STATE_XFER, "State transition instruction (FSM)" },
  { INST_LABEL, "Label definition" },
  { INST_TIMER, "Timer instruction (FSM)" },
  { INST_FSM_OPTIONS, "Options for state instruction (FSM)" },
  { INST_FSM_INP_FLAG, "FSM Input MuX value specifier" },
  { INST_FSM_OUT_FLAG, "FSM Output MuX value specifier" },
  { INST_FSM_DO_SEQ, "FSM Do-Sequential specifier" },
  { INST_FSM_GO_SEQ, "FSM Go-Sequential specifier" },
  { INST_FSM_INV_STB, "FSM invert strobe specifier" },
  { 0, NULL }};



static const descriptor_t arg_types[] = {
  { ARG_NO_ARG, "Unknown argument" },
  { ARG_IMMEDIATE, "Immediate value" },
  { ARG_NORMAL_REG, "Register" },
  { ARG_LABEL, "Relative value (Label)" },
  { ARG_FSM_IN, "FSM input specifier" },
  { ARG_FSM_OUT, "FSM output specifier" },
  { ARG_FSM_TRANSITION, "FSM transition specifier" },
  { ARG_FSM_MODIFIER, "FSM signal modifier flag" },
  { ARG_FSM_TIMER, "FSM Timer values" },
  { ARG_SPECIAL_REG, "Special Register" },
  { ARG_FSM_ONLY, "FSM Only construction" },
  { ARG_STACK, "Argument stack" },
  { 0, NULL }};



/* ************************************************************************* */
/*                    PROTOTYPES FOR STATIC FUNCTIONS                        */
/* ************************************************************************* */
static void delete_inst_list(inst_t *first_inst);
static void delete_arg_list(arg_t *first_arg);
static char *get_type_description(int type,const descriptor_t *desc);



/* ************************************************************************* */
/* PROTOTYPE   : static char *get_type_description(int type,                 */
/*                                                 const descriptor_t *desc);*/
/*                                                                           */
/* ARGUMENTS   : "type" is the value of the type descriptor to find in the   */
/*               descriptor array "desc"                                     */
/*                                                                           */
/* FUNCTION    : Prints a textual version of an internal type descriptor. The*/
/*               function is intended for debugging only.                    */
/*                                                                           */
/* RETURN VALUE: Pointer to textual description of "type"                    */
/* ************************************************************************* */
static char *get_type_description(int type,const descriptor_t *desc)
{
  int i = 0;
  while(desc[i].name != NULL) {
    if (desc[i].type_val == type) {
      return desc[i].name;
    }

    i++;
  }

  return "No descriptor found"; /* FULKOD! */
}



/* ************************************************************************* */
/* PROTOTYPE   : plist_t *create_plist();                                    */
/*                                                                           */
/* ARGUMENTS   : None                                                        */
/*                                                                           */
/* FUNCTION    : Create a new parser list                                    */
/*                                                                           */
/* RETURN VALUE: Pointer to a new parser list.                               */
/* ************************************************************************* */
plist_t *create_plist()
{
  plist_t *new_list;
  
  new_list = (plist_t *) IOPASM_MALLOC(sizeof(plist_t));

  if (new_list != NULL) {
    new_list->n_origins = 0;
    new_list->n_instructions = 0;
    new_list->org_list = NULL;
    new_list->first_inst = NULL;
    new_list->current_inst = NULL;
  }
  
  return new_list;

}



/* ************************************************************************* */
/* PROTOTYPE   : void set_next_inst(inst_t *where,inst_t *next);             */
/*                                                                           */
/* ARGUMENTS   : "where" and "next" points two to instructions.              */
/*                                                                           */
/* FUNCTION    : Inserts the instruction "next" after "where" in the linked  */
/*               list and updates internal pointers accordingly.             */
/*                                                                           */
/* RETURN VALUE: None                                                        */
/* ************************************************************************* */
void set_next_inst(inst_t *where,inst_t *next) 
{
  assert(where != NULL);
  where->next = next;
}



/* ************************************************************************* */
/* PROTOTYPE   : inst_t *copy_inst(inst_t *an_inst);                         */
/*                                                                           */
/* ARGUMENTS   : "an_inst" is a valid pointer to an instruction              */
/*                                                                           */
/* FUNCTION    : Creates a verbatim copy of the instruction pointed to by    */
/*               "an_inst".                                                  */
/*                                                                           */
/* RETURN VALUE: Returns a pointer to the copy of an instruction.            */
/* ************************************************************************* */
inst_t *copy_inst(inst_t *an_inst)
{
  inst_t *temp_inst;
  arg_t *temp_arg;
  arg_t *org_arg;

  assert(an_inst != NULL);
  
  temp_inst = (inst_t *) IOPASM_MALLOC(sizeof(inst_t));
  assert(temp_inst != NULL);
 
  if (an_inst->inst != NULL) {
    temp_inst->inst = (char *) strdup(an_inst->inst);
  }
  temp_inst->i_type = an_inst->i_type;
  temp_inst->mem_loc = 0;
  temp_inst->line_no = an_inst->line_no;
  temp_inst->file_index = an_inst->file_index;
  temp_inst->reloc_flags = an_inst->reloc_flags;
  temp_inst->mem_size = an_inst->mem_size;
  temp_inst->next = NULL;
  temp_inst->prev = NULL;
  temp_inst->argument = NULL;

  if (an_inst->code != NULL) {
    temp_inst->code = IOPASM_MALLOC(sizeof(char) * an_inst->mem_size);
    assert (temp_inst->code != NULL);
    memcpy(temp_inst->code,an_inst->code,an_inst->mem_size);
  }

  else {
    temp_inst->code = NULL;
  }

  temp_inst->current_arg = NULL;
  org_arg = an_inst->argument;
  
  while (org_arg != NULL) {
    temp_arg = copy_arg(org_arg);
    assert(temp_arg != NULL);

    add_inst_arg(temp_inst,temp_arg);
    org_arg = get_next_arg(org_arg);
  }
  
  return temp_inst;
}
  


/* ************************************************************************* */
/* PROTOTYPE   : int get_inst_count(plist_t *a_list,int inst_type);          */
/*                                                                           */
/* ARGUMENTS   : "a_list" is a valid pointer to a parse list. "inst_type" is */
/*               the type of instruction to be counted.                      */
/*                                                                           */
/* FUNCTION    : Counts the number of instructions specified by "inst_type"  */
/*                                                                           */
/* RETURN VALUE: The number of instructions equal to "inst_type" found in    */
/*               the parse list.                                             */
/* ************************************************************************* */
int get_inst_count(plist_t *a_list,int inst_type)
{
  inst_t *temp_inst;
  int i_count = 0;
  assert(a_list != NULL);
  temp_inst =get_first_inst(a_list);

  while(temp_inst != NULL) {
    if (get_inst_type(temp_inst) == inst_type) { 
      i_count++;
    } /* End of if (get_inst_type(temp_inst) ... */
    temp_inst = get_next_inst(temp_inst);
  } /* End of while(temp_inst != NULL) */

  return i_count;
}



/* ************************************************************************* */
/* PROTOTYPE   : void set_prev_inst(inst_t *where, inst_t *prev);            */
/*                                                                           */
/* ARGUMENTS   : "where" is a valid pointer to an instruction. "prev" points */
/*               to an instruction or NULL.                                  */
/*                                                                           */
/* FUNCTION    : The previous instruction as seen from "where" is set to     */
/*               "prev"                                                      */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void set_prev_inst(inst_t *where, inst_t *prev)
{
  assert(where != NULL);
  where->prev = prev;
}



/* ************************************************************************* */
/* PROTOTYPE   : inst_t *get_next_inst(inst_t *inst);                        */
/*                                                                           */
/* ARGUMENTS   : "inst" points to a valid instruction.                       */
/*                                                                           */
/* FUNCTION    : Returns the next instruction as seen from "inst"            */
/*                                                                           */
/* RETURN VALUE: Pointer to next instruction as seen from "inst"             */
/* ************************************************************************* */
inst_t *get_next_inst(inst_t *inst) { return (inst!=NULL) ? inst->next:NULL; }



/* ************************************************************************* */
/* PROTOTYPE   : inst_t *get_prev_inst(inst_t *inst);                        */
/*                                                                           */
/* ARGUMENTS   : "inst" points to a valid instruction (or NULL).             */
/*                                                                           */
/* FUNCTION    : Returns the previous instruction relative to "inst".        */
/*                                                                           */
/* RETURN VALUE: Pointer to the previous instruction or NULL.                */
/* ************************************************************************* */
inst_t *get_prev_inst(inst_t *inst) { return (inst!=NULL) ? inst->prev:NULL; }



/* ************************************************************************* */
/* PROTOTYPE   : inst_t *get_current_inst(plist_t *list);                    */
/*                                                                           */
/* ARGUMENTS   : "list" points to a valid parse list.                        */
/*                                                                           */
/* FUNCTION    : Returns the current instruction in the list.                */
/*                                                                           */
/* RETURN VALUE: pointer to the current instruction.                         */
/* ************************************************************************* */
inst_t *get_current_inst(plist_t *list) 
{
  if (list != NULL) {
    return list->current_inst; 
  }

  else {
    return NULL;
  }
}



/* ************************************************************************* */
/* PROTOTYPE   : inst_t *create_inst(char *inst,int i_type,                  */
/*                                   unsigned int addr,int reloc_info);      */
/*                                                                           */
/* ARGUMENTS   : "inst" points to textual part of instruction (if any).      */
/*               "i_type" is the instruction type and "addr" is the address  */
/*               if known. "reloc_info" determines wether the instruction is */
/*               relocatable or not.                                         */
/*                                                                           */
/* FUNCTION    : Creates an instruction (inst_t representation) of an        */
/*               instruction based on the input data.                        */
/*                                                                           */
/* RETURN VALUE: Pointer to a new instruction.                               */
/* ************************************************************************* */
inst_t *create_inst(char *inst,int i_type,unsigned int addr,int reloc_info)
{

  inst_t *new_inst;

  new_inst = (inst_t *) IOPASM_MALLOC(sizeof(inst_t));
  assert (new_inst != NULL);

  if (inst != NULL) {
    new_inst->inst = strdup(inst);
    assert (new_inst->inst != NULL);
  }
  
  else {
    new_inst->inst = NULL;
  }

  new_inst->i_type = i_type;
  new_inst->reloc_flags = (0 | reloc_info);
  new_inst->mem_loc = addr;
  new_inst->mem_size = 0;
  new_inst->line_no = 0;
  new_inst->file_index = 0;
  new_inst->argument = NULL;
  new_inst->current_arg = NULL;
  new_inst->prev = NULL;
  new_inst->next = NULL;
  new_inst->code = NULL;
  
  return new_inst;
}



/* ************************************************************************* */
/* PROTOTYPE   : arg_t *copy_arg(arg_t *arg);                                */
/*                                                                           */
/* ARGUMENTS   : "arg" points to a valid instruction.                        */
/*                                                                           */
/* FUNCTION    : Make a verbatim copy of an argument or argument stack       */
/*                                                                           */
/* RETURN VALUE: Pointer to the copy of the argument pointed to by "arg"     */
/* ************************************************************************* */
arg_t *copy_arg(arg_t *arg)
{
  arg_t *temp_arg;
  char *arg_string,*raw_string;
  assert (arg != NULL);

  temp_arg = (arg_t *) IOPASM_MALLOC(sizeof(arg_t));
  assert(temp_arg != NULL);

  temp_arg->arg_type = arg->arg_type;
  temp_arg->arg_len = arg->arg_len;
  temp_arg->raw_len = arg->raw_len;
  temp_arg->next = NULL;

  if (temp_arg->arg_type == ARG_STACK) {
    temp_arg->arg_val.arg_stack = arg->arg_val.arg_stack;
  }

  else {
    arg_string = (char *) strdup(arg->arg_val.value);
    assert(arg_string != NULL);
    temp_arg->arg_val.value = arg_string;
  }
  
  raw_string = (char *) strdup(arg->raw_arg);
  assert(raw_string != NULL);

  temp_arg->raw_arg = raw_string;
  
  return temp_arg;
}



/* ************************************************************************* */
/* PROTOTYPE   : inst_t *move_next_inst(inst_t *inst,inst_t *new_next);      */
/*                                                                           */
/* ARGUMENTS   : "inst" points to a valid instruction. "new_next" is a new   */
/*               instruction.                                                */
/*                                                                           */
/* FUNCTION    : next instruction as seen from "inst" will be "new_next"     */
/*                                                                           */
/* RETURN VALUE: Former value associated with "inst->next"                   */
/* ************************************************************************* */
inst_t *move_next_inst(inst_t *inst,inst_t *new_next)
{
  inst_t *old_next;
  
  if ((inst == NULL)||(new_next == NULL)) { 
    old_next = NULL;
  }

  else {
    old_next=((inst_t *) inst->next);
    inst->next=((struct __inst_t *) new_next); 
  }

  return old_next;
  
}



/* ************************************************************************* */
/* PROTOTYPE   : inst_t *move_prev_inst(inst_t *inst,inst_t *new_prev);      */
/*                                                                           */
/* ARGUMENTS   : "inst" points to an instruction. "new_prev" points to the   */
/*               instruction which will be placed previous to "inst".        */
/*                                                                           */
/* FUNCTION    : inserts "new_prev" prior to "inst" in the linked list.      */
/*                                                                           */
/* RETURN VALUE: Old previous instruction.                                   */
/* ************************************************************************* */
inst_t *move_prev_inst(inst_t *inst,inst_t *new_prev)
{
  inst_t *old_prev;
  
  if ((inst == NULL)||(new_prev == NULL)) { 
    old_prev = NULL;
  }

  else {
    old_prev=((inst_t *) inst->prev);
    inst->next=((struct __inst_t *) new_prev); 
  }

  return old_prev;
  
}



/* ************************************************************************* */
/* PROTOTYPE   : arg_t *create_arg(char *arg_val,int arg_len,int arg_type);  */
/*                                                                           */
/* ARGUMENTS   : "arg_val" is a textual representation of the argument and   */
/*               "arg_len is the length of "arg_val". "arg_type" is the type */
/*               of argument (literal, label etc).                           */
/*                                                                           */
/* FUNCTION    : Allocates memory and creates an argument data structure     */
/*               based on the input.                                         */
/*                                                                           */
/* RETURN VALUE: Pointer to the argument structure.                          */
/* ************************************************************************* */
arg_t *create_arg(char *arg_val,int arg_len,int arg_type)
{
  arg_t *new_arg;
  new_arg = (arg_t *) malloc (sizeof(arg_t));
  assert(new_arg != NULL);

  if ((arg_val != NULL) && (arg_len > 0)) {
    new_arg->raw_arg = (char *) strdup(arg_val);
    assert(new_arg->raw_arg != NULL);
    new_arg->arg_val.value = (char *) strdup(arg_val);
    assert(new_arg->arg_val.value != NULL);
  }
    
  new_arg->raw_len = arg_len;
  new_arg->arg_len = arg_len;
  new_arg->arg_type = arg_type;
  new_arg->next = NULL;
  
  return new_arg;
}



/* ************************************************************************* */
/* PROTOTYPE   : char *get_raw_arg(arg_t *arg);                              */
/*                                                                           */
/* ARGUMENTS   : arg points to an argument                                   */
/*                                                                           */
/* FUNCTION    : Returns the raw (unparsed) argument. Used for listing files */
/*                                                                           */
/* RETURN VALUE: character pointer to the raw argument.                      */
/* ************************************************************************* */
char *get_raw_arg(arg_t *arg)
{
  assert(arg != NULL);
  return arg->raw_arg;
}



/* ************************************************************************* */
/* PROTOTYPE   : void set_raw_arg(arg_t *arg,char *arg_val);                 */
/*                                                                           */
/* ARGUMENTS   : "arg" is a valid argument. "arg_val" is the new raw value.  */
/*               By raw we mean an unparsed argument such as "3 + 4 | 2.     */
/*                                                                           */
/* FUNCTION    : Allocates memory for and sets the raw argument in an arg_t  */
/*               to "arg_val".                                               */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void set_raw_arg(arg_t *arg,char *arg_val)
{
  assert((arg != NULL));
  
  if (arg->raw_arg != NULL) {
    IOPASM_FREE(arg->raw_arg);
  }

  arg->raw_arg = (char *)strdup(arg_val);
  assert(arg->raw_arg != NULL);
}


  
/* ************************************************************************* */
/* PROTOTYPE   : int add_raw_arg(arg_t *arg,char *string);                   */
/*                                                                           */
/* ARGUMENTS   : "arg" points to a valid argument and "string" is the string */
/*               to be added to the raw (unparsed) part of the argument.     */
/*                                                                           */
/* FUNCTION    : Adds the token to the unparsed part of the current argument */
/*               this is used for printing.                                  */
/*                                                                           */
/* RETURN VALUE: The length of the current raw argument measured in bytes.   */
/* ************************************************************************* */
int add_raw_arg(arg_t *arg,char *string)
{
  char *tmp_buffer;
  int old_len;

  assert((arg != NULL) && (string != NULL));

  old_len = arg->raw_len;
  arg->raw_len = old_len + strlen(string);

  tmp_buffer = (char *) IOPASM_MALLOC(sizeof(char) * (arg->raw_len+1));
  assert(tmp_buffer != NULL);
  
  memset(tmp_buffer,0,arg->raw_len+1);
  
  if (arg->raw_arg != NULL) {
    strcpy(tmp_buffer,arg->raw_arg);
    strcat(tmp_buffer,string);
    IOPASM_FREE(arg->raw_arg);
  }
  
  else {
    strcpy(tmp_buffer,string);
  }

  arg->raw_arg = tmp_buffer;
  return old_len;
}



/* ************************************************************************* */
/* PROTOTYPE   : arg_t *create_arg_stack(dyn_stack_t *a_stack);              */
/*                                                                           */
/* ARGUMENTS   : "a_stack" points to a valid stack.                          */
/*                                                                           */
/* FUNCTION    : Creates an argument with a stack of arguments stored in     */
/*               "a_stack".                                                  */
/*                                                                           */
/* RETURN VALUE: Pointer to a valid argument or NULL on failure.             */
/* ************************************************************************* */
arg_t *create_arg_stack(dyn_stack_t *a_stack)
{
  arg_t *new_arg;

  assert(a_stack != NULL);

  new_arg = (arg_t *) malloc (sizeof(arg_t));
  assert(new_arg != NULL);

  new_arg->raw_len = 0;
  new_arg->raw_arg = NULL;

  new_arg->arg_len = 0;
  new_arg->arg_val.arg_stack = a_stack;
  new_arg->arg_type = ARG_STACK;
  new_arg->next = NULL;
  return new_arg;
}



/* ************************************************************************* */
/* PROTOTYPE   : int add_inst(plist_t *list, inst_t *inst);                  */
/*                                                                           */
/* ARGUMENTS   : "list" is a valid instruction list. "inst" is the           */
/*               instruction we need to add.                                 */
/*                                                                           */
/* FUNCTION    : Adds the instruction pointed to by "inst" to the parse list */
/*               pointed to by "list"                                        */
/*                                                                           */
/* RETURN VALUE: Always returns 1. (should be void)                          */
/* ************************************************************************* */
int add_inst(plist_t *list, inst_t *inst)
{
  if ((list == NULL) || (inst == NULL)) return -1;

  if (list->current_inst==NULL) { /* First instruction ever to add */
    list->current_inst = inst;
    list->first_inst = inst;
    list->n_instructions++;
  }
  
  else {
    list->current_inst->next = inst;
    inst->prev = list->current_inst;
    list->n_instructions++;
    list->current_inst = inst;
  }

  return 1;    
}



/* ************************************************************************* */
/* PROTOTYPE   : int add_origin(plist_t *list, inst_t *inst);                */
/*                                                                           */
/* ARGUMENTS   : "list" is a valid list and "inst" is the instruction point- */
/*               ing to the origin instrucion to add.                        */
/*                                                                           */
/* FUNCTION    : Add an origin instruction to the current list.              */
/*                                                                           */
/* RETURN VALUE: Always returns 1. (should be void)                          */
/* ************************************************************************* */
int add_origin(plist_t *list, inst_t *inst)
{
  inst_t *add_here;

  if ((list == NULL) || (inst == NULL)) return -1;

  if (list->org_list == NULL)
    list->org_list = inst;

  else {
    add_here = list->org_list;
    while (add_here->next != NULL) add_here = add_here->next;
    
    add_here->next = inst;
    inst->prev = add_here;
  }

  return 1;
}



/* ************************************************************************* */
/* PROTOTYPE   : arg_t *get_current_arg(inst_t *inst);                       */
/*                                                                           */
/* ARGUMENTS   : "inst" points to a valid instruction                        */
/*                                                                           */
/* FUNCTION    : Returns the last "current" argument in instruction          */
/*                                                                           */
/* RETURN VALUE: Pointer to the current argument associated with "inst"      */
/* ************************************************************************* */
arg_t *get_current_arg(inst_t *inst)
{
  assert(inst != NULL);
  return inst->current_arg;
}



/* ************************************************************************* */
/* PROTOTYPE   : int add_arg(plist_t *list,arg_t *arg);                      */
/*                                                                           */
/* ARGUMENTS   : "list" points to a valid list. "arg" is a valid argument    */
/*                                                                           */
/* FUNCTION    : Adds an argument to the current instruction in the list     */
/*               specified by "list".                                        */
/*                                                                           */
/* RETURN VALUE: Always returns 1 (should be void).                          */
/* ************************************************************************* */
int add_arg(plist_t *list,arg_t *arg)
{

  if ((list == NULL) || (arg == NULL)) return -1;
  if (list->current_inst == NULL) return -1; /* No instruction added */
  
  if (list->current_inst->current_arg != NULL) {
    list->current_inst->current_arg->next = arg;
    list->current_inst->current_arg = arg;
  }
  
  else {
    list->current_inst->current_arg = arg;
    list->current_inst->argument = arg;
  }
  
  return 1; 
}



/* ************************************************************************* */
/* PROTOTYPE   : int add_inst_arg(inst_t *inst,arg_t *arg);                  */
/*                                                                           */
/* ARGUMENTS   : "inst" is a valid instruction, and "arg" is a valid ar-     */
/*               gument.                                                     */
/*                                                                           */
/* FUNCTION    : adds an argument to the instruction pointed to by "inst"    */
/*                                                                           */
/* RETURN VALUE: Always returns 1 (should be void)                           */
/* ************************************************************************* */
int add_inst_arg(inst_t *inst,arg_t *arg) 
{
  arg_t *temp_arg;

  assert((inst != NULL) && (arg != NULL));

  temp_arg = inst->argument;


  if (temp_arg == NULL) {
    inst->argument = arg;
  }


  else {
    while(temp_arg->next != NULL) {
      temp_arg = temp_arg->next;
    }
    
    temp_arg->next = arg;
  }

  return 1;

}



/* ************************************************************************* */
/* PROTOTYPE   : arg_t *get_first_arg(inst_t *inst);                         */
/*                                                                           */
/* ARGUMENTS   : "inst" points to a valid instruction.                       */
/*                                                                           */
/* FUNCTION    : Returns the first argument of an instruction.               */
/*                                                                           */
/* RETURN VALUE: Pointer to the first argument of the instruction, NULL if   */
/*               no argument has been found.                                 */
/* ************************************************************************* */
arg_t *get_first_arg(inst_t *inst)
{
  if (inst == NULL) return NULL;
  return inst->argument;
}



/* ************************************************************************* */
/* PROTOTYPE   : int get_arg_count(inst_t *inst);                            */
/*                                                                           */
/* ARGUMENTS   : "inst" is a pointer to a valid instruction.                 */
/*                                                                           */
/* FUNCTION    : Return the number of arguments associated with  the         */
/*               instructions pointed to by "inst"                           */
/*                                                                           */
/* RETURN VALUE: Number of arguments.                                        */
/* ************************************************************************* */
int get_arg_count(inst_t *inst)
{
  int t_count = 0;
  arg_t *next_arg;
  if (inst == NULL) { 
    t_count = 0;
  }

  else if (inst->argument == NULL) {
    t_count = 0;
  }

  else {
    t_count = 1;
    next_arg = inst->argument;
      
      while(next_arg->next != NULL) { 
	t_count++;
	next_arg = next_arg->next;
      }
  }

  return t_count;
}



/* ************************************************************************* */
/* PROTOTYPE   : arg_t *get_next_arg(arg_t *arg);                            */
/*                                                                           */
/* ARGUMENTS   : "arg" points to a valid argument.                           */
/*                                                                           */
/* FUNCTION    : Get the next argument relative to "arg"                     */
/*                                                                           */
/* RETURN VALUE: Returns a pointer to the next argument (trailing) the       */
/*               argument pointed to by "arg".                               */
/* ************************************************************************* */
arg_t *get_next_arg(arg_t *arg)
{
  if (arg->next == NULL) return NULL;
  else return arg->next;
  
}



/* ************************************************************************* */
/* PROTOTYPE   : int set_reloc_flag(inst_t *inst,int flag);                  */
/*                                                                           */
/* ARGUMENTS   : "inst" points to a valid instruction. "flag" is the value   */
/*               for the relocation flags (defined in p_list.h).             */
/*                                                                           */
/* FUNCTION    : Sets/changes the relocation flag of an instruction to that  */
/*               of "flag".                                                  */
/*                                                                           */
/* RETURN VALUE: Old flag value.                                             */
/* ************************************************************************* */
int set_reloc_flag(inst_t *inst,int flag)
{
  int old_flag;
  assert(inst != NULL);
  old_flag = inst->reloc_flags;
  inst->reloc_flags = inst->reloc_flags | flag;
  
  return old_flag;
}



/* ************************************************************************* */
/* PROTOTYPE   : int get_reloc_flag(inst_t *inst);                           */
/*                                                                           */
/* ARGUMENTS   : "inst" is a valid pointer to an instrucion.                 */
/*                                                                           */
/* FUNCTION    : Returns the relocation flag for the current instruction     */
/*                                                                           */
/* RETURN VALUE: Relocation flags. Values should be interpreted as specified */
/*               in (p_list.h)                                               */
/* ************************************************************************* */
int get_reloc_flag(inst_t *inst)
{
  assert(inst != NULL);
  return inst->reloc_flags;
}



/* ************************************************************************* */
/* PROTOTYPE   : int clear_reloc_flag(inst_t *inst,int flag);                */
/*                                                                           */
/* ARGUMENTS   : "inst" is a valid pointer to an instruction. "flag" is the  */
/*               flag to clear.                                              */
/*                                                                           */
/* FUNCTION    : Clears the specified flag in the instruction pointed to by  */
/*               "inst"                                                      */
/*                                                                           */
/* RETURN VALUE: Returns the old value for relocation flags.                 */
/* ************************************************************************* */
int clear_reloc_flag(inst_t *inst,int flag)
{
  int old_flag;
  
  assert(inst != NULL);
  old_flag = inst->reloc_flags;
  inst->reloc_flags = inst->reloc_flags & ~flag;
  return old_flag;
}



/* ************************************************************************* */
/* PROTOTYPE   : int check_reloc_flag(inst_t *inst,int flag);                */
/*                                                                           */
/* ARGUMENTS   : "inst" points to a valid instruction. "flag" is the flag    */
/*               which needs to be checked.                                  */
/*                                                                           */
/* FUNCTION    : check the value of the relocation flags of the instruction  */
/*                                                                           */
/* RETURN VALUE: Returns non-zero value if the flag is set. Zero otherwise.  */
/* ************************************************************************* */
int check_reloc_flag(inst_t *inst,int flag)
{
  assert(inst != NULL);
   return (inst->reloc_flags & flag);
}



/* ************************************************************************* */
/* PROTOTYPE   : char *get_inst(inst_t *inst);                               */
/*                                                                           */
/* ARGUMENTS   : "inst" points to a valid instruction.                       */
/*                                                                           */
/* FUNCTION    : return the textual part of the instruction.                 */
/*                                                                           */
/* RETURN VALUE: Pointer to the textual part of the instruction (e.g. move)  */
/* ************************************************************************* */
char *get_inst(inst_t *inst)
{
  char *inst_name;
  if (inst == NULL) { 
    inst_name = NULL; 
  }

  else {
    inst_name = inst->inst;
  }

  return inst_name;
}



/* ************************************************************************* */
/* PROTOTYPE   : inst_t *get_first_inst(plist_t *list);                      */
/*                                                                           */
/* ARGUMENTS   : "list" points to a valid parse list                         */
/*                                                                           */
/* FUNCTION    : returns the first instruction stored in the parse list.     */
/*                                                                           */
/* RETURN VALUE: Pointer to the first instruction in the parse list.         */
/* ************************************************************************* */
inst_t *get_first_inst(plist_t *list)
{
  if (list == NULL) return NULL;
  else return list->first_inst;
}



/* ************************************************************************* */
/* PROTOTYPE   : unsigned int get_inst_addr(inst_t *inst);                   */
/*                                                                           */
/* ARGUMENTS   : "inst" points to a valid instruction                        */
/*                                                                           */
/* FUNCTION    : Returns the current address of the instruction              */
/*                                                                           */
/* RETURN VALUE: Value representing the address in memory                    */
/* ************************************************************************* */
unsigned int get_inst_addr(inst_t *inst)
{
  if (inst == NULL) return 0; 
  else return inst->mem_loc;
}



/* ************************************************************************* */
/* PROTOTYPE   : void set_inst_size(inst_t *inst,int size);                  */
/*                                                                           */
/* ARGUMENTS   : "inst" points to a valid instuction "size" is the size of   */
/*               the instruction measured in bytes.                          */
/*                                                                           */
/* FUNCTION    : set the size of the current instruction (how much memory it */
/*               occupies. This is normally always measured in bytes.        */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void set_inst_size(inst_t *inst,int size) 
{  
  if (inst != NULL) { inst->mem_size = size; }
}



/* ************************************************************************* */
/* PROTOTYPE   : int get_inst_type(inst_t *inst);                            */
/*                                                                           */
/* ARGUMENTS   : "inst" is a pointer to a valid instruction.                 */
/*                                                                           */
/* FUNCTION    : Returns the type of instruction (see plist.h for details)   */
/*                                                                           */
/* RETURN VALUE: Type of instruction.                                        */
/* ************************************************************************* */
int get_inst_type(inst_t *inst)
{ 
  return (inst != NULL) ? inst->i_type:INST_UNKNOWN;
}



/* ************************************************************************* */
/* PROTOTYPE   : int get_inst_size(inst_t *inst);                            */
/*                                                                           */
/* ARGUMENTS   : "inst" is a valid pointer to an instruction.                */
/*                                                                           */
/* FUNCTION    : Get the size of "inst" measured in bytes.                   */
/*                                                                           */
/* RETURN VALUE: Size of the instruction measured in bytes.                  */
/* ************************************************************************* */
int get_inst_size(inst_t *inst) { return (inst != NULL) ? inst->mem_size: 0; }



/* ************************************************************************* */
/* PROTOTYPE   : void set_inst_addr(inst_t *inst,unsigned int addr);         */
/*                                                                           */
/* ARGUMENTS   : "inst" is a valid pointer to an instruction. "addr" is an   */
/*               address.                                                    */
/*                                                                           */
/* FUNCTION    : Sets the address of the current instrction. Measured in     */
/*               measured in bytes.                                          */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void set_inst_addr(inst_t *inst,unsigned int addr)
{

  if (inst != NULL) {
    inst->mem_loc = addr;
  }
  
}



/* ************************************************************************* */
/* PROTOTYPE   : int get_arg_type(arg_t *arg);                               */
/*                                                                           */
/* ARGUMENTS   : "arg" is a valid pointer to an argument.                    */
/*                                                                           */
/* FUNCTION    : Returns the type of argument (integer, label etc). Types    */
/*               are defined in p_list.h.                                    */
/*                                                                           */
/* RETURN VALUE: type of argument. Should be interpreted according to values */
/*               defined in plist.h                                          */ 
/* ************************************************************************* */
int get_arg_type(arg_t *arg) 
{ 
  return (arg==NULL) ? ARG_NO_ARG: arg->arg_type; 
}



/* ************************************************************************* */
/* PROTOTYPE   : void set_arg_type(arg_t *arg,int arg_type);                 */
/*                                                                           */
/* ARGUMENTS   : "arg" is a valid pointer to an argument. "arg_type" is an   */
/*               argment type as defined in p_list.h                         */
/*                                                                           */
/* FUNCTION    : Sets the argument type of "arg" to "arg_type".              */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void set_arg_type(arg_t *arg,int arg_type) 
{ 
  if (arg != NULL) arg->arg_type = arg_type; 
  else return;
}



/* ************************************************************************* */
/* PROTOTYPE   : char get_arg_len(arg_t *arg);                               */
/*                                                                           */
/* ARGUMENTS   : "arg" is a valid pointer to an argument.                    */
/*                                                                           */
/* FUNCTION    : Return the length of the textual part of the argument       */
/*                                                                           */
/* RETURN VALUE: Legnth of textual part of argument.                         */
/* ************************************************************************* */
char get_arg_len(arg_t *arg) 
{ 
  assert(arg != NULL);
  return arg->arg_len;
}



/* ************************************************************************* */
/* PROTOTYPE   : char *get_arg_val(arg_t *arg);                              */
/*                                                                           */
/* ARGUMENTS   : "arg" is a valid pointer to an argument.                    */
/*                                                                           */
/* FUNCTION    : Returns the textual part of an argument.                    */
/*                                                                           */
/* RETURN VALUE: Pointer to the text of the argument.                        */
/* ************************************************************************* */
char *get_arg_val(arg_t *arg) 
{ 
  assert (arg != NULL);
  return arg->arg_val.value;
};



/* ************************************************************************* */
/* PROTOTYPE   : dyn_stack_t *get_arg_stack(arg_t *arg);                     */
/*                                                                           */
/* ARGUMENTS   : "arg" is a valid pointer to an argument.                    */
/*                                                                           */
/* FUNCTION    : Retrieve the stack of arguments stored in an argument.      */
/*                                                                           */
/* RETURN VALUE: Pointer to a stack containing arguments and operands.       */
/* ************************************************************************* */
dyn_stack_t *get_arg_stack(arg_t *arg)
{
  assert(arg != NULL);
  return arg->arg_val.arg_stack;
}



/* ************************************************************************* */
/* PROTOTYPE   : void set_arg_val(arg_t *arg,char *arg_val,int arg_len);     */
/*                                                                           */
/* ARGUMENTS   : "arg" is a valid pointer to an argument. "arg_val" is a     */
/*               valid pointer to a string, and "arg_len" is the length of   */
/*               the string "arg_val".                                       */
/*                                                                           */
/* FUNCTION    : Sets the argument value of "arg" to that of "arg_val"       */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void set_arg_val(arg_t *arg,char *arg_val,int arg_len)
{
  if (arg != NULL) {
    arg->arg_val.value = arg_val;
    arg->arg_len = arg_len;
  }

  else return;
}


/* ************************************************************************* */
/* PROTOTYPE   : unsigned int get_inst_line(inst_t *inst);                   */
/*                                                                           */
/* ARGUMENTS   : "inst" is a valid pointer to an instruction.                */
/*                                                                           */
/* FUNCTION    : Return the line number of the instruction                   */
/*                                                                           */
/* RETURN VALUE: The line number.                                            */
/* ************************************************************************* */
unsigned int get_inst_line(inst_t *inst) 
{
  assert(inst != NULL);
  return inst->line_no;
}



/* ************************************************************************* */
/* PROTOTYPE   : void set_inst_line(inst_t *inst,unsigned int line_no);      */
/*                                                                           */
/* ARGUMENTS   : "inst" is a valid instruction pointer. "line_no" is the     */
/*               line number where the instruction was found.                */
/*                                                                           */
/* FUNCTION    : Sets the line number for the instruction "inst"             */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void set_inst_line(inst_t *inst,unsigned int line_no)
{
  assert(inst != NULL);
  inst->line_no = line_no;
}



/* ************************************************************************* */
/* PROTOTYPE   : void set_inst_code(inst_t *inst,unsigned char *code,        */
/*                                  int code_len);                           */
/*                                                                           */
/* ARGUMENTS   : "inst" points to a valid instruction, "code" is points to   */
/*               the generated code. and "code_len" is the length of the     */
/*               generated code measured in bytes.                           */
/*                                                                           */
/* FUNCTION    : Assigns the genereted code for the current instruction.     */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void set_inst_code(inst_t *inst,unsigned char *code,int code_len)
{
  unsigned char *new_area;
  assert(inst != NULL);

  if (inst->code == NULL) {
    new_area = (unsigned char *) IOPASM_MALLOC((sizeof(unsigned char) * code_len));
    assert (new_area != NULL);
    inst->code = new_area;
  }

  memcpy(inst->code,code,code_len);
}



/* ************************************************************************* */
/* PROTOTYPE   : unsigned char *get_inst_code(inst_t *inst);                 */
/*                                                                           */
/* ARGUMENTS   : "inst" points to a valid instruction                        */
/*                                                                           */
/* FUNCTION    : Return the binary code generated for the instruction.       */
/*                                                                           */
/* RETURN VALUE: Pointer to the instruction code generated for the inst.     */
/* ************************************************************************* */
unsigned char *get_inst_code(inst_t *inst)
{
  
  assert (inst != NULL);
  return inst->code;
}



/* ************************************************************************* */
/* PROTOTYPE   : static void delete_arg_list(arg_t *first_arg);              */
/*                                                                           */
/* ARGUMENTS   : "first_arg" is a valid pointer to an argument.              */
/*                                                                           */
/* FUNCTION    : Deletes an argument list associated with an instruction,    */
/*               and frees all allocated memory.                             */
/*                                                                           */
/* RETURN VALUE: none.                                                       */
/* ************************************************************************* */
static void delete_arg_list(arg_t *first_arg)
{
  
  if (first_arg->next == NULL) {
    IOPASM_FREE(first_arg->arg_val.value);
    first_arg->next = NULL;
    IOPASM_FREE(first_arg);
  }

  else {
    delete_arg_list(first_arg->next);
  }
}


 
/* ************************************************************************* */
/* PROTOTYPE   : static void delete_inst_list(inst_t *first_inst);           */
/*                                                                           */
/* ARGUMENTS   : "first_inst" is a pointer to an instruction                 */
/*                                                                           */
/* FUNCTION    : Deletes "first_inst and all subsequent instructions.        */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
static void delete_inst_list(inst_t *first_inst)
{

  while(first_inst->next != NULL) { first_inst = first_inst->next; }

  while(first_inst->prev != NULL) {
    first_inst = first_inst->prev;

    if (first_inst->next->argument != NULL) {
      delete_arg_list(first_inst->next->argument);
    }

    IOPASM_FREE(first_inst->next->inst);
    IOPASM_FREE(first_inst->next->code);
    IOPASM_FREE(first_inst->next);

  } 

  /* Delete the last instruction */
  if (first_inst->argument != NULL) {
    delete_arg_list(first_inst->argument);
  }
  
  IOPASM_FREE(first_inst->inst);
  IOPASM_FREE(first_inst->code);
  IOPASM_FREE(first_inst);
  
}



/* ************************************************************************* */
/* PROTOTYPE   : void delete_plist(plist_t *a_list);                         */
/*                                                                           */
/* ARGUMENTS   : "a_list" is a valid pointer to a parse list.                */
/*                                                                           */
/* FUNCTION    : Deletes the entire parse list, all instructions and         */
/*               arguments.                                                  */
/*                                                                           */
/* RETURN VALUE: none.                                                       */
/* ************************************************************************* */
void delete_plist(plist_t *a_list)
{
  assert ( a_list != NULL);

  delete_inst_list(a_list->first_inst);
  IOPASM_FREE(a_list);
}



/* ************************************************************************* */
/* PROTOTYPE   : int delete_arg(arg_t *arg);                                 */
/*                                                                           */
/* ARGUMENTS   : "arg" is a valid pointer to an argument.                    */
/*                                                                           */
/* FUNCTION    : delete an argument and all arguments stored in a stack      */
/*                                                                           */
/* RETURN VALUE: 0 (should be void)                                          */
/* ************************************************************************* */
void delete_arg(void *the_arg) 
{
  arg_t *arg;

  assert(the_arg != NULL);
  arg = (arg_t *) the_arg;

  if (arg->arg_type == ARG_STACK) {
    delete_stack(arg->arg_val.arg_stack);
  }

  else {
    IOPASM_FREE(arg->arg_val.value);
    IOPASM_FREE(arg->raw_arg);
    IOPASM_FREE(arg);
    }
}
  
/* ************************************************************************* */
/* PROTOTYPE   : void print_arg_val(void *arg);                              */
/*                                                                           */
/* ARGUMENTS   : "arg" is a valid pointer to an argument                     */
/*                                                                           */
/* FUNCTION    : prints the argument value.                                  */
/*                                                                           */
/* RETURN VALUE: none.                                                       */
/* ************************************************************************* */
void print_arg_val(void *arg)
{
  arg_t *the_arg;

  assert(arg != NULL);
  
  the_arg = (arg_t *)arg;
  
  if (the_arg->arg_type != ARG_STACK) {
    printf("type=0x%x, value=%s\n",the_arg->arg_type,the_arg->arg_val.value);
  }
}



/* ************************************************************************* */
/* PROTOTYPE   : int save_inst_object(FILE *out_file,inst_t *inst);          */
/*                                                                           */
/* ARGUMENTS   : "out_file" is a valid file open for output. "inst" is a     */
/*               pointer to an instruction.                                  */
/*                                                                           */
/* FUNCTION    : Saves the argument "inst" to "out_file" in a platform inde- */
/*               pendent format.                                             */
/*                                                                           */
/* RETURN VALUE: Number of objects saved                                     */
/* ************************************************************************* */
int save_inst_object(FILE *out_file,inst_t *inst)
{
  meta_object_t *temp_object;
  int i;
  uint32_t s_type;        /* Saved instruction type in network byte order    */
  uint32_t s_line_no;     /* Saved line number in network byte order         */
  uint32_t s_file_index;  /* Saved file index in network byte order          */
  assert ((out_file != NULL) && (inst != NULL));
  
  /* Convert integers to network byte order */
  s_type = htonl((uint32_t) inst->i_type);
  s_line_no = htonl((uint32_t) inst->line_no);
  s_file_index = htonl((uint32_t) inst->file_index);


  /* Calculate size of object */
  i = 3 * sizeof(uint32_t);
  i = (inst->inst != NULL) ? i+strlen(inst->inst):i;

  temp_object = create_meta_object(INST_LIST_ELEM,i);

  save_meta_object(out_file,temp_object);
  delete_meta_object(temp_object);
  
  assert (fwrite(&s_type,sizeof(uint32_t),1,out_file) == 1);
  assert (fwrite(&s_line_no,sizeof(uint32_t),1,out_file) == 1);
  assert (fwrite(&s_file_index,sizeof(uint32_t),1,out_file) == 1);

  if ((inst->inst != NULL) && (strlen(inst->inst) > 0)) {
    assert(fwrite(inst->inst,strlen(inst->inst),1,out_file) == 1);
  }
  return 1; /* An instruction is always _exactly_ 1 object */
}



/* ************************************************************************* */
/* PROTOTYPE   : int save_arg_object(FILE *out_file,void *arg);              */
/*                                                                           */
/* ARGUMENTS   : "out_file" is a valid file descriptor open for output.      */
/*               "arg" points to an argument to save.                        */
/*                                                                           */
/* FUNCTION    : Saves the argument "arg" to "out_file" in a platform inde-  */
/*               pendent format.                                             */
/*                                                                           */
/* RETURN VALUE: Returns the number of objects saved.                        */
/* ************************************************************************* */
int save_arg_object(FILE *out_file,void *arg)
{
  meta_object_t *temp_object;
  int  i;
  arg_t *temp_arg;
  uint32_t s_type;    /* saved arg type in network byte order */
  uint32_t s_raw_len; /* saved raw length in network byte order */
  uint32_t s_arg_len; /* Saved argument length in network byte order */
  int obj_count = 1;
  assert((out_file != NULL) && (arg != NULL));
  
  temp_arg = (arg_t*) arg;
  
  s_type = htonl((uint32_t)temp_arg->arg_type);
  s_raw_len = htonl((uint32_t)temp_arg->raw_len);
  s_arg_len = htonl((uint32_t)temp_arg->arg_len);

  /* Get the saved size of all objects */
  i = 3  * sizeof(uint32_t);
  i += temp_arg->arg_len; 
  i += temp_arg->raw_len;

  temp_object = create_meta_object(ARG_LIST_ELEM,i);
  save_meta_object(out_file,temp_object);
  delete_meta_object(temp_object);
  
  assert(fwrite(&s_type,sizeof(uint32_t),1,out_file) == 1);
  assert(fwrite(&s_arg_len,sizeof(uint32_t),1,out_file) == 1);
  assert(fwrite(&s_raw_len,sizeof(uint32_t),1,out_file) == 1);
  
  if (temp_arg->arg_len > 0) {
    assert(fwrite(temp_arg->arg_val.value,temp_arg->arg_len,1,out_file) == 1);
  }

  if (temp_arg->raw_len > 0) {
    assert(fwrite(temp_arg->raw_arg,temp_arg->raw_len,1,out_file) == 1);
  }

  if (temp_arg->arg_type == ARG_STACK) { /* Save stack here */
    obj_count += save_stack_object(out_file,temp_arg->arg_val.arg_stack);
  }

  return obj_count;
}



/* ************************************************************************* */
/* PROTOTYPE   : void set_inst_file_index(inst_t *inst,int file_index);      */
/*                                                                           */
/* ARGUMENTS   : "inst" points to a valid instruction. "file_index" is the   */
/*               index of the file from which the instruction was generated. */
/*                                                                           */
/* FUNCTION    : Sets the file index for the instruction "inst"              */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void set_inst_file_index(inst_t *inst,int file_index)
{
  assert(inst != NULL);
  inst->file_index = file_index;
}



/* ************************************************************************* */
/* PROTOTYPE   : int get_inst_file_index(inst_t *inst);                      */
/*                                                                           */
/* ARGUMENTS   : "inst" points to a valid instruction.                       */
/*                                                                           */
/* FUNCTION    : Get the index number for the input file from which the      */
/*               instruction was generated.                                  */
/*                                                                           */
/* RETURN VALUE: Returns the index for the filename                          */
/* ************************************************************************* */
int get_inst_file_index(inst_t *inst)
{
  assert(inst != NULL);
  return inst->file_index;
}



/* ************************************************************************* */
/* PROTOTYPE   : inst_t *load_inst_object(FILE *in_file,void *object);       */
/*                                                                           */
/* ARGUMENTS   : "in_file" is a valid input file. "object" is the object     */
/*               descriptor.                                                 */
/*                                                                           */
/* FUNCTION    : Loads an instruction object from the input file "in_file"   */
/*                                                                           */
/* RETURN VALUE: pointer to a valid instruction object.                      */
/* ************************************************************************* */
inst_t *load_inst_object(FILE *in_file,void *object)
{
  int i_type;
  char *inst;
  inst_t *temp_inst;
  int size;
  int line_no,file_index;
  int i_size;
  uint32_t l_type;
  uint32_t l_line_no;
  uint32_t l_file_index;

  assert ((in_file != NULL) && (object != NULL));
  size = get_object_size(object);
  i_size = size - (3*sizeof(uint32_t));
  inst = (char *) IOPASM_MALLOC((size_t) (i_size+1));
  assert(inst != NULL);
  memset(inst,0,(i_size+1));
  
  assert (fread(&l_type,sizeof(uint32_t),1,in_file) == 1);
  assert (fread(&l_line_no,sizeof(uint32_t),1,in_file) == 1);
  assert (fread(&l_file_index,sizeof(uint32_t),1,in_file) == 1);
  
  i_type = ntohl(l_type);
  line_no = ntohl(l_line_no);
  file_index = ntohl(l_file_index);
 
  if (size > (3 * sizeof(uint32_t))) {
    assert (fread(inst,i_size,1,in_file) == 1);
    temp_inst = create_inst(inst,i_type,0,RF_RELOCATE);
    set_inst_line(temp_inst,line_no);
    set_inst_file_index(temp_inst,file_index);
    IOPASM_FREE(inst);
  }
  
  else {
    temp_inst = create_inst(NULL,i_type,0,RF_RELOCATE);
  }

  return temp_inst;
  
    
}



/* ************************************************************************* */
/* PROTOTYPE   : void *load_arg_object(FILE *in_file);                       */
/*                                                                           */
/* ARGUMENTS   : "in_file" is a valid input file.                            */
/*                                                                           */
/* FUNCTION    : loads and creates an argument from an object file specified */
/*               by "in_file".                                               */
/*                                                                           */
/* RETURN VALUE: Pointer to the newly loaded argument.                       */
/* ************************************************************************* */
void *load_arg_object(FILE *in_file)
{
  int arg_type;
  int arg_len;
  int raw_len;
  char *raw_arg = NULL;
  char *arg = NULL;
  arg_t *temp_arg;
  dyn_stack_t *the_stack;
  uint32_t l_type;
  uint32_t l_arg_len;
  uint32_t l_raw_len;

  assert (in_file != NULL);
  

  
  assert(fread(&l_type,sizeof(uint32_t),1,in_file) == 1);
  assert(fread(&l_arg_len,sizeof(uint32_t),1,in_file) == 1);
  assert(fread(&l_raw_len,sizeof(uint32_t),1,in_file) == 1);

  /* Convert to host integers */
  arg_type = ntohl(l_type);
  arg_len = ntohl(l_arg_len);
  raw_len = ntohl(l_raw_len);

  if (arg_len > 0) {
    arg = (char *) IOPASM_MALLOC(sizeof(char) * (arg_len+1));
    assert(arg != NULL);
    memset(arg,0,arg_len+1);
    assert(fread(arg,arg_len,1,in_file) == 1);
  }

  if (raw_len > 0) {
    raw_arg = (char *) IOPASM_MALLOC(sizeof(char) * (raw_len+1));
    assert(raw_arg != NULL);
    memset(raw_arg,0,raw_len+1);
    assert(fread(raw_arg,raw_len,1,in_file) == 1);
  }

  if (arg_type == ARG_STACK) {
    the_stack = load_stack_object(in_file,
				  delete_arg,
				  load_arg_object,
				  save_arg_object);
    assert(the_stack != NULL);
    temp_arg = create_arg_stack(the_stack);
    add_raw_arg(temp_arg,raw_arg);
  }

  else {
    temp_arg = create_arg(arg,arg_len,arg_type);
    assert(temp_arg != NULL);
/*     add_raw_arg(temp_arg,raw_arg); */

    if (arg_len > 0) {
      IOPASM_FREE(arg);
    }

    if (raw_len > 0) {
      IOPASM_FREE(raw_arg);
    }
  }
  
  return (void *) temp_arg; 
}

  
   
/* ************************************************************************* */
/* PROTOTYPE   : void dump_inst_object(FILE *in_file,meta_object_t *object); */
/*                                                                           */
/* ARGUMENTS   : "in_file" is a pointer to an object file. "object" is the   */
/*               object from the object file that is to be dumped.           */
/*                                                                           */
/* FUNCTION    : Dumps (prints) an instruction object to standard out.       */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void dump_inst_object(FILE *in_file,meta_object_t *object)
{
  uint32_t i_type;
  int32_t line_no,file_index;
   
  int size;
  char inst[MAX_STRLEN];
  
  assert((in_file != NULL) && (object != NULL));

  size = get_object_size(object);

  assert (fread(&i_type,sizeof(uint32_t),1,in_file) == 1);
  assert (fread(&line_no,sizeof(uint32_t),1,in_file) == 1);
  assert (fread(&file_index,sizeof(uint32_t),1,in_file) == 1);

  fprintf(stdout,"     Line number : %d\n",ntohl(line_no));
  fprintf(stdout,"     File index  : %d\n",ntohl(file_index));
  fprintf(stdout,"     Inst. type  : 0x%d (%s)\n",ntohl(i_type),
	  get_type_description(ntohl(i_type),inst_types));
  
  if (size > (3 * sizeof(uint32_t))) {
    memset(inst,0,MAX_STRLEN);
    assert (fread(inst,size-(3*sizeof(uint32_t)),1,in_file) == 1);
    fprintf(stdout,"     Inst. value : %s\n",inst);
  }

  else {
    fprintf(stdout,"                 value: No value found\n");
  }
  
  fprintf(stdout,"--------------------------------------------------------\n");
}


/* ************************************************************************* */
/* PROTOTYPE   : void dump_arg_object(FILE *in_file,meta_object_t *object);  */
/*                                                                           */
/* ARGUMENTS   : "in_file" is a valid pointer to an open file. "object" is   */
/*               a valid object descriptor.                                  */
/*                                                                           */
/* FUNCTION    : Dumps an object entry as plain text to stdout.              */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void dump_arg_object(FILE *in_file,meta_object_t *object)
{
  uint32_t arg_type,arg_len,raw_len;
  int size;
  char arg[MAX_STRLEN];

  size = get_object_size(object);

  fread(&arg_type,sizeof(uint32_t),1,in_file);
  fread(&arg_len,sizeof(uint32_t),1,in_file);
  fread(&raw_len,sizeof(uint32_t),1,in_file);
  arg_type = ntohl(arg_type);
  arg_len = ntohl(arg_len);
  raw_len = ntohl(raw_len);

  fprintf(stdout," Size according to meta object %d\n" ,size);

  fprintf(stdout,"     Arg. type   : %d (%s)\n",arg_type,
	  get_type_description(arg_type,arg_types));

  if (arg_type == ARG_STACK) {
    dump_stack_object(in_file,object,dump_arg_object);
  }

  if (arg_len > 0) {
    memset(arg,0,MAX_STRLEN);
    assert (fread(arg,arg_len,1,in_file) == 1);
    fprintf(stdout,"     Arg. value  : %s\n",arg);
  }
  
  if (raw_len > 0) {
    memset(arg,0,MAX_STRLEN);
    assert (fread(arg,raw_len,1,in_file) == 1);
    fprintf(stdout,"     Raw Arg.    : %s\n",arg);
  }

  else {
    fprintf(stdout,"              value: No value\n");
  }
  fprintf(stdout,"--------------------------------------------------------\n");
  
}



/* ************************************************************************* */
/* PROTOTYPE   : int get_list_count(plist_t *a_list);                        */
/*                                                                           */
/* ARGUMENTS   : "a_list" points to a valid list.                            */
/*                                                                           */
/* FUNCTION    : Counts the number of instructions in a list.                */
/*                                                                           */
/* RETURN VALUE: Number of instructions in the list.                         */
/* ************************************************************************* */
int get_list_count(plist_t *a_list)
{
  
  assert(a_list != NULL);
  return   a_list->n_instructions;
}



/* ************************************************************************* */
/* PROTOTYPE   : inst_t *synth_org_inst(unsigned int addr);                  */
/*                                                                           */
/* ARGUMENTS   : "addr" is an address.                                       */
/*                                                                           */
/* FUNCTION    : Synthesizes an "org" instruction and returns a pointer to   */
/*               a syntactically correct org instruction                     */
/*                                                                           */
/* RETURN VALUE: Pointer to an org instruction.                              */
/* ************************************************************************* */
inst_t *synth_org_inst(unsigned int addr) 
{
  inst_t *tmp_inst;
  char *tmp_string;
  arg_t *tmp_arg;
  char raw_text[MAX_STRLEN];

  tmp_inst = create_inst(NULL,INST_ORIGIN,0,RF_ABSOLUTE);
  assert(tmp_inst != NULL);
  tmp_string = IOPASM_MALLOC(sizeof(char) * 10);
  assert(tmp_string != NULL);
  memset(tmp_string,0,10);

  snprintf(tmp_string,10,"%u",addr);
  snprintf(raw_text,MAX_STRLEN," ;(inserted by iopasm)");

  tmp_arg = create_arg(tmp_string,strlen(tmp_string),ARG_IMMEDIATE);
  add_inst_arg(tmp_inst,tmp_arg);
  add_raw_arg(tmp_arg,raw_text);

  return tmp_inst;
}



/* ************************************************************************* */
/* PROTOTYPE   : inst_t *synth_align_inst(unsigned int bit_val);             */
/*                                                                           */
/* ARGUMENTS   : "bit_val" is the alignment value measured in bits.          */
/*                                                                           */
/* FUNCTION    : Synthesizes an "align" instruction and returns a pointer to */
/*               a syntactically correct alignment instruction.              */
/*                                                                           */
/* RETURN VALUE: Pointer to an align instruction.                            */
/* ************************************************************************* */
inst_t *synth_align_inst(unsigned int bit_val) 
{
  inst_t *tmp_inst;
  char *tmp_string;
  arg_t *tmp_arg;
  char raw_text[MAX_STRLEN];

  tmp_inst = create_inst(NULL,INST_ALIGN,0,RF_ABSOLUTE);
  assert(tmp_inst != NULL);
  tmp_string = IOPASM_MALLOC(sizeof(char) * 10);
  assert(tmp_string != NULL);
  memset(tmp_string,0,10);

  snprintf(tmp_string,10,"%u",bit_val);
  snprintf(raw_text,MAX_STRLEN," ;(inserted by iopasm)"); 

  tmp_arg = create_arg(tmp_string,strlen(tmp_string),ARG_IMMEDIATE);
  add_inst_arg(tmp_inst,tmp_arg);
  add_raw_arg(tmp_arg,raw_text);

  return tmp_inst;
}



/* ************************************************************************* */
/* PROTOTYPE   : void insert_inst(inst_t *prior_to,inst_t *insert_this);     */
/*                                                                           */
/* ARGUMENTS   : "prior_to" is a valid instruction so is "insert_this".      */
/*                                                                           */
/* FUNCTION    : inserts the instruction "insert_this" prior to "prior_to".  */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void insert_inst(inst_t *prior_to,inst_t *insert_this)
{
  inst_t *prev_inst;
  
  assert ((prior_to != NULL) && (insert_this !=NULL));
  prev_inst = prior_to->prev;
  
  insert_this->next = prior_to;
  prior_to->prev = insert_this;
  prev_inst->next = insert_this;
  insert_this->prev = prev_inst;
}



/* ************************************************************************* */
/* PROTOTYPE   : void delete_code(plist_t *a_list);                          */
/*                                                                           */
/* ARGUMENTS   : "a_list" points to a valid parser list.                     */
/*                                                                           */
/* FUNCTION    : Iterates over each instruction and erases all code          */
/*               stored in each and every instruction.                       */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void delete_code(plist_t *a_list)
{
  inst_t *temp_inst;

  assert(a_list != NULL);
  
  temp_inst = get_first_inst(a_list);

  while (temp_inst != NULL) {

    if (temp_inst->code != NULL) {
      temp_inst->mem_size =0;
      IOPASM_FREE(temp_inst->code);
      temp_inst->code = NULL;
    }
    temp_inst = get_next_inst(temp_inst);
  } /* end of while (temp_inst != NULL) */
}
