/* ************************************************************************* */
/* save_operations.c : Save operations                                       */
/*                                                                           */
/*    Copyright 2002-2004 Axis Communications AB.                            */
/*                                                                           */
/*                                                                           */
/*    This program is free software; you can redistribute it and/or modify   */
/*    it under the terms of the GNU General Public License as published by   */
/*    the Free Software Foundation; either version 2 of the License, or      */
/*    (at your option) any later version.                                    */
/*                                                                           */
/*    This program is distributed in the hope that it will be useful,        */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*    GNU General Public License for more details.                           */
/*                                                                           */
/*    You should have received a copy of the GNU General Public License      */
/*    along with this program; if not, write to the Free Software            */
/*    Foundation, Inc., 59 Temple Place - Suite 330,                         */
/*    Boston, MA 02111-1307, USA.                                            */
/*                                                                           */
/*                                                                           */
/*    Send bug reports and inquiries to Johan.Jorgensen@axis.com             */
/*                                                                           */
/* Short revision history:                                                   */
/*                                                                           */
/*   Date      Description                                          Sign.    */
/*  =======================================================================  */
/*  2003-02-05 Module Implemented                                    JohanJ  */
/*  2003-03-17 Changed behaviour of list_file                        JohanJ  */
/* ************************************************************************* */



/* ************************************************************************* */
/*                      INCLUDES USED BY THIS MODULE                         */
/* ************************************************************************* */
#include <assert.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "cg_fe.h"
#include "dyn_array.h"
#include "dyn_stack.h"
#include "iopasm_err.h"
#include "iopasm_global.h"
#include "labels.h"
#include "meta_object.h"
#include "p_list.h"
#include "save_operations.h"


/* ************************************************************************* */
/*                       TYPES DEFINED IN THIS MODULE                        */
/* ************************************************************************* */


/* ************************************************************************* */
/*                     DEFINITIONS LOCAL TO THIS MODULE                      */
/* ************************************************************************* */



/* ************************************************************************* */
/*                              CONSTANT VARIABLES                           */
/* ************************************************************************* */



static void dump_only_state(char *dst,int len,inst_t *inst);
inst_t *find_listable_inst(plist_t *optimized_list,inst_t *an_inst);

/* ************************************************************************* */
/* PROTOTYPE   : static void save_fname_object(FILE *file,char *string,      */
/*                                             int len);                     */
/*                                                                           */
/* ARGUMENTS   : "file" is the object file. "string" is the string that      */
/*               contains the filename and "len" is the length of the file-  */
/*               name.                                                       */
/*                                                                           */
/* FUNCTION    : save a filename object to the object file                   */
/*                                                                           */
/* RETURN VALUE:  None.                                                      */
/* ************************************************************************* */
static void save_fname_object(FILE *file,char *string,int len)
{
  meta_object_t *object;
  assert(file != NULL);
  
  object = create_meta_object(SRC_FILENAME_ELEM,len);
  assert(object != NULL);
  
  save_meta_object(file,object);
  delete_meta_object(object);
    
  assert(fwrite(string,len,sizeof(char),file) == 1);


}

/* ************************************************************************* */
/* PROTOTYPE   : inst_t *find_listable_inst(plist_t *optimized_list,         */
/*                                          inst_t *an_inst);                */
/*                                                                           */
/* ARGUMENTS   : "optimized_list" is the optimized code list. "an_inst" is   */
/*               the instruction to find.                                    */
/*                                                                           */
/* FUNCTION    : Finds an instruction in the optimized list.                 */
/*                                                                           */
/* RETURN VALUE: Pointer to the instruction in the optimized list.           */
/* ************************************************************************* */
inst_t *find_listable_inst(plist_t *optimized_list,inst_t *an_inst)
{
  int line_no;
  int i_type;
  int file_index;
  int tmp_line,tmp_index,tmp_type;
  inst_t *tmp_inst;

  assert((optimized_list != NULL) && (an_inst != NULL));

  line_no = get_inst_line(an_inst);
  file_index = get_inst_file_index(an_inst);
  i_type = get_inst_type(an_inst);

  tmp_inst = get_first_inst(optimized_list);

  while(tmp_inst != NULL) {
    tmp_line = get_inst_line(tmp_inst);
    tmp_index = get_inst_file_index(tmp_inst);
    tmp_type = get_inst_type(tmp_inst);

    if ((tmp_type == i_type) && 
	(tmp_line == line_no) &&
	(file_index == tmp_index)) {
      return tmp_inst;
    }
    
    tmp_inst = get_next_inst(tmp_inst);
  } /* End of while(tmp_inst != NULL) */

  return NULL;
}

/* ************************************************************************* */
/* PROTOTYPE   : static void left_comment(FILE *out_file,char *text);        */
/*                                                                           */
/* ARGUMENTS   : "file" is a file to write to. "text" is the text to format  */
/*                                                                           */
/* FUNCTION    : Write a left-justified C-comment to "out_file"              */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
static void left_comment(FILE *out_file,char *text)
{
  char out_string[MAX_STRLEN];
  int i;
  assert ((out_file != NULL) && (text != NULL));
  
  memset(out_string,0,MAX_STRLEN);
  strcpy(out_string,"/* ");
  strncat(out_string,text,72);

  i = strlen(out_string);
  while(i < 76) {
    out_string[i++] = ' ';
  }
  
  strcat(out_string," */\n");
  fprintf(out_file,out_string);
}



/* ************************************************************************* */
/* PROTOTYPE   : static void comment_delim(FILE *out_file);                  */
/*                                                                           */
/* ARGUMENTS   : "out_file" is the file to write to.                         */
/*                                                                           */
/* FUNCTION    : Writes a comment delimitor to file "out_file"               */
/*                                                                           */
/* RETURN VALUE: None                                                        */
/* ************************************************************************* */
static void comment_delim(FILE *out_file)
{
  int i;
  char work_string[MAX_STRLEN];

  assert (out_file != NULL);

  /* Create a ***-line */
  memset(work_string,0,MAX_STRLEN);
  strcpy(work_string,"/* ");
  for (i=3;i<76;i++) { work_string[i] = '*'; }
  strcat(work_string," */");
  assert(fprintf(out_file,"%s\n",work_string) > 0);
}



/* ************************************************************************* */
/* PROTOTYPE   : static void header_hdr(FILE *out_file);                     */
/*                                                                           */
/* ARGUMENTS   : "out_file" is the file to write the comment-header to.      */
/*                                                                           */
/* FUNCTION    : Writes a commented header to file "out_file"                */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
static void header_hdr(FILE *out_file)
{
  char work_string[MAX_STRLEN];

  assert(out_file != NULL);
  memset(work_string,0,MAX_STRLEN);
  comment_delim(out_file);
  
  memset(work_string,0,MAX_STRLEN);
  
  snprintf(work_string,MAX_STRLEN,
	   "This file is autogenerated by IOPASM Version %d.%d",
	   IOPASM_MAJOR_VERSION,IOPASM_MINOR_VERSION);
  
  left_comment(out_file,work_string);
  left_comment(out_file,"DO NOT EDIT THIS FILE - All changes will be lost!");
  
  comment_delim(out_file);
    
}



/* ************************************************************************* */
/* PROTOTYPE   : static void centered_comment(FILE *out_file,char *text);    */
/*                                                                           */
/* ARGUMENTS   : "out_file" is the file to write to and "text" is the text   */
/*               to write as comment.                                        */
/*                                                                           */
/* FUNCTION    : Write a centered comment to "out_file"                      */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
static void centered_comment(FILE *out_file,char *text)
{
  int i,j;
  char work_string[MAX_STRLEN];

  assert ((out_file != NULL) && (text != NULL));

  j = (72 - strlen(text)) / 2;
  i = 3;
  comment_delim(out_file);

  /* Write text on a centered line */
  memset(work_string,0,MAX_STRLEN);
  strcpy(work_string,"/* ");
  for (i=3;i<j+3;i++) { work_string[i] = ' '; };
  strncat(work_string,text,72);
  i = strlen(work_string);
  for (;i<76;i++) {work_string[i] = ' '; };
  strcat(work_string," */");
  assert(fprintf(out_file,"%s\n",work_string) > 0);
  comment_delim(out_file);
}



/* ************************************************************************* */
/* PROTOTYPE   : static void dump_timer(char *dst,int len,inst_t *inst);     */
/*                                                                           */
/* ARGUMENTS   : "dst" is the destination to write to. "len" is the length   */
/*               of the "dst" string and "inst" is the instruction to dump   */
/*                                                                           */
/* FUNCTION    : Reformat a timer instruction to human readable format.      */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
static void dump_timer(char *dst,int len,inst_t *inst)
{
  int i,j;
  arg_t *arg;


  assert((dst != NULL) && (inst != NULL));
  memset(dst,0,len);

  j = 0;
  i = snprintf(dst,len,"%s ","timer ");
  arg = get_first_arg(inst);
  
  while(arg != NULL) {
    switch(get_arg_type(arg)) {
    case ARG_TIMER_KEEP:
      i += snprintf(dst+i,len-i,"KEEP %s",get_raw_arg(arg));
      arg = get_next_arg(arg);
      arg = get_next_arg(arg);
      break;
      
    default:
      i += snprintf(dst+i,len-i,"%s",get_raw_arg(arg));
      arg = get_next_arg(arg);
      if (j ==0) { 
	arg = get_next_arg(arg); 
      }
      break;
    }
	   
    if (arg != NULL) {
      i += snprintf(dst+i,len-i,": ");
      j++;
    }
    
    else {
      i += snprintf(dst+i,len-i,"\n");
    }
  }
}


/* ************************************************************************* */
/* PROTOTYPE   : static void dump_only_state(char *dst,int len,inst_t *inst);*/
/*                                                                           */
/* ARGUMENTS   : "dst" points to a valid destination string with length      */
/*               as specified by "len". "inst" is the only state to dump     */
/*                                                                           */
/* FUNCTION    : Create a textual representation of an only state            */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
static void dump_only_state(char *dst,int len,inst_t *inst)
{
  arg_t *arg;
  int i;
  
  assert((dst != NULL) && (inst != NULL));
  memset(dst,0,len);

  i = 0;
  arg = get_first_arg(inst);
  snprintf(dst+i,len-i,"only %s",get_arg_val(arg));
}
  
/* ************************************************************************* */
/* PROTOTYPE   : static void dump_state_xfer(char *dst,int len,inst_t *inst) */
/*                                                                           */
/* ARGUMENTS   : "dst" points to a valid destination string with length      */
/*               as specified by "len". "inst" is the only state to dump     */
/*                                                                           */
/* FUNCTION    : Create a textual representation of a state transition       */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
static void dump_state_xfer(char *dst,int len,inst_t *inst)
{
  int i;
  arg_t *arg;

  assert((dst != NULL) && (inst != NULL));

  memset(dst,0,len);

  i = 0;
  arg = get_first_arg(inst);
  
  while(arg != NULL) {
    i += snprintf(dst+i,len-i,"%s",get_arg_val(arg));
    arg = get_next_arg(arg);

    if (arg != NULL) {
      i += snprintf(dst+i,len-i,"%s"," : ");
    }
    
    else {
      i += snprintf(dst+i,len-i,"\n");
    }

  }
}



/* ************************************************************************* */
/* PROTOTYPE   : static void dump_inst(char *dst,int len,inst_t *inst);      */
/*                                                                           */
/* ARGUMENTS   : "dst" points to a valid destination string with length      */
/*               as specified by "len". "inst" is the only state to dump     */
/*                                                                           */
/* FUNCTION    : */
/*                                                                           */
/* RETURN VALUE: */
/* ************************************************************************* */
static void dump_inst(char *dst,int len,inst_t *inst)
{
  int i;
  arg_t *arg;
  assert((dst != NULL) && (inst != NULL));

  memset(dst,0,len);
  
  switch (get_inst_type(inst)) {

  case INST_ALIGN:
    i = snprintf(dst,len,"%s ",".align ");
    break;
      
  case INST_ORIGIN:
    i = snprintf(dst,len,"%s ",".org ");
    break;
    
  default:
    i = snprintf(dst,len,"%s ",get_inst(inst));
    break;
  } /* end of switch(get_inst_type(inst)) */
    
  arg = get_first_arg(inst);
  
  if (arg == NULL) {
    i += snprintf(dst+i,len-i,"\n");
  }
  
  while(arg != NULL) {
    i += snprintf(dst+i,len-i,"%s",get_raw_arg(arg));
    arg = get_next_arg(arg);
    
    if (arg != NULL) {
      i += snprintf(dst+i,len-i,"%s",", ");
    }
    
    else {
      i += snprintf(dst+i,len-i,"\n");
    }
    
  } /* End of  while(arg != NULL) { */
}


/* ************************************************************************* */
/* PROTOTYPE   : static void hex_dump(char *dst,int dst_len,                 */
/*                                    unsigned char *buff,int start,         */
/*                                    int len);                              */
/*                                                                           */
/* ARGUMENTS   : "dst" points to a valid destination string with length      */
/*               as specified by "len". "inst" is the only state to dump     */
/*                                                                           */
/* FUNCTION    : Dump code in hexadecimal format                             */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
static void hex_dump(char *dst,int dst_len,unsigned char *buff,int start,int len)
{
  int i;
  int j;
  int k;
  char out_byte;
  assert(dst_len > (3*len+1));
  memset(dst,0,dst_len);
  j = len;
  k = 0;

  while ((j-8) < 0 ) {
    snprintf(dst+(3*k),dst_len-(3*k),"   ");
    j++;
    k++;
  }

  for (i=start;i<(start + len);i++) {
    out_byte = (buff == NULL) ? 0:buff[i];
    snprintf(dst+(3*k),dst_len-(3*k),"%02x ",(out_byte & 0xFF));
    k++;
  }
}

/* ************************************************************************* */
/* PROTOTYPE   : static void save_state(FILE *file,cg_context_t *ctxt,       */
/*                                      inst_t *first, inst_t *last);        */
/*                                                                           */
/* ARGUMENTS   : "file" points to an open file in which to save. "ctxt" is   */
/*               the current code generation context. "first" points to the  */
/*               first instruction in the state and "last" points to the last*/
/*                                                                           */
/* FUNCTION    : Save a textual (human readable) representation of an FSM    */
/*               state.                                                      */
/*                                                                           */
/* RETURN VALUE: None. (internal helper function).                           */
/* ************************************************************************* */
static void save_state(FILE *file,cg_context_t *ctxt,inst_t *first,
		       inst_t *last)
{
  unsigned int pc;
  int timer_offset;
  int seq_offset;
  int xfer_offset;
  int size;
  arg_t *arg;
  inst_t *old_first;
  char hex_buff[MAX_STRLEN];
  unsigned char *buffer;
  char i_buff[MAX_STRLEN];
  
  assert((file != NULL) && (ctxt != NULL) &&
	 (first != NULL) && (last != NULL));



  pc = get_inst_addr(first);
  pc = cg_convert_fsm_addr(ctxt,pc);
  size = get_inst_size(first);
  old_first = first;

  xfer_offset = 4;
  timer_offset = 4;
  seq_offset = 4;

  buffer = get_inst_code(first);
  hex_dump(hex_buff,MAX_STRLEN,buffer,0,4);
  fprintf(file,"%4.4x   %s  %s:  \n",pc,
	  hex_buff,get_inst(first));
  /* We need to check if there are any timers or sequential instructions before
   * we can generate the list file, as the order in which the instructions
   * appear might differ from that of the generated code 
   */

  while(old_first != last) {
    if (get_inst_type(old_first) == INST_SEQ) {
      timer_offset += 4;
      xfer_offset +=4;
    }

    if (get_inst_type(old_first) == INST_TIMER) {
      xfer_offset += 4;
    }

    old_first = get_next_inst(old_first);
  }
    
  while(first != last) {
    switch(get_inst_type(first)) {
      
    case INST_FSM_HALT:
      fprintf(file,"                                 fsm_halt\n");
      break;
      
    case INST_FSM_DO_SEQ:
      fprintf(file,"                                 %s\n","do_seq");
      break;
      
    case INST_FSM_GO_SEQ:
      fprintf(file,"                                 %s\n","go_seq");
      break;
      
    case INST_FSM_INV_STB:
      fprintf(file,"                                 %s\n","inv_stb");
      break;
      
    case INST_FSM_STB:
      arg = get_first_arg(first);
      fprintf(file,"                                 %s = %s\n","stb",get_raw_arg(arg));
      break;
      
    case INST_FSM_OUT_FLAG:
      arg = get_first_arg(first);
      fprintf(file,"                                 %s = %s\n","outp",get_raw_arg(arg));
      break;
      
    case INST_FSM_INP_FLAG:
      arg = get_first_arg(first);
      fprintf(file,"                                 %s = %s\n","inp",get_raw_arg(arg));
      break;

    case INST_FSM_EVENT_MASK:
      arg = get_first_arg(first);
      fprintf(file,"                                 %s = %s\n","emask",get_raw_arg(arg));
      break;

    case INST_FSM_UPDATE_MASK:
      arg = get_first_arg(first);
      fprintf(file,"                                 %s = %s\n","umask",get_raw_arg(arg));
      break;
            
    case INST_FSM_REUPDATE:
      fprintf(file,"                                 reupdate\n");
      break;

    case INST_SEQ:
      hex_dump(hex_buff,MAX_STRLEN,buffer,seq_offset,get_inst_size(first));
      dump_inst(i_buff,MAX_STRLEN,first);
      fprintf(file,"       %s  seq %s",hex_buff,i_buff);
      seq_offset += get_inst_size(first);
      break;
	      
    case INST_TIMER:
      hex_dump(hex_buff,MAX_STRLEN,buffer,timer_offset,4);
      dump_timer(i_buff,MAX_STRLEN,first);
      fprintf(file,"       %s  %s",hex_buff,i_buff);
      timer_offset += 4;
      break;

    case INST_STATE_XFER:
      hex_dump(hex_buff,MAX_STRLEN,buffer,xfer_offset,3);
      dump_state_xfer(i_buff,MAX_STRLEN,first);
      fprintf(file,"       %s  %s",hex_buff,i_buff);
      xfer_offset += 3;
      break;

    case INST_ONLY_STATE:
      dump_only_state(i_buff,MAX_STRLEN,first);
      fprintf(file,"                                 %s",i_buff);
      fprintf(file,"\n");
    }
    first = get_next_inst(first);
  } 
  fprintf(file,"\n");
}

/* ************************************************************************* */
/* PROTOTYPE   : static void rotate_binary_buffer(cg_context_t *ctxt,        */
/*                                                unsigned char *bin_buffer);*/
/*                                                                           */
/* ARGUMENTS   : "ctxt" points to a valid code generation context.           */ 
/*               "bin_buffer" is the binary buffer to operate on.            */
/*                                                                           */
/* FUNCTION    : rotate buffer so that hexadecimal printing will be correct. */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
static void rotate_binary_buffer(cg_context_t *ctxt,unsigned char *bin_buffer) 
{
  unsigned char tmp_byte;
  int i,j,l,mem_size,row_len;

  assert((ctxt != NULL) && (bin_buffer != NULL));

  row_len = cg_get_row_len(ctxt);
  mem_size = cg_get_memory_size(ctxt);

  j = 0; i = 0;
  l = 0;

  while (l < mem_size) {
    j = l + row_len - 1;

    for (i=l;i<l+row_len/2;i++) {
      tmp_byte = bin_buffer[j];
      bin_buffer[j] = bin_buffer[i];
      bin_buffer[i] = tmp_byte;
      j--;
    } /* End of for (i=l+k-1;i<l+row_len;i--) { */

    l += row_len;
  } /* End of while(... */

}



/* ************************************************************************* */
/* PROTOTYPE   : void save_symbol_file(char *file_name,int file_type,        */
/*                                     cg_context_t *ctxt, plist_t *a_list); */
/*                                                                           */
/* ARGUMENTS   : "file_name" is the name of the symbol file and "file_type"  */
/*               denotes the type of symbol file (header or plain). "ctxt"   */
/*               is the valid code-generation context and "a_list" is        */
/*               a valid pointer to the parser list.                         */
/*                                                                           */
/* FUNCTION    : Save a symbol file in the specified format.                 */
/*                                                                           */
/* RETURN VALUE: none.                                                       */
/* ************************************************************************* */
void save_symbol_file(char *file_name,int file_type,cg_context_t *ctxt,
		      plist_t *a_list)
{
  array_t *labels;
  FILE *sym_file = NULL;
  char *pretty_name;
  int i;

  assert((ctxt != NULL) && (a_list != NULL));
  labels = cg_get_label_array(ctxt);
  sym_file = fopen(file_name,"wt");
  
  if (sym_file == NULL) {
    perror("Cannot open symbolfile for writing"); /* use IOPASM_ERR */
  }
  
  else if (file_type == SYM_FILE_FMT_SYMBOLS) { /* Save symbol file as plain
						 * file usable by iopasm */
    array_iterator((void *)ctxt,labels,(void *)sym_file,
		   write_label_as_symbols);
    fclose(sym_file);
  }

  else if (file_type == SYM_FILE_FMT_HEADER) { /* Save symbol file as C-
						* header file */

    assert ((pretty_name = strdup(file_name)) != NULL);
    
    /* create pretty-name */
    for (i=0;i<strlen(pretty_name);i++) {
      if (pretty_name[i] =='.') { 
	pretty_name[i] ='_'; 
      }
      if (isalpha(pretty_name[i])) {
	pretty_name[i] = toupper(pretty_name[i]);
      }
    }
    
    header_hdr(sym_file);

    fprintf(sym_file,"\n\n\n#ifndef __%s__\n",pretty_name);
    fprintf(sym_file,"#define __%s__\n\n\n",pretty_name);

    if (cg_get_namespace(ctxt) != NULL) {
      centered_comment(sym_file,"DEFINE MACROS FOR USE WITH NAME SPACE");
      fprintf(sym_file,"#define USE_LABEL(NAMESPACE,LABEL) ##NAMESPACE##_##LABEL##\n\n\n");
    }

    centered_comment(sym_file,"LABEL VALUE DEFINITIONS");
    array_iterator(ctxt,labels,(void *)sym_file,write_label_as_header);
   
    fprintf(sym_file,"\n#endif /* end of __%s__ */\n",pretty_name);
    fprintf(sym_file,"/* End of %s */\n",file_name);

    free(pretty_name);
    fclose(sym_file);
  }

  else { 
    iopasm_error(ctxt,NULL,ERR_UNKNOWN_SYMFILE_FORMAT,NULL);
  }
}    




/* ************************************************************************* */
/* PROTOTYPE   : void save_binary_file(char *bin_fname,cg_context_t *ctxt,   */
/*		                       plist_t *a_list,                      */
/*                                     unsigned char *bin_buffer);           */
/*                                                                           */
/* ARGUMENTS   : "bin_fname" points to the name of the binary file. "ctxt"   */
/*               is the valid context. "a_list" points to the current parse- */
/*               list and "bin_buffer" is the binary buffer                  */
/*                                                                           */
/* FUNCTION    : Saves a binary file that can be loaded into the memory of   */
/*               the I/O processor.                                          */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void save_binary_file(char *bin_fname,cg_context_t *ctxt,
		      plist_t *a_list,unsigned char *bin_buffer)
{
  int mem_size;

  FILE *bin_file = NULL;
  assert((bin_fname != NULL) && (ctxt != NULL) && (a_list != NULL));
  
  bin_file = fopen(bin_fname,"wb");
  
  if (bin_file == NULL) {
    iopasm_error(ctxt,NULL,ERR_CANNOT_WRITE_FILE,"Output file name: %s\n",
	       bin_fname);
  }

  mem_size = cg_get_memory_size(ctxt);

  rotate_binary_buffer(ctxt,bin_buffer);

  fwrite(bin_buffer,mem_size,sizeof(unsigned char),bin_file);
  fclose(bin_file);

}


/* ************************************************************************* */
/* PROTOTYPE   : void save_vlog64_file(char *vlog64_fname,                   */
/*                                     cg_context_t *ctxt,                   */
/*      		               plist_t *a_list,                      */
/*                                     unsigned char *bin_buffer);           */
/*                                                                           */
/* ARGUMENTS   : "vlog64_fname" points to the name of the vlog64 file.       */
/*               and "ctxt" is the valid code generation context. "a_list"   */
/*               points to the valid parse list, and bin_buffer is the       */
/*               to the valid binary buffer.                                 */
/*                                                                           */
/* FUNCTION    : Saves the assembled code in verilog-64 format.              */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void save_vlog64_file(char *vlog64_fname,cg_context_t *ctxt,
		      plist_t *a_list,unsigned char *bin_buffer)
{
  int i = 0;
  int mem_size = 0;
  FILE *vlog64_file;

  assert ((vlog64_fname != NULL) && (ctxt != NULL) && (bin_buffer !=NULL) &&
	  (a_list != NULL));

  mem_size = cg_get_memory_size(ctxt);

  vlog64_file = fopen(vlog64_fname,"wt");

  if (vlog64_file == NULL) {
        iopasm_error(ctxt,NULL,ERR_CANNOT_WRITE_FILE,"Output file name: %s\n",
	       vlog64_fname);
  }

  
  while (i < mem_size) {
    fprintf(vlog64_file,"%02x%02x%02x%02x_%02x%02x%02x%02x\n",
	    bin_buffer[i],bin_buffer[i+1],bin_buffer[i+2],bin_buffer[i+3],
	    bin_buffer[i+4],bin_buffer[i+5],bin_buffer[i+6],bin_buffer[i+7]);
    i += 8;
	    
  }
  fclose(vlog64_file);
   
}


/* ************************************************************************* */
/* PROTOTYPE   : void save_vlog256_file(char *vlog256_fname,                 */
/*                                      cg_context_t *ctxt,                  */
/* 		                        plist_t *a_list,                     */
/*                                      unsigned char *bin_buffer);          */
/*                                                                           */
/* ARGUMENTS   : "vlog256_fname" points to the name of the vlog64 file.      */
/*               and "ctxt" is the valid code generation context. "a_list"   */
/*               points to the valid parse list, and bin_buffer is the       */
/*               to the valid binary buffer.                                 */
/*                                                                           */
/* FUNCTION    : Saves the assembled code in verilog 256 format.             */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void save_vlog256_file(char *vlog256_fname,cg_context_t *ctxt,
		      plist_t *a_list,unsigned char *bin_buffer)
{
  int i = 0;
  int mem_size = 0;
  int k;
  FILE *vlog256_file;

  assert ((vlog256_fname  != NULL) && (ctxt != NULL) && (bin_buffer !=NULL) &&
	  (a_list != NULL));

  
  vlog256_file = fopen(vlog256_fname,"wt");

  if (vlog256_file == NULL) {
        iopasm_error(ctxt,NULL,ERR_CANNOT_WRITE_FILE,"Output file name: %s\n",
	       vlog256_fname);
  }


  mem_size = cg_get_memory_size(ctxt);
	  
  while (i < mem_size) {
    for (k=0;k<32;k+=4) {

      if (k < 28) {
	fprintf(vlog256_file,"%02x%02x%02x%02x_",
		bin_buffer[i],bin_buffer[i+1],bin_buffer[i+2],bin_buffer[i+3]);
      }

      else {
	fprintf(vlog256_file,"%02x%02x%02x%02x\n",
		bin_buffer[i],bin_buffer[i+1],bin_buffer[i+2],bin_buffer[i+3]);
      }
      i += 4;
    }

  }
 
  fclose(vlog256_file);
}


/* ************************************************************************* */
/* PROTOTYPE   : void save_list_file(char *list_fname,                       */
/*                                   cg_context_t *ctxt,plist_t *a_list,     */
/*                      	     plist_t *optimized_list);               */
/*                                                                           */
/* ARGUMENTS   : "list_fname" is the pointer to the name of the list file    */
/*               "ctxt" is the current code generation context, "a_list"     */
/*               is the unoptimized parser list and "optimized_list" is the  */
/*               parser list containing the optimized code                   */
/*                                                                           */
/* FUNCTION    : List the output in human readable format.                   */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void save_list_file(char *list_fname,cg_context_t *ctxt,plist_t *a_list,
		    plist_t *optimized_list)
{
  arg_t *arg;
  char *raw_arg;
  inst_t *loop_inst;
  inst_t *first;
  
  inst_t *last;
  int pc = 0;
  FILE *list_file;
  char hex_buff[MAX_STRLEN];
  char i_buff[MAX_STRLEN];
  unsigned char *i_ptr;
  unsigned char *bin_buffer;
  int opt_flag = FALSE;
  assert((list_fname != NULL) && (ctxt != NULL) && (a_list != NULL));

  bin_buffer = cg_get_code_buffer(ctxt);

  list_file = fopen(list_fname,"wt");
  if (list_file == NULL) {
    iopasm_error(ctxt,NULL,ERR_CANNOT_WRITE_FILE,"Output file name: %s\n",
		 list_fname);
  }
  
  if (check_option("optimize-fsm",(int *)1,OPT_INTEGER)) {
    opt_flag = TRUE;
  }

  else {
    opt_flag = FALSE;
  }

  loop_inst = get_first_inst(a_list);
  first = loop_inst;

  while (loop_inst != NULL) {
    
    if (opt_flag == TRUE) {
      first = find_listable_inst(optimized_list,loop_inst);
    }
    
    else {
      first = loop_inst;
    }

    pc = get_inst_addr(first);

    
    switch(get_inst_type(first)) {
      
    case INST_LABEL:
      if (get_inst_type(get_next_inst(first)) == INST_STATE_BEGIN) {
	fprintf(list_file,"\n");
	last = get_next_inst(first);
	
	while(get_inst_type(last) != INST_STATE_END) {
	  last = get_next_inst(last);
	}
	
	save_state(list_file,ctxt,first,last);
      
      first = last ;
      }
      else {
	fprintf(list_file,"\n%4.4X  ",pc);
	fprintf(list_file,"                           %s:\n",
		get_inst(first));
      }

      /* skip past any states in loop_inst */
      if (get_inst_type(get_next_inst(loop_inst)) == INST_STATE_BEGIN) {
	while(get_inst_type(loop_inst) != INST_STATE_END) {
	  loop_inst = get_next_inst(loop_inst);
	} /* End of while(get_inst_type(loop_inst) != INST_STATE_END) */
      } /* End of if ((get_inst_type(get_next_inst(loop_inst)) ... */
      break;

    case INST_DWORD:
      i_ptr = get_inst_code(first);
      pc = get_inst_addr(first);
      fprintf(list_file,"%4.4X  ",pc);
      hex_dump(hex_buff,MAX_STRLEN,i_ptr,0,4);
      memset(i_buff,0,MAX_STRLEN);
      arg = get_first_arg(first);
      raw_arg = get_raw_arg(arg);
      snprintf(i_buff,MAX_STRLEN,".dword %s\n",raw_arg);

      fprintf(list_file," %s  %s",hex_buff,i_buff);
      break;


    case INST_SEQ:
    case INST_ORIGIN:
    case INST_ALIGN:
      i_ptr = get_inst_code(first);
      pc = get_inst_addr(first);
      fprintf(list_file,"%4.4X  ",pc);
      hex_dump(hex_buff,MAX_STRLEN,i_ptr,0,get_inst_size(first));
      dump_inst(i_buff,MAX_STRLEN,first);
      fprintf(list_file," %s  %s",hex_buff,i_buff);
      break;


    }
    
    loop_inst = get_next_inst(loop_inst);
  }

  fclose(list_file);

} /* end of save_list_file(...) */



/* ************************************************************************* */
/* PROTOTYPE   : */
/*                                                                           */
/* ARGUMENTS   : */
/*                                                                           */
/* FUNCTION    : */
/*                                                                           */
/* RETURN VALUE: */
/* ************************************************************************* */
void save_object_file(char *file_name,cg_context_t *ctxt,plist_t *list)
{
  array_t *labels;
  FILE *bin_file = NULL;
  inst_t *first_inst;
  arg_t *first_arg;
  unsigned int obj_count = 0;
  char *fname;
  int i;
  assert((file_name != NULL) && (ctxt != NULL) && (list != NULL));

  bin_file = fopen(file_name,"wb");
  
  if (bin_file == NULL) {
    iopasm_error(ctxt,NULL,ERR_CANNOT_WRITE_FILE,"Object file name: %s\n",
	       file_name);
  }
  
  labels = cg_get_label_array(ctxt);
  save_object_header(bin_file,ctxt,obj_count);
  i = 0;
  fname = cg_get_file_by_index(ctxt,i);

  while (fname != NULL) {
    fname = cg_get_file_by_index(ctxt,i);
    if (fname != NULL) {
      save_fname_object(bin_file,fname,strlen(fname));
      obj_count++;
      i++;
    }
  }

  obj_count += get_array_size(labels);
  array_iterator((void *)ctxt,labels,bin_file,write_label_as_object);
  first_inst = get_first_inst(list);
  
  while (first_inst != NULL) {
    
    save_inst_object(bin_file,first_inst);
    obj_count++;

    first_arg = get_first_arg(first_inst);
    
    while(first_arg != NULL) {
      save_arg_object(bin_file,first_arg);
      first_arg = get_next_arg(first_arg);
      obj_count++;
    }

    first_inst = get_next_inst(first_inst);
  }

  /* Rewrite header with correct # of objects */
  save_object_header(bin_file,ctxt,obj_count);
  fclose(bin_file);
}  


/* ************************************************************************* */
/* PROTOTYPE   : void save_c_macros(char *file_name,cg_context_t *ctxt);     */
/*                                                                           */
/* ARGUMENTS   : "file_name" is the name of the files to contain the macros  */
/*               "ctxt" is the code generation context.                      */
/*                                                                           */
/* FUNCTION    : Saves all valid instructions and register representations   */
/*               as C-macros. These macros can be used to generate code in   */
/*               C-programs.                                                 */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void save_c_macros(char *file_name,cg_context_t *ctxt)
{
  FILE *output_file;
  char *pretty_name;
  int i;

  assert ((file_name != NULL) && (ctxt != NULL));
  assert((output_file = fopen(file_name,"wt")) != NULL);
  
  assert ((pretty_name = strdup(file_name)) != NULL);

    /* create pretty-name */
    for (i=0;i<strlen(pretty_name);i++) {
      if (pretty_name[i] =='.') { 
	pretty_name[i] ='_'; 
      }
      if (isalpha(pretty_name[i])) {
	pretty_name[i] = toupper(pretty_name[i]);
      }
    }

    header_hdr(output_file);

    assert(fprintf(output_file,"\n\n\n#ifndef __%s__\n",pretty_name) > 0);
    assert(fprintf(output_file,"#define __%s__\n\n\n",pretty_name) > 0);
    
    centered_comment(output_file,"REGISTER DEFINITIONS");
    save_register_macros(output_file,cg_get_reg_table(ctxt),cg_get_scope_name(ctxt));
    assert(fprintf(output_file,"\n\n\n") > 0);
   
    centered_comment(output_file,"ADDRESS MACROS");
    assert(fprintf(output_file,
		   "#define MK_DWORD_ADDR(ADDR) (ADDR >> 2)\n") > 0);
    assert(fprintf(output_file,
		   "#define MK_BYTE_ADDR(ADDR) (ADDR)\n") > 0);

    assert(fprintf(output_file,"\n\n\n") > 0);
    centered_comment(output_file,"INSTRUCTION MACROS");
    save_inst_macros(output_file,cg_get_inst_table(ctxt),cg_get_scope_name(ctxt));

    assert(fprintf(output_file,"\n#endif /* end of __%s__ */\n",pretty_name) > 0);
    assert(fprintf(output_file,"/* End of %s */\n",file_name) >0); 
    free(pretty_name);
    fclose(output_file);
  
}


