/* ************************************************************************* */
/* stage_1.c: First stage - filter input through CPP and write intermediate  */
/*            file.                                                          */
/*                                                                           */
/*    Copyright 2002-2006 Axis Communications AB.                            */
/*                                                                           */
/*                                                                           */
/*    This program is free software; you can redistribute it and/or modify   */
/*    it under the terms of the GNU General Public License as published by   */
/*    the Free Software Foundation; either version 2 of the License, or      */
/*    (at your option) any later version.                                    */
/*                                                                           */
/*    This program is distributed in the hope that it will be useful,        */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*    GNU General Public License for more details.                           */
/*                                                                           */
/*    You should have received a copy of the GNU General Public License      */
/*    along with this program; if not, write to the Free Software            */
/*    Foundation, Inc., 59 Temple Place - Suite 330,                         */
/*    Boston, MA 02111-1307, USA.                                            */
/*                                                                           */
/*                                                                           */
/*    Send bug reports and inquiries to Johan.Jorgensen@axis.com             */
/*                                                                           */
/*   Date      Description                                          Sign.    */
/*  =======================================================================  */
/*  2002-10-09 Module implemented                                    JohanJ  */
/*  2004-06-19 Now uses IOPASM_MALLOC and IOPASM_FREE macros         JohanJ  */
/* ************************************************************************* */



/* ************************************************************************* */
/*                      INCLUDES USED BY THIS MODULE                         */
/* ************************************************************************* */
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/stat.h>
#include <time.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include "iopasm_err.h"
#include "iopasm_global.h"
#include "dyn_array.h"
#include "cg_fe.h"
#include "debug.h"




/* ************************************************************************* */
/*                       TYPES DEFINED IN THIS MODULE                        */
/* ************************************************************************* */


/* ************************************************************************* */
/*                     DEFINITIONS LOCAL TO THIS MODULE                      */
/* ************************************************************************* */
#define MAX_CPP_ARGS 16



/* ************************************************************************* */
/*                              CONSTANT VARIABLES                           */
/* ************************************************************************* */
static const char *cpp_stat_paths[] = {
  "/lib/cpp", 
  "/usr/bin/cpp",
  NULL };



/* ************************************************************************* */
/*                    The template to use in order to run CPP                */
/* ************************************************************************* */
static const char *cpp_arg_template[] = {
  "cpp",             /* Path to the cpp found                                */
  "-nostdinc",       /* No standard includes in I/O processor                */
  "-no-gcc",         /* No gcc specific options for iopasm                   */
  "-DNO_ASM_MACROS", /* Do not include assembler macros in output            */
  "-undef",          /* Do not define any nonstandard macros                 */
  "-D__IOPASM__",    /* Define to identify IOPASM                            */
  NULL               /* _MUST_ BE NULL                                       */ 
};

static const char *builtin_includes[] = {
  "-I/n/asic/design/io/io_proc/sw/include",
  "-I/n/asic/design/io/io_proc/sw/guinness/include",
  NULL /* MUST BE NULL */ 
};


/* ************************************************************************* */
/*                    PROTOTYPES FOR STATIC FUNCTIONS                        */
/* ************************************************************************* */




/* ************************************************************************* */
/* PROTOTYPE   : char *create_tmpfile()                                      */
/*                                                                           */
/* ARGUMENTS   : none                                                        */
/*                                                                           */
/* FUNCTION    : Generates a temporary filename                              */
/*                                                                           */
/* RETURN VALUE: A pointer to a temporary filename                           */
/* ************************************************************************* */
char *create_tmpfile()
{
  char *fn_template="/tmp/iopasm_XXXXXX.s";
  char *fn_name;
  int i,j;
  int f;


  srand((unsigned int)time(NULL));
 
  i = 0;
  f = 0;

  while (fn_template[f] != 'X') { 
    f++; 
  }

  fn_name = (char *) strdup(fn_template);
    
  while (i<6) { /* This is where we try to generate a random ascii integer */
    j = (int) ((63.0*rand()/(RAND_MAX+1.0))+64);
    fn_name[f] = (char )j;
    f++;
    i++;
  }

  return fn_name;
}


/* ************************************************************************* */
/* PROTOTYPE   : int exec_cpp(cg_context_t ctxt,char *input_file,            */
/*                            char *output_file,array_t *includes);          */
/*                                                                           */
/* ARGUMENTS   : "ctxt" points to a valid context. "input_file" is the name  */
/*               of the file to read. "output_file" is the name of the       */
/*               output file once cpp has finished. "includes" is an array   */
/*               containing include-paths to be used.                        */
/*                                                                           */
/* FUNCTION    : Runs the C pre-processor (cpp).                             */
/*                                                                           */
/* RETURN VALUE: Return 1 if cpp was sucessfully run.                        */
/* ************************************************************************* */
int exec_cpp(cg_context_t *ctxt,char *input_file,
	     char *output_file, array_t *includes)
{
  const char *output_flag = "-o";
  pid_t pid;
  int wait_status;
  pid_t wait_result;
  struct stat stat_buf;
  int i,j, mx_index;
  int cpp_to_use = -1;
  char **cpp_args;
  char **arch_macros;

  assert((ctxt != NULL) && (input_file != NULL) && 
	 (output_file != NULL) && (includes != NULL));


  j = sizeof(cpp_arg_template) / sizeof(char *); /* ...-1 would be wrong! */
  j += get_array_size(includes); /* We're using j as a temporary variable here */
  j += cg_get_cpp_macro_count(ctxt);

  if (check_option("with-builtin-includes",(void *)1,OPT_INTEGER)) {
    j += sizeof(builtin_includes) / sizeof(char *);
  } /* Ends if (check_option("nostdinc",(void *)0,OPT_INTEGER)) */

  /* We also need 3 pointers for additional:
   *
   * -o switch
   * output_filename
   * input_filename
   *
   */
  
  j += 3;
  mx_index = j; /* This is just a debug-variable */

  assert((cpp_args = (char **) IOPASM_MALLOC(j * sizeof(char *))) != NULL);

  memset(cpp_args,0,j * sizeof(char *));
  
  /* Prepare for CPP execution. First set up all arguments */
  i = 0;

  while (cpp_stat_paths[i] != NULL) {
    if (stat(cpp_stat_paths[i],&stat_buf) == 0) {
      cpp_to_use = i;
      }
    i++;
  }
  
  i = 0;
  while(cpp_arg_template[i] != NULL) {
    cpp_args[i] = (char *) cpp_arg_template[i];
    i++;
  }
  
  /* Add architecture definition */
  arch_macros = cg_get_cpp_macros(ctxt);
  
  j = 0;
  while(arch_macros[j] != NULL) {
    cpp_args[i] = arch_macros[j];
    i++; j++;
  }
  
  
  if (cpp_to_use < 0) {
    iopasm_error(NULL,NULL,ERR_CPP_STAGE_FAILED,NULL,"\n");
  }
  
  else {
    cpp_args[0] = (char *) cpp_stat_paths[cpp_to_use];
  } /* End of !if (cpp_to_use < 0) */
  
  
    /* Add builtin include paths if required */
  if (check_option("with-builtin-includes",(void *)1,OPT_INTEGER)) {
    j = 0;
    while(builtin_includes[j] != NULL) {
      cpp_args[i++] = (char *) builtin_includes[j++];
    }
  } /* Ends if (check_option("nostdinc",(void *)0,OPT_INTEGER)) */
  
    /* Add Extra includes from command line */
  if (get_array_size(includes) > 0) {
    j = 0;
    while(get_array_elem(includes,j) != NULL) {
      cpp_args[i++] = (char *) get_array_elem(includes,j++);
    }
  } /* Matches if (get_array_size(includes) > 0) */
  
    /* Add input and output files last */
  
  cpp_args[i++] = input_file;
  cpp_args[i++] = (char *) output_flag;
  cpp_args[i++] = output_file;
  
  cpp_args[i] = NULL;
  
  fprintf(stderr,"Running cpp like this: ");
  /* Debug starts here */
  i = 0;
  while (cpp_args[i] != NULL) {
    fprintf(stderr,"%s ",cpp_args[i]);
    i++;
  } 
    
  fprintf(stderr,"\n"); /* Debug Ends here */ 
  
  
  pid = fork();
  
  if (pid == 0) { /* Executed by child only */ 

    i = execv(cpp_args[0],cpp_args);
    /* Anything below this line should not be executed */
    iopasm_error(NULL,NULL,ERR_CPP_EXEC_FAILED,NULL);
  }

  else if (pid == -1) { /* Error */
    iopasm_error(NULL,NULL,ERR_FORK_FAILED,NULL);
  }
  
  else { /* Parent */
    DEBUG(3,"Waiting for child %d",pid);
    wait_result = waitpid(pid,&wait_status,0);
    if ((WIFEXITED(wait_status) == 1) &&
        (WEXITSTATUS(wait_status) > 0)) { /* Error in CPP stage */
      iopasm_error(NULL,NULL,ERR_CPP_STAGE_FAILED,NULL,"This is most likely due to a syntax error in the C macros used in the source file %s",
		   input_file);
    }

  }
  return 1;
} /* End of exec_cpp(...) */

