/* ************************************************************************* */
/* state_descriptor.c : General description of a state used by the FSM       */
/*                      optimizer.                                           */
/*                                                                           */
/*    Copyright 2002-2006 Axis Communications AB.                            */
/*                                                                           */
/*                                                                           */
/*    This program is free software; you can redistribute it and/or modify   */
/*    it under the terms of the GNU General Public License as published by   */
/*    the Free Software Foundation; either version 2 of the License, or      */
/*    (at your option) any later version.                                    */
/*                                                                           */
/*    This program is distributed in the hope that it will be useful,        */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*    GNU General Public License for more details.                           */
/*                                                                           */
/*    You should have received a copy of the GNU General Public License      */
/*    along with this program; if not, write to the Free Software            */
/*    Foundation, Inc., 59 Temple Place - Suite 330,                         */
/*    Boston, MA 02111-1307, USA.                                            */
/*                                                                           */
/*                                                                           */
/*    Send bug reports and inquiries to Johan.Jorgensen@axis.com             */
/*                                                                           */
/*   Date      Description                                          Sign.    */
/*  =======================================================================  */
/*  2004-12-08 Module implented                                     JohanJ   */
/* ************************************************************************* */



/* ************************************************************************* */
/*                      INCLUDES USED BY THIS MODULE                         */
/* ************************************************************************* */
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include "iopasm_global.h"
#include "iopasm_types.h"
#include "p_list.h"

/* ************************************************************************* */
/*                       TYPES DEFINED IN THIS MODULE                        */
/* ************************************************************************* */
struct __state_t__ {
  char *state_name;    /* Name of state (label defining it */
  inst_t *state_start; /* Start of state (label) */
  inst_t *state_end;   /* end of state meta_instruction */
  int state_size;      /* Size of state in bytes */
};

/* ************************************************************************* */
/*                     DEFINITIONS LOCAL TO THIS MODULE                      */
/* ************************************************************************* */



/* ************************************************************************* */
/*                              CONSTANT VARIABLES                           */
/* ************************************************************************* */



/* ************************************************************************* */
/*                    PROTOTYPES FOR STATIC FUNCTIONS                        */
/* ************************************************************************* */



/* ************************************************************************* */
/* PROTOTYPE   : state_t *create_state(char *state_name,inst_t *start,       */
/*                                     inst_t *end,int size);                */
/*                                                                           */
/* ARGUMENTS   : "state_name" is the name of the state to create, "start" is */
/*               the starting point in the list and "end" is the ending point*/
/*               size is the size of the state in bytes.                     */
/*                                                                           */
/* FUNCTION    : Creates a state descriptor.                                 */
/*                                                                           */
/* RETURN VALUE: Pointer to a new state descriptor. NULL on failure          */
/* ************************************************************************* */
state_t *create_state(char *state_name,inst_t *start,inst_t *end,int size)
{
  state_t *temp_state;
  assert((state_name != NULL));
  
  temp_state = IOPASM_MALLOC(sizeof(state_t));
  assert(temp_state != NULL);

  temp_state->state_name = (char *) strdup(state_name);
  assert(temp_state->state_name != NULL);

  /* Set start and end of a state */
  temp_state->state_start = start;
  temp_state->state_end = end;
  
  temp_state->state_size = size;

  return temp_state;
}



/* ************************************************************************* */
/* PROTOTYPE   : state_t *copy_state(state_t *src);                          */
/*                                                                           */
/* ARGUMENTS   : "src" points to the state to copy.                          */
/*                                                                           */
/* FUNCTION    : Copies a state descriptor.                                  */
/*                                                                           */
/* RETURN VALUE: Pointer to a copy of the state descriptor.                  */
/* ************************************************************************* */
state_t *copy_state(state_t *src)
{
  state_t *tmp_state;
  
  tmp_state = create_state(src->state_name,
			   src->state_start,
			   src->state_end,
			   src->state_size);
  assert(tmp_state != NULL);

  return tmp_state;
}



/* ************************************************************************* */
/* PROTOTYPE   : void delete_state(state_t *a_state);                        */
/*                                                                           */
/* ARGUMENTS   : "a_state" is a pointer to a valid state descriptor          */
/*                                                                           */
/* FUNCTION    : Deletes a state descriptor                                  */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void delete_state(void *a_state)
{
  state_t *the_state;
  assert(a_state != NULL);
  
  the_state = (state_t *) a_state;
  IOPASM_FREE(the_state->state_name);
  IOPASM_FREE(the_state);
}



/* ************************************************************************* */
/* PROTOTYPE   : void set_state_end(state_t *a_state,inst_t *state_end);     */
/*                                                                           */
/* ARGUMENTS   : "a_state" points to a state and "state_end" points to the   */
/*               instruction marking the end of the state.                   */
/*                                                                           */
/* FUNCTION    : Mark the end of the state in the state descriptor.          */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void set_state_end(state_t *a_state,inst_t *state_end)
{
  assert((a_state != NULL) && (state_end != NULL));
  a_state->state_end = state_end;
}



/* ************************************************************************* */
/* PROTOTYPE   : void set_state_start(state_t *a_state,inst_t *state_start); */
/*                                                                           */
/* ARGUMENTS   : "a_state" points to a state and "state_start" points to the */
/*               instruction marking the start of the state.                 */
/*                                                                           */
/* FUNCTION    : Set the start of the state in the state descriptor.         */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void set_state_start(state_t *a_state,inst_t *state_start)
{
  
  assert((a_state != NULL) && (state_start != NULL));
  a_state->state_start = state_start;
}



/* ************************************************************************* */
/* PROTOTYPE   : void set_state_name(state_t *a_state,char *state_name);     */
/*                                                                           */
/* ARGUMENTS   : "a_state" points to a state and "state_name" points to the  */
/*               the name of the state.                                      */
/*                                                                           */
/* FUNCTION    : Set the name of the state.                                  */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void set_state_name(state_t *a_state,char *state_name)
{
  assert((a_state != NULL) && (state_name != NULL));

  a_state->state_name = (char *) strdup(state_name);
  assert(a_state->state_name != NULL);
}



/* ************************************************************************* */
/* PROTOTYPE   : void set_state_size(state_t *a_state,unsigned int size);    */
/*                                                                           */
/* ARGUMENTS   : "a_state" points to a state. "size" is the size of the      */
/*               state in bytes.                                             */
/*                                                                           */
/* FUNCTION    : Set the size of the state                                   */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void set_state_size(state_t *a_state,unsigned int size) 
{
  assert(a_state != NULL);
  a_state->state_size = size;
}



/* ************************************************************************* */
/* PROTOTYPE   : char *get_state_name(state_t *a_state);                     */
/*                                                                           */
/* ARGUMENTS   : "a_state" points to a valid state name.                     */
/*                                                                           */
/* FUNCTION    : returns the name of the state                               */
/*                                                                           */
/* RETURN VALUE: Pointer to the state name                                   */
/* ************************************************************************* */
char *get_state_name(state_t *a_state)
{
  assert((a_state != NULL));
  return a_state->state_name;
}



/* ************************************************************************* */
/* PROTOTYPE   : inst_t *get_state_start(state_t *a_state);                  */
/*                                                                           */
/* ARGUMENTS   : "a_state" points to a valid state name.                     */
/*                                                                           */
/* FUNCTION    : Returns the start of the state.                             */
/*                                                                           */
/* RETURN VALUE: Pointer to the instruction marking the start of the state.  */
/* ************************************************************************* */
inst_t *get_state_start(state_t *a_state) 
{
  assert(a_state != NULL);
  return a_state->state_start;
}



/* ************************************************************************* */
/* PROTOTYPE   : inst_t *get_state_end(state_t *a_state);                    */
/*                                                                           */
/* ARGUMENTS   : "a_state" points to a valid state name.                     */
/*                                                                           */
/* FUNCTION    : returns the end of the state                                */
/*                                                                           */
/* RETURN VALUE: pointer to the instruction marking the end of the state.    */
/* ************************************************************************* */
inst_t *get_state_end(state_t *a_state)
{
  assert(a_state != NULL);
  return a_state->state_end;
}



/* ************************************************************************* */
/* PROTOTYPE   : unsigned int get_state_size(state_t *a_state);              */
/*                                                                           */
/* ARGUMENTS   : "a_state" points to a valid state name.                     */
/*                                                                           */
/* FUNCTION    : returns the size of the state in bytes.                     */
/*                                                                           */
/* RETURN VALUE: Size of the size of the state measured in bytes.            */
/* ************************************************************************* */
unsigned int get_state_size(state_t *a_state)
{
  assert(a_state != NULL);
  return a_state->state_size;
}



/* ************************************************************************* */
/* PROTOTYPE   : int cmp_states(void *s1,void *s2);                          */
/*                                                                           */
/* ARGUMENTS   : "s1" points to one state and "s2" to another.               */
/*                                                                           */
/* FUNCTION    : Compares two states                                         */
/*                                                                           */
/* RETURN VALUE: 0 if s1 == s2, >1 if s1 > s2 and <1 if s1 < s2              */
/* ************************************************************************* */
int cmp_states(void *s1,void *s2)
{
  state_t *state_1,*state_2;
  
  assert((s1 != NULL) && (s2 != NULL));
  state_1 = (state_t *)s1;
  state_2 = (state_t *)s2;
  return strcmp(state_1->state_name,state_2->state_name);
}



/* ************************************************************************* */
/* PROTOTYPE   : void clone_state(state_t *dest,state_t *src);               */
/*                                                                           */
/* ARGUMENTS   : "dest" is a destination state and "src" is the source       */
/*                                                                           */
/* FUNCTION    : Clone a state by creating verbatim copies from src to dst   */
/*                                                                           */
/* RETURN VALUE: None.                                                       */
/* ************************************************************************* */
void clone_state(state_t *dest,state_t *src) 
{
  assert((dest != NULL) && (src != NULL));

  if (strcmp(dest->state_name,src->state_name) == 0) {
    dest->state_start = src->state_start;
    dest->state_end = src->state_start;
    dest->state_size = src->state_size;
  }
};
